;;; -*- Mode: LISP; Syntax: Common-lisp; Package: QPE -*-

;;;; QPE Benchmarking system

;; Copyright (c) 1989, Kenneth D. Forbus, University of Illinois
;; All rights reserved.

(in-package 'QPE)

(defvar *qpe-benchmark-data* nil) ;; Information accumulated for report.
(defvar *QPE-example-data* nil) ;information about the current example. 

;; Because of long runs and wrap-around in internal timers,
;; also cache the time of day. Benchmarking timers have
;; the form (<regular timer> . <universal time integer>)
;; for easy comparison. 

(defun store-times (timer1 timer2 &optional (phrase ""))
  (let* ((delta-time (make-time-difference (car timer2) (car timer1)))
	 (real-time (cdr delta-time))
	 (run-time (car delta-time)))
    (push (list phrase run-time real-time (cdr timer1) (cdr timer2)) *QPE-example-data*)
    (values run-time real-time)))

(defun snarf-current-timers ()
  (cons (mark-time) (get-universal-time)))

(defun get-readable-time ()
  ;; Probably dead code.  Want to keep internal time around
  ;; so that if run-time loses, can subtract clock time.
  (multiple-value-bind (second minute hour date month year)
      (get-decoded-time)
    (list (format nil "~D:~D:~D" hour minute second)
	  (format nil "~D//~D//~D" date month year))))

;;;; Running benchmarks
;;
;; Call BENCHMARK-QPE to be queried step-by-step.
;; Call RUN-QPE-BENCHMARKS to get version with keyword arguments.

(defvar *scenarios* nil) ;for particular run

(proclaim '(special *start-time* *qpe-example-data* *qstate* *derived-qstate*
		    *GC-ON?* *GC-EACH-TIME?* *benchmark-physics-files*
		    *qpe-benchmark-data* *last-benchmark-parameters*
		    *Dump-result-data?*
		    *QC-qstate* *derived-qstate* *resolve-qstate*))

(defun benchmark-qpe (&optional (formats '(:text :raw))  ;; do everything by default
		      &aux result-file comments GC-During GC-Between Examples)
  (format t "~%First filename for results:")
  (setq result-file (read))
  (setq result-file (format nil "~A" result-file)) ;; Safely stringify
  (format t "~%Line of comments (end with RETURN):~%")
  (setq comments (read-line))
  (format t "~%Dump envisioner results for each example?")
  (cond ((yes-or-no-p) (setq *dump-result-data?* t))
	(t (setq *dump-result-data?* nil)))	 
  (format t "~%Run with GC On?")
  (setq GC-During (yes-or-no-p))
  (format t "~%GC between runs?")
  (setq GC-Between (yes-or-no-p))
  (format t "~%Assuming ~A as domain file(s).  Okay?" *benchmark-physics-files*)
  (unless (yes-or-no-p)
    (format t "~%Please enter list of domain file name(s):")
    (setq *benchmark-physics-files* (read)))
  (setq examples *benchmark-scenarios*) 
  (format t "~%Include all examples?")
  (unless (yes-or-no-p)
    (dolist (example *benchmark-scenarios*)
      (format t "~%Include ~A?" (car example))
      (unless (yes-or-no-p)
	(setq examples (remove example examples)))))
  (format t "~%Are you really ready to go?")
  (when (yes-or-no-p)
    (setq *last-benchmark-parameters*
	  (list result-file :Comments comments
			:GC-During GC-during :GC-Between GC-Between
			:Formats formats :Examples examples))
    (apply #'run-qpe-benchmarks *last-benchmark-parameters*)))

(defun run-qpe-benchmarks (Result-file
			    &key (Comments "") (GC-During nil)
			    (GC-Between nil) (Examples *Benchmark-scenarios*)
			    (formats '(:TEXT :RAW)))
  (setq *GC-ON?* GC-during *GC-EACH-TIME?* GC-Between)
  (cond (GC-During #+Rel6 (scl:gc-on)
		   #+Rel7 (sys:gc-on)
		   #-Symbolics (gc-on)
		   )
	(t #+Rel6 (scl:gc-off)
	   #+Rel7 (sys:gc-off)
	   #-Symbolics (gc-off)))
  (setq *start-time* (summarize-current-time))
  (store-initial-benchmark-data)
  (dolist (example Examples)
   (when GC-Between
	  #+Symbolics (si:gc-immediately)
	  #-Symbolics (gc))
    (Start-QPE-Example-Data example) ;clear accumulators
    (benchmark-qpe-example example)
    (finish-QPE-example-data)) 
  (store-end-benchmark-data)
  (dump-qpe-benchmark-results comments result-file formats))

(defun rerun-qpe (&optional (result-file (car *last-benchmark-parameters*))
			    (comments (caddr *last-benchmark-parameters*)))
  (apply #'run-qpe-benchmarks
	 `(,result-file :COMMENTS ,comments
	   ,@ (cdddr *last-benchmark-parameters*))))  

;;;; The data for each example is stored as an alist.

(defvar *QPE-Example-Name* nil)

(defun start-qpe-example-data (example)
  ;; Packages up runtime information
  (setq *QPE-Example-Name* (car example))
  (setq *QPE-example-data* nil) ;clear old information
  (push `(:Name ,(car example)) *QPE-example-data*)
  (push `(:Short-name ,(cadr example)) *QPE-Example-Data*)
  (push `(:File ,(caddr example)) *QPE-Example-Data*)
  (push `(:start-time ,(mark-time)) *QPE-Example-Data*)
  )
  
(defun finish-qpe-example-data ()
  ;; Adds in the rest of the statistics
  ;; ***** Add in resolve qstate
  (push `(End-Time ,(mark-time)) *QPE-Example-Data*)
  (when *Dump-result-data?*
    (dump-envisioner-results
      (caddr *last-benchmark-parameters*)
	     (concatenate 'string *QPE-benchmark-data-directory*
			  (format nil "~A" 
				  (car *last-benchmark-parameters*)) "-" 
			  (cadr (assoc 'short-name
				       *QPE-Example-Data*))
			  ".LPT")))
  (multiple-value-bind (sit-count class-count)
      (count-envisioned-situations)
    (push `(:NSituations ,sit-count) *QPE-Example-Data*)
    (push `(:NSclasses ,class-count) *QPE-Example-Data*))
  (push `(:Qstate ,(length *QC-qstate*)) *QPE-Example-Data*)
  (push `(:Derived-Qstate ,(length *derived-qstate*)) *QPE-Example-Data*)
  (push `(:Preconditions ,(length *preconditions*)) *QPE-Example-Data*)
  (push `(:QPE-Stats ,(with-output-to-string (foo)
		       (summarize-qpe-statistics *QPE-Example-Name*
					       foo))) *QPE-Example-Data*)
  (push `(:ATMS-Stats ,(adb::list-statistics)) *QPE-Example-Data*)
  (push *QPE-Example-Data* *QPE-Benchmark-Data*))

;; These two are here in case something interesting comes up.
;; For instance, how much free space did one start with and how
;; much is left at the end.  Some implementations seem to have
;; faulty garbage collectors.

(defun store-initial-benchmark-data ()
  (setq *QPE-Benchmark-Data* nil))

(defun store-end-benchmark-data () nil)

;;; Dumping the results

(defun dump-qpe-benchmark-results (comments file formats)
  ;;; Several formats are desirable: Text, Lisp, TeX, and REFLEX
  ;;; Text first.
 (let ((*end-time* (summarize-current-time nil)))
  (when (member :Text formats)
  (let ((outfile  (concatenate 'string *QPE-benchmark-data-directory* file ".txt")))
    (with-open-file (stream outfile :direction :output)
      (format stream ";-*- TEXT -*-~%~A, -- ~D examples, QPE Version ~A~%"
	      (summarize-current-time nil)
	      (length *QPE-benchmark-data*)
	      *version-number*)
      (format stream "~%~%Started ~A, finished ~A."
	      *start-time* (summarize-current-time nil))
      (format stream "~%~A, ~A.~%"
	      (if *GC-on?* "Garbage Collection on" "Garbage Collection off")
	      (if *GC-each-time?* "full GC between examples"
		  "no GC between examples"))
      #+rel6 (scl:print-herald :stream stream)
      #+rel7 (sys:print-herald :stream stream)
      #-Symbolics (summarize-location stream)
      (format stream "~%~%~A~%~%" comments)
      (dump-qpe-benchmark-data (reverse *QPE-benchmark-data*) stream))))
  (when (member :RAW formats)
    (let ((outfile (concatenate 'string *QPE-benchmark-data-directory* file ".dat")))
      (with-open-file (stream outfile :direction :output)
	(format stream "~%(:HEADER :WHEN ~A :N-EXAMPLES ~D :VERSION ~A)"
		*end-time*
		(length *qpe-benchmark-data*)
		*version-number*)
	(format stream "~%(:WHEN :START ~A :FINISHED ~A)"
		*start-time* *end-time*)
	(format stream "~%(:GC-STATUS :INCREMENTAL ~A :FULL-BETWEEN ~A)"
		*GC-on?* *GC-each-time?*)
	(format stream "~%(:COMMENTS |~A| )" comments) ;; Need to turn this into a string!
	(dump-raw-qpe-benchmark-data (reverse *qpe-benchmark-data*) stream))))))

;;;; More dumping -- Text mode
(defun dump-qpe-benchmark-data (data stream)
  ; Print out a table consisting of the data for each example.  The right
  ; format is Situations, Sclasses, loading, PS/VS, resolve, limit, total.
  ; Also compute some other interesting statistics,
  ; such as Seconds/situation, % paging, etc.
  (let ((secs-sit nil)
	(sits nil)
	(sclasses nil)
	(examples nil)
	(qstates nil)
	(derived-qstates nil)
	(preconditions nil))
(format stream "~%=================================================")
(format stream "~% Times broken down by example:")
(format stream "~%~%~6:< ~>~15:<Expanding~>~15:<QC/PC~>~15:<Resolve~>~15:<Limit~>~15:<Total~>")
  (dolist (datum data)
    (let ((example (cadr (assoc ':NAME datum)))
	  (file (cadr (assoc ':FILE datum)))
	  (loading (cdr (assoc 'Expand datum)))
	  (PSVS (cdr (assoc 'PSVS datum)))
	  (resolve (cdr (assoc 'resolve datum)))
	  (limit (cdr (assoc 'limit-analysis datum)))
	  (nsits (cadr (assoc ':Nsituations datum)))
	  (nsclasses (cadr (assoc ':Nsclasses datum)))
	  (qstate (cadr (assoc ':Qstate datum)))
	  (derived-qstate (cadr (assoc ':Derived-Qstate datum)))
	  (precondition (cadr (assoc ':Preconditions datum)))
	  (total 0)
	  (realtime-total 0)
	  (sec-sit 0))
      ;; Times are stored as (<run time> <real time> <start time> <end time>)
      ;; The first two are in internal time units, the rest in universal time format.
      (setq total (convert-internal-time (+ (car loading) (car psvs) (car resolve) (car limit)))
	    sec-sit (// total nsits)
	    realtime-total (convert-internal-time (+ (cadr loading) (cadr psvs)
						     (cadr resolve) (cadr limit))))
      ;;Store information for second table
      (push example examples)
      (push nsits sits)
      (push nsclasses sclasses)
      (push sec-sit secs-sit)
      (push qstate qstates)
      (push derived-qstate derived-qstates)
      (push precondition preconditions)
      ;Two lines of display, one for run times, the other for paging.
      (format stream "~%   ~A, from ~A" example file)
    (format stream "~%~15:<Run time:~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
	    (convert-internal-time (car loading))
	    (convert-internal-time (car psvs)) (convert-internal-time (car resolve))
	    (convert-internal-time (car limit)) total)
    (format stream "~%~15:<Real time:~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
	    (convert-internal-time (cadr loading))
	    (convert-internal-time (cadr psvs))
	    (convert-internal-time (cadr resolve)) (convert-internal-time (cadr limit))
	    realtime-total)
    ;; Time of day
    (format stream "~%~15:<Start:~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
	    (extract-time (third loading))
	    (extract-time (third psvs))
	    (extract-time (third resolve))
	    (extract-time (third limit)))
    ;; Date (for LONG runs!)
    (format stream "~%~15:<~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
	    (extract-date (third loading))
	    (extract-date (third psvs))
	    (extract-date (third resolve)) (extract-date (third limit)))
    ;; Same for end
    (format stream "~%~15:<End:~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
	    (extract-time (fourth loading))	  
	    (extract-time (fourth psvs))
	    (extract-time (fourth resolve))
	    (extract-time (fourth limit)))
    (format stream "~%~15:<~>~15:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
	    (extract-date (fourth loading))
	    (extract-date (fourth psvs))
	    (extract-date (fourth resolve))
	    (extract-date (fourth limit)))))
  ;Now do summary table
  (format stream "~%~% Summary of runtime data")
  (format stream "~%~%~40:<Example~>~15<# Sits~>~15<# Sclasses~>~15<Sec/Sit~>")
  (mapcar #'(lambda (name nsits nsclasses sec-sit)
	      (format stream "~%~40:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
		      name nsits nsclasses sec-sit))
	  (reverse examples)
	  (reverse sits)
	  (reverse sclasses)
	  (reverse secs-sit))
 ;;Dump yet more stuff
  (format stream "~%~%Some background data:") 
(format stream "~%~%~40<Example~>~15<PC's~>~15<Qstate~>~15<Ext. Qstate~>")
(mapcar #'(lambda (name nsits nsclasses sec-sit)
	      (format stream "~%~40:<~A~>~15:<~A~>~15:<~A~>~15:<~A~>"
		      name nsits nsclasses sec-sit))
	  (reverse examples)
	  (reverse preconditions)
	  (reverse qstates)
	  (reverse derived-qstates))
;;; Now some atms statistics
(format stream "~%=================================================")
(format stream "~%~% QPE statistics, by example:")
(dolist (datum data)
  (let ((example (cadr (assoc ':NAME datum)))
	(stats (cadr (assoc ':QPE-Stats datum))))
    (print-new-page stream)
    (print-divider stream)
    (format stream "~%For ~A:" example)
    (print-divider stream)
    (format stream "~A" stats)))))

;;;; Lisp mode dumping
(defun dump-raw-qpe-benchmark-data (data stream)
  ;;; Spits out the statistics in raw form.
  (dolist (datum data) (pprint data stream))) ;; pretty-print it, at least.

(defun benchmark-qpe-example (example)
  #+Symbolics (funcall tv:selected-window :set-more-p nil) ;turn off more processing
  (setq *domain-files* *benchmark-physics-files*)
  (setq *scenario-files* (cddr example))
  (summarize-current-time *standard-output*)
  (format t "~%Beginning ~A..." (car example))
  (benchmark-it))

(defun benchmark-it (&optional
		      start-time total-time start-clock end-clock)
  (do-it-for-the-record t)
  (setq start-time (snarf-current-timers))
  (setq total-time (make-time-difference (car start-time) (car start-time)))
  (setq start-clock (summarize-current-time nil))
  (expand-scenario-using-domain-model)
  (multiple-value-bind (rt dt)
      (store-times start-time (snarf-current-timers) 'Expand)
    (setq total-time (make-time-sum total-time (make-time rt dt))))
  (setq start-time (snarf-current-timers))
  (install-QC-PC-assumptions *Assume-Pcs*)
  (multiple-value-bind (rt dt)
      (store-times start-time (snarf-current-timers) 'PSVS)
    (setq total-time (make-time-sum total-time (make-time rt dt))))
  (setq start-time (snarf-current-timers))
  (resolve-influences)
  (multiple-value-bind (rt dt)
   (store-times start-time (snarf-current-timers) 'RESOLVE)
   (setq total-time (make-time-sum total-time (make-time rt dt))))
  (setq start-time (snarf-current-timers))
  (limit-analysis)
  (setq end-clock (snarf-current-timers))
  (multiple-value-bind (rt dt)
      (store-times start-time end-clock 'Limit-analysis)
      (setq end-clock (cdr end-clock))
    (setq total-time (make-time-sum total-time (make-time rt dt))))
  (push (list 'Total (car total-time) ;; total run time
	      (cdr total-time) ;; total real time
	      start-clock
	      end-clock) *QPE-example-data*))

(defun do-it-for-the-record (&optional (initialize? nil))
  ;; Setting up for the real work
  (adb::init initialize?)
  (if initialize? (reload-QPE-rules))
  (dolist (physics-file *domain-files*) (load physics-file))
  (dolist (test-file *scenario-files*) (load test-file)))

(defun round-it (num)
  ;;Want to have no more than two digits past the decimal.
  (// (round (* 100.0 num)) 100.0)) 
