;;; -*- Mode: LISP; Syntax: Common-lisp; Package: QPE -*-

(in-package 'QPE)

;;;; CAREFUL mode
;;
;; Like a flight recorder, this module taps into the state of QPE, producing
;; descriptions of various events that are normally hard to monitor.
;;
;; Strewn through the code are calls to (ADD-RECORDER-EVENT . <event>),
;; which add the specified event onto the *EVENT-LIST*.  The format for events is
;; (<type> . <data>). 

;;  Some utilities for exploring the event list are included below.
;;  Right now they are very simple, but should be data-driven eventually.

;; Careful mode will also enables various cross-checks of datastructure
;; integrity, and make sure that certain invariants remain unviolated.
;; These cross-checks are not needed if everything is working perfectly
;; and there is nothing weird about the model, but when "pushing the
;; edge of the envelope" both of these conditions are often false.

(proclaim '(special *Careful*))

(defvar *event-list* nil) ;; List of interesting events
(adb:init-form (setq *event-list* nil))

(defmacro When-Careful (&rest body)
  `(when *careful* ,@ body))

(defmacro add-recorder-event (&rest event-description)
  `(push (list ,@ event-description) *event-list*))

;;;; Flight recorder queries

(defun types-of-events ()
  (reverse (remove-duplicates (mapcar #'car *event-list*))))

(defun fr-event-breakdown (&optional (stream *standard-output*) 
				  &aux alist)
  (dolist (event *event-list*)
    (let ((entry (assoc (car event) alist)))
      (unless entry (setq entry (cons (car event) 0))
	      (push entry alist))
      (incf (cdr entry))))
  (cond (alist
	 (format stream "~%~45<Event type~>~5<#~>")
	 (dolist (entry alist)
	   (format stream "~%~45<~A~>~5<~D~>" (car entry) (cdr entry))))
	(t (format stream "~% QPE Flight recorder not on."))))

(defun events-of-type (ty)
  (reverse 
    (remove-if-not #'(lambda (event) (eq (car event) ty)) *event-list*)))

(defun show-fr-events (&optional (type nil) (filter nil))
  (let ((events (if type (events-of-type type)
		    (reverse *event-list*))))
    (when filter (setq events (remove-if-not filter events)))
    (do ((evs events (cdr evs))
	 (num 1 (1+ num)))
	((null evs) (format t "~% ~D events total." (1- num)))
      (format t "~%   ~D ~A" num (car evs)))))

(defun list-fr-events (&optional (type nil) (filter nil))
  (let ((events (if type (events-of-type type)
		    (reverse *event-list*))))
    (when filter (setq events (remove-if-not filter events)))
    events))

;;; Switch system -- useful for keeping interfaces under control

(defmacro if-switches-set (switch-message-list &rest body)
  (car (make-switches-set-form switch-message-list body)))

(defun make-switches-set-form (switch-message-list body)
  (cond ((null switch-message-list) `((values T (progn ,@ body))))
	(t `((cond ((and (boundp ',(caar switch-message-list))
			 ,(caar switch-message-list))
		    ,@ (make-switches-set-form (cdr switch-message-list) body))
		   (t (values NIL ,(cdar switch-message-list))))))))

(defmacro clear-switches (&rest switch-list)
  `(progn ,@ (mapcar #'(lambda (switch) `(setq ,switch nil)) switch-list)))

(defmacro set-switch (switch val) `(setq ,switch ,val))