;;-*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Defining process and individual views
;;
;; ORGANIZATION: All statements in a domain description are translated into ADB rules
;;               and facts. The rules all trigger on :INTERN, so that the constraints
;;               they impose are immediately available.
;; (Since we are exploring all consistent states anyway, the standard ATMS consumer mechanism
;;   is unnecessary overhead.)
;;
;; As per the theory, since there is so much shared structure between processes and individual
;; views we shall speak of "conditionalized descriptions", whose instances are CDI's, whenever
;; refering to some operation or datastructure that is valid for both types.
;;
;; For efficiency, we do NOT decompose the process description into explicit statements about
;; preconditions, quantity conditions, etc. as GIZMO did.  Instead, two rules are created for
;; each CD, one which finds instances of it, and the other which installs the constraints
;; associated with it.
;;
;; The "Many Worlds" interpretation of the ATMS allows a conceptual simplification
;; which also results in more compact code.  Situations are represented as environments. 
;; Facts thus refer to individuals, and are implicitly scoped into possible worlds by
;; consistency tests, rather than making distinct copies of facts to talk about slices.
;; There are certain facts that cannot be said in this nomenclature (i.e., comparatives
;; between distinct times), but those are not necessary for generating an envisionment.
;; 

(defmacro define-cd-constructor (cd-name cd-type implement-fun)
  ;; Since there is a great deal of syntactic commonality between Individual Views and
  ;; Processes (since Processes were figured out first and simplified to make Individual
  ;; Views), call their union "conditionalized descriptions", or "cd"'s.  This macro creates
  ;; macros that build CDI's.  Future versions of QPE may well include more of them,
  ;; such as encapsulated histories and hieararchial processes.
`(defmacro ,cd-name (pform &rest specs)
   ;; begin by extracting the various fields
   (let ((pname (cond ((listp pform) (car pform)) (t pform)))
	 (prest (cond ((listp pform) (cdr pform)) (T NIL)))
	 (individuals (cadr (member 'individuals specs)))
	 (preconditions (cadr (member 'preconditions specs)))
	 (quantityconditions (or (cadr (member 'quantityconditions specs))
				 (cadr (member 'Quantity-Conditions specs))))
	 (relations (cadr (member 'relations specs)))
	 (influences (cadr (member 'influences specs)))
	 (local-vars nil)
	 (free-vars nil))
     ;; now do some rudimentary syntax checks
     ;; All fields, if there, must be lists.
     (unless individuals
       (error "Missing Individuals field: ~A." pname))
     (unless (listp individuals)
       (error "Non-list Individuals field: ~A." pname)) 
     (when (and quantityconditions (not (listp quantityconditions)))
       (error "Non-list Quantity Conditions: ~A" pname))
     (when (and preconditions (not (listp preconditions)))
       (error "Non-list Preconditions: ~A." pname))
     (when (and relations ; There need not be relations. 
		(not (listp relations)))
       (error "Non-list Relations: ~A." pname))
     ,@ (if (eq cd-type 'process)
	    `((when (null influences)
		(error "Must have direct influences: ~A." pname))
	      (unless (listp influences)
		(error "Non-list Influences: ~A." pname)))
	    `((when influences
		(error "Only processes can have direct influences: ~A." pname))))
     (setq local-vars (cons '?self (test-individual-specs individuals pname)))
     (unless local-vars
       (error "~% ~A ~A must have SOME variables, somewhere."
	      ',cd-type pname))
     (when (setq free-vars
		 (set-difference (get-variables preconditions) local-vars))
       (error "~% ~A ~A has free variables ~A in Preconditions."
	      ',cd-type pname free-vars))
     (when (setq free-vars
		 (set-difference (get-variables quantityconditions) local-vars))
       (error "~% ~A ~A has free variables ~A in Quantity Conditions."
	      ',cd-type pname free-vars))
     (when (setq free-vars
		 (set-difference (get-variables relations) local-vars))
       (error "~% ~A ~A has free variables ~A in Relations field."
	      ',cd-type pname free-vars))
     (when (setq free-vars
		 (set-difference (get-variables influences) local-vars))
       (error "~% ~A ~A has free variables ~A in Influences field."
	      ',cd-type pname free-vars))
     (test-preconditions-specs preconditions pname)
     (test-quantityconditions-specs quantityconditions pname)
     (test-relations-specs relations pname)
     (test-influences-specs influences pname)
     (,implement-fun pname prest ',cd-type individuals 
      preconditions quantityconditions
      relations influences))))

(define-cd-constructor defprocess 
		       process 
		       implement-cdi-from-specification)

(define-cd-constructor defview
		       individual-view 
		       implement-cdi-from-specification)

;;;; Testing the Individuals field.
;;
;; I've signficantly expanded the syntax here, in response to feedback.
;; Each individual specification can now look like:
;; 1. (<var> :TYPE <type>), meaning "<var> can be anything of type <type>".
;; 2. (<var> :FORM <type>), meaning "<var> is bound to something which matches
;;      <pattern>".  Any variables appearing in <pattern> are available for use internally,
;;       thus providing a means of de-structuring.
;; 3. :TYPE and :FORM may be combined.
;; 4. optional: :CONDITIONS <patterns> provides additional patterns
;;    which must be successfully matched before an instance will be built.  If used,
;;    :CONDITIONS must be the last item in the individual spec.
;; 5. :TEST runs a piece of lisp code, and carries on if the test returns non-nil.
;;    I'm including this for debugging purposes, syntax checking tools should flag its use.
;;
;; As before, the modeler should assume that all specs are evaluated in order.
;; No forward references are allowed.

(defun keyword? (x) (and (symbolp x)
			 (string-equal "KEYWORD"
				       (package-name (symbol-package x)))))

(defun test-individual-specs (in cd-name &aux vars-so-far)
  (dolist (spec in (nreverse vars-so-far))
    (setq vars-so-far (check-individual-spec spec cd-name vars-so-far))))

(defun check-individual-spec (spec cd-name vars-so-far &aux vars)
  (unless (listp spec) (error "~% Non-list in Individuals of ~A." cd-name))
  (unless (variable? (car spec)) (error "~% Non-variable in Individuals of ~A: ~A"
					cd-name spec))
  (when (member (car spec) vars-so-far)
    (adb:if-debug-flag cdi-installer
       (format t "~% Warning: Redundant variable definitions in ~A: ~A.~% See ~A."
	   	 cd-name (car spec) spec)))
  (push (car spec) vars-so-far) ;; must be new to get this far.
  ;; Split out parsing the form
  (multiple-value-bind (lossage-type ignore1 ignore2 form-spec conditions tests bind)
      (extract-ispec-fields spec)
      (declare (ignore ignore1 ignore2)) ;; Lucid weirdness
    (when lossage-type ;; when things are really screwed up
      (error " ~A:~A of ~A." lossage-type spec cd-name))
    ;; update the variables with any new ones introduced by the form.
    (setq vars-so-far (adb::order-preserving-union (get-variables form-spec) vars-so-far))
    ;; We don't want random things to occur in the conditions, so each condition is either
    ;; a constant or mentions at least one variable which has been bound before.
    (when conditions
      (let ((mentioned? nil)
	    (vars nil)
	    (form-vars (get-variables form-spec)))
	(dolist (condition conditions)
	  (when (setq vars (get-variables condition))
	    (unless (some #'(lambda (var) (member var vars-so-far))
			  vars) 
	      (error "~%Floating condition ~A:~A in ~A." condition spec cd-name))
	    (if (or (member (car spec) vars)
		    (intersection form-vars vars)) (setq mentioned? t))
	    (setq vars-so-far (adb::order-preserving-union vars vars-so-far))))
	(unless (or mentioned? (intersection (get-variables form-spec) vars-so-far))
	  (format t "~%>>Warning: No condition mentions ~A: ~A in ~A."
		 (car spec) spec cd-name))))
    ;; No forward references allowed in BIND
    (setq vars (set-difference (get-variables bind) vars-so-far))
    (when vars
      (error "~% Forward reference found in :BIND of ~A:~% See ~A in ~A."
	     cd-name vars spec))
    ;; Now see if the tests are free of forward references
    (setq vars (set-difference (get-variables tests) vars-so-far))
    (when vars
      (error "~% Forward reference found in :TEST of ~A:~% See ~A in ~A."
	     cd-name vars spec))
    vars-so-far))

;;;; Oh, how I hate syntax!

(defun extract-ispec-fields (ispec &aux var type form conditions tests
				   bind)
  ;; Returns lossage type, var, type, form, conditions, and tests.
  ;; This is used for syntax checking and rule construction.
  (setq var (car ispec) ispec (cdr ispec))
  ;; Look for :TYPE specifications
  (when (setq type (member ':TYPE ispec))
    (when (member ':TYPE (cddr type)) ;; make sure no duplicates
      (return-from extract-ispec-fields
	(values "Multiple :TYPEs in Individuals spec.")))
    (setq type (cadr type)))
  ;; Find and check conditions.
  (when (setq conditions (member ':CONDITIONS ispec))
    (setq conditions (cdr conditions))
    (when (some #'keyword? conditions)
      (return-from extract-ispec-fields
	(values ":CONDITIONS must occur last.")))
    (unless (every #'listp conditions)
      (return-from extract-ispec-fields
	(values ":CONDITIONS must be ADB patterns."))))
  (when (setq bind (member ':BIND ispec))
    (setq bind (cdr bind))
    (when (cdr bind)
      (return-from extract-ispec-fields 
	(values "Improper use of :BIND in Individuals spec.")))
    (setq bind (car bind))
    (when (adb::contains-occurance-of bind var)
      (return-from extract-ispec-fields
	(values ":BIND cannot include occurrence of variable.")))
    (unless (get-variables bind)
      (return-from extract-ispec-fields
	(values ":BIND must include at least one variable."))))
  ;;Make sure type of ispec is properly defined
  (unless (or type conditions bind)
    (return-from extract-ispec-fields 
      (values "Must be either :TYPE, :CONDITIONS, or :BIND in Individuals spec.")))
  ;; Look for :FORM specifications
  (when (setq form (member ':FORM ispec))
    (when (member ':FORM (cddr form))
      (return-from extract-ispec-fields
	(values "Multiple :FORMs in Individuals spec.")))
    (setq form (cadr form))
    (when (adb::contains-occurance-of form var)
      (return-from extract-ispec-fields
	(values ":FORM cannot include occurrence of variable.")))
    (unless type
      (return-from extract-ispec-fields
	(values ":FORM currently must be used with type.")))) ;;; JWC 
  ;; Find and check tests
  (when (setq tests (member ':TEST ispec))
    (when (member ':TEST (cdr tests))
      (return-from extract-ispec-fields
	(values "Multiple :TESTs in Individual spec.")))
    (setq tests (cadr tests)))
  (when (and bind (or type conditions form tests))
    (return-from extract-ispec-fields 
      (values
	":BIND cannot occur with :TYPE or :CONDITIONS in Individuals spec.")))
  (values nil var type form conditions tests bind))

;;; Generating CDI-finding rules
;;
;; The extended individuals syntax complicates this.  The problem is that an individual
;; spec can now expand into several ADB rule triggers as well as intermediate variables.

;;; ****** Still need to put in false test on preconditions.

(proclaim '(special *Vars-to-destructure*))
 
(defun make-adb-rules-from-ispecs (ispecs tforms body
					  &aux trigger inner-triggers)
  ;; Build rule by recursing down individual specs, introducing new triggers as needed.
  (cond ((null ispecs)
	 ;; Check and see if any of the antecedents is FALSE.  If so, don't execute, because
	 ;; it cannot happen!
	 `((unless (or ,@ (mapcar #'(lambda (tform)
				     `(tms::false?
					(get-tms-node 
					  ,(adb::quotize tform))))
				 tforms))
	    ,@  body)))
	(t (multiple-value-bind (ignore var type form conditions tests bind)
	       (extract-ispec-fields (car ispecs))
	     ;;; If there are conditions, must make sure test is executed AFTER them.
	     ;;; Conditions will occur inside of anything else.
	     (when tests (setq tests (list ':TEST tests)))
	     (cond (conditions
		    (setq inner-triggers
			  (nconc (mapcar #'adb::desugar-pattern
					 conditions) tests)
			  tests nil))
		   (t (setq inner-triggers nil)))
	     (cond (type
		    (setq trigger (cons (list type (if form form var)) ':TRUE))
		    (cond (form (push (list var form) *vars-to-destructure*)
		`((adb:rule :INTERN (,trigger ,@ tests)
			(nlet ((,var ,form))
			      ,@ (if conditions
				   `((adb:rule :INTERN
					      ,inner-triggers
				     ,@ (make-adb-rules-from-ispecs
					  (cdr ispecs)
					  (append (mapcar #'adb::desugar-pattern
							  conditions)
						  (cons trigger tforms))
					  body)))
				   (make-adb-rules-from-ispecs
				     (cdr ispecs)
				     (cons trigger tforms)
				     body))))))
			  (conditions
			   `((adb:rule :INTERN (,trigger ,@ inner-triggers)
				       ,@ (make-adb-rules-from-ispecs
					    (cdr ispecs)
					    (append (mapcar #'adb::desugar-pattern
							    conditions)
						    (cons trigger tforms))
					    body))))
			  (t `((adb:rule :INTERN (,trigger ,@ tests)
				,@ (make-adb-rules-from-ispecs
				  (cdr ispecs)
				  (cons trigger tforms)
				  body))))))
		   (bind (push (list var bind) *vars-to-destructure*)
			 `((nlet ((,var ,bind))
				 ,@ (make-adb-rules-from-ispecs
				      (cdr ispecs)
				      tforms body))))
		   (conditions
		    `((adb:rule :INTERN ,inner-triggers
			       ,@ (make-adb-rules-from-ispecs
				    (cdr ispecs)
				    (append (mapcar #'adb::desugar-pattern
						    conditions)
					    tforms)
				    body))))
		   (t (error
		"~%Good heavens, we have really lost here! Call Forbus!")))))))

;;;; Helpers for rule translation

(defun compute-ispecs-for-antecedents (ispecs &aux antes)
  ;; ******** Must change this when allowing equality statements in antecedents
  (dolist (ispec ispecs)
    (multiple-value-bind (ignore var type form conditions)
	(extract-ispec-fields ispec)
      (cond (type (cond (form (push (adb::desugar-pattern (list type form)) antes))
			(t (push (adb::desugar-pattern (list type var)) antes)))))
      (dolist (condition conditions)
	(push (adb::desugar-pattern condition) antes))))
  antes)

;;;; Other syntactic checks.
;;
;; These are mostly trivial, and done to simplify later code
;; There is always a temptation to do predicate type checking,
;; but that can be built in later.  

(defun check-list-of-lists (ll)
(cond ((null ll) T)
      ((listp (car ll)) (check-list-of-lists (cdr ll)))
      (t NIL)))

(defun test-preconditions-specs (pc cd)
  (unless (check-list-of-lists pc)
    (error "Ill-formed Preconditions: ~A" cd)))

(defun test-quantityconditions-specs (qc cd)
  (unless (check-list-of-lists qc)
    (error "Ill-formed Quantity Conditions: ~A." cd)))

(defun test-relations-specs (rs cd)
  (unless (check-list-of-lists rs)
    (error "Ill-formed Relations: ~A." cd)))

(defun test-influences-specs (ins cd)
  (unless  (check-list-of-lists ins)
    (error "Ill-formed influences: ~A." cd))
  (unless (every #'(lambda (in) (member (car in) '(I+ I- CTRANS))) ins)
    (error "Non-direct influence statement in Influences: ~A." cd)))

;;;; Creating rules to find process and view instances

(defmacro instance-long-name (itype) ;; Was a subst.  
  `(case ,itype
    ((process pi) 'process-instance)
    ((individual-view vi) 'view-instance)))

(defun implement-cdi-from-specification 
  (pname prest ptype individuals preconditions quantity-conditions
	 relations influences &aux specs local-vars)
  (setq *Vars-to-destructure* nil) ;; Clear out accumulator
  (setq specs (compute-ispecs-for-antecedents individuals))
  (setq local-vars (test-individual-specs individuals pname))
  `(progn 'compile
	   ;First generate the rule to find them.
	  (adb::rule :INTERN ( (Expand-Scenario . :TRUE) )
		 ,@ (make-adb-rules-from-ispecs individuals nil
	       ;assert that an instance has been found, which will trigger
	       ;asserting its consequences.
	`((nlet ((?cdi (:eval (build-cdi-name ',ptype ',pname
					     (list ,@ local-vars)
					     (cons ',pname ,(adb::quotize prest))))))
		      (adb::if-debug-flag cdi-finder
					 (format t 
				       "~% CDI: Found ~A, an instance of ~A ~A"
						 ?cdi ',pname 
						 (list ,@ local-vars)))
		      (adb::if-debug-flag cdi-break
					 (break "~% CDI: Break on finding CDI"))
		      (adb::rassert! ((,(instance-long-name ptype)
				 ,pname ?cdi
				 ,@ local-vars)
				. :TRUE))
		      ;; Assert existence in both directions.  Use Rjustify instead of Rassume
		      ;; for run-time efficiency.
		      (adb::rjustify ((exists ?cdi) . :TRUE)
				,specs)
		      ,@ (mapcar #'(lambda (f)
				     `(adb::rjustify ((exists ?cdi) . :FALSE)
						,(list (flip-truths f))))
				 specs)))))
	  ,(multiple-value-bind (normals results evals local-names creates-object?)
	       (translate-relations relations '(((Active ?self) . :TRUE)) ptype
				    (format nil ":~A-RELATIONS" pname))
	     (multiple-value-bind (nnormals nresults nevals ignore)
		 (translate-relations
		   (first-order-sublis local-names influences) 
				      '(((Active ?self) . :TRUE)) ptype
				      (format nil ":~A-INFLUENCES" pname))
	       (make-cdi-consequence-rule 
		 ptype pname prest individuals preconditions quantity-conditions
		 (translate-relations preconditions '(((Active ?self) . :TRUE))
				      ptype)
		 (translate-relations quantity-conditions 
				      '(((Active ?self) . :TRUE)) ptype)
		 (remove-conjunctions (append normals nnormals))
		 (remove-conjunctions (append results nresults))
		 (append evals nevals) creates-object?)))))

;;;; Installing consequences of process and view instances
;
;Again much simpler than poor old GIZMO, because we are free to avoid thinking
;about slices!

(defun make-cdi-consequence-rule
       (i-type type-name vars individuals preconditions quantity-conditions
	translated-preconditions
	translated-quantity-conditions
	normal-relations ;justified by instance being active
	self-contained-relations ;carries own justification
	eval-relations ;things to just plug in
	creates-object?)
  (let ((pform (append
		 (list (instance-long-name i-type) ;when an instance is found
		       type-name
		       '?self)
 ;; This hair is to ensure all the variables mentioned in a :FORM spec are available.
;		 (mapcar #'(lambda (iform) 
;			     (let* ((var (car iform))
;				    (entry (assoc var *vars-to-destructure*)))
;			       (if entry (cadr entry) var))) individuals)
		 ;; Experiment -- make the form include ALL variables
		 (test-individual-specs individuals type-name)
		 ))
	(EQE '((Enforce Quantity-Existence) . :TRUE))
	(not-EQE '((Enforce Quantity-Existence) . :FALSE)))
    `(progn 'compile
	    (adb::rule :INTERN ((Expand-Scenario . :TRUE) ;; Control
				(,pform . :TRUE))
		  (adb::nlet ((?pform ,pform)  ;splice them in, if necessary
			      ,@ *vars-to-destructure*)
			;status is either active or inactive
			;active only if there is an instance and if
			;the preconditions and quantity conditions are true.
			(adb::if-debug-flag cdi-installer
			  (format T "~% CDI: Installing consequences of ~A"
				  ?pform))
			(adb::if-debug-flag cdi-asserts
				       (adb::set-debug-flag 'adb::justifications t)
				       (adb::set-debug-flag 'adb::nogoods t))
			(adb::rjustify ((Active ?self) . :TRUE)
				  ((,pform . :TRUE)
				   ((Exists ?self) . :TRUE)
			   ;;; If it creates an object, must believe EQE for this to hold.
				   ,@ (if creates-object? (list EQE))
				   ,@ translated-preconditions
				   ,@ translated-quantity-conditions)
				  'CDI-construction)
			;;Make justifications for being inactive.
			;; Current theory: Only QC changes need to be conditioned on EQE.
			;; This may have to change when envisionments are augmented.
			(adb::rjustify ((Active ?self) . :FALSE)
				       ((,pform . :FALSE))
				      'CDI-construction)
			(adb::rjustify ((Active ?self) . :FALSE)
				       (((Exists ?self) . :FALSE))
				      'CDI-construction)
			,@ (mapcar #'(lambda (pc)
				     `(adb::rjustify ((Active ?self) . :FALSE)
						     (,pc)
						    'CDI-Activation))
				     (flip-truths
				       translated-preconditions))
			,@ (mapcar #'(lambda (qc)
				     `(adb::rjustify ((Active ?self) . :FALSE)
						     ,(if creates-object?
							  (list qc EQE)
							  (list qc))
						'CDI-Activation))
				   (flip-truths
				     translated-quantity-conditions))
			;;; If the CDI creates an individual, then it needs to be active
			;;; whenever EQE is false in order to check ZE continuity. 
			,@ (if  creates-object?
			       `((adb::rjustify ((Active ?self) . :TRUE)
					   (,not-EQE) 'EQE-semantics)))
			;indexing information, for matching when you don't need all
			;the information about the instance
			(adb::rjustify (((,(instance-long-name i-type) ,type-name)
				    ?self) . :TRUE)
				  ((,pform . :TRUE))
				  'CDI-construction)
			;In this version the qstate and preconditions are just one
			;big list.  Here are where the forms to install them at runtime
			;are created.
			,@ (install-Qstate-information quantity-conditions)
			,@ (install-Precondition-information preconditions)
			;need to enforce existence constraints
;			(adb::rjustify ((Active ?self) . :FALSE)
;				  (((exists ?self) . :FALSE))
;				  'CDI-construction)
			;assert the predicate form.
			(adb::rjustify ((,type-name ,@ vars) . :TRUE)
				  (((active ?self) . :TRUE))
				  'CDI-predicate-form)
			;; Assert the negation of the predicate form
			(adb::rjustify ((,type-name ,@ vars) . :FALSE)
				  (((active ?self) . :FALSE))
				  'CDI-predicate-form)
			;install definitional relationships
			,@ (mapcar #'(lambda (entry)
				       `(adb::rjustify
					  ((?self 
					     ,(adb::unvariablize (car entry) *user-package*)
						  ,(car entry)) . :TRUE)
						 ((?pform . :TRUE))
						 'CDI-construction))
				   individuals)
			,@ eval-relations
			;install relations
			,@ (mapcan #'(lambda (fact)
				       (justify-facts-with-connectives
					 fact
					 '(((Active ?self) . :TRUE))
					 'CDI-Activation))
				   normal-relations)
			;install self-contained-relations
			,@ (mapcar #'(lambda (fact) `(adb::rjustify ,@ fact)) 
				   self-contained-relations)
		      (adb::if-debug-flag cdi-asserts
			  (adb::set-debug-flag 'adb::justifications nil)
			  (adb::set-debug-flag 'adb::nogoods nil)))))))



;;;; Naming view and process instances

(proclaim '(special *vi-list* *pi-list* *vi-counter* *pi-counter*))

(defvar *vi-list* nil) ;view instances for current problem
(defvar *pi-list* nil) ;process instances for current problem
(defvar *vi-counter* 0)
(defvar *pi-counter* 0)

(adb::Init-Form (setq *vi-list* nil *pi-list* nil
		      *vi-counter* 0 *pi-counter* 0))

;;; The entries on these lists take the form
;;   (<symbol> <defining form> <user form>)
;;    where <symbol> is a symbol in the user package, like PI3,
;;          <defining form> is (<type> . <exhaustive, in order list of individual vars>)
;;          and <user form> is the CADR of the defining form.
;;          The variables in <defining form> may be a proper superset of those in <user form>.
;;          IMPORTANT: It is up to the modeller to include enough information in <user form> to
;;          distinguish between different instances!

(defmacro build-cdi-name-of (generic-name arguments user-form prefix counter cache)
  `(let ((entry (find (cons ,generic-name ,arguments) ,cache :test #'equal
		       :key #'cadr))
	 (name nil))
     (cond (entry (car entry));; Already there -- just return the name
	   (t (setq name (intern (format nil "~A~A" ,prefix ,counter) 
				 *user-package*))
	      (setq entry (list name (cons ,generic-name ,arguments) ,user-form))
	      (push entry ,cache)
	      (incf ,counter)
	      name))))

(defun build-cdi-name (type generic-name arguments user-form)
  (case type
    (process (build-cdi-name-of generic-name arguments user-form "PI"
				*PI-COUNTER* *PI-LIST*))
    (individual-view (build-cdi-name-of generic-name arguments user-form
					"VI" *VI-COUNTER* *VI-LIST*))))

;;;; User code for fetching view and process instances 

(defun gather-pis () (mapcar #'car *pi-list*))
(defun gather-vis () (mapcar #'car *vi-list*))

(defun lookup-pi (name) (caddr (assoc name *pi-list*)))
(defun lookup-vi (name) (caddr (assoc name *vi-list*)))

(defun lookup-cdi-user-form (name)
  ;; Takes symbol, returns user form of associated CDI
  (let ((entry (lookup-vi name)))
    (cond (entry (return-from lookup-cdi-user-form entry))
	  (t (lookup-pi name)))))

(defun lookup-cdi-long-form (name)
  ;; Takes symbol, returns user form of associated CDI
  (let ((entry (assoc name *vi-list*)))
    (cond (entry (cadr entry))
	  (t (cadr (assoc name *pi-list*))))))

(defun lookup-cdi-name (form)
  ;; Takes user form, returns all names associated with it. 
  ;; Assumes constants, maybe make version with variables?
  (let ((entry (remove-if-not #'(lambda (entry) (equal (caddr entry) form)) *vi-list*)))
    (cond (entry (mapcar #'car entry))
	  (t (setq entry (remove-if-not #'(lambda (entry) (equal (caddr entry) form)) *pi-list*))
	     (mapcar #'car entry)))))

;;;; Installing Qstate information
;;
;; The Qstate consists of the union of quantity space information.
;;Much of it comes from explicit quantity conditions in the view and process
;;descriptions.  Other parts come from resolving influences and limit analysis
;;(rate assumptions), and now, correspondences!  The correspondence information goes
;;on the variable *derived-qstate*, so that it can be taken properly into account.
;;
;; The Qstate and the preconditions are central to state.  Combinations of their assumptions
;; form the environments corresponding to situations.

(defvar *qc-qstate* nil)
(defvar *derived-qstate* nil)
(defvar *preconditions* nil)
(adb::init-form (setq *qc-qstate* nil))
(adb::init-form (setq *derived-qstate* nil))
(adb::init-form (setq *preconditions* nil))

(defun install-qstate-information (qcs)
  (delete nil
	(mapcar #'(lambda (qc)
		    (cond ((not (listp qc)) nil)
			  ((eq (car qc) 'not)
			   (cond ((and (listp (cadr qc))
				       (member (caadr qc) 
					     '(Less-Than Equal-To Greater-Than)))
				  (install-qstate-form (cadadr qc) (caddr (cadr qc))))))
			  ((member (car qc) '(Less-Than Equal-To Greater-Than))
			   (install-qstate-form (cadr qc) (caddr qc)))))
		qcs)))

(defun install-qstate-form (first second)
  `(pushnew (apply #'fetch-ineq (find-qrel ,(adb::quotize first)
					   ,(adb::quotize second)))
	      *qc-qstate* :test #'equal))

(defun install-qc-qstate-pair (first second)
  (pushnew (apply #'fetch-ineq (find-qrel first second))
	   *qc-qstate* :test #'equal))

(defun install-derived-qstate-pair (first second)
  (pushnew (apply #'fetch-ineq (find-qrel first second))
	   *derived-qstate* :test #'equal))

;;;; Installing precondition information
;;
;; Preconditions are either asserted or assumed.  Either way, QPE must be able to
;; get its fingers on them easily.

(defun install-precondition-information (pcs)
  (mapcar #'(lambda (pc)
	      `(pushnew ,(adb::quotize
			   (car (adb::desugar-pattern pc)))
			*preconditions*
			:test #'equal))
	  pcs))

(defun strip-truth (form)
  (cond ((not (listp form)) form)
	((member (cdr form) '(:TRUE :FALSE))
	 (if (and (listp (car form))
		  (eq (caar form) 'NOT)) (cadar form)
	     (car form)))
	(t form)))

;;; Representing objects (more syntactic sugar)

(defmacro defentity (predicate &rest consequences)
  "Defining type predicates on individuals"
  (let ((p-form (if (symbolp predicate) `(,predicate ?SELF)
		     predicate))
	 (i-form (if (symbolp predicate) '?SELF
		     (if (listp predicate) (cadr predicate)
			 (error "~%Bad argument to defentity -- ~A." predicate))))
	 (informant (concatenate 'string ":DEFENTITY-"
			    (if (symbolp predicate)
				(symbol-name predicate)
				(if (symbolp (car predicate))
				    (symbol-name (car predicate))
				    (error "~%Bad argument to defentity -- ~A." 
					   predicate))))))
    ;;this uses the QPE syntaxer to simplify the modeller's life.
    ;;The first argument is either a symbol, which is used to stand for the name of
    ;;the predicate with the default argument being ?SELF, or a list whose CAR is
    ;;the predicate and whose CADR is a form which must match the individual description 
    (multiple-value-bind (facts stand-alone-results evaluation-forms)
      (translate-relations consequences (list (cons p-form  :TRUE)) nil
			   informant)
  `(adb::rule :INTERN ( (Expand-Scenario . :TRUE) ;; Add a bit more control.
		       (,p-form . :TRUE) ) ;install upon mere reference.
	 ;if the object doesn't exist, the predicate cannot be true of it.
	     (adb::rjustify  (,p-form . :FALSE)
			    (((exists ,i-form) . :FALSE))
			    ,informant)
	 ;if the predicate is true the object must exist
	 (adb::rjustify ((exists ,i-form) . :TRUE)
		  ((,p-form . :TRUE)) ,informant)
	 ;install things true because the predicate holds
	 ,@ (mapcan #'(lambda (fact)
			(justify-facts-with-connectives fact `((,p-form . :TRUE)) informant))
		    facts)
	 ;install other consequences 
	 ,@ (mapcar #'(lambda (s-a-r) `(adb::rjustify ,@ s-a-r))
		    stand-alone-results)
	 ;do any additional work required
	 ,@ evaluation-forms))))

(defmacro defpredicate (predicate &rest consequences)
  "Defining type predicates on individuals"
  (let ((p-form (if (symbolp predicate) `(,predicate ?SELF)
		     (if (listp predicate) predicate
			 (error "~% Defpredicate must have symbol or list -- ~A" predicate))))
	 (informant (format nil ":defpredicate-~A" 
			    (if (symbolp predicate) (string-downcase (symbol-name predicate))
				(if (symbolp (car predicate))
				    (string-downcase (symbol-name (car predicate)))
				    (error "~% Defpredicate must be symbol -- ~A"
					   (car predicate)))))))
    ;;this uses the QPE syntaxer to simplify the modeller's life.
    ;;The first argument is either a symbol, which is used to stand for the name of
    ;;the predicate with the default argument being ?SELF, or a list whose CAR is
    ;;the predicate and whose CADR is a form which must match the individual description 
    (multiple-value-bind (facts stand-alone-results evaluation-forms)
      (translate-relations consequences (list (cons p-form  :TRUE)) nil informant)
  `(adb::rule :INTERN ( (Expand-Scenario . :TRUE)
		       (,p-form . :TRUE) ) ;install upon mere reference.
	 ;install things true because the predicate holds
	 ,@ (mapcan #'(lambda (fact)
			(justify-facts-with-connectives fact `((,p-form . :TRUE))
							informant))
		    facts)
	 ;install other consequences 
	 ,@ (mapcar #'(lambda (s-a-r) `(adb::rjustify ,@ s-a-r ,@ (unless (third s-a-r)
								    (list informant))))
		    stand-alone-results)
	 ;do any additional work required
	 ,@ evaluation-forms))))

;;;; Display routines

(defun show-cdis (&optional (stream *standard-output*))
  (show-vis stream)
  (show-pis stream))

(defun show-pis (&optional (stream *standard-output*))
  (cond (*pi-list*
	 (format stream "~%Process instances:")
	 (dolist (entry (reverse *pi-list*))
	   (format stream "~%~A = ~A"
		   (car entry) (adb::lprinc-string (caddr entry)))))
	(t (format stream "~%No process instances."))))

(defun show-vis (&optional (stream *standard-output*))
  (cond (*vi-list* 
	 (format stream "~%View instances:")
	 (dolist (entry (reverse *vi-list*))
	   (format stream "~%~A = ~A" 
		   (car entry)(adb::lprinc-string (caddr entry)))))
	(t (format stream "~%No view instances."))))
