;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;; Character-oriented interface for QPE.
;; Ugly, old fashioned, but boy it's portable...

;; Todo:
;; ******* integrate syntax checker?
;; ******* display commands? <-- Yes -- all the "show" stuff
;; ******* Inspector modes?

(proclaim '(special *cmenu-prompt* *qpe-command-menu*))

(defun qpe ()
  ;; Toplevel menu
  (let ((*cmenu-prompt* (format nil "QPE v~A>" *version-number*)))
    (run-command-menu *qpe-command-menu* "QPE Commands")))

(defvar *qpe-command-menu*
'(("Reset" (adb::init) :documentation "Clears entire environment.")
  ("ATMoSphere" (ATmoSphere-menu-options) :documentation "Clearing caches, low-level stuff.")
  ("Parameters" (select-qpe-parameters) :documentation "Sets up parameters for envisionment.")
  ("Modes" (select-qpe-modes) :documentation "Switches which affect QPE's operation.")
  ("Flags" (select-qpe-flags) :documentation "Controls amount of information displayed.")    
  ("Envision" (run-qpe) :documentation "Performs an envisionment assuming current parameters.")
  ("Show" (show-qpe-results) :documentation "Shows various aspects of the envisionment and model.")
  ("Report" (dump-qpe) :documentation "Saves QPE results to disk in various forms.")))

;; This is the easiest:

(defun run-qpe ()
  (start-qpe) ;; Initialize ATMoSphere and flags.
  (multiple-value-bind (okay? reason)
    (load-domain)
    (unless okay?
      (format t "~% Failed -- ~A" reason)
      (return-from run-qpe nil)))
  (multiple-value-bind (okay? reason)
   (load-scenario)
   (unless okay?
	   (format t "~% Failed -- ~A" reason)
	   (return-from run-qpe nil)))
  (multiple-value-bind (okay? reason)
   (envision)
    (unless okay?
      (format t "~% Failed -- ~A" reason)
      (return-from run-qpe nil))))

;;;; Selecting parameters

;; Make this a list for user extendability

(defvar *qpe-parameters*
'(("Domain files" *domain-files* :list-of-strings)
  ("Scenario files" *scenario-files* :list-of-strings)))

(defun select-qpe-parameters ()
  (let ((old-domain-files *domain-files*)
	(old-scenario-files *scenario-files*))
  (cmenu-choose-values *qpe-parameters*
		       "Choose domain model and scenario")
  ;; See if reloading is necessary.
  (unless (eq old-domain-files *domain-files*)
    (adb::clear-packet :QPE-domain-model))
  (unless (eq old-scenario-files *scenario-files*)
    (adb::clear-packet :QPE-scenario-model))))

;; The mode information has the format 
;; (<var> <key> <if on> <if off>)
;; which we massage into cmenu format.

(defun select-qpe-modes ()
  (let ((item-list
	  (mapcar #'(lambda (entry)
		      (list (second entry) (first entry)
			    :one-of '(("ON" t) ("OFF" nil))
			    :documentation
			    (format nil "T = ~A; NIL = ~A."
				    (third entry) (fourth entry))))
		  *qpe-modes*)))
    (cmenu-choose-values item-list "Current QPE modes"
			 "These switches can affect the results and output")))

;; The debugging flags have a similar format:
;; (<flag> <key> <var>) where <flag> is what is actually stored
;; in the debug table, <key> is the pretty version, and <var>
;; is a gensym indicating its desired state.

(defun select-qpe-flags ()
  (let ((item-list (mapcar #'(lambda (entry)
			       (cond ((and (boundp 'adb::*debug-table*)
					   (member (car entry) adb::*debug-table*))
				      (set (third entry) t))
				     (t (set (third entry) nil)))
			       (list (second entry) (third entry)
				     :one-of '(("ON" t) ("OFF" nil))
				     (car entry)))
			   adb::*debug-flag-registry*)))
    (cmenu-choose-values item-list "Debugging flags"
       "These control what information is printed during operation.")
    (dolist (entry item-list)
	    (cond ((symbol-value (second entry)) ;; set flag on
		   (adb::set-debug-flag (fifth entry) T))
		  (t (adb::set-debug-flag (fifth entry) nil)))
	    (makunbound (second entry)))))

(defvar *dump-file* "Your file name here")
(defvar *dump-comments* "Your remarks here")

(defun dump-qpe ()
  (unless *analysis-complete*
	  (format t "~% The envisionment hasn't been generated yet.")
	  (return-from dump-qpe nil))
  (do ((ready? nil)) (nil)
    (setq ready? (cmenu-choose '(("Change parameters" :FIDDLE)
				 ("Generate report" :REPORT))))
    (case ready?
      (:PUNT (return-from dump-qpe t))
      (:FIDDLE
       (cmenu-choose-values '(("Formatting" *report-mode* 
			       :one-of (("LaTeX" :LATEX)("ascii" :ASCII))
			       :documentation
             "LaTeX prettiest, ascii directly usable.")
			      ("Destination file" *dump-file* :string
			       :documentation "Must be a legitimate file name.")
			      ("Comments" *dump-comments* :string
			       :documentation "One line of remarks.")
			      ("Include situations?" *dump-situations*
			       :one-of (("Yes" t)("No" nil))
			       :documentation
			       "Useful for getting summaries")
			      ("Include conjunctive limit hypotheses?" *dump-conjunctive-lhs*
			       :one-of (("Yes" t) ("No" nil))
			       :documentation
			       "Good for brief report.")
			      ("Include details of limit hypotheses?" *dump-lh-details*
			       :one-of (("Yes" t) ("No" nil))
			       :documentation
			       "Good for brief report."))))
      (:REPORT 
       (dump-envisioner-results *dump-comments* *dump-file*))
      (t (return-from dump-qpe t)))))


(defun show-qpe-results ()
  ;; Don't check results flag, since some are useful before envisioner is run
  (do ((status nil)) (nil) 
    (setq status (cmenu-choose '(("Quantities" (show-quantities))
				 ("Views" (show-vis))
				 ("Processes"  (show-pis))
				 ("Processes/views" (show-cdis))
				 ("Choice sets" (show-choice-sets))
				 ("Quantity Space information" (show-qstate))
				 ("Direct Influences" (show-possible-directs)) 
				 ("Indirect Influences" (show-possible-indirects)) 
				 ("Summarize PS/VS" (tabulate-psvs))
				 ("Limit Hypotheses" (show-lhs))
				 ("State transitions" (show-transitions))
				 ("Flight Recorder" (fr-event-breakdown))
				 ("FR details" (cmenu-show-fr-events)))))
    (cond ((or (eq status :PUNT) (null status)) (return-from show-qpe-results t))
	  (t (eval status)))))

(defun cmenu-show-fr-events (&aux what)
  (setq what (cmenu-choose (mapcar #'(lambda (etype)
				       (list (format nil "~A" etype) etype))
				   (types-of-events))))
  (when what
    (show-fr-events what)))

;;;; ATMoSphere submenu

(defun ATMoSphere-menu-options ()
  (let ((*cmenu-prompt* (format nil "ATMoSphere v~A>" adb::*version-number*)))
    (run-command-menu *ATMoSphere-command-menu* *cmenu-prompt*)))

(defvar *ATMoSphere-command-menu*
'(("Reset" (adb::init) :documentation "Clears entire environment.")
  ("Summarize packet contents"
   (adb::show-packets) :documentation "Describes contents of cached information.")
  ("Clear scenario model" (adb::clear-packet :QPE-Scenario-Model)
   :documentation "Clears cached scenario model.")
  ("Clear domain model" (adb::clear-packet :QPE-Domain-Model)
   :documentation "Clears cached domain model.")
  ("Clear QPE rules" (adb::clear-packet :QPE-rules)
   :documentation "Clears internal QPE rules.")))