;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Installing correspondences
;;
;; For speed, unwind the justifications as fully as possible.
;;  Fixed correspondences are unwound in the syntaxer.
;;  Special run-time correspondences are found by the pattern-directed system.
;;  An example of a construct which requires a run-time correspondence is
;;  FUNCTION-SPEC

(defun install-runtime-correspondence (cor)
  (let ((pairs (check-correspondence-pairs (cddr cor))))
    (unless pairs
      (error "~%Bad run-time correspondence:~%~A." cor))
    (dolist (pair pairs)
      (install-derived-qstate-pair (car pair) (cadr pair)))
    (dolist (j (make-correspondence-justifications (cadr cor)
						   (cddr cor)))
      (adb::justify (truthify (car j)) (cons (cons cor :TRUE) 
						 (mapcar #'truthify (cdr j)))))))

(defun truthify (thing)
  (if (consp thing)
      (if (listp (cdr thing))
	  (cons thing ':TRUE) ;; Make it true
	  thing) ;; Otherwise assume it is okay
      (cons thing ':TRUE))) ;; Atomic proposition case

;; Error checking
(defun check-correspondence-pairs (pairs)   ;; error checking
  (check-correspondence-pairs1 pairs nil))

(defun check-correspondence-pairs1 (pairs result)
  (cond ((null pairs) result)
	((or (not (listp pairs))
	     (not (listp (car pairs)))
	     (and (not (listp (caar pairs)))
		  (not (symbolp (caar pairs))))
	     (and (not (listp (cadar pairs)))
		  (not (symbolp (cadar pairs))))) NIL)
	(t (check-correspondence-pairs1 (cdr pairs) (cons (car pairs) result)))))

;;;;; Top level correspondence justification construction
;;; John Collins developed the reverse cases and alot of the cases left out in GIZMO.
;;; Experimental mode: only include qprop's whose sign is relevant to the change.

(defvar *all-qprop-correspondences* T)
;; When T, the justification for each case of the correspondence
;; includes qprop's for each quantity involved.  When NIL, only
;; qprop's whose sign is needed to establish the new relationship
;; are mentioned explicitly in the justification.  This mode switch
;; is an experiment; if correspondences are declared "correctly", 
;; they will never be true in an environment where not every qprop
;; is around.  Not mentioning them is more efficient because the
;; pairs whose values equal the keyed value don't have to spin off
;; justifications for all combinations of qprop- and qprop.  BUT,
;; if the correspondence can be true when not all of the qprops
;; are, expect some subtle bugs!
;;
;; Maybe there should be a post-test that can be applied as a
;; diagnostic to a situation, testing to see if this is true for
;; each correspondence.

(defmacro strip-off-a-if-need-be (form) ;; Was defsubst
  `(let ((res ,form))
    (cond ((and (listp res) (member (car res) '(A D))) (cadr res))
	  (t res))))

(defun make-correspondence-justifications (result pairs)
  ;; There are five kinds of justifications that must be built.
  ;; 1. When all the input value pairs are equal, the result pair is equal.
  ;; 2. When the result pair and all but one of the input pairs are equal,
  ;;    then the other input pair is equal.
  ;; 3. When all but one of the input pairs are equal, the result pair has an inequality
  ;;    appropriate for the sign of the Qprop and the inequality of the other input pair.
  ;; 4. When all but one of the input pairs are equal, the other input pair has an inequality
  ;;    appropriate for the sign of the Qprop and the inequality of the result pair.
  ;; 5. When the result pair and all but two of the input pairs are equal,
  ;;    then the other two input pairs have the same or opposite inequalities,
  ;;    depending on the signs of their Qprops.
  ;;
  ;;  What is returned is a list of justifications whose CAR is the consequence
  ;;  and whose CDR is the list of antecedents
  (let ((results nil)
	(all-precursors (make-correspondence-precursors result pairs)))
    (if *all-qprop-correspondences*
	(setq results
	      (mapcar #'(lambda (precursors)
			  `(,(find-qrel-assertion (car result) (cadr result) '=)
			    ,@ (mapcar #'(lambda (pair)
					   (find-qrel-assertion (car pair) (cadr pair) '=))
				       pairs)
			    ,@ precursors))
		      all-precursors))
	(setq results
	      (list `(,(find-qrel-assertion (car result) (cadr result) '=)
		      ,@ (mapcar #'(lambda (pair)
				     (find-qrel-assertion (car pair) (cadr pair) '=))
				 pairs)))))
    ;;; Here we install the <-
    ;;; Must do this for each "input" in order to make transitivity work.
    (dolist (pair pairs)
      (cond (*all-qprop-correspondences*
	     (dolist (precursor all-precursors)
	       (push (equality-correspondence-justification
		       pair result (remove pair pairs) precursor)
		     results)))
	    (t (push (equality-correspondence-justification
		       pair result (remove pair pairs) nil)
		     results))))
    ;;; here is where the inequality justifications are created. 
    (dolist (pair pairs)
      (let ((rest (remove pair pairs)))
	(dolist (flip? '(T nil))
	  (dolist (invert? '(T nil))
	    (setq results
		  (nconc (make-correspondence-inequality-justifications
			   result pair rest flip? invert?)
			 (if rest
			     (make-other-correspondence-inequality-justifications
			       result pair rest flip? invert?))
			 results))))))
    results))

;;;;; Making inequality justifications

(defun make-correspondence-inequality-justifications
       (result pair rest &optional (flip? nil) (invert? NIL) &aux precursors)
  ;; returns lists of the form (<conclusion> . <antecedents>)
  (setq precursors 
	(when *all-qprop-correspondences*
	  (make-correspondence-precursors result rest)))
  (cond (precursors
	 (mapcan #'(lambda (precursors)
		     (list (make-correspondence-inequality-justification
			     result pair rest flip? invert? precursors)
			   (make-reverse-correspondence-inequality-justification
				   result pair rest flip? invert? precursors)))
		     precursors))
	(t (list (make-correspondence-inequality-justification
		   result pair rest flip? invert? nil)
		 (make-reverse-correspondence-inequality-justification
		   result pair rest flip? invert? nil)))))

(defun make-correspondence-inequality-justification 
       (result pair rest flip? invert? precursors)
  `(,(find-qrel-assertion (car result) (cadr result)
			  (if (eq invert? flip?)
			      'Greater-Than 'Less-Than))
    ((,(cond (invert? 'Qprop-) (t 'Qprop))
      ,(strip-off-a-if-need-be (car result))
      ,(strip-off-a-if-need-be (car pair))) . :TRUE)	;particular Qprop
    ,@ (if *all-qprop-correspondences* ;EXPERIMENTAL MODE -- NIL
	   precursors) ;; other qprops
    ,(find-qrel-assertion (car pair) (cadr pair) 
			  (if flip? 'Less-Than 'Greater-than))
    ,@ (make-correspondence-equality-assertions rest)))

(defun make-reverse-correspondence-inequality-justification
       (result pair rest flip? invert? precursors)
  ;; Adds the missing rules to propagate inequalities backwards.
  `(,(find-qrel-assertion (car pair) (cadr pair)
			  (if (eq invert? flip?)
			      'Greater-Than 'Less-Than))
		((,(if invert? 'Qprop- 'Qprop)
		  ,(strip-off-a-if-need-be (car result))
		     ,(strip-off-a-if-need-be (car pair))) . :TRUE) ;particular Qprop
		 ,@ (if *all-qprop-correspondences* ;EXPERIMENTAL MODE -- NIL
			precursors) ;; other qprops
		 ,(find-qrel-assertion (car result) (cadr result) 
				       (if flip? 'Less-Than 'Greater-than))
		 ,@ (make-correspondence-equality-assertions rest)))

(defun make-other-correspondence-inequality-justifications
       (result pair1 rest flip? invert? &aux rest-of-rest results precursors)
  ;; Relates two input pairs when the result pair is equal.
  (setf results nil)
  (dolist (pair2 rest)
    (setf rest-of-rest (remove pair2 rest))
    (setf precursors (if *all-qprop-correspondences*
			 (make-correspondence-precursors result rest-of-rest)))
    (dolist (switch? '(t nil))
      (push (make-one-more-correspondence-inequality-justification
	      result pair1 pair2 rest-of-rest flip? invert? switch? precursors)
	    results)))
  results)

(defun make-one-more-correspondence-inequality-justification
       (result pair1 pair2 rest flip? invert? switch? precursors)
  `(,(find-qrel-assertion (car pair1) (cadr pair1)
			  (if (xor invert? flip? switch?)
			      'Greater-than 'Less-Than))
		((,(if invert? 'Qprop- 'Qprop)
		  ,(strip-off-a-if-need-be (car result))
		  ,(strip-off-a-if-need-be (car pair1))) . :TRUE)
                ((,(if switch? 'Qprop- 'Qprop)
		  ,(strip-off-a-if-need-be (car result))
		  ,(strip-off-a-if-need-be (car pair2))) . :TRUE)
		 ,@ precursors ;; other qprops
		 ,(find-qrel-assertion (car pair2) (cadr pair2) 
				       (if flip? 'Less-Than 'Greater-than))
		 ,@ (make-correspondence-equality-assertions (cons result rest))))

(defun xor (&rest booleans)
  (oddp (length (remove nil booleans))))

;;;; Equality case
(defun equality-correspondence-justification (result vary other-pairs precursors)
 `(,(find-qrel-assertion (car result) (cadr result) '=)
   ,(find-qrel-assertion (car vary) (cadr vary) '=)
   ,@ (mapcar #'(lambda (p)
		  (find-qrel-assertion (car p) (cadr p) '=))
	      other-pairs)
   ,@ precursors))

(defun make-correspondence-precursors (start pairs)
 (make-correspondence-precursors1 (strip-off-a-if-need-be
					       (car start)) pairs))

(defun make-correspondence-precursors1 (res pairs)
  ;; Returns combinatorial list of qprop sets.
  (cond ((null pairs) nil)
	((null (cdr pairs))
	 (list (list `((Qprop ,res ,(strip-off-a-if-need-be (caar pairs)))
		       . :TRUE))
	       (list `((Qprop- ,res ,(strip-off-a-if-need-be (caar pairs)))
		       . :TRUE))))
	(t (mapcan #'(lambda (other-qprops)
		       (list (cons `((Qprop ,res ,(strip-off-a-if-need-be (caar pairs)))
				     . :TRUE)
				   other-qprops)
			     (cons `((Qprop- ,res ,(strip-off-a-if-need-be (caar pairs)))
				     . :TRUE)
				   other-qprops)))
		   (make-correspondence-precursors1 res (cdr pairs))))))

(defun make-correspondence-equality-assertions (pairs)
  (mapcar #'(lambda (pair) 
	      (find-qrel-assertion (car pair) (cadr pair) '=)) pairs))

;;;; Debugging

(defun print-correspondence-assertions (c-statement)
  (dolist (as (mapcar #'(lambda (r)
			 `(-> (And (,c-statement . :TRUE) ,@ (cdr r)) ,(car r)))
		      (make-correspondence-justifications (cadr c-statement)
							  (cddr c-statement))))
    (adb::lprint as)))

;;;; Ordered correspondences (jcollins)

(defun make-ordered-correspondence-results (pair1 pair2 &aux results)
  (dolist (rel '(> = <) results)
    (push (make-ordered-correspondence-result pair1 pair2 rel) results)
    (push (make-ordered-correspondence-result pair2 pair1 rel) results)))
      
(defun make-ordered-correspondence-result (pair1 pair2 rel)
  `(,(make-quoted-qrel-assertion (car pair1) (cadr pair1) rel)
    (,(make-quoted-qrel-assertion (car pair2) (cadr pair2) rel)
     ,@ **antecedents**)
    :ORDERED-CORRESPONDENCE))

(defun make-quoted-qrel-assertion (N1 N2 rel)
  (cond ((or (eq N1 'ZERO)
	     (eq N2 'ZERO))
	 (make-qrel-assertion N1 N2 rel))
	(t `(:eval (make-qrel-assertion ,(quotize N1) ,(quotize N2) ',rel)))))

(defun make-ordered-correspondence-justifications (pair1 pair2)
  (dolist (result (make-ordered-correspondence-results pair1 pair2))
    (eval `(rjustify ,(car result) ,(cadr result) :ORDERED-CORRESPONDENCE))))