;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;; More debugging stuff.

(defun ineq-prospects (ineq)
  (list (tms::node-status (ineq-< ineq))
	(tms::node-status (ineq-> ineq))
	(tms::node-status (ineq-= ineq))
	(tms::node-status (ineq-bt ineq))))

(defun check-ineq-state-size ()
  (let ((ineqs (gather-assumed-qstate))
	(total 1)
	(current 0))
    (dolist (ineq ineqs (format t "~%Worst case = ~D situations." total))
      (setq current (length (delete-if #'(lambda (entry);; what about falses?
					   (eq 'tms::out entry))
				       (ineq-prospects ineq))))
      (setq total (* total current))
      (format t "~% ~D for ~A" current ineq))))

;;; The important information to summarize a choice set is
;;; (<name of choice set> <total possible> <actually possible>)
;;; 

(defun summarize-ineq-choice-sets (&aux entry results prospects)
  (dolist (ineq (gather-assumed-qstate) results)
    (setq prospects (ineq-prospects ineq))
    (push (setq entry (list ineq (length prospects) 0))
	  results)
    (dolist (prospect prospects)
      (cond ((eq prospect 'tms::out))
	    (t (incf (caddr entry)))))))

(defun summarize-other-choice-sets (&aux rentry results)
  (dolist (entry *choice-sets* results)
    ;; format of choice sets is (<variable> . <values>)
    (push (setq rentry (list (lprinc-string (car entry)) (length (cdr entry)) 0))
	  results)
    (dolist (value (cdr entry))
      (cond ((adb::true? value) (setf (caddr rentry) 1) (return))
	    ((adb::false? value))
	    (t (incf (caddr rentry) 1))))))

(defun say-choice-sets-summary (header summary &optional (stream *standard-output*)
				&aux worst far-worst)
  (format stream "~%For ~A: ~D choice sets total." header (length summary))
  (setq worst 1 far-worst 1)
  (dolist (entry summary)
    (format stream "~%  ~D out of ~D for ~A." (third entry) (second entry) (car entry))
    (setq worst (* worst (third entry))
	  far-worst (* far-worst (second entry))))
  (format stream "~%     Contribution to worst case = ~D (out of ~D)"
	  worst far-worst)
  (values worst far-worst))

(defun check-psvs-size (&optional (stream *standard-output*))
  ;;; This combines the results of choice sets and inequalities
  ;;; to provide a worst-case estimate of state size
  (multiple-value-bind (iworst ifworst)
      (say-choice-sets-summary "inequality choices" 
			       (summarize-ineq-choice-sets) stream)
    (multiple-value-bind (oworst ofworst)
	(say-choice-sets-summary "other choices" 
				 (summarize-other-choice-sets) stream)
      (format stream "~%Total worst case = ~D (out of ~D)."
	      (* iworst oworst) (* ifworst ofworst))
      (values (* iworst oworst) (* ifworst ofworst)))))

;;;; Hairy debugging stuff for influence resolution

(proclaim '(special *debug-influence-tables* *debug-sclass*))

(defun test-ambiguous-influence-code (sc unknowns)
  unknowns
  (when (cdr (sclass-assumptions sc))
  ;; Check to see if the order tables are different 
    (let ((tables nil))
      (dolist (sit (sclass-assumptions sc))
	(push (multiple-value-list (compute-situation-influence-table sit))
	      tables))
      (unless (null (cdr (remove-duplicates tables :test #'equal)))
	(format t "~% Difference found in influences for situations under ~A:"
		sc)
	(setq *Debug-Influence-tables* (nreverse tables))
	(setq *debug-sclass* sc)
	(break "Bug in ambigous influence code")))))

;;;; Yet more debugging utilities for influence resolution

(defun check-resolutions (sit &optional (stream *terminal-io*))
  (let ((envs (sclass-assumptions sit)))
    ;; ****** Need to find a faster way to do this!
    ;; ****** Put a uniquifier on quantity structs
    (dolist (q (sort (copy-list (gather-quantities)) #'adb::alphalessp))
      (do ((nodes (tms::class-nodes
		    (let ((this `(s (D ,q))))
		    (adb::get-class-tms-node (adb::lookup-class 
			    (adb::get-assertion-class this) t) this)))
		  (cdr nodes))
	   (got-it? nil))
	  ((or (null nodes) got-it?)
	   (unless got-it?
	     (terpri stream)
	     (adb::lprinc (cons `(s (D ,q)) '???) stream)))
	(when (tms::union-envss envs (tms::label (car nodes)))
	  (terpri stream)
	  (adb::lprinc (adb::view-node (car nodes)) stream)
	  (setq got-it? t))))))

(defun check-IR (&optional (break? nil))
  (dolist (s *ps-vs*)
    (terpri)
    (print-sclass s)
    (if break? (read-char))
    (check-resolutions s)))

;;;; Some stuff for the true paranoid.

(defun check-ps-vs-consistency (&aux winners losers)
  (map-sclasses #'(lambda (sit) 
		    (if (some #'(lambda (x) (tms::env-contradictory x))
			       (sclass-assumptions sit))
			(push sit losers)
			(push sit winners))))
     (values winners losers))

(defun check-ineqs-in (e)
  (adb::print-status1 '((greater-than . ?x) . :TRUE) (list ':IN e))
  (adb::print-status1 '((Equal-to . ?x) . :TRUE) (list ':IN e))
  (adb::print-status1 '((less-than . ?x) . :TRUE) (list ':IN e))
  e)

(defun check-lh-integrity (lh)
  (dolist (start (lh-end-envs lh))
    (if (tms::env-contradictory start)
	(format t "~%   ~A has contradictory end env (~A)."
		lh start)))
  (dolist (start (lh-start-envs lh))
    (if (tms::env-contradictory start)
	(format t "~%   ~A has contradictory start env (~A)."
		lh start))))

(defun check-lhs () (dolist (lh *limit-hypotheses*) (check-lh-integrity lh)))

;; More debugging stuff

(defun verify-sclass-integrity (sclass)
  ;; For debugging.  Tests to see if all the environments in
  ;; an Sclass are consistent with that Sclass
  (let ((sits (sclass-assumptions sclass)))
    ;; Check consistency of environments
    (dolist (sit sits)
      (if (tms::env-contradictory sit)
	  (format t "~%  Situation ~A of ~A is contradictory." sit sclass)))
    ;; Check existence
    (let ((sci (sclass-individuals sclass))
	  (all (gather-individuals)))
      (dolist (sit sits)
	(let ((actuals (remove-if-not #'(lambda (i) (adb::implied-by? i sit)) all)))
	  (unless (same-elements? actuals sci)
	    (format t "~%Individuals in ~A out of synch with ~A." sit sclass)
	    (let ((losers (set-difference actuals sci)))
	      (when losers
		(format t "~%   ~A exist in ~A but not in ~A." losers sit sclass))
	      (when (setq losers (set-difference sci actuals))
		(format t
		   "~%   ~A exist in ~A but not in ~A." losers sclass sit)))))))
    ;;***** FINISH THIS OFF!
    ;; Check VS
    ;; Check PS
    ;; Check Ds values
))

(defun verify-sitenv-integrity (&optional (stream *standard-output*)
					  &aux (nenvs 0))
  (dolist (env *ps-vs-envs* nenvs)
    (when (tms::env-contradictory env)
      (format stream "~%;;;   Situation ~A has become contradictory." env))
    (incf nenvs)))


;;;; Checking for syntaxer bugs

(proclaim '(special *loser*))

(defun scout-out-syntaxer-bugs ()
  ;; first test -- are backquotes around?
  #+Symbolics
  (when (setq *loser* (adb::lookup-class 'si:xr-bq-list))
    (format t "~% Something probably isn't being evaled right -- SI:XR-BQ-LIST is a class")
    (break "Syntax scouter"))
  ;; Next, look for bugs in individual classes.
  (dolist (class adb::*class*)
    (let ((l-val nil)
	  (s-val nil))
      (dolist (as (class-instances class))
	(when (listp (adb::lisp-form as))
	  (cond ((listp (cdr (adb::lisp-form as))) (setq l-val t))
		(t (setq s-val t)))))      
      (when (and l-val s-val)
	(format t "~% Class ~A has mixed values in assertions."
		(class-name class))))))

;;;; Some inequality-related stuff

(defun check-ineqs (&optional (stream *standard-output*) &aux constant varying)
  (dolist (ineq *ineqs*)
    (if (constant-ineq? ineq) (push ineq constant)
	(push ineq varying)))
  (cond (constant 
	 (format stream "~%  ~D inequality relations are constant." (length constant))
	 (dolist (con constant)
	   (format stream "~%    ~A" con)))
	(t (format stream "~% No inequality relations are constant.")))
  (cond (varying 
	 (format stream "~%  ~D inequality relations can vary." (length varying)))
	(t (format stream "~% No inequality relations can vary.")))
  (when (and constant varying)
    (format stream "~% ~D inequality relations total." (+ (length constant)
							  (length varying)))))

;;;; Influence resolution debugging

(defun current-ds-values (env &optional (break? nil)
			      &aux elist val)
  (setq elist (list env))
  (dolist (q (gather-quantities))
    (setq val (get-ds-value q elist))
    (unless val (setq val "??"))
    (format t "~% ~A" (adb::lprinc-string `((s (d ,q)) . ,val)))
    (if break? (read-char))))

(defun direct-influences-on (q env)
  (let ((entry (assoc q *dp-influencers* :test #'equal)))
    (cond ((null entry)
	   (format t "~% ~A never directly influenced." (adb::lprinc-string q)))
	  (t (format t "~% For ~A:" (adb::lprinc-string q))
	     (dolist (subentry (cdr entry))
	       (when (every #'(lambda (ante) (implied-by? ante env)) (car subentry))
		 (dolist (ispec (cdr subentry))
		   (format t "~%  ~A" (adb::lprinc-string ispec)))))))))

(defun indirect-influences-on (q env)
  (let ((entry (assoc q *Ip-influencers* :test #'equal)))
    (cond ((null entry)
	   (format t "~% ~A never indirectly influenced." (adb::lprinc-string q)))
	  (t (format t "~% For ~A:" (adb::lprinc-string q))
	     (dolist (subentry (cdr entry))
	       (when (every #'(lambda (ante) (implied-by? ante env)) (car subentry))
		 (dolist (ispec (cdr subentry))
		   (format t "~%  ~A" (adb::lprinc-string ispec)))))))))


(defun influences-to (q val)
  (adb::why `((,(case val (0 'Equal-to) (-1 'Less-Than) (1 'Greater-than))
			  (D ,q) ZERO) . :TRUE)))

(defun count-ineqs ()
  (let ((sum 0)
	(assumed 0))
    (dolist (ineq *ineqs*)
      (incf sum)
      (if (ineq-assumed? ineq) (incf assumed)))
    (values sum assumed)))

;;;; Sniffing out package problems

;; One symptom of such problems is the existence
;; of two ADB classes which are the same name but
;; in different packages.  The program below finds them.

(defun check-for-overlapping-classes ()
  (let ((class-list nil)
	(key nil)
	(entry nil))
    (dolist (class (adb::list-classes))
      (setq key (format nil "~A" (adb::class-name class)))
      (setq entry (assoc key class-list :test #'string=))
      (cond (entry ;; got one!
		   (format t "~% Duplicate classes for ~A!" key)
		   (push class (cdr entry)))
	    (t (push (setq entry (list key class)) class-list))))
    (remove-if-not #'cddr class-list)))

;;; Putting things in the user package for debugging

(defun usify-qpe ()
  (eval '(defun user::print-sclass (x) (qpe::print-sclass x)))
  (eval '(defun user::print-lh (x) (qpe::print-lh x)))
  (eval '(defun user::show-cdis (x) (qpe::show-cdis x)))
  (eval '(defun user::show-pis (x) (qpe::show-pis x)))
  (eval '(defun user::show-qstate (x) (qpe::show-qstate x)))
  (eval '(defun user::print-env (x) (qpe::print-env x)))
  (eval '(defun user::e (x) (qpe::e x))))

;;;; Showing transitions at the sclass level

(defun show-transitions (&optional (transition-table *next-state-table*))
  (pprint (desituate-transition-table transition-table)))

(defun desituate-transition-table (ttable)
  (mapcar #'(lambda (entry)
	      (cons (env-sclass (car entry))
		    (mapcar #'(lambda (subentry)
				(cons (car subentry)
				      (env-sclass (cdr subentry)))) (cdr entry)))) ttable))


(proclaim '(special *lhs-used*))

(adb::Init-form (setq *lhs-used* nil))

(defun find-lhs-used (&optional (table *next-state-table*) &aux marker)
  ;; grovels over a transition table and ascertains which limit hypotheses
  ;; actually result in valid transitions
  (setq *lhs-used* nil
	marker (cons nil nil))
  (dolist (entry table *lhs-used*)
    (dolist (pairs (cdr entry))
      (unless (eq (get-lh-prop (car pairs) :lh-used?) marker)
	(push (car pairs) *lhs-used*)
	(put-lh-prop (car pairs) marker :lh-used?)))))

;;;; Showing results of temporal projection

(defun show-transition (before lh after &optional (stream *standard-output*))
  (format t "~% From ~A to ~A via ~A:~%" before after lh)
  (princ-lh lh nil stream t)
  (show-before-after before after stream))

(defun show-before-after (before after &optional (stream *standard-output*))
  ;; Maps down the before assumptions, maps down the corresponding ones
  ;; on the other side.
  (labels ((get-datum-generic (asn) 
	     (let ((form (car (view-assumption asn))))
	       (cond ((inequality-form? form) (values (cdr form) t))
		     (t (values form nil)))))
	   (interpret-datum-value (asn)
	     (let ((form (if (tms::assumption? asn) (view-assumption asn) nil)))
	       (cond ((or (null form) (not (listp form))) "?")
		     ((inequality-form? (car form))
		      (get-inequality-relation-name (caar form)))
		     (t (case (cdr form)
			  (:TRUE "T")
			  (:FALSE "F")
			  (-1 "-")
			  (1 "+")
			  (0 "0")
			  (t "?")))))))
    (let ((old (tms::env-assumptions before))
	  (new (tms::env-assumptions after))
	  (used nil))
      ;; These can be out of order since one may contain an assumption the other doesn't.
      ;; Put the relationships first since those are briefer.
      (dolist (asnO old)
;;	(format t "~% ~A: ~A ~A" asnO (get-datum-generic asnO) (interpret-datum-value asnO))
	(multiple-value-bind (old-form ineq?)
	    (get-datum-generic asnO)
	  (let ((asnN (find old-form new :key #'get-datum-generic 
			  :test #'equal)))
	  (format stream "~% ~A ~A ~A" 
		  (interpret-datum-value asnO) (interpret-datum-value asnN)
		  (if ineq? (concatenate 'string
					 (qnum-string (fetch-number (car old-form)))
					 "," (qnum-string (fetch-number (cadr old-form))))
		      (lprinc-string old-form)))
	  (push asnN used))))
      ;; Now show the new assumptions which have nothing corresponding in old.
      (dolist (asnN new)
	(unless (member asnN used)
	  (multiple-value-bind (new-form ineq?)
	      (get-datum-generic asnN)
	  (format stream "~% ? ~A ~A"
		  (interpret-datum-value asnN)
		  (if ineq? (concatenate 'string
					 (qnum-string (fetch-number (car new-form)))
					 "," (qnum-string (fetch-number (cadr new-form))))
		      (lprinc-string new-form)))))))))

;;;; Figuring out why transitions got ruled out
;; This is best used as a subroutine -- see CHECK-TRANSITIONS-FOR below.

(defun why-not-transition? (before lh after &optional (stream *standard-output*)
				   &aux entry subentry dead?)
  ;; See if the lh is really implied by the before state
  (when (and (setq entry (assoc before *next-state-table*))
	     (some #'(lambda (subentry) (and (eq (car subentry) lh)
					     (eq (cdr subentry) after)))
		   (cdr entry)))
    (format stream "~% Transition from ~A to ~A via ~A is believed to occur."
	    before after lh)
    (return-from WHY-NOT-TRANSITION? :OKAY))
  (unless  (implies-limit-hypothesis? before lh)
    (format stream "~% ~A not applicable to before state ~A." lh before)
    (return-from WHY-NOT-TRANSITION? :BEFORE))
  (unless (implies-limit-hypothesis-result? after lh)
    (format stream "~% ~A not applicable to after state ~A." lh after)
    (return-from WHY-NOT-TRANSITION? :AFTER))
  ;; Assume limit analysis has been run, since we need the ineq partitioning.
  (when (setq dead? (tms::get-env-prop after :DEAD))
    ;; Bring result state back to life, at least temporarily
    (format stream "~%    [Potential result state ~A was dead.]" after)
    (tms::put-env-prop after (cadr dead?) 'Sclass))
  (multiple-value-bind (okay? reason datum)
      (continuity-check before after)
    (unless (or okay? (eq reason ':ZERO))
      (format stream "~% Continuity violated between ~A and ~A: ~A, ~A"
	      before after reason datum)
      (format stream "~%   Before: A's = ~A, D's = ~A; After: A's = ~A, D's = ~A"
	      (inequality-under datum before)
	      (inequality-under (ineq-other datum) before)
	      (inequality-under datum after)
	      (inequality-under (ineq-other datum) after))
      (if dead? (tms::put-env-prop after nil 'Sclass))
      (return-from WHY-NOT-TRANSITION? :CONTINUITY)))
  ;; Now we have to use the information accumulated for debugging,
  ;; since the rest of the checks are non-local.
  ;; First do a reality check and make sure this transition was once
  ;; considered.
  (setq entry (assoc before *contin-next-states*))
  (setq subentry (some #'(lambda (s) (and (eq (car s) lh) (eq (cdr s) after))) (cdr entry)))
  (unless subentry 
    (format stream 
	    "~% Serious bug: ~A to ~A via ~A is locally plausible but was not computed!!."
	    before after lh)
    (if dead? (tms::put-env-prop after nil 'Sclass))
    (return-from WHY-NOT-TRANSITION? :BUG))
  (setq subentry (assoc before *close-next-states*))
  (unless (and subentry (some #'(lambda (s) (and (eq (car s) lh) (eq (cdr s) after)))
			      (cdr subentry)))
    ;; Nailed by closeness, For now just comment on the fact
    (format stream
	    "~% ~A ruled out as result from ~A via ~A using closeness."
	    after before lh)
    (if dead? (tms::put-env-prop after nil 'Sclass))
    (return-from WHY-NOT-TRANSITION? :CLOSENESS))
  ;; Negation by failure -- it must have lost in ECL
  (format stream "~% Only thing left ruling out ~A from ~A via ~A is ECL."
	  after before lh)
  (if dead? (tms::put-env-prop after nil 'Sclass))
  (return-from WHY-NOT-TRANSITION? :ECL))

;;; More limit analysis debugging stuff.

(defun check-transitions-for (before &optional (stream *standard-output*))
  (let ((entry (assoc before *first-next-states*))
	(dead? (tms::get-env-prop before :DEAD)))
    ;; If dead, temporarily replace pointer back to sclass
    (when dead? (tms::put-env-prop before (cadr dead?) 'Sclass))
    (unless entry
      (setq entry (assoc before *situation-lh-table*))
      (cond (entry ;; had some, but no next state
	     (format stream "~% ~A suggested ~A, but no consistent next states found."
		     before (cdr entry)))
	    (t (format stream "~% No limit hypotheses locally suggested for ~A." before)))
      (return-from check-transitions-for :NO-CONSISTENT-LHS))
    (format stream "~%Possible transitions involving ~A include:" before)
    (pprint (cdr entry) stream)
    (dolist (next-pair (cdr entry))
      (why-not-transition? before (car next-pair) (cdr next-pair) stream))
    (when dead? (tms::put-env-prop before nil 'Sclass))
    before))

(defun find-dead-enders (&aux deaders lhs)
  (map-situations #'(lambda (sit)
		      (setq lhs (assoc sit *situation-lh-table*))
		      (when (cdr lhs)
			(unless (assoc sit *next-state-table*)
			  (push sit deaders)))))
  deaders)

;;;; Testing the basics.

(adb::store-in-packet :DM1
		 (defquantity-type Color Individual)
		 (defquantity-type Speed Individual)
		 (defentity physob
		   (quantity (color ?self))
		   (quantity (speed ?self)))
		 (defview (forward ?phob)
		   Individuals ((?phob :type physob))
		   Quantityconditions ((greater-than (A (speed ?phob)) zero))
		   Relations ((less-than (A (color ?phob)) zero))))

(adb::store-in-packet :SM1
		 (deffact (physob Foo)))

(defun test-simple ()
  (adb::clear-packet :QPE-domain-model)
  (adb::merge-packets :QPE-domain-model :DM1)
  (adb::clear-packet :QPE-scenario-model)
  (adb::merge-packets :QPE-scenario-model :SM1)
  (start-qpe)
  (load-domain)
  (load-scenario)
  (envision))