;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; QPE data abstractions and globals
;;
;; Many globals are local to one module, and are stored with it.
;; This file contains stuff that, if changed, should cause most of the system
;; to be re-compiled.

(defvar *qpe-modes* nil
  "QPE modes affect the actual operation of the program.")
(defvar *QPE-Program-Counter* nil
  "Information about state of computation for flight recorder.")

(defvar *new* nil) ;; For experimenting with ATMS clause system


;;;; Limit hypotheses

(defstruct (lh (:predicate limit-hypothesis?)
	       (:print-function (lambda (lh s ignore) (format s "~A" (lh-name lh)))))
  (name nil)			; backpointer to symbol
  (index 0)			; integer uniquifier, for sorting
  (relations nil)		; list of (<ineq> <old relation> <new relation>)
  (D-rels nil)	                 ;  Derivative relationships (atms nodes)
  (Equality-type nil)		; :FROM, :TO, or :MIXED
;;; ATMS aspects
  (old-assumptions nil)		; Assumptions of current state that will change
  (new-assumptions nil)		; What these assumptions become
  (start-envs nil)		; what it holds under
  (end-envs nil)		        ; what must be true when it is over
;;; Indexing properties
  (single-change nil)		; NIL -> conjunctive, single change o.w.
  (includes nil)		; what lh's it includes, if conjunctive
  (included-in nil)		; what lh's it is contained in
  (singles nil)			; Unit lh's participating in conjunctive LH
  (plist nil))			; Usual storage bin for sparse information 

(defun get-lh-prop (lh pname) (getf (lh-plist lh) pname))

(defun put-lh-prop (lh pval pname) (setf (getf (lh-plist lh) pname) pval))

(defun lh-next-relations (lh)
  (if (limit-hypothesis? lh)
      (mapcar #'(lambda (entry)
		  (list (caddr entry) (ineq-n1 (car entry))
			(ineq-n2 (car entry)))) (lh-relations lh))
      (error "~%~A not a limit hypothesis." lh)))

(defvar *lh-index* nil) ;; includes both lim-hyps and clim-hyps.
(defvar *lh-counter* 0) ;; Counter for uniquifing lh's (both single and conjunctive)

(adb::init-form (progn
		 (dolist (lh-pointer *lh-index*)
		   (makunbound lh-pointer)) ;prevent constipation
		 (setq *lh-index* nil)
		 ;; Zero out the counter whenever ADB is initialized.
		 (setq *lh-counter* 0)))

(defun generate-lh-name (base-name)
  ;; Generate unique name, but enforce consistent numbering scheme across runs
  (let ((lhname (intern (format nil "~A~D" base-name *lh-counter*))))
    (incf *lh-counter*)
    (push lhname *lh-index*)
    (values *lh-counter* lhname)))

(defun lh-sort-predicate (lh1 lh2) (< (lh-index lh1) (lh-index lh2)))

;;;; Sclass definition
;; An sclass represents a class of situations with identical PS, VS, and Ds values.
;; A situation is an sclass + a specific set of inequalities and preconditions from the
;; Qstate and the global set of preconditions.  There are usually several situations for
;; each Sclass,  so the Sclass structure is a convenient abstraction for manipulating and
;; understanding envisionments.
 
(defvar *sclass-index* nil) ;; Index for instances of sclasses
(defvar *sclass-counter* 0) ;; Counter for uniquifing sclasses
(defvar *sclass-table* nil) ;; Table indexed by individuals, PS, and VS.

(adb::init-form
  (progn (dolist (sclass-pointer *sclass-index*)
	   (makunbound sclass-pointer)) ;prevent constipation 
	 (setq *sclass-index* nil) ;clear index
	 (setq *sclass-table* nil)
	 (setq *sclass-counter* 0))) ;; Zero out the counter whenever ADB is initialized

(defun generate-sclass-name ()
  (let ((sit (intern (format nil "S~D" *sclass-counter*))))
    (incf *sclass-counter*)
    (push sit *sclass-index*)
    sit))
;;;; ****** Generating indexing functions via macrology?

(defstruct (sclass (:predicate sclass?)
		   (:print-function (lambda (sclass s ignore)
				      (format s "~A" (sclass-name sclass)))))
  (name nil)			; Backpointer to symbol
  (assumptions nil)		; a list of ENVs corresponding to situations
  (status 'EMPTY)		;EMPTY/PS-VS-COMPLETE/R-COMPLETE
  (plist nil)		; Property list for markers and other information
;;; Properties of this Sclass
  (individuals nil)		;individuals which exist in it
  (view-instances nil)		;view instances that exist in it
  (view-structure nil)		;active view instances
  (process-instances nil)	;possible processes
  (process-structure nil)	;processes which are active
  (Ds-values nil)		;results of influence resolution
;;; Relationships between sclasses
  (Unknown-Ds-values nil)	; Ds values which are unspecified due to ambiguity
  (R-completions nil)		; Sclasses which assign values to ambiguous Ds'
  (R-completion-of nil)		; Backpointer to Sclass with ambiguity
;;; Transition information
  (Limit-hypotheses nil)	; list of applicable limit hypotheses
  (next nil)			; Alist of situation, lh, and next state
  (previous nil)		; Alist of situation, lh, and previous state
  (duration nil)		; INTERVAL, INSTANT, or NIL
  )

(defun put-sclass-prop (sclass pval pname)
  (setf (getf (sclass-plist sclass) pname) pval))

(defun get-sclass-prop (sclass pname)
  (getf (sclass-plist sclass) pname))

(defun env-sclass? (e &aux sclass)
  ;;Returns sclass corresponding to a situation if one exists
  ;; otherwise, returns nil.
  (unless (tms::env-contradictory e)
    (setq sclass (tms::get-env-prop e 'sclass))
    (if (sclass? sclass) sclass nil)))

(defun env-sclass (e)
  ;; Like env-sclass, but assumes env is a situation
  ;; and provides an error if it isn't.
  (cond ((tms::env-contradictory e)
	 (error "~%~A is contradictory -- cannot be a situation." e))
	(t (let ((sclass (tms::get-env-prop e 'sclass)))
	   (if (sclass? sclass) sclass
	       (error "~%~A does not correspond to a situation." e))))))

;;;; Indexing sclasses

;;The notion of what is a distinct situation is slightly different
;;in this implementation.  Before the Qstate was an essential part.
;;Now the only fields which count in the equality tests are, in order,
;;   individuals
;;   view structure
;;   process structure
;;   Ds values
;;
;;Additional distinctions may be necessary, for instance, if rate assumptions
;;need to be introduced during limit analysis.  But these will do for now.
;;Since all keys along a path must be specified, can be very simple.
;;
;;The problem with using the old scheme is that there are many possible
;;ways for something to be false, even given a particular set of basis facts.

(defun lookup-sclass (keys)
  (adb::nested-alist-lookup keys *sclass-table*))

(defun insert-sclass (keys sclass)
  (setq *sclass-table* 
	(adb::nested-alist-insert keys sclass *sclass-table*)))

(proclaim '(special *PS-VS*))

(defun map-sclasses (fun &optional (sits *ps-vs*))
  (dolist (sit sits)
    (cond ((eq (sclass-status sit) 'R-Complete)
	   (funcall fun sit)
	   (let ((other (get-sclass-prop sit ':Clone)))
	     (if other (funcall fun other))))
	  (t (map-sclasses fun (sclass-r-completions sit))))))

(defun map-situations (fun)
  (map-sclasses #'(lambda (scl)
		    (dolist (sit (sclass-assumptions scl))
		      (funcall fun sit)))))

(defun count-situations ()
  (let ((*count* 0))
    (map-sclasses
      #'(lambda (s) 
	  (setq *count* (+ *count* (length (sclass-assumptions s))))))
    *count*))

(defun count-sclasses ()
  (let ((*count* 0))
    (map-sclasses #'(lambda (ignore) (incf *count*)))
    *count*))

(defun list-all-sclasses ()
  (let ((*situations* nil))
    (map-sclasses #'(lambda (s) (push s *situations*)))
    *situations*))

;;;; Cloning sclasses

(defun clone-sclass (parent &optional (use-name? nil) &aux clone)
  ;;; R-completions and ECL sometimes require building
  ;;; new situations.  Here we "clone" the old situation,
  ;;; keeping all the static information possible.
  (setq clone (make-sclass :NAME (if use-name?
				     (intern (format nil "~A-I"
						     (sclass-name parent)))
				     (generate-sclass-name))
			   :INDIVIDUALS (sclass-individuals parent)
			   :VIEW-INSTANCES (sclass-view-instances parent)
			   :PROCESS-INSTANCES (sclass-process-instances parent)
			   :PROCESS-STRUCTURE (sclass-process-structure parent)
			   :VIEW-STRUCTURE (sclass-view-structure parent)
			   :DS-VALUES (copy-list (sclass-Ds-values parent))
			   :UNKNOWN-DS-VALUES (copy-list
						(sclass-unknown-Ds-values parent))
			   :STATUS (sclass-status parent)))
  (set (sclass-name clone) clone)
  clone)

;;;; Some general accessor functions.

(defun get-individuals ()
  (apply #'append (adb::fetch '((individual ?I) . :TRUE) :bindings '(?I))))

(defun get-pis () 
  (apply #'append (adb::fetch '((process-instance ?type ?name . ?rest) . :TRUE)
			 :bindings '(?name))))

(defun get-vis ()
  (apply #'append (adb::fetch '((view-instance ?type ?name . ?rest) . :TRUE)
			 :bindings '(?name))))

(proclaim '(special *basis* *preconditions*))

(defun gather-preconditions () *preconditions*)

(defvar *adb-individual-marker* nil
  "Synchronization marker for individuals")
(defvar *individuals* nil
  "List of possible quantity-conditioned individuals")

(defun gather-individuals ()
  (unless (and (boundp '*individuals*)
	       (listp *individuals*)
	       (eq *adb-individual-marker* adb::*current-adb*))
    (setq *individuals* 
	  (let ((individuals nil))
	    ;only the individuals whose existence can change are in the
	    ;sclass-individuals field, so the union of these will suffice
	    (dolist (s *ps-vs*)
	      (dolist (individual (sclass-individuals s))
		(pushnew individual individuals :test #'equal)))
	    individuals))))

;;;; Mode flags
;;
;;The "flags" item on the menu varies how much information QPE reports.
;;Modes actually change the way QPE operates.
;; All the ones I've thought of so far are binary, so that will be assumed.   Otherwise,
;; more alist hacking would be required.  

(defmacro DefQPEmode
	  (mode-name Menu-line True-Output-line False-Output-Line initial-value)
  `(progn (pushnew (list ',mode-name ,menu-line
			 ,true-output-line ,false-output-line
			 ',initial-value)
		*QPE-Modes* :test #'equal)
	  (defvar ,mode-name ',initial-value)))


;;;; QPE modes
;;; These can fundamentally change the way QPE works.
;;; Some of them just control report generation.  Others
;;; can change the engine's operation profoundly, and are
;;; provided for experimental purposes.

(defQPEMode *Prune-Asymptotic-Approaches* "Prune asymptotic approach states"
  ;; QP theory says "T"
	    "Asymptotic approach states pruned."
  "Asymptotic approach states allowed."
  t)

(defQPEMode *Allow-Unstable-Trans* "Allow unstable transitions"
  ;; QP theory says "nil"
  "Unstable transitions allowed."
  "Unstable transitions not allowed."
  nil)

(defQPEMode *Case2-ECL-ON* "Equality Change Law, Case 2"
  ;; QP theory says "T"
  "Equality Change Law, Case 2: ON"
  "Equality Change Law, Case 2: OFF"
  t)

(defQPEmode *Allow-Coincidences* "Allow simultaneous changes to and from =."
  ;; QP theory says "T"
  "Simultaneous changes to and from =: ON."
  "Simultaneous changes to and from =: OFF."
  t)

(defQPEmode  *Augment-Derivative-Comparisons*
  ;; QP theory says "T".  Turn this off only if you are desperate for some
  ;; kind of answer on a very large problem.  (Even then, you would be
  ;; better off using operating assumptions and exploring the behaviors
  ;; incrementally.)
	   "Ensure all relevant D relations known."
	   "Derivative Augments: ON"
	   "Derivative Augments: OFF"
	   T)

(defQPEMode *Allow-discontinuous-directs* "Allow discontinuous direct influences"
  ;; QP theory is agnostic on this one.  There are certain reasonable intuitive models
  ;; (such as NST) which do not work properly when this is set to NIL.  I consider
  ;; classical continuity somewhat harmful, but the switch is here for those who wish
  ;; to be strict about it.
  "Discontinuous direct influences allowed."
  "Discontinuous direct influences not allowed."
  t)

(defQPEMode *Resolve-Simple-Multiple-Directs*
   ;; QP theory is agnostic.  You get slightly more information about what it takes
   ;; to resolve the ambiguity with this on, but at the cost of extra assumptions, and
   ;; thus extra work, for the rest of the computation.
	    "Introduce inequality asn's for pairs of conflicting direct influences"
  "Add assumptions for resolving pairs of direct influences"
  "No inequalities added for pairs of conflicting direct influences"
  T)

(DefQPEmode *Assume-PCs* "Vary preconditions"
  ;; This should be "T", to handle variations imposed by external representation
  ;; systems.  This switch was of limited use in debugging some early models, but
  ;; will probably be flushed in the next version. 
  "Preconditions assumed." "Preconditions asserted."  t)

(defQPEMode *differential-situation-display* "Display situations by differences"
  ;; Controls printing, theoretically unimportant.
  "Display situations by differences" "Display situations by environments"
  t)

(defQPEMode *Silence-zero-Ds-values* "Skip printing zero Ds values"
  ;; Controls printing, theoretically unimportant.
  "Zero Ds values suppressed." "Zero Ds values printed." nil)

(defQPEmode *Careful* "Careful mode"
  ;; Performs extra "reality checks" and gathers lots of extra information.
  ;; This information is stored in the QPE Flight Recorder, and is accessible
  ;; from various menu systems.
  ;; I like running with it on, but it can add signficant overhead.
  "Careful mode ON"
  "Careful mode OFF" nil)

;;; Vestigal code.  Will always use net-influence quantities, never use labelling.
;;; The former is just too handy, the latter too expensive and uninformative.

;(defQPEMode *Resolve-Multiple-Directs-By-Labeling*
;	    "Disambiguate multiple direct influences via inequalities"
;  "Add inequalities for multiple direct influences"
;  "No inequalities added for multiple direct influences"
;  nil)
;
;(defQPEMode *Resolve-Multiple-Directs-By-NI-quantity*
;	    "Disambiguate multiple direct influences via net-influence quantity."
;  "Net-Influence quantities introduced as needed."
;  "No Net-Influence quantities introduced." T)

