;;; -*- Mode: LISP; Syntax: Common-lisp; Package: QPE; -*-

;;;; Internal event scheduler for QPE

;;; Copyright (c) 1989, Kenneth D. Forbus, University of Illinois
;;; All rights reserved

(in-package 'QPE)

;;; When using QPE as a module, or when experimenting with augments,
;;; it is often handy to be able to perform operations within stages.
;;; For instance, if one were searching a space of modeling assumptions to
;;; find an envisionment that is rich enough to make an accurate prediction
;;; yet not drown in details, you might want to check the states generated after
;;; influence resolution to see whether or not it is worth going on to limit
;;; analysis.  This code provides a place to hang such hooks.

;;; This should only be directly hacked by the macho QPE user.

;;; Interface:

;;; (schedule-qpe-event <thunk> <modifier> <operation> <priority>)
;;; where <thunk> is a procedure of no arguments,
;;;       <modifier> is :BEFORE or :AFTER
;;;       <operation> is a named QPE operation (like :LIMIT-ANALYSIS)
;;;       <priority> is an integer
;;; This will cause <thunk> to be executed either before or after the named
;;; operation.  <priority> specifies the order in which thunks are executed,
;;; with smaller numbers being executed before larger ones.

;;; (show-qpe-internal-events &optional (stream *standard-output*))
;;;  prints the contents of the table for debugging purposes.

;;; Internals:
;;; (run-qpe-events <modifier> <operation>) runs the thunks associated with this
;;;  operation.  This is what various QPE operations call.

(defvar *internal-events* nil) ;; Format is (<operation> <befores> <afters>),
;; where the befores and afters take the form (<priority> . <thunk>), and
;; they are sorted in the order to execute them.

(adb::init-form (setq *internal-events* nil)) ;; You have to reload them each time!

;; ******* Should provide some sort of interface to edit the event table.


(defvar *internal-event-types*
	'(:LOAD-DOMAIN ;; Obvious
	   :SCENARIO-EXPANSION ;; obvious
	   :PSVS-ASNS ;; Assuming the initial basis set
	   :PSVS-IC ;; Interpretation construction for view and process structures
	   :IR ;; Overall influence resolution
	   :IR-AMBIG ;; Ambiguous influence resolution (within IR)
	   :LIMIT-ANALYSIS ;; obvious
	   ))

(defun schedule-qpe-event (thunk modifier operation priority
				 &aux entry queue)
  (unless (member operation *internal-event-types*)
    (error "~% ~A not an internal QPE event.~%~A ~A ~A"
	   operation operation priority thunk))
  (unless (integerp priority)
    (error "~% Priority of ~A not a integer: QPE internal event ~A, ~A ~A."
	   priority thunk modifier operation))
  (setq entry (assoc operation *internal-events*))
  (unless entry (setq entry (list operation nil nil))
	  (push entry *internal-events*))
  (case modifier
    (:BEFORE (setf (cadr entry) (add-event-thunk priority thunk (cadr entry))))
    (:AFTER (setf (caddr entry) (add-event-thunk priority thunk (caddr entry))))
    (t (error "~% ~A illegal event modifier -- must be :BEFORE or :AFTER~%~A ~A ~A"
	      modifier operation priority thunk))))


(defun show-qpe-internal-events (&optional (stream *standard-output*))
  (format stream "~%QPE Internal Events table")
  (dolist (event-entry *internal-events*)
    (format stream "~% ~A" (car event-entry))
    (cond ((cadr event-entry)
	   (format stream "~%   Before: ")
	   (print-comma-list (cadr event-entry) stream)
	   (format stream "~%           " (cadr event-entry)))
	  (t (format stream "~%   Before: None.")))
        (cond ((caddr event-entry)
	   (format stream "~%    After: ")
	   (print-comma-list (caddr event-entry) stream)
	   (format stream "~%           " (cadr event-entry)))
	  (t (format stream "~%    After: None.")))))

(defun print-comma-list (list-of-stuff &optional (stream *standard-output*))
  ;; Thank you, Quux.
  ;; Should get fancy about nested indenting, but what the heck.
  (format stream "   ~{~<~%   ~1:; ~S~>~^,~}.~%" list-of-stuff))

;;;; Internals

(defun add-event-thunk (priority thunk queue)
  (cond ((null queue) (list (cons priority thunk)))
	((< priority (caar queue)) (cons (cons priority thunk) queue))
	(t ;; you are here
	 (do ((q queue (cdr q)))
	     ((null (cdr q))
	      (setf (cdr q) (cons (cons priority thunk) nil)))
	   (when (< priority (caadr q))
	     (setf (cdr q) (cons (cons priority thunk) (cdr q)))
		     (return q)))
	 queue)))

(defun run-qpe-events (modifier operation)
  (let ((entry (assoc operation *internal-events*)))
    (unless entry (return-from EXECUTE-EVENT-THUNKS nil))
    (case modifier
      (:BEFORE (execute-event-thunks (cadr entry)))
      (:AFTER (execute-event-thunks (third entry)))
      (t (error "~%Unknown event modifier ~A of ~A."
		modifier operation)))))

(defun execute-event-thunks (queue)
  (dolist (entry queue) (funcall (cdr entry))))

;;;; Persistent events
;; These live across initializations of ATMoSphere
;; Good for debugging

(defvar *persistent-internal-events* nil)

(defun add-persistent-ie (name thunk modifier operation priority)
    (unless (member operation *internal-event-types*)
      (error "~A not an internal QPE event.~%~A ~A ~A"
	     operation name priority thunk))
    (unless (integerp priority)
      (error "Priority of ~A not a integer: QPE internal event ~A, ~A ~A."
	     priority name modifier operation))
    (unless (or (eq modifier ':BEFORE) (eq modifier ':AFTER))
      (error "Invalid modifier for QPE internal event: ~A, of ~A." modifier name)) 
  (push (list name modifier operation priority thunk)
	*persistent-internal-events*))

(defun remove-persistent-ie (name)
  (setq *persistent-internal-events*
	(delete (assoc name *persistent-internal-events*)
		*persistent-internal-events*)))

(defun recycle-persistent-events ()
  (dolist (pe *persistent-internal-events*)
    (schedule-qpe-event (fifth pe) (second pe) (third pe)
			(fourth pe))))

(defun show-persistent-events (&optional (stream *standard-output*))
  (cond ((null *persistent-internal-events*)
	 (format stream "~%No persistent thunks are defined."))
	(t (format stream "~%There are ~D persistent internal events:" 
		   (length *persistent-internal-events*))
	   (dolist (pe *persistent-internal-events*)
	     (format stream "~%   ~A: ~A ~A; ~D."
		     (first pe) (second pe) (third pe) (fourth pe))))))

(adb::after-init-form (recycle-persistent-events))

