;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Limit Analysis
;;
;;  Thanks to the organization of QPE, we can assume 
;;  all sclasses have been generated up to r-completeness.
;;  This means that any situation which is consistent must already
;;  exist, simplifying consistency testing for temporal inheritance. 

;; Old algorithm:
;; 1. For each element of the Assumed Qstate, 
;;         a. Find consistent combinations of inequalities and Ds values
;;         b. Create single-change limit hypotheses as appropriate.
;;             i.e., the start envirionment must be a subset of at least
;;             one existing situation.
;; 2. For each situation, find the set of single-change limit hypotheses implied by it.
;; 3. For each combination of situation/single change hypotheses, compute what simultaneous
;;      effects are possible.
;; 4. Evaluate the limit hypotheses as follows:
;;    4.1 If temporal inheritance does not yield a known situation, flush the hypothesis.
;;    4.2 If any continuity checks fail, flush it.
;;    4.3 If these tests are passed, install the transition in the Next-State table.
;; 5. Assign durations based on the Equality Change Law.

;;;; YET TO DO:
;;
;; Incorporate magnitudes in the transition-finding process.
;; (Hasn't been urgent demand, so many other things to do...)

;;;; Globals for limit analysis

(proclaim '(special *limit-hypotheses*
		  *single-limit-hypotheses*
		  *conjunctive-limit-hypotheses*))

(defvar *Limit-Qstate* nil)

(adb::init-form (makunbound '*single-limit-hypotheses*))
(adb::init-form (makunbound '*conjunctive-limit-hypotheses*))
(adb::init-form (makunbound '*limit-hypotheses*))
(adb::init-form (setq *Limit-Qstate* nil))

(defvar *Conjunctive-LH-table* nil
  "Cache of conjunctive limit hypotheses.  Key is list of INEQ's.")

(adb::init-form (when (null *conjunctive-lh-table*)
		  (setq *conjunctive-lh-table*
			(make-hash-table :test #'equal))))

(proclaim '(special *Allow-Unstable-Trans* *Case2-ECL-ON*
		    *Allow-Coincidences* *Augment-Derivative-Comparisons*))

;;;; Limit Analysis top level
;;

(proclaim '(special *first-next-states*  ;; for debugging
		    *contin-next-states*  ;; for debugging
		    *close-next-states* ;; for debugging
		    *conjunctive-lh-nogood-table*
		    *next-state-table*
		    *situation-lh-table*
		    *asymp-next-states*))

(adb::init-form (setq *first-next-states* nil
		      *contin-next-states* nil
		      *close-next-states* nil
		      *next-state-table* nil
		      *situation-lh-table* nil))

(defun limit-analysis ()
  (When-Careful
    (setq *QPE-Program-Counter* "Limit"))
  (run-qpe-events :BEFORE :LIMIT-ANALYSIS)
  (flush-inconsistent-sclasses)
  (when *Augment-Derivative-Comparisons*
    (augment-derivative-comparisons))
  (setq *single-limit-hypotheses* nil
	*conjunctive-limit-hypotheses* nil
	*limit-hypotheses* nil)
  (clrhash *conjunctive-lh-table*)
  (clrhash *conjunctive-lh-nogood-table*)
  (index-situations)
  (find-single-change-limit-hypotheses)
  (apply-lhs-to-situations) ;; Builds *situation-lh-table*
  (generate-conjunctive-limit-hypotheses) ;; From entries in situation-lh table
  (setq *limit-hypotheses* (append *single-limit-hypotheses*
				   *conjunctive-limit-hypotheses*))
;;  (check-lhs) ;; For ridiculous bugs.
  (find-next-states)
  (setq *first-next-states* (copy-tree *next-state-table*))
  ;; This version prunes by continuity first, then checks closeness.
  (prune-via-continuity)
  (setq *next-state-table* (delete-if-not #'cdr *next-state-table*))
  (setq *contin-next-states* (copy-tree *next-state-table*))
  ;; Closest pruning now a seperate step.
  (setq *next-state-table* (prune-via-closest))
  (setq *close-next-states* (copy-tree *next-state-table*))
  (apply-equality-change-law)
  (setq *next-state-table* (delete-if-not #'cdr *next-state-table*))
  (install-duration-information)
  (when-careful ;; Interesting statistic.
    (dolist (entry (check-for-ambiguous-transitions))
      (Add-Recorder-Event :Ambiguous-Transition
			  (car entry) ;; start
			  (cadr entry) ;; Transition
			  (cddr entry) ;; next states
			  )))
  (unsimplify-lh-envs) ;; Needed due to internal simple env hack
  ;; ****** PATCH.  This might be able to go earlier in the code, in particular
  ;; ****** before the duration checking, to avoid throwing away splits.
  ;; ****** Also, it does gross linear searches to detect the initial losing
  ;; ****** states, when the information cached with sclasses should be used
  ;; ****** instead.
  (setq *asymp-next-states* (copy-tree *next-state-table*))
  (when *Prune-Asymptotic-Approaches*
    (prune-asymptotic-approach-states))
  (run-qpe-events :AFTER :LIMIT-ANALYSIS))

(defun lhtable-statistics (table &aux sum count max min
				 max-set min-set)
  (setq sum 0 count 0)
  (dolist (entry table)
    (let ((len (length (cdr entry)))) 
      (incf count)
      (incf sum len)
      (cond (max-set (if (> len max) (setq max len)))
	    (t (setq max-set t max len)))
      (cond (min-set (if (< len min) (setq min len)))
	    (t (setq min-set t min len)))))
  (values count sum (if (= count 0) 0
			(/ (float sum) count))
	  (if max-set max 0)
	  (if min-set min 0)))


;;;; Ensuring completeness of deriviative information

;;; Jcollins points out that individuating single-change LH's on the difference between
;;  D[Q1] and D[Q2] makes more sense than on Ds values, since fewer LH's will be required
;;  to represent the same change in inequalities.
;;
;;  To do this, we must ensure that all the necessary derivative relationships are known in
;;  every situation.  Empirically, we've had models where some situations cannot be
;;  consistently extended by these derivative relationships, indicating a state which
;;  is actually inconsistent.  We don't have a good characterization of such circumstances
;;  yet, hence the flight recorder. 
;;
;; Why not just make all such derivative relationships assumed?  Typically many of them have
;; been determined as a consequence of other assumptions.  And assumptions are NOT cheap.
;; Consequently, we add extra assumptions only when needed.  
 
(defun augment-derivative-comparisons ()
  (let ((d-comps nil)  ;; Comparisons found.
	(to-prune nil)) ;; Sclasses pruned, if any.
  ;; Find the relevant derivative comparisons in derived qstate.
  ;; Any comparisons in assumed qstate, being already assumed, are known.
    (dolist (ineq *derived-qstate*)
      (when (and (derivative? (ineq-num1 ineq))
		 (derivative? (ineq-num2 ineq)))
	(push ineq d-comps)))
    (when d-comps
      (if-debug-flag
	limit-details
	(format t "~%    LA: Derivatives relations to check:")
	(dolist (ineq d-comps)
	  (format t "~%         ~A" ineq)))
    (map-sclasses
      #'(lambda (sclass)
	  (let ((new-sits nil)
		(result nil))
	    (if-debug-flag limit-details
			   (format t
			   "~%     LA: Current Sclass = ~A, Situations = ~A."
				   sclass (sclass-assumptions sclass)))
	    (dolist (sit (sclass-assumptions sclass))
	      (setq result (mapcar #'tms::simple-to-env
				   (extend-situation-via-inequalities
				     sit d-comps)))
	      (cond ((null result)
		     ;; Situation has no consistent extension
		     (if-debug-flag limit-details
		       (format t "~%      LA:   No consistent extension for ~A."
			       sit))
		     (prune-situation sit sclass :No-Derivative-Augment))
		    ((and (null (cdr result))
			  (eq (car result) sit))) ;; No extensions are needed
		    (t ;; must update situation markers. 
		     ;; Install a forwarding pointer for temporal inheritance.
		     (unmark-env-sclass sit)
		     (tms::put-env-prop sit result ':Forwarding-Pointer)
		     (dolist (nsit result)
		       (mark-env-with-sclass nsit sclass)
		       (tms::put-env-prop nsit sit ':Forwarded-from))))
	      (if-debug-flag limit-details
			     (format t "~%     LA:   Extension of ~A is ~A."
				     sit result))
	      (setq new-sits (nconc new-sits result)))
	    (cond ((null new-sits)
;		   (error "~%Unexpected sclass pruning -- ~A."
;			  sclass)
		   (prune-sclass sclass :No-Derivative-Augment)
		   (push sclass to-prune))
		  (t (if-debug-flag limit-details
				    (format t
				      "~%     LA: New situations for ~A: ~A."
					    sclass new-sits))
		     (setf (sclass-assumptions sclass) new-sits))))))
    (setq *PS-VS* (nset-difference *PS-VS* to-prune))
    (flush-inconsistent-sclasses))))

;;;; Helpers for augmenting via derivative relationships

(defun ineq-known-in? (ineq env)
  (or (tms::generic-true-in? (ineq-> ineq) env)
      (tms::generic-true-in? (ineq-= ineq) env)
      (tms::generic-true-in? (ineq-< ineq) env)
      (tms::generic-true-in? (ineq-bt ineq) env)))

(defun extend-situation-via-inequalities (sit ineqs)
  (cond ((null ineqs) (list sit))
	((ineq-known-in? (car ineqs) sit)
	 (extend-situation-via-inequalities sit (cdr ineqs)))
	(t ;; must expand this one
	 (assume-derivative-pair-if-needed (car ineqs))
	 (let ((result nil)
	       (nsits (extend-env-by-ineq sit (car ineqs))))
	   (if-debug-flag limit-details
	    (format t "~%   Extending by ~A yields ~A." (car ineqs) nsits))
	   (dolist (nsit nsits)
	     (setq result (nconc (extend-situation-via-inequalities
				   nsit (cdr ineqs))
				 result)))
	   result))))

(defun assume-derivative-pair-if-needed (ineq)
  (let ((asn (assume-ineq-if-needed ineq)))
    (when asn ;; nil if already assumed.
      ;; If not assumed by now, then it cannot be on either QC-qstate or
      ;; resolve-qstate.  Hence it should be put on LIMT-QSTATE.
      (push ineq *limit-qstate*))))

(defun extend-env-by-ineq (env ineq &aux new result)
  (unless (ineq-assumed? ineq)
    (error "~%~A not assumed: extend-env-by-ineq (~A)." ineq env))
  (when (ineq-bt-a ineq)
    (setq new (tms::generic-cons-env env (ineq-bt-a ineq)))
    (unless (tms::generic-env-contradictory new) (push new result)))
  (when (ineq-lt-a ineq)
    (setq new (tms::generic-cons-env env (ineq-lt-a ineq)))
    (unless (tms::generic-env-contradictory new) (push new result)))
  (when (ineq-eq-a ineq)
    (setq new (tms::generic-cons-env env (ineq-eq-a ineq)))
    (unless (tms::generic-env-contradictory new) (push new result)))
  (when (ineq-gt-a ineq)
    (setq new (tms::generic-cons-env env (ineq-gt-a ineq)))
    (unless (tms::generic-env-contradictory new) (push new result)))
  result)

;;;; Finding single-change hypotheses

;; This version also interleaves the possible next relationships computation with
;; finding single-change limit hypotheses

;;; ****** Should amount-derivative comparisons be considered illegal?

(defun find-single-change-limit-hypotheses ()
  (dolist (ineq (gather-assumed-qstate))
      (when (and (amount? (ineq-num1 ineq))
		 (not (derivative? (ineq-num2 ineq)))  ;; i..e, amount or special.
		 (not (constant-ineq? ineq))) ;; not universally true
	(if-debug-flag limit-details
	    (format t "~%     LH: Analyzing ~A." ineq))
	(cond ((or (magnitude? (ineq-num1 ineq))
		   (magnitude? (ineq-num2 ineq)))
	       (format t "~% WARNING: Possible transitions involving ~A won't be detected."
		       (ineq-string ineq))
	       (format t "~%          This should change in future versions of QPE.  Sorry.")
	       (when-careful
		 (add-recorder-event 
		   :MAGNITUDE-TRANSITION-POSSIBLE ineq)))
	      (t (find-possible-limit-hypothesis-for ineq)))))
  (setq *single-limit-hypotheses* (nreverse *single-limit-hypotheses*)))

(defun find-possible-limit-hypothesis-for (ineq)
  ;; Guarenteed to be concerned with amounts by this call.
  ;; ***** Should probably replace *enforce-env* with *basis-env*!
  (let* ((d1 (make-derivative (quantity-of (ineq-num1 ineq))))
	 (d2 (make-derivative (quantity-of (ineq-num2 ineq))))
	 (D-ineq (fetch-ineq d1 d2)) (enforcer (list *enforce-env*))
	 (envs nil) (nenvs nil)(end-envs nil))
;; This version is based on using the derivative inequalities,
;; rather than the Ds values, for efficiency.
;;  The transition table looks like this:
;;
;; N1 < N2 -- if D[N1] > D[N2], can become =.
;; N1 = N2 -- if D[N1] > D[N2], will become >
;;                  D[N1] < D[N2], will become <.
;; N1 > N2 -- if D[N1] < D[N2], can become =.
    (when (setq envs (tms::union-envss (tms::label (ineq-< ineq))
					       enforcer))
        ;; if A[N1] < A[N2] is possible
      (when (and (tms::label (ineq-> D-ineq)) ;; and D[N1] > D[N2] is possible
		 (tms::label (ineq-= ineq)))  ;; and A[N1] > A[N2] is possible
	(when (setq nenvs
		    (tms::union-envss envs
				     (tms::label (ineq-> D-ineq)))) ;;and together,
	  ;; Now ensure consequences of the change can be consistent
	  (when (setq end-envs (tms::union-envss (tms::label (ineq-> D-ineq))
						 (tms::union-envss (tms::label (ineq-= ineq))
								   enforcer)))
								   
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'LESS-THAN 'EQUAL-TO 'GREATER-THAN 'GREATER-THAN
	      nenvs end-envs))
	  ;; Other possibility -- D becomes =.
	  (when (setq end-envs (tms::union-envss (tms::label (ineq-= D-ineq))
						 (tms::union-envss (tms::label (ineq-= ineq))
								   enforcer)))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'LESS-THAN 'EQUAL-TO 'GREATER-THAN 'EQUAL-TO
	      nenvs end-envs)))))
    (when (setq envs (tms::union-envss (tms::label (ineq-> ineq))
				       enforcer))
      ;; if A[N1] > A[N2] is possible
      (when (tms::label (ineq-< D-ineq)) ;; and D[N1] < D[N2> is possible
	(when (setq nenvs
		    (tms::union-envss envs
				     (tms::label (ineq-< D-ineq)))) ;;and together,
	  ;; Check both possibilities
	  (when (setq end-envs (tms::union-envss (tms::label (ineq-< D-ineq))
						 (tms::union-envss (tms::label (ineq-= ineq))
								   enforcer)))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'GREATER-THAN 'EQUAL-TO 'LESS-THAN 'LESS-THAN
	      nenvs end-envs))
	  (when (setq end-envs (tms::union-envss (tms::label (ineq-= D-ineq))
						 (tms::union-envss (tms::label (ineq-= ineq))
								   enforcer)))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'GREATER-THAN 'EQUAL-TO 'LESS-THAN 'EQUAL-TO
	      nenvs end-envs)))))
    (when (setq envs (tms::union-envss (tms::label (ineq-= ineq)) enforcer))
      (when (tms::label (ineq-< D-ineq))
	(when (setq nenvs (tms::union-envss envs
					   (tms::label (ineq-< D-ineq))))
	  (when (setq end-envs (tms::union-envss (tms::label (ineq-< D-ineq))
						 (tms::union-envss (tms::label (ineq-< ineq))
								   enforcer)))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'EQUAL-TO 'LESS-THAN 'LESS-THAN 'LESS-THAN nenvs end-envs))
	  (when (and *Allow-Coincidences*
		     (setq end-envs (tms::union-envss (tms::label (ineq-= D-ineq))
						      (tms::union-envss
							(tms::label (ineq-< ineq))
							enforcer))))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'EQUAL-TO 'LESS-THAN 'LESS-THAN 'EQUAL-TO nenvs end-envs))))
      (when (tms::label (ineq-> D-ineq))
	(when (setq nenvs (tms::union-envss envs
					   (tms::label (ineq-> D-ineq))))
	  (when (setq end-envs (tms::union-envss (tms::label (ineq-> D-ineq))
						 (tms::union-envss (tms::label (ineq-> ineq))
								   enforcer)))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'EQUAL-TO 'GREATER-THAN 'GREATER-THAN 'GREATER-THAN nenvs end-envs))
	  (when (and *Allow-Coincidences* 
		     (setq end-envs (tms::union-envss (tms::label (ineq-= D-ineq))
						 (tms::union-envss (tms::label (ineq-> ineq))
								   enforcer))))
	    (make-single-change-limit-hypothesis
	      ineq D-ineq 'EQUAL-TO 'GREATER-THAN 'GREATER-THAN 'EQUAL-TO
	      nenvs end-envs))))
      ;;; For "spontaneous generation" -- Experiment
      ;; A less inflammatory phrasing is "transitions from unstable equilibriums"
      (when *Allow-Unstable-Trans*
	(when (tms::label (ineq-= D-ineq))
	  (when (setq nenvs (tms::union-envss envs
					     (tms::label (ineq-= D-ineq))))
	    (when (setq end-envs (tms::union-envss (tms::label (ineq-> D-ineq))
						   (tms::union-envss
						     (tms::label (ineq-> ineq)) enforcer)))
	      (make-single-change-limit-hypothesis
		ineq D-ineq 'EQUAL-TO 'GREATER-THAN 'EQUAL-TO 'GREATER-THAN nenvs end-envs
		':Unstable))
	    (when (setq end-envs (tms::union-envss (tms::label (ineq-< D-ineq))
						   (tms::union-envss
						     (tms::label (ineq-< ineq)) enforcer)))
	      (make-single-change-limit-hypothesis
		ineq D-ineq'EQUAL-TO 'LESS-THAN 'EQUAL-TO 'LESS-THAN nenvs end-envs
		':Unstable))))))))

;;;; Creating single-change limit hypotheses

;; Used to be more filtering (orphan test + applicability of end env).
;; The latter now occurs in FIND-POSSIBLE-LIMIT-HYPOTHESES-FOR.
;; The former is irrelevant now that conjunctive LH's are generated locally to
;; a situation, and orphans could be valuable for explaining why a transition
;; couldn't happen. 

;; The reason for accumulating END-ENVS now instead of END-ENV is that not every
;; derivative relationship need be an assumption.

(defun make-single-change-limit-hypothesis (ineq D-ineq before after D-before D-after
					    start-envs end-envs &optional (unstable? nil))
  (let ((lh (multiple-value-bind (index name)
		(generate-lh-name 'LH)
	      (make-lh :NAME name
		       :INDEX index
		       :START-ENVS start-envs
		       :END-ENVS end-envs
		       :RELATIONS (list (list ineq before after))
		       :EQUALITY-TYPE (cond ((eq before 'EQUAL-TO) ':FROM)
					    ((eq after 'EQUAL-TO) ':TO)
					    (t (error
				 "~%Transition not involving equality: ~A, ~A -> ~A; ~A ~A ~A."
  ineq before after (list D-ineq D-before D-after) start-envs end-envs)))
			 :D-RELS (list D-ineq D-before D-after)
			 :OLD-ASSUMPTIONS (delete nil ;; Some derivative relations aren't asns
						  (list (get-ineq-asn ineq before)
							(get-ineq-asn D-ineq D-before)))
			 :NEW-ASSUMPTIONS (delete nil 
						  (list (get-ineq-asn ineq after)
							(get-ineq-asn D-ineq D-after)))
			 :SINGLE-CHANGE T))))
    (push lh *single-limit-hypotheses*)
    (adb::if-debug-flag limit ;upgrade this to a neater print statement
		    (format t "~%Found Limit Hypothesis ~A" (lh-name lh))
		    (terpri)
		    (princ-lh lh))
    (setf (lh-singles lh) (list lh))
    (put-ineq-prop ineq T ':LH-participant?)
    (put-lh-prop lh unstable? :UNSTABLE?)
    (set (lh-name lh) lh)))

(defun get-ineq-asn (ineq rel)
  (case rel
    (LESS-THAN (ineq-lt-a ineq))
    (GREATER-THAN (ineq-gt-a ineq))
    (EQUAL-TO (ineq-eq-a ineq))
    (UNRELATED (ineq-bt-a ineq))))

;;;; Finding which singleton limit hypotheses apply to a situation
;;
;; Iterate over each situation, checking to see which limit hypotheses
;; have a start-envs which is implied by the situation.
;;   For each of them, install them in the situation-lh table
;;    as a possible transition.
;;
;; Later, the conjunctives will be found by finding all consistent combinations of these
;; singletons. 

(defvar *situation-lh-table* nil) ;; (<situation> . <candidate limit hypotheses>)
(adb::init-form (setq *situation-lh-table* nil))

(defun apply-lhs-to-situations ()
  (setq *situation-lh-table* nil)
  (map-sclasses #'(lambda (sclass)
		  (dolist (sit (sclass-assumptions sclass))
		    (let ((lhs nil))
		      (dolist (lh *single-limit-hypotheses*)
			(when (implies-limit-hypothesis? sit lh)
			  (push lh lhs)))
		      (push (cons sit lhs) *situation-lh-table*))))))

(defun implies-limit-hypothesis? (env lh)
  (some #'(lambda (start-e) (tms::generic-subset-env? start-e env)) (lh-start-envs lh)))

(defun implies-limit-hypothesis-result? (env lh) ;; For debugging
  (some #'(lambda (start-e) (tms::generic-subset-env? start-e env)) (lh-end-envs lh)))

;;;; Generating conjunctive limit hypotheses
;; Conjunctions of limit hypotheses denote simultaneous changes.
;; These hypotheses are created by a bottom-up lattice merge, filtering
;; via the end-envs for consistency.
;; 
;; For efficiency, combine just those hypotheses which are applicable to some situation.
;; No additional testing for applicability is needed, just consistency.
;; The *CONJUNCTIVE-LH-TABLE* provides a cache. 

(defun generate-conjunctive-limit-hypotheses ()
  (setq *conjunctive-limit-hypotheses* nil)
  (dolist (entry *situation-lh-table*) ;; Entries have form (<sit> . <lh's>)
    (let ((clhs (generate-situation-clhs (cdr entry))))
      (setf (cdr entry) (nconc (cdr entry) clhs))))
  (setq *conjunctive-limit-hypotheses*
	(nreverse *conjunctive-limit-hypotheses*)))

(defun generate-situation-clhs (single-changes)
  (cond (*Allow-Coincidences*
	 (generate-next-conjunctive-limit-hypotheses single-changes 2))
	(t (nconc
	     (generate-next-conjunctive-limit-hypotheses
	       (remove-if #'(lambda (lh) (eq (lh-equality-type lh) ':From))
			  single-changes) 2)
	     (generate-next-conjunctive-limit-hypotheses
	       (remove-if #'(lambda (lh) (eq (lh-equality-type lh) ':To))
			  single-changes) 2)))))

(defun fetch-conjunctive-lh (singles)
  (gethash singles *conjunctive-lh-table*))

(defun conjunctive-lh? (lh) (lh-includes lh))

;;; This generation algorithm is a bit of an experiment.  It assumes there may be many
;;; hypotheses, but few of them will be mutually consistent.  

(defun generate-next-conjunctive-limit-hypotheses (current-hypos size)
  ;; Given a set of limit hypotheses of length N-1, this code generates those of length N.
  ;; Filtering is accomplished by creating a alist whose keys are sorted lists of assumptions,
  ;; and whose item is the corresponding environment and the list of
  ;; LH's under it in the lattice.
;  (format t "~%>> Next level, hypotheses = ~A" current-hypos)
;  (format t "~%>> Lengths are ~A." (mapcar #'(lambda (x) (length (lh-singles x)))
;					   current-hypos))
 (when current-hypos
  (do ((h1 (car current-hypos) (car rest))
       ;; Iterate over all pairs of hypotheses at this level.
       (rest (cdr current-hypos) (cdr rest))
       (level-table nil)
       (new-lhs nil))
      ((null rest)
       (setq new-lhs (make-conjunctive-lhs-from-level-table level-table))
;       (format t "~% Level table:")
;       (pprint level-table)
;       (setq *NEW-LHS* new-lhs)
;       (break "Check *NEW-LHS*")
       (append new-lhs
	       (generate-next-conjunctive-limit-hypotheses new-lhs
							   (1+ size))))
    (do ((h2 (car rest) (car others))
	 (others (cdr rest) (cdr others))
	 (singles nil)
	 (entry nil))
	((null h2))
      (setq singles ;; Copy once for safety
	    (safe-union (lh-singles h1) (lh-singles h2)))
      (when (= (length singles) size) ;; generate them systematically
      (setq singles (sort singles #'lh-sort-predicate))
      ;; This sort's okay, # of elements is always small.  (If large, you will lose
      ;; for other reasons!)
      (setq entry (assoc singles level-table :test #'equal))
      (cond (entry ;; Join in the lattice, so just add these two in.
	     (pushnew h1 (cdddr entry))
	     (pushnew h2 (cdddr entry)))
	    (t (let ((conj-lh (fetch-conjunctive-lh (car entry))))
		 (cond (conj-lh 
			(push `(,singles) level-table))
		       ((lh-combination-nogood? singles)) ;; avoid redundant checks.
		       (t (let ((new-ends (tms::generic-union-envss (lh-end-envs h1)
								    (lh-end-envs h2))))
			    (cond (new-ends
				   (push `(,singles ,new-ends ,h1 ,h2)
					 level-table))
				  (t (add-lh-nogood singles))))))))))))))

(defun safe-union (l1 l2)
  (union (copy-list l1) (copy-list l2)))

;;;; Nogood detection for LH's
;; The idea is to minimize calls to UNION-ENVSS by detecting combinations which
;; together are nogood.  This is the simplest version, just cache results of identical calls.

(defvar *Conjunctive-LH-Nogood-Table* nil
  "Cache of nogood singleton limit hypotheses.")

(adb::init-form (cond ((null *conjunctive-lh-nogood-table*)
		       (setq *conjunctive-lh-nogood-table*
			     (make-hash-table :test #'equal)))
		      (t (clrhash *conjunctive-lh-nogood-table*))))

(defun add-lh-nogood (singles)
  (setf (gethash singles *conjunctive-lh-nogood-table*) t))

(defun lh-combination-nogood? (singles)
  (gethash singles *conjunctive-lh-nogood-table*))

;;; Helpers for generating conjunctive limit hypotheses
;; Since we are doing this now for situation-lh entries,
;; must check to see if the LH has already been created.

(defun make-conjunctive-lhs-from-level-table (table &aux lh lhs)
  (dolist (entry table)
    ;; Recall format of table entries is
    ;;(<single lh's> <end-envs> . <below in lattice>)
    ;; Theorem: Since the only singleton LH's that are rubbed together
    ;;          are those which are already implied by some particular
    ;;          situation, the union of the start-envs will never be empty. 
    ;; Consequence: Don't even bother computing the start-envs for conjunctives.
    ;;              It is a humongous waste of energy, and looking at the nodes would
    ;;              provide the same information to external systems if they need it. 
    (setq lh (fetch-conjunctive-lh (car entry)))
    (unless lh
    (multiple-value-bind (index name)
	(generate-lh-name 'CLH)
      (setq lh (make-lh :NAME name :INDEX index
			:RELATIONS (reduce #'nunion
					   (mapcar #'(lambda (lh)
						   (copy-list (lh-relations lh)))
					     (car entry)))
			:D-Rels (reduce #'nunion
				     (mapcar #'(lambda (lh)
						 (copy-list (lh-D-rels lh)))
					     (car entry)))
			:OLD-ASSUMPTIONS (reduce #'nunion
					   (mapcar #'(lambda (lh)
						       (copy-list 
							 (lh-old-assumptions lh)))
						   (car entry)))
			:NEW-ASSUMPTIONS (reduce #'nunion
					   (mapcar #'(lambda (lh)
						       (copy-list 
							 (lh-new-assumptions lh)))
						   (car entry)))
			:START-ENVS NIL ;; ******* Put list of nodes in here for debugging??
			:END-ENVS (cadr entry)
			:EQUALITY-TYPE
			(do ((lhs (car entry) (cdr lhs))
			     (from? nil)
			     (to? nil))
			    ((null lhs)
			     (if from?
				 (if to? ':MIXED ':FROM)
				 ':TO))
			  (case (lh-equality-type (car lhs))
			    (:FROM (setq from? t))
			    (:TO (setq to? t))
			    (:MIXED (setq from? t to? t))
			    (t (error "~%Equality type for ~A unmarked."
				      (car lhs)))))))
	(set name lh) ;; Install backpointer
	(push lh *conjunctive-limit-hypotheses*)
	(setf (gethash (car entry) *conjunctive-lh-table*) lh)
	;; Make the lattice interconnections here
	(setf (lh-singles lh) (car entry))
	(setf (lh-includes lh) (cddr entry))
	(dolist (lower (cddr entry))
	  (push lh (lh-included-in lower)))
;	(format t "~% Newly created:") (short-princ-lim-hyp lh nil *terminal-io*
;							    t)
;	(format t "~%Singles = ~A, includes = ~A."
;		(lh-singles lh) (lh-includes lh))
	))
	(push lh lhs))
  lhs)

;;;; Finding candidate next states
;;  The theory underlying this computation is summarized in "The problem of existence"
;;  paper.  Using the situation-lh table, the next-state-table is constructed by
;;  ascertaining which applicable limit hypotheses result in a consistent transition
;;  to a legal state.

(proclaim '(special *next-state-table*))

(defvar *next-state-table* nil)
(adb::init-form (setq *next-state-table* nil))

(defun find-next-states ()
  (setq *next-state-table* nil)	;;build a new table for simplicity
  (dolist (sit-lhs *situation-lh-table*)
    (when (cdr sit-lhs) ;;if there are transitions
      (let* ((sit (car sit-lhs))
	     (entry (cons sit nil))
	     (newentries nil))
	(multiple-value-bind (base-env ineq-assumptions)
	    (partition-situation-assumptions sit)
	  (dolist (lh (cdr sit-lhs))
	    (setq newentries 
		  (find-changed-situation sit base-env ineq-assumptions lh))
	    (when newentries
	      (setf (cdr entry) (nconc (mapcar #'(lambda (nsit)
						   (cons lh nsit))
					       newentries) (cdr entry))))))
	  (push entry *next-state-table*))))
  *next-state-table*)

(proclaim '(special *base-env* *ineq-asns* *omega*))

(defun debug-next-state (env lh)
  (multiple-value-bind (base-env ineq-asns)
      (partition-situation-assumptions env)
    (setq *base-env* base-env *ineq-asns* ineq-asns
	  *omega* (set-difference ineq-asns (lh-old-assumptions lh)))
    (format t "~%    [[qpe::*base-env* and qpe::*ineq-asns* hold intermediate results]]")
    (find-changed-situation env base-env ineq-asns lh)))

;;;; Partitioning situation assumptions
;; This procedure decomposes the assumptions which comprise a situation env
;; according to whether or not they are inequalities or not.

(defun partition-situation-assumptions (env &aux
					    base-env base-assumptions
					    ineq-assumptions)
  ;; All assumptions about infinitesimal differences (i>, i<) must be discarded, since
  ;; they are computed for each situation anew.  All statements which are not inequalities
   ;; must be cached, since they will remain constant across the transition (computable PC's
  ;; are never part of the situation env, by definition every situation must imply either it
  ;; or its negation.).  
  (dolist (as (tms::env-assumptions env))
    (let ((asn-form (adb::view-assumption as)))
    (case (caar asn-form)
      ((UNRELATED GREATER-THAN LESS-THAN EQUAL-TO)
       (unless (or (derivative? (cadar asn-form))
		   (derivative? (caddar asn-form)))
	 ;; Inherit inequalities between amounts, but not if derivatives are involved.
	 (push as ineq-assumptions)))
      ((i> i<)) ;; Throw them away.
      (t (push as base-assumptions)))))
  (adb::if-debug-flag limit-details
   (format t "~%     LA: Partitioning ~A." env))
  (adb::if-debug-flag limit-details
  (format t "~%     LA: Base assumptions are:")
  (dolist (ba base-assumptions)
    (format t "~%           ~A" (adb::lprinc-string (view-assumption ba))))
  (format t "~%     LA: Inequality assumptions are:")
  (dolist (in ineq-assumptions)
    (format t "~%     ~A" (adb::lprinc-string (view-assumption in)))))
  (setq base-env (tms::find-or-make-env base-assumptions))
  (adb::if-debug-flag Limit-details
    (format t "~%     LA: For ~A:" env)
    (format t "~%          Base = ~A" base-env)
    (format t "~%          Inequality assumptions:")
    (dolist (asn ineq-assumptions)
      (format t "~%           ~A" (adb::lprinc-string (view-assumption asn)))))
  (values base-env ineq-assumptions))

;; A common operation in the next-state computation is fetching 
;; related inequality assumptions.  These hooks into the INEQ structs
;; and ineq-asn-table simplify this process.

(defun ineq-asn-ineq (asn) (gethash asn *ineq-asn-table*)) ;; May be unnecessary

(defun ineq-asn-bt (asn) (ineq-bt-a (gethash asn *ineq-asn-table*)))
(defun ineq-asn-lt (asn) (ineq-lt-a (gethash asn *ineq-asn-table*)))
(defun ineq-asn-eq (asn) (ineq-eq-a (gethash asn *ineq-asn-table*)))
(defun ineq-asn-gt (asn) (ineq-gt-a (gethash asn *ineq-asn-table*)))

;;;; Determines what situation can arise next, if any.

(defun find-changed-situation (original-env base-env ineq-assumptions lh)
  (let ((solutions nil) (new-env nil)
	(omega (set-difference ineq-assumptions (lh-old-assumptions lh))))
    ;; Iterate over the assumptions that suffice to ensure the LH's conditions.
    (dolist (end-env (lh-end-envs lh))
      (setq new-env (tms::generic-union-env base-env end-env))
      (cond ((or (null new-env) (tms::generic-env-contradictory new-env))
	     )
	    (t (dolist (new-solution 
			 (filter-by-omega (fetch-situations-subset-of new-env)
						omega))
		 (pushnew new-solution solutions)))))
    (adb::if-debug-flag limit-details
     (format t "~%    NS: ~A on ~A has ~D possible extensions." lh original-env
	     (length solutions)))
    solutions))

(defun old-find-changed-situation (original-env base-env ineq-assumptions lh)
  (let ((nenv base-env) (solutions nil))
    ;; Add in the new assumptions.  These are mandatory.
    (adb::if-debug-flag limit-details
      (format t "~%     NS: Examining ~A from ~A." lh original-env))
    (dolist (as (lh-new-assumptions lh)) (setq nenv (tms::cons-env nenv as)))
    (adb::if-debug-flag limit-details
      (format t "~%     NS: After adding in change assumptions, ~A is ~A." base-env
	   nenv))
    ;;; If contradictory at this point, we lose.
    (cond ((tms::env-contradictory nenv)
	   (adb::if-debug-flag limit-details
	     (format t "~%    NS: ~A not consistent from ~A." (lh-name lh) base-env))
	   nil)
	  (t ;;; Otherwise fetch candidate situations from index
	   (setq solutions (filter-by-omega (fetch-situations-subset-of nenv)
					    (set-difference ineq-assumptions
							    (lh-old-assumptions lh))))
	   (adb::if-debug-flag limit-details
	     (format t "~%    NS: ~A on ~A has ~D possible extensions." lh original-env
		     (length solutions)))
	   solutions))))

(defun filter-by-omega (situations stable-asns &aux form ineq)
  ;; Do a deKleerism for speed (and keep your fingers crossed)
  (mapl #'(lambda (sits)
	    (unless (every #'(lambda (asn)
	       (setq form (car (adb::view-assumption asn)))
	       (or (eq (car form) 'UNRELATED) ;; was unrelated before
		   (eq (inequality-under      ;; is unrelated now
			 (setq ineq (gethash asn *ineq-asn-table*))
			 (car sits)) ':UNRELATED)
		   (and *allow-discontinuous-directs* ;; JWC
			(involves-net-influence? form)) 
		   (member asn (tms::env-assumptions (car sits))))) ;; holds unchanged
			       stable-asns)
	      (setf (car sits) nil)))
	situations)
  (delete nil situations))

(defun involves-net-influence? (form &aux temp)
  (and (setq temp (cdr form))
       (listp temp)
       (setq temp (car temp))
       (listp temp)
       (setq temp (cdr temp))
       (listp temp)
       (setq temp (car temp))
       (listp temp)
       (eq (car temp) 'NET-INFLUENCE)))       

(defun check-omega-on (after omega &optional (stream *standard-output*))
  (dolist (asn omega)
    (unless (or (eq (caar (adb::view-assumption asn)) 'UNRELATED)
		(eq ':UNRELATED (inequality-under (gethash asn *ineq-asn-table*)
						  after))
		(member asn (tms::env-assumptions after))
		(involves-net-influence? (car (adb::view-assumption asn))))
      (format stream "~%~A is actually ~A."
	      (adb::lprinc-string (adb::view-assumption asn))
	      (inequality-under (gethash asn *ineq-asn-table*) after)))))

;;; Pruning by closeness

(defun prune-via-closest (&aux best-solutions new-table)
  (setq new-table nil)
  (dolist (entry *next-state-table*)
    (let* ((old (car entry))
	   (new-entry (cons old nil))
	   (lh-entries (sort-by-lh-entries (cdr entry))))
      (dolist (lh-entry lh-entries) ;; each lh-entry is (<lh> . <list of next states>)
	(cond ((null (cddr lh-entry))
	       (if-debug-flag limit-details
			      (format t "~%  NS: Unique next state for ~A occuring at ~A: ~A."
				      (car lh-entry) old (cadr lh-entry)))
	       (push (cons (car lh-entry) (cadr lh-entry)) (cdr new-entry)))
	      (t ;; Otherwise, multiple potential solutions.
	       (if-debug-flag limit-details
		(format T "~%  NS: Possible solution set for ~A at ~A: ~A."
			(car lh-entry) old (cdr lh-entry)))
	       (setq best-solutions (pick-closest-results old (cdr lh-entry)))
	       (if-debug-flag limit-details
		 (format T "~%  NS: Minimal solutions for ~A at ~A: ~A"
			 (car lh-entry) old best-solutions))
	       (When-careful 
		 (when (set-difference (cdr lh-entry) best-solutions)
		   (Add-Recorder-Event :Closeness-Filter-Results old (car lh-entry)
				       (cdr lh-entry) best-solutions)))
	       (dolist (next best-solutions)
		 (push (cons (car lh-entry) next) (cdr new-entry))))))
      (push new-entry new-table)))
  new-table)

(defun sort-by-lh-entries (lh-alist &aux table entry)
  (dolist (simple-entry lh-alist table)
    (setq entry (assoc (car simple-entry) table))
    (unless entry
      (setq entry (cons (car simple-entry) nil))
      (push entry table))
    (push (cdr simple-entry) (cdr entry))))

;;;; Indexing situations by assumptions

;;;; ****** Change the inspector to use this table

(defun fetch-situations-subset-of (env)
  (fetch-situations-containing (tms::generic-env-assumptions env)))

(defun fetch-situations-containing (asns &aux index result)
  (setq result :first)
  (dolist (asn asns result)
    (unless (ignorable-situation-assumption? asn)
      (setq index (fetch-situations-under-asn asn))
      (cond ((eq result :first) (setq result (copy-list index)))
	    ((setq result (adb::order-preserving-intersection-nuke1
			    result index :ordering #'tms::env-orderp
			    :test #'eq)))
	    (t (return-from FETCH-SITUATIONS-CONTAINING nil))))))


;;; A special-case optimization would be to not use preconditions in indexing.
;;; However, the wave of the future lies in explicit modeling assumptions, so
;;; making this distinction would at best be a temporary speed-up. 

(defun ignorable-situation-assumption? (asn)
  (let ((form (adb::view-assumption asn)))
    (and (listp form)
	 (equal '(Enforce quantity-existence) (car form)))))

(defvar *sit-asn-table* nil)

(adb:init-form (cond (*sit-asn-table* (clrhash *sit-asn-table*))
		     (t (setq *sit-asn-table* 
			      (make-hash-table :test #'eq :size 200)))))

(defun fetch-situations-under-asn (asn) ;; ******* macroize this later
  (gethash asn *sit-asn-table* nil))

(defun index-situations ()
  (clrhash *sit-asn-table*) ;; Be cautious
  (map-situations #'(lambda (sit)
		      (dolist (asn (tms:env-assumptions sit))
			(unless (ignorable-situation-assumption? asn)
			  ;; Situations are unique, so push will do. 
			  (setf (gethash asn *sit-asn-table*)
				(push sit (gethash asn *sit-asn-table*)))))))
  ;; Now sort entries for rapid intersection tests
  (maphash #'(lambda (key val) (setf (gethash key *sit-asn-table*)
				     (sort val  #'tms::env-orderp)))
	   *sit-asn-table*))

;;;; Picking "closest" solution

(defun pick-closest-results (original nexts &aux max count solutions old-asns)
  ;; The "closest" solution is the one with the most assumptions
  ;; in common with the old situation.
  (setq max 0 old-asns (tms::env-assumptions original))
  (dolist (env nexts)
    (setq count 0)
    (dolist (a (tms::env-assumptions env))
      (if (member a old-asns) (incf count)))
    (cond ((> count max)
	   (When-Careful		  
	     (when (> max 0)
	       (Add-Recorder-Event :Non-Close-Solution original
				   max solutions count env
				   original)
	       ;(format t "~% Found non-close solution in pick-closest-results:")
	       ;(format t "~% Old count = ~A, old solutions = ~A" max solutions)
	          ;(format t "~% New count = ~A, new solution = ~A" count env)
	       ))
	   (setq max count solutions (list env)))
	  ((= count max) (push env solutions))
	  (t ;; (< count max) 
	   (When-Careful		  
	     (when (> max 0)
	       (Add-Recorder-Event :Non-Close-Solution original
				   max solutions count env
				   original)
	       ;(format t "~% Found non-close solution in pick-closest-results:")
	       ;(format t "~% Old count = ~A, old solutions = ~A" max solutions)
	       ;(format t "~% New count = ~A, new solution = ~A" count env)
	       )))))
  solutions)

(defun inequality-asn? (asn) (member (caar (adb::view-assumption asn))
				     '(GREATER-THAN LESS-THAN EQUAL-TO)))

;;;; Debugging
(defun explain-next-states (&optional (break? t) &aux fstring)
  (setq fstring "~%Type any character to continue...")
  (dolist (entry *next-state-table*)
    (print-sclass (tms::get-env-prop (car entry) 'sclass))
    (cond ((or (not break?) (y-or-n-p "Show env?"))
	   (adb::print-env (car entry))))
    (when break?
      (format t fstring)
      (read-char))
    (cond ((null (cdr entry))
	   (format t "~% No limit hypotheses applicable to ~A" (car entry)))
	  (t ;there are some limit hypotheses.
	   (format t "~% There are ~D limit hypotheses for ~A."
		   (length (cdr entry))
		   (car entry))
	     (dolist (pair (cdr entry))
	       (print-lh (car pair))
	       (format t "~%   The result is ~A." (cdr pair))
	       (print-sclass (tms::get-env-prop (cdr pair) 'sclass))
	       (cond ((or (not break?)
			  (y-or-n-p "Show env?"))
		      (adb::print-env (cdr pair))))
	       (when break? 
		 (format t fstring)
		 (read-char)))))
    (when break?
      (format t "~%Any character to continue...")
      (read-char))))

(proclaim '(special *next-sclass-table*))

(defun make-next-sclass-table ()
  (setq *next-sclass-table*
	(mapcar #'(lambda (entry)
		  (cons (env-sclass (car entry))
			(mapcar #'(lambda (pair)
				  (cons (car pair) (env-sclass (cdr pair))))
				(cdr entry))))
		*next-state-table*)))

;;;; Continuity filtering
;;
(defvar *type-a-ineqs* nil) ;; list of (<ineq for A's> . <ineq for D's>)
(defvar *other-ineqs* nil) ;; Everything else

(adb::Init-form (setq *type-a-ineqs* nil))
(adb::Init-form (setq *other-ineqs* nil))

(defun prune-via-continuity (&optional (table *next-state-table*)
			      &aux start-env legals)
;;; Performs Ds, Qstate, and zero existence continuity tests.
;;; First, split up the ineqs by cases for
  (partition-ineqs-for-continuity)
  (dolist (entry table)
    (setq start-env (car entry) legals nil)
    (dolist (subentry (cdr entry))
      (multiple-value-bind (okay? reason datum)
	  (continuity-check start-env (cdr subentry))
	(when okay? (push subentry legals))
	(adb::if-debug-flag continuity
	 (if okay? (format t "~%    CONT: O.K.: ~A to ~A."
			   start-env (cdr subentry))
	     (format t "~%    CONT: Failed: ~A to ~A: ~A, ~A."
		     start-env (cdr subentry) reason datum)))))
    (setf (cdr entry) (nreverse legals))))

(defun continuity-check (start end)
  (let ((vanished! (find-vanished-individuals start end))
	(appeared! (find-vanished-individuals end start)))
    (when vanished!
      (unless (check-zero-existence-continuity start end vanished! t)
	(When-Careful 
	  (Add-Recorder-Event :CONTINUITY-PRUNING :ZERO-EXISTENCE
			      start end vanished!))
	(return-from continuity-check (values nil :ZERO vanished!))))
    (when nil appeared! ;; ***** Re-install and experiment.
      (unless (check-zero-existence-continuity start end appeared! nil)
	(When-Careful 
	  (Add-Recorder-Event :CONTINUITY-PRUNING :ZERO-EXISTENCE
			      start end appeared!))
	(return-from continuity-check (values nil :ZERO appeared!))))
    (continuity-satisfied? start end)))

;;;; Sorting ineq's for speedy continuity checking
;; Every scrap of inequality information is scrutinized.

(defun partition-ineqs-for-continuity (&aux other d-n2)
  (setq *type-a-ineqs* nil)
  (dolist (ineq *ineqs*)
    (cond ((eq (ineq-type ineq) ':A)
	   (setq d-n2 (quant-derivative (qnum-quantity (ineq-n2 ineq)))))
	  ((and (eq (ineq-type ineq) ':Special)
		(eq (qnum-role (ineq-n1 ineq)) ':A))
	   (setq d-n2 (ineq-n2 ineq)))
	  (t (setq d-n2 nil)))
    (when (and d-n2
	       (setq other
		     (grab-ineq 
		       (quant-derivative (qnum-quantity (ineq-n1 ineq)))
		       d-n2)))
      (setf (ineq-other ineq) other)
      (setf (ineq-other other) ineq)
      (push ineq *type-a-ineqs*)))
  (setq *other-ineqs* (remove-if #'ineq-other *ineqs*)))

(defun continuity-satisfied? (start end)
  (dolist (ineq *type-a-ineqs*)
    (unless (or (and *allow-discontinuous-directs*
		     (net-influence-ineq? ineq))
		(type-a-consistent? (inequality-under ineq start)
				    (inequality-under (ineq-other ineq) start)
				    (inequality-under ineq end)
				    (inequality-under (ineq-other ineq) end)))
      (When-Careful
	(Add-Recorder-Event :CONTINUITY-PRUNING :TYPE-A start end ineq))
      (return-from continuity-satisfied? (values nil :TYPE-A ineq))))
  (dolist (ineq *other-ineqs* t)
    (unless (or (and *allow-discontinuous-directs*
		     (net-influence-ineq? ineq))
		(continuous-relations? (inequality-under ineq start)
				       (inequality-under ineq end)))
      (When-Careful
	(Add-Recorder-Event :CONTINUITY-PRUNING :TYPE-B start end ineq))
     (return-from continuity-satisfied? (values nil :TYPE-B ineq)))))

(defun inequality-under (ineq env)
  (cond ((tms::generic-true-in? (ineq-< ineq) env) ':Less-than)
	((tms::generic-true-in? (ineq-= ineq) env) ':Equal-to)
	((tms::generic-true-in? (ineq-> ineq) env) ':Greater-than)
	((tms::generic-true-in? (ineq-bt ineq) env) ':Unrelated)
	(t ':UNKNOWN)))

(defun opposite-relation (rel)
  (case rel
    (:Less-than ':Greater-than)
    (:Greater-than ':less-than)
    (:equal-to ':equal-to)
    (:> ':<)
    (:= ':=)
    (:< ':>)
    (:>= ':<=)
    (:<= ':>=)
    (t (error "~% ~A not a valid inequality relationship." rel))))

(defun unavailable? (rel) (or (eq rel ':unknown) (eq rel ':unrelated)))

(defun continuous-relations? (old new)
  (or (unavailable? old) (unavailable? new)
      (eq old ':equal-to) (eq new ':equal-to)
      (eq old new)))

(defun type-a-consistent? (old-a old-d new-a new-d)
  (unless (continuous-relations? old-a new-a)
    (return-from type-a-consistent? nil))
  (unless (continuous-relations? old-d new-d)
    (return-from type-a-consistent? nil))
  (when (eq new-a ':equal-to)
    (unless (or (unavailable? old-d)
		(unavailable? old-a))
      (unless (eq old-d (opposite-relation old-a))
	(return-from type-a-consistent? nil))))
  (when (eq old-a ':equal-to)
    (unless (unavailable? new-a)
      (cond (*allow-unstable-trans*
	     (unless (or (unavailable? new-d)
			 (and *allow-coincidences*
			      (eq new-d ':equal-to)
			      (not (eq old-d ':equal-to))
			      (not (eq old-d (opposite-relation new-a))))
			 (eq new-d new-a))
	       (return-from type-a-consistent? nil)))
	     (t (unless (or (unavailable? old-d)
			    (eq old-d new-a))
		  (return-from type-a-consistent? nil))
		(unless (or *Allow-Coincidences*
			  (eq new-a ':equal-to)
			  (eq new-a new-d))
		  (return-from type-a-consistent? nil))))))
  t)

;;;;; Debugging continuity code

(defun tabulate-type-a-cases (&optional (noprint nil) (stream *standard-output*))
  (let ((cases '((:Less-than . "<") (:Equal-to . "=") (:Greater-than . ">")))
	(count 0) (answer (list "" "" "" ""))
	(counters (list 0 0 0 0)))
;    (format stream "~%Unstable transitions ~A."
;	    (if *allow-unstable-trans* "allowed" "forbidden"))
;    (format stream "~%Coincidences ~A."
;	    (if *allow-coincidences* "allowed" "forbidden"))
    (cond (noprint
      (format stream 
	  "~%~6:< ~>~28:<Cases:~>~7:<-Unst~>~7:<-Unst~>~7:<+Unst~>~7:<+Unst~>")
      (format stream 
	  "~%~6:< ~>~28:< ~>~7:<+Coin~>~7:<-Coin~>~7:<+Coin~>~7:<-Coin~>"))
	  (t 
      (format stream 
  "~%~6:< ~>~7:<Old A~>~7:<Old D~>~7:<New A~>~7:<New D~>~7:<-Unst~>~7:<-Unst~>~7:<+Unst~>~7:<+Unst~>")
      (format stream 
	  "~%~6:< ~>~28:< ~>~7:<+Coin~>~7:<-Coin~>~7:<+Coin~>~7:<-Coin~>")))
  (dolist (Old-A cases)
    (dolist (Old-D cases)
      (dolist (New-A cases)
	(dolist (New-D cases)
	  (incf count)
	  (dotimes (i 4)
	    (let ((*allow-coincidences* (evenp i))
		  (*allow-unstable-trans* (> i 1)))
		(cond ((type-a-consistent? 
			 (car old-a) (car old-d) (car new-a) (car new-d))
		       (incf (nth i counters))
		       (setf (nth i answer) "Yes"))
		      (t (setf (nth i answer) "No")))))
	  (unless noprint
	    (format stream
    "~%~6:< ~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>"
		    (cdr old-a) (cdr old-d) (cdr new-a) (cdr new-d)
		    (car answer) (cadr answer) (caddr answer) (cadddr answer)))))))
  (format stream
	  "~%~6:< ~>~28:<~A cases, consistent:~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>"
	  count (car counters) (cadr counters)
	  (caddr counters) (cadddr counters))))

(defun explain-continuity-event (event &optional (stream *standard-output*))
  ;; format of event is 
  ;;(:continuity-pruning <:type-A or :type-B>
  ;;		       <start> <end> <ineq>)
  (unless (eq :continuity-pruning (car event)) ;; Skip it
    (return-from explain-continuity-event nil))
  (let ((start (third event))
	(end (fourth event))
	(ineq (fifth event)))
    (format stream "~%Transition from ~A to ~A ruled out by ~A continuity test,~%    ~A"
	    start end (cadr event) ineq)
    (case (cadr event)
      (:TYPE-A 
	(format stream "~%~25:<~A~>~20:<~A~>" start end)
	(format stream "~%~5:<A:~>~20:<~A~>~20:<~A~>" (inequality-under ineq start)
		(inequality-under ineq end))
	(format stream "~%~5:<D:~>~20:<~A~>~20:<~A~>"
		(inequality-under (ineq-other ineq) start) 
		(inequality-under (ineq-other ineq) end)))
      (:TYPE-B 
	(format stream "~%~25:<~A~>~20:<~A~>" start end)
	(format stream "~%~5:<~>~20:<~A~>~20:<~A~>" (inequality-under ineq start)
		(inequality-under ineq end))))))

(defun compare-type-a-relations (start end &optional (stream *standard-output*))
  (format stream
	  "~%  ~40:<N1~>~40:<N2~>~7:<Old A~>~7:<Old D~>~7:<New A~>~7:<New D~>~7:<Okay?~>")
  (dolist (ineq *type-a-ineqs*)
    (format stream
       "~%  ~40:<~A~>~40:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>"
       (ineq-n1 ineq) (ineq-n2 ineq)
       (pretty-ineq-name (inequality-under ineq start))
       (pretty-ineq-name (inequality-under (ineq-other ineq) start))
       (pretty-ineq-name (inequality-under ineq end))
       (pretty-ineq-name (inequality-under (ineq-other ineq) end))
       (type-a-consistent? (inequality-under ineq start)
			   (inequality-under (ineq-other ineq) start)
			   (inequality-under ineq end)
			   (inequality-under (ineq-other ineq) end)))))

(defun compare-other-relations (start end &optional (stream *standard-output*))
    (format stream
	  "~%  ~40:<N1~>~40:<N2~>~7:<Old~>~7:<New~>~7:<Okay?~>")
  (dolist (ineq *other-ineqs*)
    (format stream
       "~%  ~40:<~A~>~40:<~A~>~7:<~A~>~7:<~A~>~7:<~A~>"
       (ineq-n1 ineq) (ineq-n2 ineq)
       (pretty-ineq-name (inequality-under ineq start))
       (pretty-ineq-name (inequality-under ineq end))
       (continuous-relations? (inequality-under ineq start)
			      (inequality-under ineq end)))))

(defun pretty-ineq-name (rel)
  (case rel
    (:LESS-THAN "<")
    (:GREATER-THAN ">")
    (:EQUAL-TO "=")
    (:UNKNOWN "?")
    (:UNRELATED "!")
    (t "barf")))

(defun situation-inequalities (sit &optional (stream *standard-output*))
  (format stream
	  "~%  ~40:<N1~>~40:<N2~>~7:<A~>~7:<D~>")
  (dolist (ineq *type-a-ineqs*)
    (format stream
       "~%  ~40:<~A~>~40:<~A~>~7:<~A~>~7:<~A~>"
       (ineq-n1 ineq) (ineq-n2 ineq)
       (pretty-ineq-name (inequality-under ineq sit))
       (pretty-ineq-name (inequality-under (ineq-other ineq) sit))))
  (format stream
	  "~%  ~40:<N1~>~40:<N2~>~7:<Relation~>")
  (dolist (ineq *other-ineqs*)
    (format stream
       "~%  ~40:<~A~>~40:<~A~>~7:<~A~>"
       (ineq-n1 ineq) (ineq-n2 ineq)
       (pretty-ineq-name (inequality-under ineq sit)))))

(proclaim '(special *from* *to*))

(defun debug-compare-situations (s1 s2)
  (setq *from* (car (sclass-assumptions s1)))
  (setq *to* (car (sclass-assumptions s2))))

;;;; "Zero existence" case of continuity checking.
;;
;; This code captures the reasoning that one uses when thinking about
;; a situation at 0- or 0+.  In essense, we don't enforce the Quantity-Existence
;; law, and thus allow things to get mapped across as they would be otherwise.

;;; ****** Jcollins claims that, for symmetry, this should be done when
;;; ******  individuals appear as well.

(defun check-zero-existence-continuity (start-env end-env changed forward?
					&aux benv to-replace to-keep)
  (adb::if-debug-flag zero-existence
    (format t "~%     ZE: Checking ~A --> ~A, since ~A ~A."
	    start-env end-env changed (if forward? "vanished" "appeared")))
  ;;; First subtract out the UNRELATED assumptions that include the vanished individual.
  (dolist (as (tms::env-assumptions (if forward? end-env start-env)))
    (let ((form (adb::view-assumption as)))
      (cond ((or (and (listp (car form))
		      (eq (caar form) 'UNRELATED))
		  (and (listp (car form))
		       (involves-net-influence? (car form)))) ;; New relationship needed
	     (push as to-replace))
	    (t ;; This one should be kept.
	     (push as to-keep)))))
  (adb::if-debug-flag zero-existence 
	  (format t "~%     ZE: Keeping ~D assumptions:" (length to-keep))
	  (dolist (asn to-keep)
	    (format t "~%        ~A" (adb::lprinc-string (adb::view-assumption asn)))))
  (adb::if-debug-flag zero-existence 
	  (format t "~%     ZE: Replacing ~D assumptions:" (length to-replace))
	  (dolist (asn to-replace)
	    (format t "~%        ~A" (adb::lprinc-string (adb::view-assumption asn)))))
  (setq benv (subtract-QE-enforcement (tms::find-or-make-env to-keep)))
  (adb::if-debug-flag zero-existence
		 (format t "~%     ZE: Basic environment is ~A" benv))
  ;;; Now we have to figure out what inequalities will hold.  These aren't
  ;;; legal situations, so we can't just look them up.
  ;;; Unclear what it means for there to be multiple solutions in this case,
  ;;; but we'll call it legal if ANY solution satisfies continuity.
  ;;; Use depth-first construction for efficiency.
  (when benv
    (find-continuous-extension benv (if forward? start-env end-env) forward?
			     (mapcar #'(lambda (asn)
					 (let ((ineq (ineq-asn-ineq asn))
					       (choices nil))
					   (when (ineq-bt-a ineq)
					     (push (ineq-bt-a ineq) choices))
					   (when (ineq-lt-a ineq)
					     (push (ineq-lt-a ineq) choices))
					   (when (ineq-eq-a ineq)
					     (push (ineq-eq-a ineq) choices))
					   (when (ineq-gt-a ineq)
					     (push (ineq-gt-a ineq) choices))
					   choices)) to-replace))))

;;;; Helpers for zero existence continuity testing

(defun subtract-QE-enforcement (env)
  (tms::make-generic-env 
    (cons (adb::fact-assumption '((Enforce Quantity-Existence) . :FALSE))
	  (remove (adb::fact-assumption '((Enforce Quantity-Existence) . :TRUE))
		  (tms::env-assumptions env)))))

(defun fetch-corresponding-inequality-assumptions (asns env)
  ;; Finds which assumptions of env involve the same inequalities mentioned in asns.
  (mapcar #'(lambda (asn)
	      (find (ineq-asn-ineq asn) ;; the number pair
		    (tms::env-assumptions env)
		    :test #'equal
		    :key #'(lambda (oasn) (ineq-asn-ineq oasn))))
	  asns))

(defun find-vanished-individuals (start-env end-env &aux end-individuals Vanished!)
  (unless (env-sclass start-env) (error   "~%~A not a situation." start-env))
  (unless (env-sclass end-env) (error   "~%~A not a situation." end-env))
  (setq end-individuals (sclass-individuals (env-sclass end-env)))
  (dolist (old (sclass-individuals (env-sclass start-env)))
    (unless (member old end-individuals :test #'equal) (push old vanished!)))
  vanished!)

(defun find-continuous-extension (solution fixed-env forward? choice-sets &aux new)
  (cond ((null choice-sets)
	 (adb::if-debug-flag zero-existence
           (format t "~%          ZE: Possible extension: ~A." solution))
	 (if forward? (multiple-value-bind (okay? type what)
			  (continuity-satisfied? fixed-env solution)
			(cond (okay? 
			       (adb::if-debug-flag zero-existence
				(format t "~%  ..Satisfies continuity."))
			       solution)
			      (t (adb::if-debug-flag zero-existence
				   (format t "  ..Failed: ~A on ~A." type what))
				 nil)))
	     (multiple-value-bind (okay? type what)
			  (continuity-satisfied? solution fixed-env)
			(cond (okay? 
			       (adb::if-debug-flag zero-existence
				(format t "~%     ZE: Satisfying continuity: ~A" solution))
			       solution)
			      (t (adb::if-debug-flag zero-existence
				   (format t "  Failed: ~A on ~A." type what))
				 nil)))))
	(t (dolist (choice (car choice-sets))
	     (setq new (tms::generic-cons-env solution choice))
	     (unless (tms::generic-env-contradictory new)
	       (setq new (find-continuous-extension new fixed-env forward?
						    (cdr choice-sets)))
	       (when new (return-from FIND-CONTINUOUS-EXTENSION new)))))))

;;;; Equality Change Law
;;
;;Case 1 -- Limit hypotheses that represent changes FROM equality
;;          occur instantly.
;;Case 2 -- Limit hypotheses that represent changes TO equality between
;;          quantities that were influenced away from equality for only an
;;          instant also occur instantly.
;;Otherwise, all changes take an interval of time.

;;; Some implications are:
;; 1. If a Case 1 transition is possible from a state, then that state can only
;;    last an instant.
;; 2. If a Case 2 transition is possible, then that state may split into two, since
;;    any state is potentially an initial state.  One state will last an instant, the other
;;    will last for an interval, unless a Case 1 transition is possible from it.
;; 3. Any limit hypothesis which does not contain the maximal set of ECL hypotheses
;;    for that state cannot occur.

;; Algorithm:
;; 1. For each limit hypothesis, mark the subset of changes it predicts which
;;    satisfy Case 1 ECL.  Filter transitions which do not contain the maximal subset.
;; 2. For each situation, check if any of its limit hypotheses involve Case 1 ECL.
;;    If so, add it to a queue of states to check for Case 2 ECL.
;; 3. For each queued element, let NEXT = state reached, LAST = state before.
;;    Call a comparison between numbers for which ECL occurs an "ECL  changed pair".
;;      3.1 Find the maximal set(s) of ECL changed pair.
;;      3.2 For each maximal set (typically one),
;;          3.2.1 Clone the sclass.
;;          3.2.2 Update state transition pointers to it.
   
(proclaim '(special *next-state-extensions*))

(defvar *next-state-extensions* nil
  "States created by Case 1 of Equality Change Law")
(adb::init-form (setq *next-state-extensions* nil)) 

(defun apply-equality-change-law ()
  (mark-case-1-limit-hypotheses) ;; Detectable locally
  (apply-case-1-conclusions) ;; Prune next-state table accordingly.
  (setq *next-state-extensions* nil)
  (when *Case2-ECL-ON*
    (apply-case-2-ECL)
      ;; This will clone sclasses as needed, adding them to *next-state-extensions*
    (setq *next-state-table* (nconc *next-state-table* *next-state-extensions*))))

;;;;  Applying case 1 ECL

(defun mark-case-1-limit-hypotheses ()
;; Limit hypotheses satisfying ECL Case 1 can be recognized locally.
  (dolist (lh *limit-hypotheses*)
    (multiple-value-bind (Case-1? number-pairs) (case-1-lh? lh)
      (when Case-1?
	(adb::if-debug-flag ECL (format t "~%     ECL: ~A satisfies Case 1" lh))
	;; Cache information needed for computing case2 transitions.
	(put-lh-prop lh number-pairs ':CASE1-ASSUMPTIONS)))))

(defun case-1-lh? (lh &aux infs)
  (dolist (entry (lh-relations lh))
    ;; Recall the entries are (<ineq> <old> <new>)
    (when (eq (cadr entry) 'EQUAL-TO) ; got one
      (push (cons (car entry) (case (caddr entry)
				(LESS-THAN 'i<) (GREATER-THAN 'i>))) infs)))
  (if infs (values t infs) (values nil nil)))

(defun apply-case-1-conclusions ()
  ;; Invert the case1-assumptions associated with the lh's for each state,
  ;; and keep only those lh's which include the maximal set(s).
  (dolist (entry *next-state-table*)
    (let ((C1-asn-sets nil) (cache nil)
	  (c1 nil) (centry nil))
      (When-Careful (setq cache (copy-tree (cdr entry)))) ;; For debugging
      (dolist (pair (cdr entry))
	(when (setq c1 (get-lh-prop (car pair) ':case1-assumptions))
	  (setq centry (assoc c1 c1-asn-sets :test #'equal))
	  (unless centry (setq centry (cons c1 nil)) (push centry c1-asn-sets))
	  (setf (cdr centry) (cons pair (cdr centry)))))
      (when c1-asn-sets ;; There are some case1 hypotheses
	;; This next step could be sped up by sorting them, if necessary.  Probably not
	;; worth it.
	(setq c1-asn-sets (remove-if #'(lambda (e1)
					 (some #'(lambda (e2)
						   (proper-subset? (car e1) 
								   (car e2)))
					       c1-asn-sets)) c1-asn-sets))
	;; Cache these sets to simplify case2 code.
	(tms::put-env-prop (car entry) (copy-list c1-asn-sets) ':Case1-sets)
	(if-debug-flag ECL (format t "~%     ECL: Maximal sets for ~A = ~A."
				   (car entry) (mapcar #'car c1-asn-sets)))
	(setf (cdr entry) (reduce #'(lambda (s1 s2) (nunion s1 s2 :test #'equal))
				  (map 'list #'cdr c1-asn-sets))))
      (when-careful
	(dolist (centry cache)
	  (unless (assoc (car centry) (cdr entry))
	    (Add-Recorder-Event :LH-FLUSHED :ECL-CASE-1
				(car entry) (car centry) (cdr centry))))))))

(defun proper-subset? (l1 l2) 
  (and (subsetp l1 l2 :test #'equal) (not (subsetp l2 l1 :test #'equal))))


;;;; Applying case 2 E.C.L.
;; 

(defun apply-case-2-ECL ()
  (dolist (entry (invert-case-1-pointers))
    ; format = (<env> . (<case1 asn's> . (<old env> . <lh>)))
    (let ((next-pointers (assoc (car entry) *next-state-table*))
	  (case1-changes (create-union-sets (mapcar #'car (cdr entry))))
	  (case2-lhs nil))
      (dolist (lh-pair (cdr next-pointers))
	(let ((case2-changes (case-2-lh? (car lh-pair) case1-changes))
	      (asn-entry nil))
	  (when case2-changes 
	    ;; Store the hypotheses indexed under the set of infinitesimals which must be
	    ;; introduced for the transition to occur.
	    (setq asn-entry (assoc case2-changes case2-lhs :test #'equal))
	    (unless asn-entry (setq asn-entry (cons case2-changes nil))
		    (push asn-entry case2-changes))
	    (push lh-pair (cdr asn-entry)))))
      (when case2-lhs
	(install-Case2-clones entry next-pointers case2-lhs)))))

(defun create-union-sets (sets &aux result)
  (dolist (set sets result)
    (dolist (el set) (pushnew el result :test #'equal))))  

;;;; Helpers for Case2 ECL

(defun Invert-Case-1-pointers (&aux case1-sets queue)
  ;; Being minimalist, we want to avoid installing i>, i< assumptions when we don't have to.
  ;; This means finding the states reached by a Case 1 transition
  ;;   (since those are states which potentially have a noticable infinitesimal difference),
  ;; and checking their lh's to see if any
  ;; involve number pairs that are very close.  The entries of QUEUE,
  ;; our result, take the form
  ;; (<env> . (<case1 asn's> . (<old env> . <lh>))) 
  (dolist (entry *next-state-table* queue)
    (when (setq case1-sets (tms::get-env-prop (car entry) ':Case1-sets))
      (dolist (sentry case1-sets)
	;; form is (<set of case1 asn's> . <pairs of form (lh . next)>)
	(dolist (next-env-pair (cdr sentry))
	  (let ((env-entry (assoc (cdr next-env-pair) queue)))
	    (unless env-entry  ;; Make it if needed
	      (push (setq env-entry (cons (cdr next-env-pair) nil)) queue))
	    (let ((set-entry (assoc (car sentry) (cdr env-entry) :test #'equal)))
	      (unless set-entry (push (setq set-entry (cons (car sentry) nil))
				      (cdr env-entry)))
	      (push (cons (car entry) (car next-env-pair)) (cdr set-entry)))))))))

(defun case-2-lh? (lh case1-list &aux case2-rels)
  ;; Checks to make sure the lh mentiones the appropriate inequalities.
  (dolist (rel (lh-relations lh)) ;; rel = (<ineq> <old-relationship> <new relationship>)
    (let ((case1 (assoc (car rel) case1-list)))
      (when (and case1 (case (cdr case1)
			 (I> (eq (cadr rel) 'Greater-Than))
			 (I< (eq (cadr rel) 'Less-Than))))
	(push rel case2-rels))))
  case2-rels)

;;;; Installing Case 2 E.C.L. results
;; 
;; Case 2 only applies to states which came about due to some
;; case 1 transition (yes, in principle one should put infinitesimal
;; differences in the qstate, but why be grotesque?).  Thus we need
;; to clone the Sclass involved, to express the additional assumptions.

(defun install-Case2-clones (case2-entry next-pointers case2-lhs)
  case2-lhs next-pointers case2-entry ;; fake out compiler
  ;; Here is where we clone the sclass.  Or will, when awake!  
  (break "Got some Case 2 LH's...Please record data and call Forbus!"))

;;;; Installing limit hypothesis information

;;; There are several fields of the SCLASS that must be filled in:
;;; LIMIT-HYPOTHESES, NEXT, and PREVIOUS. 
;;; 
;;; LIMIT-HYPOTHESES is an alist whose keys are envs and whose datums are the CAR's of
;;; the corresponding entry in the next-state table.
;;;
;;; NEXT and PREVIOUS should be alists whose keys are envs and whose datums are
;;; pairs of (<LH> <ENV>).
;;;

(defun install-duration-information ()
  (map-sclasses
    #'(lambda (s)
	(multiple-value-bind (instants intervals)
	    (split-sclass-duration s)
	  (cond (instants
		 (if intervals
		     (split-sclass-by-duration s instants intervals)
		 (setf (sclass-duration s) ':INSTANT)))
		(intervals
		 (setf (sclass-duration s) ':INTERVAL))
		(t (error   "~%  ~A has no situations." s))))))
  ;; Seperate step, to be indepenent of the mechanics of map-sclasses.
  (map-sclasses #'update-sclass-transition-pointers))

(defun update-sclass-transition-pointers (s)
(dolist (env (sclass-assumptions s))
  (let ((entry (assoc env *next-state-table*))
	(lhs nil))
    (when entry 
      (push (copy-list entry) (sclass-next s))
      (dolist (pair (cdr entry))
	(let ((psclass (env-sclass (cdr pair)))
	      (penv (cdr pair))
	      (lh (car pair))
	      (old nil))
	  (pushnew lh lhs)
	  (setq old (assoc penv (sclass-previous psclass)))
	  (cond (old (setf (cdr old) (cons (cons lh env) (cdr old))))
		(t (push (list penv (cons lh env))
			 (sclass-previous psclass))))))
      (push (cons env lhs) (sclass-limit-hypotheses s))))))

;;;; Splitting and cloning sclasses

(defun split-sclass-duration (sclass &aux instants intervals)
  (dolist (env (sclass-assumptions sclass))
    (cond ((or (tms::get-env-prop env ':Case1-sets)
	       (and *Case2-ECL-On*
		    (tms::get-env-prop env ':Case2-ECL)))
	   (push env instants))
	  (t (push env intervals))))
  (values instants intervals))

(defun split-sclass-by-duration (old instants intervals)
  (let ((clone (get-sclass-prop old ':CLONE))) ;; Avoid duplication
    (unless clone
      (setq clone (clone-sclass old t))
      (put-sclass-prop clone old ':CLONE-OF)
      (put-sclass-prop old clone ':CLONE))
    (When-Careful
	(Add-Recorder-Event :Duration-Sclass-Cloning
			    old clone instants intervals))
    (dolist (sit instants)
      (setf (sclass-assumptions old)
	    (delete sit (sclass-assumptions old)))
      (tms::put-env-prop sit clone 'sclass)
      (push sit (sclass-assumptions clone)))
    (setf (sclass-duration old) ':INTERVAL)
    (setf (sclass-duration clone) ':INSTANT)))

;;;; A statistical check

(defun classify-ineq-assumptions (&optional (stream *standard-output*) &aux total)
;; See how many assumptions actually end up being part of a valid situation.  
;; Those which don't can be ignored when building choice sets for temporal inheritance.
  (let ((asns-mentioned nil)
	(mentioned 0))
  (map-situations #'(lambda (sit)
		      (dolist (asn (tms::env-assumptions sit))
			(if (inequality-assumption? asn)
			    (pushnew asn asns-mentioned)))))
  (setq total (* 4 (length (gather-assumed-qstate)))
	mentioned (length asns-mentioned))
  (cond ((= total 0)
	 (format stream "~% No assumed qstate"))
	(t (format stream
	  "~% ~D assumptions mentioned out of ~D total (~5F%)"
	  mentioned total (* 100 (/ (float mentioned) total)))))))

(defun inequality-assumption? (asn)
  (let ((form (view-assumption asn)))
    (and (listp (car form))
	 (member (caar form) '(LESS-THAN GREATER-THAN EQUAL-TO
						  UNRELATED)))))

(defun unsimplify-lh-envs ()
  ;; Turns the simple envs into real envs for
  ;; printing and dumping.
  (dolist (lh *limit-hypotheses*)
    (setf (lh-end-envs lh) (mapcar #'tms::simple-to-env (lh-end-envs lh)))
    (setf (lh-start-envs lh)
	  (mapcar #'tms::simple-to-env (lh-start-envs lh)))))

(defun check-for-ambiguous-transitions (&aux result)
  (dolist (entry *next-state-table* result)
    (let ((start (car entry))
	  (alist nil))
      (dolist (pair (cdr entry))
	(let ((aentry (assoc (car pair) alist)))
	  (unless aentry
	    (push (setq aentry (cons (car pair) nil)) alist))
	  (push (cdr pair) (cdr aentry))))
      (dolist (lh-entry alist)
	(when (cddr lh-entry) ;; Got one
	  (push (cons start lh-entry) result))))))
 
;;;; More debugging

(defun explain-sclass-gap (start end)
  ;; When you think there should be SOME transition between
  ;; start and end, this function examines the flight recorder
  ;; to explain why not.
  (let* ((start-sits (sclass-assumptions start))
	 (end-sits (sclass-assumptions end))
	 (events (list-fr-events :continuity-pruning
			      #'(lambda (e)
				  (and (member (third e) start-sits)
				       (member (fourth e) end-sits))))))
    (format t "~%There were ~d potential transitions." (length events))
    (dolist (event events) (explain-continuity-event event))))

;;;; Pruning asymptotic approach states

;; Here's the idea: A state for which there are locally consistent
;; limit hypotheses, yet all of them are ruled out by consistency
;; criteria, represents an asymptotic approach state.  Such states
;; are illegal under QP theory, and hence should be pruned out.
;;
;; This can recursively kill states, of course, since some states
;; now do not transition to anything, even though they did previously.
;;  
;; To provide at least some information for explanations, each state killed
;; in this way will be marked with the reason for it's death.
;; Furthermore, transitions to dead states will be marked as dead on their
;; starting Sclass, so that an explanation for their death can be constructed.

;; ****** Do this much faster by using local information cached with the
;; ****** Sclass...

(defvar *dead-states* nil)
(adb::init-form (setq *dead-states* nil))

(defun asymptotic-approach? (sit &aux lhs)
  (and (setq lhs (cdr (assoc sit *situation-lh-table*)))
       (not (every #'(lambda (lh) (get-lh-prop lh :UNSTABLE?)) lhs))
       (null (assoc sit *next-state-table*))))

(defun prune-asymptotic-approach-states (&aux zombies)
  (map-situations #'(lambda (sit) (if (asymptotic-approach? sit) (push sit zombies))))
  ;; Install debug flag here.
  (do ((queue (cdr zombies) (nconc (cdr queue) newly-dead))
       (z (car zombies) (car queue))
       (newly-dead nil nil)
       (graveyard nil))
      ((null z) (flush-inconsistent-sclasses)
       graveyard)
    (push z graveyard)
    (setq newly-dead (kill-asymptotic-approach z))))

(defun kill-asymptotic-approach (sit &aux sclass losers newly-dead)
  ;; Don't need to remove it from the situation-lh table, since
  ;; that has happened by now.
  (when (env-sclass? sit) ;; Make sure isn't already dead.
    (setq sclass (env-sclass sit))
    (tms::put-env-prop sit `(:ASYMPTOTIC ,(env-sclass sit)) ':DEAD) ;; keep backpointers
    (put-sclass-prop sclass 
		     (cons sit (get-sclass-prop sclass ':DEAD-SITUATIONS))
		     :DEAD-SITUATIONS)
    (dolist (pentry (cdr (assoc sit (sclass-previous (env-sclass sit)))))
      (setq newly-dead (mark-transition-as-dead (cdr pentry) (car pentry) sit
						`(:DEAD ,sit)))
      (if newly-dead (push newly-dead losers)))
    (prune-situation sit sclass :ASYMPTOTE-CHECK)
    (push sit *dead-states*)
    losers))

(defun mark-transition-as-dead (before lh after reason &aux entry subentry key sclass)
  (when (env-sclass? before)
    (setq sclass (env-sclass before))
    (put-sclass-prop sclass 
		     (cons (cons (setq key (cons lh after)) reason)
			   (get-sclass-prop sclass :DEAD-LHS))
		     :DEAD-LHS))
  ;; Leave the internal sclass pointers about transitions, 
  ;; since they record what locally one might think can happen.
  ;; The next-state-table is the real key.
  (setq entry (assoc before *next-state-table*))
  (setq subentry (find key (cdr entry) :test #'equal))
  (unless subentry (error 
     "~% Indexing error in *NEXT-STATE-TABLE* ~A should have transition to ~A via ~A"
     before lh after))
  (setf (cdr entry) (delete subentry (cdr entry)))
  (when (null (cdr entry)) ;; We've lost this one, too.
    (setq *next-state-table* (delete entry *next-state-table*))
    before))


