;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Precondition manipulation for QPE

;; Preconditions are divided into several classes.

;; Modeling Assumptions: These remain fixed over a particular run.
;; Examples might include whether or not to consider thermal properties,
;; or steady-state assumptions.  These will be asserted to be either true or
;; false, as per user (QPE-user, that is, which might be a program) request. 

;; Computable Relationships: As before, we presume that in every state where these
;; are relevant that enough information is included in the state description to
;; ascertain these.

;; Manipulable assumptions: Some theory of action external to QP theory can change
;; these.  Consequently, statement classes declared to be manipulable must be treated as
;;  assumptions rather than assertions. 

;;; The external factors of a domain are declared by naming choice sets.
;; Sometimes one knows exactly what the members of this set are in advance.
;; For that case, the CHOICES primitive is useful for declaring these.
;; (In any case, the choices will be gathered at run-time, and right now
;;  no error checking is performed to see if the CHOICES declared are
;;   exhaustive.)  

;; Unless something is marked as manipulable, these choices may either be allowed
;; to vary or are fixed by the user.   

(defvar *named-choice-sets* nil)
(adb::init-form (setq *named-choice-sets* nil))

;; Dennis DeCoste wanted the first element to be the default if
;; preconditions are asserted.  Makes sense to me, so I've installed
;; his modifications.
(defvar *named-choice-sets-defaults* nil) 
(adb::init-form (setq *named-choice-sets-defaults* nil))

;;;; QPE Syntaxer support for varying preconditions

;;; The predicate CHOICE-SET declares a function to be something for which
;;;  varying PC's should be established.

;; The predicate CHOICES implies CHOICE-SET, but further names the possible values.
;; **** Should enforce exhaustivity someday.

;; An optional default can be supplied by adding :DEFAULT <choice>.  If an explicit
;; default is not specified, then the first element of the choice set is used. 

(deftranslation Choices
		(trans-rel1 `(Choice-set (,(cadr form) . (:quote ?var))))
  :eval `((pushnew ,(adb::quotize (cons (cadr form)
					(if (eq (fourth form) :DEFAULT)
					    (if (member (fifth form) (third form)
							:test  #'equal) 
						(fifth form)
						(error 
						  "Default value for ~A must be one of ~A: ~A"
						  (cadr form) (third form) (fifth form)))
					    (car (third form)))))
		   *named-choice-sets-defaults* :test #'equal :key #'car)
	  ,@ (mapcar #'(lambda (choice)
			 `(adb::referent ,(adb::quotize
					    (cons (cadr form) choice)) T))
		     (caddr form))))

(defTranslation Choice-set `(,form . :TRUE)
  ;; Now takes a pattern rather than a function
  :eval `((pushnew ,(adb::quotize (adb::desugar-pattern (cadr form)))
		   *Named-Choice-Sets*
		   :test #'equal)))

(defTranslation Switch `(,form . :TRUE)
  :eval `((pushnew ,(adb::quotize (adb::desugar-pattern (cadr form))) *Named-Choice-Sets*
		   :test #'equal)))

;; Added :default option here as well, and eliminated need to explicitly
;; assert seperate "switch" statement.

(defTranslation Settings
		`(AND (,form . :TRUE)
		      ,(trans-rel1 `(Switch ,(cadr form)))
		      ,@ (mapcar #'(lambda (choice)
				     `((has-setting ,(cadr form) ,choice) . :TRUE))
				 (caddr form)))
  ;; Introduce a predicate form, SETTING, for use in preconditions and other
  ;; circumstances where "nothing but the truth will do".
  :results (nconc (mapcar #'(lambda (choice)
			       `((SETTING ,(cadr form) ,choice) ((,(cadr form) . ,choice))
				 :Truthifying-function-term))
			   (caddr form))
		   (mapcan #'(lambda (choice)
			       (mapcar #'(lambda (other)
					   `((not (SETTING ,(cadr form) ,choice))
					     ((,(cadr form) . ,other))
					     :Truthifying-function-term))
				       (remove choice (third form))))
			   (caddr form)))
    :eval `((pushnew ,(adb::quotize (cons (cadr form)
					  (if (eq (fourth form) :DEFAULT)
					      (if (member (fifth form) (third form)
							  :test #'equal)
						  (fifth form)
						  (error 
						    "~%Switch ~A can't have default ~A: ~A"
						    (cadr form) (fifth form) (third form)))
					      (car (third form)))))
		     *named-choice-sets-defaults* :test #'equal :key #'car)
	    ,@ (mapcar #'(lambda (choice)
			 `(adb::referent ,(adb::quotize
					   (cons (cadr form) choice)) T))
		     (caddr form))))

;;;; Gathering choice sets

(defvar *choice-sets* nil)
(adb::init-form (setq *choice-sets* nil))

(defun gather-choice-sets ()
  (setq *choice-sets* nil)
  (dolist (cset *named-choice-sets*)
    (unless (assoc cset *choice-sets* :test #'equal)
      (let ((choices (fetch cset)))
	(push (cons cset choices) *choice-sets*))))
  (dolist (pc *preconditions*)
    (unless (member pc *named-choice-sets* :test #'equal)
      (let ((choices (fetch pc)))
	(unless choices ;; Default is propostional
	  (setq choices (list (cons pc :TRUE)
			      (cons pc :FALSE))))
	(push (cons pc choices) *choice-sets*)))))

;;;; PS-VS construction hooks

(defun make-precondition-assumptions (&optional (assume-pcs? t)
						&aux l?)
  (adb::if-debug-flag basics
      (format t "~% PSVS: ~A preconditions & externals"
	      (if assume-pcs? "Assuming" "Asserting")))
  (gather-choice-sets)
  (dolist (cset *choice-sets*)
    (setq l? (listp (car cset)))
    (cond ((and l? (computable-relation-type? (caar cset))))
	  ((computable-relation-type? (car cset))) ;; Do nothing.
	  ((and l? (modeling-assumption-type? (caar cset)))
	   (make-modeling-assumption cset))
	  ((modeling-assumption-type? (car cset))
	   (make-modeling-assumption cset))
	  ((and l? (manipulable-relation-type? (caar cset)))
	   (assume-choice-set cset))
	  ((manipulable-relation-type? (car cset))
	   (assume-choice-set cset))
	  (assume-pcs? (assume-choice-set cset))
	  (t (assert-choice-set cset)))))

(defvar *asn-choice-sets* nil)
(adb::init-form (setq *asn-choice-sets* nil))

(defun assume-choice-set (cset)
  (multiple-value-bind (trues falses unknowns)
      (classify-choice-set-elements (cdr cset))
    (declare (ignore falses))
    (cond (trues
	   (cond ((cdr trues)
		  (error "~% Choice set for ~A violates exclusion: ~A"
			 (adb::lprinc-string (caar cset))
			 (mapcar #'(lambda (x) (adb::lprinc-string (cdr x)))
				 trues)))
		 (unknowns ;; enforce exclusion here?
		  (dolist (unk unknowns)
		    (adb::nogood (list unk) :Assume-choice-set-exclusion)))))
	  ((null (cdr unknowns)) ;; Last alternative must be true
	   (adb::if-debug-flag PCS (format t "~%   PCS: Asserting only choice ~A for ~A."
					   (adb::lprinc-string (car unknowns))
					   (adb::lprinc-string (car cset))))
	   (adb::assert! (car unknowns)))
	  (t   (adb::if-debug-flag pcs
				   (format t "~%   PCS: Assuming ~A one of ~A"
					   (adb::lprinc-string (car cset)) unknowns))
	       ;; This ONEOF is dangerous, since it assumes no overlap (I think).
	       (adb::oneof unknowns)
	       (push (mapcar #'adb::fact-assumption unknowns)
		     *asn-choice-sets*)))))

(defun classify-choice-set-elements (cset-els &aux trues falses unknowns
					      node)
  (dolist (choice cset-els (values trues falses unknowns))
    (setq node (adb::get-tms-node choice))
    (if (tms::true? node) (push choice trues)
	(if (tms::false? node) (push choice falses)
	    (push choice unknowns)))))

(defun assert-choice-set (cset)
  (let ((default (assoc (car cset) *named-choice-sets-defaults* :test #'equal)))
    (cond (default
	   (adb::if-debug-flag pcs
			       (format t "~%   PCS: Asserting default ~A = ~A."
				       (adb::lprinc-string (car cset))
				       (adb::lprinc-string default)))
	   (adb::assert! default))
	  (t (adb::if-debug-flag pcs
				 (format t "~%   PCS: Asserting ~A."
					 (adb::lprinc-string (cons (car cset) :TRUE))))
	     ;; ****** Really should allow defaults to be specified as part of the
	     ;; ****** domain model.
	     (adb::assert! (cons (car cset) :TRUE))))))

(defun get-precondition-assumptions ()
  ;; repackaging old stuff for upward compatability
  (copy-list *asn-choice-sets*))

;;;;; Choice set debugging

(defun show-choice-sets (&optional (stream *standard-output*)
				   &aux counter)
  ;; ******* Add this to default QPE report
  (setq counter 0)
  (dolist (cset *choice-sets*)
    (format stream "~% ~D. ~A."
	    (incf counter) 
	    (adb::lprinc-string (cons 'set-members (cdr cset))))))

(defun show-modeling-choices (&optional (stream *standard-output*)
					&aux counter)
  ;; ****** Add this to default QPE report.
  (setq counter 0)
  (dolist (choice *modeling-choices* counter)
    (format stream "~% ~D. ~A" (incf counter)
	    (adb::lprinc-string choice))))

(defun worst-case-cset-size ()
  (let ((lengths (mapcar #'(lambda (cset)
			     (multiple-value-bind (trs fls unks)
				    (classify-choice-set-elements (cdr cset))
			       (declare (ignore fls))
			       (cons (length (cdr cset))
				     (cond (trs 1)
					   (t (length unks))))))
			 *choice-sets*))
	(counter 0))
    (dolist (csize lengths)
      (format t "~%~D: ~A out of ~A."
	      (incf counter) (cdr csize) (car csize)))
    (values (apply #'* (mapcar #'car lengths))
	    (apply #'* (mapcar #'cdr lengths)))))

;;;;; Recognizing types of relations

(defvar *computable-relations* nil)
(adb::init-form (setq *computable-relations* nil)) 

(defmacro DefComputable-Relation (rel)
  ;;; This just says that if an instance of this relationship occurs in the preconditions
  ;;; of a CDI, it still should not be asserted nor the subject of assumptions, because there
  ;;; will be enough state information lying around to cause it to be deducible
  ;;; BIZARRE BUGS CAN OCCUR IF THIS IS USED INCORRECTLY.  BE CAREFUL! 
  (let ((form `(progn 'COMPILE
		      (push ',rel *computable-relations*))))
    (adb::eval-or-packetize form)))

(defun computable-relation-type? (rel &aux assert)
  (or (eq rel 'Setting) ;; Assume it will be done by some choice set
      (and (symbolp rel) (member rel *computable-relations*))
      (and (setq assert (adb::referent rel nil))
	   (adb::get-assert-prop assert :QPE-computable-relationship))))

(defun manipulable-relation-type? (rel &aux assert)
  (or (and (symbolp rel) (get rel 'QPE-manipulable-relationship))
      (and (setq assert (adb::referent rel nil))
	   (adb::get-assert-prop assert :QPE-manipulable-relationship))))

(defun Modeling-assumption-type? (rel &aux assert)
  (and (setq assert (adb::referent rel nil))
	   (adb::get-assert-prop assert :Modeling-assumption-type)))

;;;; Modeling assumptions

;; These may have user preferences set on the alist
;; *Modeling-choices*.  If so, assert.  O.W., assume them.

(defvar *modeling-choices* nil)
(adb::init-form (setq *modeling-choices* nil))

(defmacro defModelingAssumptionType (rel)
  ;; Probably bogus
  `(progn (adb::rule :INTERN (((,rel . ?x) . ?y) :var ?f1)
		     ;; Assume binary.
		     (unless (or (eq ?y :TRUE) (eq ?y :FALSE))
		       (error "~A not allowed form of modeling assumption."
			      ?f1))
		     (adb::referent (cons (cons ',rel ?x) :TRUE) T)
		     (adb::referent (cons (cons ',rel ?x) :FALSE) T)
		    (pushnew (car ?f1) *Named-Choice-Sets* :test #'equal))
	  ,@ (cond ((symbolp rel)
		    `((setf (get ',rel 'Modeling-assumption-type) T)
		      (adb::init-once-form (remprop ',rel 'Modeling-assumption-type))))
		   (t `((adb::put-assert-prop (referent (cons ',rel ?x) T) T
					      :Modeling-assumption-type))))))

(defmacro defModelingAssumption (form)
  ;; Form will be the class of the assumption type.
  `(progn 'compile (record-modeling-assumption ',form)))

(defun record-modeling-assumption (form)
  (adb::referent (cons form :TRUE) t)
  (adb::referent (cons form :FALSE) t)
  (pushnew form *named-choice-sets* :test #'equal)
  (adb::put-assert-prop (referent form T) T :Modeling-assumption-type))

(defmacro Set-Model-Assumption (asn-form val)
  `(progn 'compile (record-modeling-assumption ',asn-form)
	  (adb::replace-alist-value ',asn-form ',val *modeling-choices*)))

(defun make-modeling-assumption (cset)
  (let ((user-choice (assoc (car cset) *modeling-choices*)))
    (cond (user-choice
	   (cond ((or (adb::false? user-choice)
		      (adb::true? user-choice)) ;; Already done
		  (adb::if-debug-flag ASNS
		     (format t "~%     ASN: ~A already fixed."
			     (adb::lprinc-string user-choice))))
		 (t (adb::if-debug-flag ASNS
		    (format t "~%     ASN: Asserting ~A."
			    (adb::lprinc-string user-choice)))
		    (adb::assert! user-choice))))
	  (t (adb::if-debug-flag ASNS
		(format t "~%     ASN: Assuming {TRUE, FALSE} for ~A."
			(adb::lprinc-string (car cset))))
	     (assume-choice-set cset)))))

(defun install-modeling-assumptions ()
  (dolist (user-choice *modeling-choices*)
    (adb::if-debug-flag ASNS
       (format t "~%     ASN: Asserting ~A." (adb::lprinc-string user-choice)))
    (adb::assert! user-choice)))