;;; -*- Mode: LISP; Syntax: Common-lisp; Package: QPE; -*-

;;;; Defining perspectives

;; Perspectives have individuals and relations, which work just
;; like those in views and processes.  Furthermore, the rules which
;; instantiate them abort if any of the antecedents are false, again
;; like CD's.  However, no instance names are created and no preconditions
;; or quantity conditions are allowed.  This provides a mechanism for
;; defining aspects of models where existence and negation are automatically
;; handled properly.

(in-package 'QPE)

(defvar *perspectives* nil)
(adb::init-form (setq *perspectives* nil))

(defmacro defPerspective (form &rest specs)
  (let ((pname (cond ((listp form) (car form)) (t form)))
	(rest (when (listp form) (cdr form)))
	(individuals (cadr (member 'individuals specs)))
	(relations (cadr (member 'relations specs)))
	(local-vars nil)
	(free-vars nil))
    (when (member 'preconditions specs)
      (error "~% Perspectives cannot have preconditions: ~A." 
	     (adb::lprinc-string form)))
    (when (member 'influences specs)
      (error "~% Perspectives cannot have influences: ~A." 
	     (adb::lprinc-string form)))
    (when (or (member 'quantityconditions specs)
	      (member 'quantity-conditions specs))
      (error "~% Perspectives cannot have quantity conditions: ~A." 
	     (adb::lprinc-string form)))
    (unless individuals
      (error "Missing individuals field: ~A." 
	     (adb::lprinc-string form)))
    (unless (listp individuals)
      (error "Individuals field must be a list: ~A." 
	     (adb::lprinc-string form)))
    (when (and relations (not (listp relations)))
      (error "Relations field must be a list: ~A." 
	     (adb::lprinc-string form)))
    (setq local-vars (test-individual-specs individuals pname))
    (unless local-vars
      (error "~% Must have at least one variable, somewhere: ~A." 
	     (adb::lprinc-string form)))
    (when (setq free-vars
		(set-difference (get-variables relations) local-vars))
      (error "~%Free variables in Relations field of ~A: ~A."
	     (adb::lprinc-string form) free-vars))
    (test-relations-specs relations form)
    (implement-perspective pname rest form individuals relations local-vars)))

;;;; Implementing perspectives

(defun implement-perspective (name rest form individuals relations local-vars
				   &aux specs)
  (setq *vars-to-destructure* nil)
  (setq specs (compute-ispecs-for-antecedents individuals))
  `(progn 'compile
	  (adb::rule :INTERN ( (Expand-Scenario . :TRUE) )
		     ,@ (make-adb-rules-from-ispecs individuals nil
      ;; Now this is what is inside.  Bind ?self to be the form, with
      ;; vars substituted in.
    `((adb::if-debug-flag perspectives
	 (format t "~%     PERS: Found perspective: ~A"
			(adb::lprinc-string ,(adb::quotize form))))
	      (adb::rjustify (,form . :TRUE) ,specs :Perspective-Definition)
	      ,@ (mapcar #'(lambda (ante)
			     `(adb::rjustify (not ,form)
					     ,(list (flip-truths ante))
					     :Perspective-Negation))
			 specs)
	      ,@ (multiple-value-bind (normals results evals 
				    local-names creates-object?)
		     (translate-relations relations `((,form . :TRUE))
					  nil :perspective)
		   (build-perspective-consequences form
		     normals results evals local-names creates-object?)))))))

(defun build-perspective-consequences (form normals results evals
					       local-names creates-object?)
  `((push ,(adb::quotize form) *perspectives*)
    ,@ (mapcan #'(lambda (conseq)
		   (justify-facts-with-connectives 
		     conseq `((:EVAL (CONS ,(adb::quotize form)
					   :TRUE))) :Perspective-Enabled))
	       normals)
    ,@ (mapcar #'(lambda (conseq)
		   `(adb::rjustify ,@ conseq)) results)
    ,@ evals))
