;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Finding View and Process structures

(defvar *PS-VS* nil) ;list of distinct S-complete Sclasses.
(defvar *PS-VS-Envs* nil)
;;*BASIS* is the environment which contains background assumptions
;; This provides a place to hang global assumptions which must be explicit
;; for some other system.
(defvar *Basis* nil) ;; Background assumptions underlying the envisionment (generally empty).
(defvar *Enforce-Env* nil) ;; Assumption that quantity/existence relationship is enforced.
(defvar *Envision-Env* nil) ;; Combination of these.

(adb::Init-Form (setq *BASIS* tms::*empty-env*)) ;reset as needed
(adb::Init-Form (setq *ENVISION-ENV* nil))
(adb::Init-Form (setq *PS-VS* nil))
(adb::Init-Form (setq *PS-VS-ENVS* nil))

(adb::Init-Form ;; Assume that quantity existence is in force or not.
  ;;; This assumption is suspended when checking continuity if an object stops existing.  
  (progn 
    (adb::assumep '(Enforce Quantity-Existence))
    (setq *enforce-env* (adb::create-environment
			  `(((enforce quantity-existence) . :TRUE))))))

(proclaim '(special *PS-VS-ENVS*))

;;;;; Model expansion and the first round of assumption-making

(defvar *domain-expander-forms* nil) ;; User-hooks
(adb::init-form (setq *domain-expander-functions* nil))

(defmacro defExpansionForm (form)
  ;; If necessary, add :BEFORE and :AFTER options.
  `(push ',form *domain-expander-forms*))

(defun expand-scenario-using-domain-model ()
  ;; The purpose of this function is to provide finer control over
  ;; model expansion versus model loading.
  ;;  (When ADB has rule caching this will be very useful!)
  ;; If the user declares something to be false, we want not to make
  ;; a big fuss involving it.  But this means holding back the intern rules
  ;; until the whole set of example files has been read.  This just
  ;; tosses in the signal to start executing the rules, and then runs any user-supplied
  ;; domain building functions.
  (run-qpe-events :BEFORE :SCENARIO-EXPANSION)
  ;; First install the modeling assumptions
  (install-modeling-assumptions)
  ;; Cry HAVOC!
  (adb:referent '(Expand-Scenario . :TRUE) t)
  ;; and unleash the dogs of war
  (adb::run)
  (dolist (form *domain-expander-forms*) (eval form))
  (adb::run)
  (if-debug-flag Expand-break
		 (break "~% Scenario now expanded."))
  ;; Close quantities (a form of negation-by-failure, in this case
  ;; negate them if there is no justification which will support them.)
  (close-quantities)
  (run-qpe-events :AFTER :SCENARIO-EXPANSION))

(Defun Install-QC-PC-assumptions (&optional (assume-pcs? nil))
  (When-Careful
    (setq *QPE-Program-Counter* "QC-PC"))
  ;; Makes necessary assumptions for Qstate and preconditions, and augments derived qstate
  ;; so that as much derivative information will be computed as possible.
  (run-qpe-events :BEFORE :PSVS-ASNS)
  (make-precondition-assumptions assume-pcs?)
  ;; Mark derivative information as something we are interested in.
  (expand-derived-qstate-by-derivatives *QC-Qstate*)
  (adb::if-debug-flag basics (format t "~% PSVS: Making Qstate assumptions.."))
  (make-qstate-assumptions *QC-Qstate* ':QC-QSTATE)
  (run-qpe-events :AFTER :PSVS-ASNS))

(defun make-qstate-assumptions (qstate informant)
  (dolist (qs qstate) (assume-ineq-if-needed qs informant)))

;;;; Expanding qstate to include derivatives (experiment)
;;    Given that some inequality involving amounts is on the QC-Qstate,  it
;;  is important to look at the relationship between their derivatives as well in
;;  order to catch continuity violations.  So we place these elements on the
;;  derived qstate, so they will be available for transitivity checking.

(defun expand-derived-qstate-by-derivatives (other-qstate)
  ;; This expansion should probably take place with the resolution and limit
  ;; contributions to assumed qstate as well, hence the input variable.
  (setq *derived-qstate* (remove-if #'(lambda (pair) ;; begin by filtering
					(member pair other-qstate))
				    *derived-qstate*))
  (dolist (pair other-qstate)
    (cond ((and (amount? (ineq-num1 pair))
		(or (amount? (ineq-num2 pair))
		    (special-quantity? (ineq-num2 pair))))
	   (pushnew (fetch-ineq
		      (make-derivative (quantity-of (ineq-num1 pair)))
		      (make-derivative (quantity-of (ineq-num2 pair))))
		 *derived-qstate* :test #'equal)))))

(defun characterize-qref (n)
  (cond ((derivative? n) ':D)
	((amount? n) ':A)
	((special-quantity? n) ':SPECIAL)
	(t ':BOGUS)))

(proclaim '(special *amounts* *derivatives* *mixed* *bogus*))

(defun check-assumed-qstate ()
  ;; Looks for amounts which have not had corresponding derivative relations
  ;; installed.
  (let ((amounts nil) (derivatives nil)
	(mixed nil) (bogus nil) (s1 nil) (s2 nil))
    (dolist (ineq (gather-assumed-qstate))
      (setq s1 (characterize-qref (ineq-num1 ineq))
	    s2 (characterize-qref (ineq-num2 ineq)))
      (case s1
	(:A (case s2
	      ((:A :SPECIAL) (push ineq amounts))
	      (:D (push ineq mixed))
	      (:BOGUS (push ineq bogus))))
	(:D (case s2
	      ((:D :SPECIAL) (push ineq derivatives))
	      (:A (push ineq mixed))
	      (:BOGUS (push ineq bogus))))
	(:SPECIAL  ; Shouldn't happen, currently
	    (push ineq bogus))
	(:BOGUS (push ineq bogus))))
    (format t
   "~% ~D total, ~D amounts, ~D derivatives, ~D mixed entries, ~D bogus entries."
	(length (gather-assumed-qstate)) (length amounts) (length derivatives)
	(length mixed) (length bogus))
    (dolist (aentry amounts)
      (let ((nentry (list (quantity-of (ineq-num1 aentry))
			  (quantity-of (ineq-num2 aentry)))))
      (unless (find nentry derivatives :test #'equal
		    :key #'(lambda (dpair)
			   (list (quantity-of (ineq-num1 dpair))
				 (quantity-of (ineq-num2 dpair)))))
	(format T "~% No derivative entry for ~A." aentry))))
    (format t "~% Amount entries:")
    (dolist (ae amounts)
      (format t "~%     ~A, ~A" (ineq-num1 ae) (ineq-num2 ae)))
    (format t "~% Derivative entries:")
    (dolist (ae derivatives)
      (format t "~%     ~A, ~A" (ineq-num1 ae) (ineq-num2 ae)))
    (when mixed
      (format t "~% Mixed entries:")
      (dolist (ae mixed)
	(format t "~%     ~A, ~A" (ineq-num1 ae) (ineq-num2 ae))))
    (when bogus
      (format t "~% Bogus entries:")
      (dolist (ae bogus)
	(format t "~%     ~A, ~A" (ineq-num1 ae) (ineq-num2 ae))))
    (setq *amounts* amounts
	  *derivatives* derivatives
	  *mixed* mixed
	  *bogus* bogus)))

;;;;;  Building S-classes

(defun construct-Sclasses (&optional (env *BASIS*))
  (let ((qstate (filter-ineq-choice-sets ;; skip those already fixed
		 (gather-assumed-qstate t)))) ;; use re-ordering
    (setq *ENVISION-ENV* (tms::union-env env *ENFORCE-ENV*))
    (when (tms::env-contradictory *ENVISION-ENV*)
      (format t "~% Assumptions underlying envisionment are contradictory.")
      (format t "~% Basis = ~A." env)
      (format t "~% Enforce = ~A." *ENFORCE-ENV*)
      (format t "~% Union = ~A." *ENVISION-ENV*)
      (error  "~% Contradictory simulation assumptions: ~A." *ENVISION-ENV*))
    (run-qpe-events :BEFORE :PSVS-IC)
    ;; Now get the partial situations
    (setq *PS-VS-ENVS* (get-psvs-solutions *ENVISION-ENV* qstate))
    (adb::if-debug-flag basics 
			(format t "~% PSVS: ~D extensions." (length *PS-VS-ENVS*)))
    ;; and construct the right structs for them
    (construct-ps-vs *PS-VS-ENVS*)
    ;; Next record the obvious results with the situations.
    (Add-unambiguous-Ds-information)
    (run-qpe-events :AFTER :PSVS-IC)
    ;; Find the R-completions as needed.
    (Resolve-ambiguous-Ds-values)))

;;;; Interpretation construction for PS/VS

(proclaim '(special *ASSUME-PCS*))

(defun get-psvs-solutions (env qstate)
  ;;; ***** Should install flight-recorder info here -- size of search space.
  (adb::find-assumption-extensions env
		       ;; Do PC's first, since they impose more constraint on rates.
		       (nconc (get-precondition-assumptions)
			      (mapcar #'(lambda (ineq)
					 ;; (filter-choice-set -- shouldn't be needed now
					    (ineq-choice-set ineq)
					    ;; )
					  )
				      qstate))))

(defun filter-choice-set (as-set)
  ;; might be dead code w/new ATMS
  (remove-if #'(lambda (asn) (tms::false? asn)) as-set))

;;;; Generating the Sclass datastructures
;;
;; Situations are represented as environments.
;; Each Sclass is an equivalence class on situations.  Two situations are
;; in the same Sclass if their individuals, view structure, process structure,
;; and Ds values are identical.  Notice that what can vary are the quantity
;; spaces.  (These in turn mean different limit hypotheses may be relevant, so
;; limit analysis must rely on the finer structure of situations.)

(defun construct-ps-vs (sit-es)
  (When-Careful
    (setq *QPE-Program-Counter* "PS-VS"))				  
  ;;; First step is to get lists consisting of the different possible constiuents of
  ;;; S-classes.  Then for each situation, find the particular values of these
  ;;; parameters, and see if there is already an existing S-class that fits them.
  ;;; If so, add the new situation to the assumptions field of that S-class.  If not,
  ;;; create a new S-class to correspond to that assumption.
(let ((is (mapcar #'(lambda (ind) 
		      (list ind (adb::get-tms-node `((Exists ,ind) . :TRUE))))
		  (get-individuals)))
      (vis (build-cdis-status-cache (get-vis)))
      (qpis (build-cdis-status-cache (get-pis)))
      (sit nil))
    (setq *PS-VS* nil *sclass-index* nil)
    (adb::if-debug-flag basics (format t "~% PSVS: Constructing Sclasses.."))
      (dolist (e sit-es) ;; Filter out that which does not exist.
	  ;;; Here we find out what this env actually implies about the S-class it belongs to.
	(let* ((individuals (filter-by-existence-in e is))
	       (vs  (find-cdis-status-in e (filter-by-existence-in e vis)))
	       (ps (find-cdis-status-in e (filter-by-existence-in e qpis)))
	       (key (list individuals vs ps)))
	  (setq sit (lookup-sclass key))
	  (cond (sit (push e (sclass-assumptions sit)))
		(t   (setq sit (make-sclass
				 :NAME (generate-sclass-name)
				 :ASSUMPTIONS (list e)
				 :INDIVIDUALS (mapcar #'car individuals)
				 :PROCESS-STRUCTURE ps
				 :VIEW-STRUCTURE vs
				 :STATUS 'PS-VS-COMPLETE))
		     (set (sclass-name sit) sit)
		     (insert-sclass key sit)
		     (push sit *PS-VS*)))
	  ;; Mark the situation as an S-class.  The importance of this marking is that if
	  ;; we generate all situations in advance is to simplify limit analysis.  If the
	  ;; result of temporal inheritance is not an env marked as an S-class, then there
	  ;; is some inconsistency in that limit hypothesis.
	  (mark-env-with-sclass e sit)))
      (adb::if-debug-flag basics (format t "~% PSVS: Initial Sclasses = ~A" *ps-vs*))
	  *PS-VS*))

(defun mark-env-with-sclass (e sit)
  (tms::put-env-prop e sit 'sclass))

(defun unmark-env-sclass (e)
  (tms::put-env-prop e nil 'sclass))

;;;; Helpers for Construct-PS-VS
;;
;; ***** Introduce macrology for these?

(defun filter-by-existence-in (e possibles)
  (remove-if-not #'(lambda (i) (tms::true-in? (cadr i) e)) possibles))

(defun build-cdis-status-cache (cdis)
  (mapcar #'(lambda (vi)
	      (list vi (adb::get-tms-node `((Exists ,vi) . :TRUE))
		    (adb::get-tms-node `((Active ,vi) . :TRUE))
		    (adb::get-tms-node `((Active ,vi) . :FALSE))))
		   cdis))

(defun find-cdis-status-in (e cdis)
  (mapcar #'(lambda (qpi)
	      (cons (car qpi)
		    (if (tms::true-in? (caddr qpi) e)
			'ACTIVE
			(if (tms::true-in? (cadddr qpi) e)
			    'INACTIVE
			    (error  "~%~A not found in ~A -- find-cdis-status-in"
					 qpi e)))))
	  cdis))

;;;;; Pruning inconsistent situations.
;;
;; Can't just do MAP-SCLASSES here, since we need to update the sclass graph.

(defun prune-sclass (sclass informant)
  (if-debug-flag Pruning
     (format t "~%     Pruning ~A via ~A." sclass informant))
  (When (sclass-assumptions sclass)
    (error "~% Sclass ~A incorrectly pruned -- still has situations ~A."
	   sclass (sclass-assumptions sclass)))
  (When-Careful
    (Add-Recorder-Event
      :Sclass-Pruned sclass informant *QPE-Program-Counter*))
  (setf (getf (sclass-plist sclass) ':Inconsistent-Reason) informant))

(defun prune-situation (sit sclass informant)
  ;;; Must pass sclass in, since contradictory envs can't be situations.
  (if-debug-flag Pruning
     (format t "~%     Pruning ~A via ~A." sit informant))
  (When-Careful
    (Add-Recorder-Event
      :Sit-Pruned sit sclass informant *QPE-Program-Counter*))
  (setf (sclass-assumptions sclass)
	(delete sit (sclass-assumptions sclass)))
  (unmark-env-sclass sit)) ;; Remove pointer to sclass, to avoid fooling limit analysis.
  
(defun flush-inconsistent-sclasses ()
  (dolist (sc *PS-VS*)
    (unless (prune-sclass-situations sc :Contradiction-Check)
      (prune-sclass sc :Contradiction-Check)
      (setq *PS-VS* (delete sc *PS-VS*)))))
   
(defun Prune-sclass-situations (sclass informant)
  (let ((nenvs (remove-if #'(lambda (e) 
			      (cond ((tms::env-contradictory e)
				     (prune-situation e sclass informant)
				     t)))
			    (sclass-assumptions sclass))))
    (cond (nenvs (setf (sclass-assumptions sclass) nenvs)
		 (cond ((eq (sclass-status sclass) 'PS-VS-Complete)
			;; Must check extensions -- if none consistent, this loses as well.
			(let ((okay nil))
			  (dolist (rcp (sclass-r-completions sclass))
			    (if (prune-sclass-situations rcp informant)	(push rcp okay)))
			  (setf (sclass-r-completions sclass) (nreverse okay)))
			(cond ((sclass-r-completions sclass) sclass)
			      (t ;; got to prune it
			       (setf (sclass-assumptions sclass) nil)
			       (prune-sclass sclass :NO-R-COMPLETIONS)
			       nil)))
		       (t sclass)))
	  (t ;; this sclass should be flushed
	   (when (eq (sclass-status sclass) ':PS-VS-Complete)
	     ;; must also flush r-completions -- these will all be inconsistent, since
	     ;; all inferences are monotonic.
	     (dolist (rsc (sclass-r-completions sclass))
	       (prune-sclass-situations rsc informant)))	   
	   nil))))

;;;; Analyzing sclass internal structure

;; For every sclass there is a "core environment" which
;; consists of all the shared assumptions of the situations
;; which comprise the sclass.  (If there is only one situation
;; in an sclass then core = situation env.)  All other situations
;; can be understood by how they differ from the core.

(defun sclass-core (sc &aux cache)
  (unless (setq cache (get-sclass-prop sc ':Core-info))
    (classify-sclass-situations sc)
    (setq cache (get-sclass-prop sc ':Core-info)))
  (car cache))

(defun sclass-situation-differences (sc &aux cache)
  (unless (setq cache (get-sclass-prop sc ':Core-info))
    (classify-sclass-situations sc)
    (setq cache (get-sclass-prop sc ':Core-info)))
  (cdr cache))

(defun classify-sclass-situations (sc &aux core diffs)
  (setq core (envs-common-env (sclass-assumptions sc)))
  (dolist (env (sclass-assumptions sc))
    (push (cons env (env-residue env core)) diffs))
  (put-sclass-prop sc (cons core (nreverse diffs)) ':Core-info))

(defun sclass-diffs (sc &optional (stream *standard-output*))
  (format stream "For Sclass ~A:" (sclass-name sc))
  (let ((core (sclass-core sc))
	(diffs (sclass-situation-differences sc)))
    (format stream "~%  Core environment is ~A." core)
    (print-env core stream)
    (format stream "~%  There are ~D situations:" (length diffs))
    (dolist (diff diffs)
      (format stream "~%   ~A differs by:" (car diff))
      (dolist (asn (cdr diff))
	(format stream "~%       ~A" (lprinc-string (view-assumption asn))))
      (format stream "."))
    sc))

(proclaim '(special *sit-dist*))

(defun situation-distribution (&optional (stream *standard-output*))
  (setq *sit-dist* nil)
  (map-sclasses #'(lambda (sc) (push (cons sc (length (sclass-assumptions sc)))
				     *sit-dist*)))
  (setq *sit-dist* (sort *sit-dist* #'(lambda (x y) (> (cdr x) (cdr y)))))
  (format *standard-output* "~% ~D situations in ~D sclasses."
	  (reduce #'(lambda (num entry) (+ num (cdr entry)))
		  *sit-dist* :initial-value 0)
	  (length *sit-dist*))
  (dolist (entry *sit-dist*)
    (format stream "~%  ~5:<~A~> = ~D" (car entry) (cdr entry)))
  (format stream "~%-------"))
