;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Definitions of quantities and inequalities
;;; Includes constructors, inequality system
;;

;;; ******* Make qnums appear in Sclasses instead of forms -- change print function to do
;;; ******* something reasonable, and use Quant's in Ds lists. 
;;; ******* Will require changing the constructors MAKE-DERIVATIVE, QUANTITY-OF, etc.
;;; ******* These could move around structs.

;;; New experiment: Attempt to minimize justifications. 
;; This is an incremental step towards flushing "soft" inequalities.

(defvar *quantity-types* nil) ;list of declared quantity types
(defvar *explicit-functions* nil) ;list of declared explicit functions
(defvar *zero-struct* nil) ;; Qnumber struct for zero.

(adb::init-form ;purge quantity types when re-loading 
  (dolist (qt *quantity-types*
	      (setq *quantity-types* nil))
    (remprop qt 'Quantity-type?)
    (remprop qt 'Quantity-Arguments)))

(adb::init-form ;purge explicit functions when re-loading
  (dolist (ef *explicit-functions*
	      (setq *explicit-functions* nil))
    (remprop ef 'Explicit-function-specs)))

(adb::after-init-form ;Build new zero struct after initalization 
  (progn (setq *zero-struct* (build-number 'zero))
	 (setf (qnum-role *zero-struct*) ':Special)))

;;;; Quantity types -- specification and checking
 
(defmacro defQuantity-Type (name &rest arguments)
  ;define a symbol to be a type of quantity by saying which
  ;of its arguments are individuals or constants.  This information
  ;is used to go backwards from an instance of a quantity to the
  ;individuals its existence logically depends on.
  (let ((form `(progn (pushnew ',name *quantity-types*)
		      (setf (get ',name 'Quantity-Type?)  T)
		      (setf (get ',name 'Quantity-Arguments) ',arguments))))
    (eval form)
    (adb::eval-or-packetize form)))

(defun quantity-type-arguments (qtype) (get qtype 'Quantity-arguments))

(defun quantity-type? (x)
  (and (symbolp x) (get x 'Quantity-Type?)))

(defun quantity-individual-args (form)
  (do ((types (quantity-type-arguments (car form)) (cdr types))
       (args (cdr form) (cdr args))
       (individuals nil))
      ((null args) individuals)
    (when (eq (car types) 'individual)
      (push (car args) individuals))))

(defun check-quantity-declaration (qform)
  (let ((qtype (quantity-type? (car qform))))
    (unless qtype
      (error "~%~A unknown quantity type -- ~A." (car qform) qform))
    (unless (= (length (cdr qform))
	       (length (quantity-type-arguments (car qform))))
      (error "~%Wrong number of args in quantity statement: ~A." qform))))


;;;; Selectors & constructors for quantity assertions

(defun quantity-of (form)
  (cond ((not (listp form)) form) ;handles ZERO case
	((member (car form) '(A D s m)) (quantity-of (cadr form)))
	(t form)))

(defun make-derivative (Qty)
  (if (eq Qty 'ZERO) 'ZERO (list 'D Qty)))

(defun make-amount (qty)
  (if (eq qty 'ZERO) 'ZERO (list 'A qty)))

(defun magnitude? (qref)
  (cond ((eq qref 'ZERO) nil)
	((not (listp qref)) nil)
	(t (case (car qref)
	     ((D s) nil)
	     (m t)
	     (t nil)))))

(defun sign? (qref)
  (cond ((eq qref 'ZERO) nil)
	((not (listp qref)) nil)
	(t (case (car qref)
	     ((D m) nil)
	     (s t)
	     (t nil)))))

(defun derivative? (qref)
  (cond ((eq qref 'ZERO) nil)
	((not (listp qref)) nil)
	((eq (car qref) 'D) t)
	((member (car qref) '(s m)) (derivative? (cadr qref)))
	(t nil)))

(defun amount? (qref)
  (cond ((eq qref 'ZERO) nil)
	((not (listp qref)) nil)
	((eq (car qref) 'A) t)
	((member (car qref) '(s m)) (amount? (cadr qref)))
	(t nil)))

(defun special-quantity? (q) (eq q 'zero))

(defun inequality-form? (form)
  (and (listp form)
       (or (eq (car form) 'GREATER-THAN)
	   (eq (car form) 'equal-to)
	   (eq (car form) 'less-than))))

;;;; Structures
;;
;; Gizmo used DEBACLE's facilty for data-directed reference to install
;; specialized datastructures as 1st-class database items.  In ADB we'll
;; compromise by adding these to the plist of an assert.  While access will
;; be slightly slower, in most algorithms one wants to fetch the structure
;; once and access it several times.  Also, once one is in the quantity system,
;; there is no real reason to leave until final results are to be installed.

(defstruct (quant (:predicate QP-quantity?)
		  (:print-function (lambda (quant s ignore)
				     (format s "~A" (quant-string quant)))))
  (assert nil)
  (string "")
  (amount nil)
  (derivative nil)
  (just-node nil) ;; cache
  (plist nil))

(defstruct (qnum (:predicate QP-number?)
		 (:print-function (lambda (qnum s ignore)
				    (format s "~A" (qnum-string qnum)))))
  (assert nil) ;pointer back to ADB
  (string "")  ;for pretty-printing
  (type nil)   ;NUMBER or MAGNITUDE
  (ineqs1 nil) ;Alist of ineq structs where this number is N1
  (ineqs2 nil) ;Alist of ineq structs where this number is N2
  (plist nil)
  (quantity nil)
  (role nil))

(defun view-qnum (qnum) (adb::lisp-form (qnum-assert qnum)))
(defun put-qnum-prop (qnum val prop) (setf (getf (qnum-plist qnum) prop) val))
;;(defun get-qnum-prop (qnum prop) (getf (qnum-plist qnum) prop))
(defmacro get-qnum-prop (qnum prop) `(getf (qnum-plist ,qnum) ,prop))


;;;; Struct for internals of inequalities

(defstruct (ineq (:predicate QP-inequality?)
		 (:print-function (lambda (ineq s ignore)
				    (format s "~A" (ineq-string ineq)))))
  (n1 nil)					;First number
  (n2 nil)					;Second number
  (string nil)
  (> nil)					; Node for greater-than
  (= nil)					; Node for equal-to
  (< nil)					; Node for less-than
  (bt nil)				   ; Node for unrelated
  (~= nil)				   ; Cache for transitivity algorithms. 
  (assumed? nil)			   ; If non-nil, yes.
  (lt-a nil)				   ; Assumption for less-than, if any
  (eq-a nil)				   ; Assumption for equal-to, if any
  (gt-a nil)				   ; Assumption for greater-than, if any
  (bt-a nil)				   ; Assumption for unrelated, if any.
  (~> nil)				    ; Cache for transitivity algorithms.
  (~< nil)				    ; Cache for transitivity algorithms.
  (~bt nil)				    ; Cache for transitivity algorithms.
  (new-cycles nil)			    ; Cycles found involving ineq.
  (choice-set nil)			    ; Why re-build constantly?
  (plist nil)
  (type nil)   ;; One of :A, :D, :mixed, :special
  (other nil)) ;; Other ineq

;; These return the list version, for easier processing
(defun ineq-num1 (ineq) (view-qnum (ineq-n1 ineq))) 
(defun ineq-num2 (ineq) (view-qnum (ineq-n2 ineq)))
(defun put-ineq-prop (ineq val prop) (setf (getf (ineq-plist ineq) prop) val))
(defun get-ineq-prop (ineq prop) (getf (ineq-plist ineq) prop))

;;;;  Constructors
;; Unlike GIZMO, where quantities were specialized datastructures,
;; in QPE the specialized information is hung off the property list of
;; the corresponding ADB assertion. 

;; Free-floating numbers are now FORBIDDEN.  All numbers must be part
;; of some quantity.  (Yes, "duration" should be treated as a number.
;; But QPE is not a universal qualitative physics system. 
;; It is an envisioner, and everything that does not contribute to that
;; goal has been streamlined away.... )

(defvar *quants* nil) ;; list of quant records
(proclaim '(special *quants*))
(adb::Init-form (setq *quants* nil))

(defvar *quant-table* nil) ;; hash table of <key> to quantity structs
(proclaim '(special *quant-table*))
(adb::Init-form (cond (*quant-table* (clrhash *quant-table*))
		      (t (setq *quant-table* (make-hash-table :size 500
							      :test #'equal)))))

(defun construct-quantity (quantity &aux q owner individuals)
  ;; owner now is ALWAYS of the form (QUANTITY <name>) 
  ;;Check to make sure individual is not floating
  (setq individuals (destructure-individuals quantity))
  (unless (or individuals (special-quantity? quantity))
    (error "~%~A is not associated with any individual." quantity))
  ;; Enforce existence here, justify UNRELATED and BT assertions
  ;; in terms of false ownership. 
  (setq owner (cons (list 'QUANTITY quantity) :TRUE))
  (unless (special-quantity? quantity)
    (setq *quantities*
	  (adb::ordered-insert quantity #'adb::alphalessp *quantities*)))
  (dolist (ind individuals)
    (adb:justify `(,(car owner) . :FALSE)
		 `(((Exists ,ind) . :FALSE)) 'Quantity-Existence))
 (when-careful
    (let* ((assert (adb:referent quantity t))
	   (loser (adb:get-assert-prop assert 'quantity)))
      (when loser
	(error "~% Quantity ~A already created!" loser))))
 (setq q (make-quant :assert (adb:referent quantity t)
		      :string (compute-qnum-string quantity)
		      :amount (construct-number `(a ,quantity) ':A)
		      :derivative (construct-number `(d ,quantity) ':D)
		      :just-node (get-tms-node owner)))
 (setf (qnum-quantity (quant-amount q)) q)
 (setf (qnum-quantity (quant-derivative q)) q)
 ;; Index all the ways it might be wanted
 (setf (gethash quantity *quant-table*) q)
 (setf (gethash (quant-assert q) *quant-table*) q)
 (push q *quants*)
  q)

;;;; Constructing numbers

(defun construct-number (num role)
  ;;; For efficiency, magnitudes are only constructed upon request.
  (let ((ns (build-number num)))
    (setf (qnum-role ns) role)
    (construct-qrel num 'zero)
    (unless (and (listp num) ;; Not creating As explicitly any more.
		 (eq (car num) 'A)) ;; inequalities w/zero provide same info anyway.
      (construct-sign num))
    ns))

(defvar *qnums* nil) ;; List of number records.
(proclaim '(special *qnums*))
(adb::Init-Form (setq *qnums* nil))

(defvar *qnum-table* nil) ;; Hash table of Qnums
(proclaim '(special *qnum-table*))
(adb::Init-Form (cond (*qnum-table* (clrhash *qnum-table*))
		      (t (setq *qnum-table*
			       (make-hash-table :size 1000
						:test #'equal)))))

(defun build-number (form) ;; Actually build the struct
  (let ((num (make-qnum :assert (adb:referent form t)
			:string (compute-qnum-string form)
			:type (number-subtype form))))
    (push num *qnums*)
    (setf (gethash form *qnum-table*) num)
    (setf (gethash (qnum-assert num) *qnum-table*) num)
    num))

;;;; Building signs

(defun construct-sign (num)
  (let ((sign (list 'S num))
	(owner (quantity-of num)))
    (cond ((eq num 'ZERO) (adb:assert! `(,sign . 0)))
	  ((not (listp num)) (cerror "~%Punting sign creation..."
				     "~%~A bad name for number." num))
	  (t ;if quantity isn't there, due to existence or other changes, then s[num] = bt
	   (adb:justify `(,sign . bt)
			  `(((Quantity ,owner) . :false)) 'Quantity-Owner-law)
	   ;; Justify Ds values in terms of inequalities.
	   (adb::justify `(,sign . 0)
			 `(((equal-to ,num zero) . :TRUE)) 'Sign-Law)
	   (adb::justify `(,sign . -1)
			 `(((less-than ,num zero) . :TRUE)) 'Sign-Law)
	   (adb::justify `(,sign . 1)
			 `(((greater-than ,num zero) . :TRUE)) 'Sign-Law)))))

;;; Computing pretty-names for internal number and inequality structs.
;;
;; Now that new LPRINT is here, simplify.

(defun compute-qnum-string (ql) (adb::lprinc-string ql))

(defun compute-number-string (num) (adb::lprinc-string num))

(defun compute-ineq-string (n1 n2)
  (format nil "~A >=< ~A" (compute-qnum-string n1)
	  (compute-qnum-string n2)))

(defun construct-magnitude (num)
  ;; The transitive algorithms will be much simpler if magnitudes are treated identically
  ;; to numbers.  Justifications are run from the inequalities to simplify life.
  ;; which are linked from the signs, but such is life...
  (let ((mag nil))
    (cond ((and (listp num) (eq (car num) 'M)) 
	   (setq mag num num (cadr num)))
	  (t (setq mag (cons 'm num))))
    (let* ((ms (build-number mag))
;	   (q (fetch-quantity (quantity-of num)))
	   (lt `(less-than ,mag zero))
	   (gt `(greater-than ,mag zero))
	   (eq `(equal-to ,mag zero)))
      (cond ((eq num 'zero) (assert! eq))
	    (t (setf (qnum-role ms) :Mag) ;; May want something more complicated here.
	     (construct-qrel mag 'zero)
	          ;When resolving direct influences, sometimes must look at magnitudes.
	          ;Install this information dependent on the signs, to simplify the flow of information.
	       (adb::assert! `(,lt . :FALSE))
	       (adb::justify `(,eq . :TRUE) `(((equal-to ,num zero) . :TRUE))
			     'Magnitude-law)
	       (adb::justify `(,gt . :TRUE) `(((greater-than ,num zero) . :TRUE))
			     'Magnitude-law)
	       (adb::justify `(,gt . :TRUE) `(((less-than ,num zero) . :TRUE))
			     'Magnitude-law)
	       ms)))))

;;;; Constructing inequalities

(defun construct-qrel (n1 n2 &aux unrel qrel ineq)
  ;; This is because jcollins and I are lazy and don't want to fix the
  ;; transitivity code right now.  Could handle this case seperately,
  ;; making the relationship either equal-to or unrelated, as appropriate.
  (when (equal n1 n2)
    (error "Comparing a number to itself: ~A." (adb::lprinc-string n1)))
  (setq qrel (find-qrel n1 n2)) ; place arguments in the correct order. 
  (setq unrel (adb::referent `((unrelated ,@ qrel) . :TRUE) t))
  ;; Test to see if it has already been created.
  (unless (adb::get-assert-prop unrel 'inequality-initialized?)
    (adb::if-debug-flag quantity
		   (format t "~% QBOX: Inequality built: ~A, ~A" n1 n2))
    ;; Mark it now to prevent inappropriate re-entry
    (adb::put-assert-prop unrel t 'inequality-initialized?)
    ;; The usual ordering relations have been extended by UNRELATED,
    ;; which indicates that some individual does not exist.  Hence for
    ;; every pair of numbers at most one ordering relationship hold, always. 
    (let* ((q1 (quantity-of n1))
	   (q2 (quantity-of n2))
	   (q1j `(quantity ,q1))
	   (q2j `(quantity ,q2))
	   (lt `(less-than ,@ qrel))
	   (gt `(greater-than ,@ qrel))
	   (eq `(equal-to ,@ qrel))
	   (un `(unrelated ,@ qrel)))
      (unless (and (special-quantity? q1) (special-quantity? q2))
	;;; If both special, assume defined elsewhere in the engine.
	(install-internal-inequality-constraints lt gt eq un)
         ;; Unrelated cannot be if both quantities are legitmate
      (adb::justify `(,un . :false)
		    `(,@ (unless (special-quantity? q1)
			   `((,q1j . :TRUE)))
		      ,@ (unless (special-quantity? q2)
			   `((,q2j . :TRUE)))))
         ;Add the entries to support the inequality lattice. 
      (setq ineq (add-qrel-index-information n1 n2 lt gt eq un))
      (unless (special-quantity? q1)
	(adb::justify `(,un . :TRUE) `((,q1j . :FALSE)) 'unrelated-definition))
      (unless (special-quantity? q2)
	(adb::justify `(,un . :TRUE) `((,q2j . :FALSE)) 'unrelated-definition))
         (when (equal n1 n2) ;special case
	(adb::assert! `(,lt . :FALSE))
	(adb::assert! `(,gt . :FALSE))))))
  (values qrel ineq))

;;; Internal constraints on inequalities

(defun install-internal-inequality-constraints (lt gt eq un)
  ;; Weakening this to at-most-one-via-nogoods is a disaster,
  ;; although that isn't obvious a priori.
  ;; Originally it was: (adb::taxonomy (list lt gt eq un))
  ;; Theory: since unrelated nodes have external justifications, the real taxonomizing
  ;; should be between the LT, EQ, and GT nodes.  Each positive instance justifies
  ;; the negations of the other two, in order to catch contradictions.  But the pointers back
  ;; seem awfully expensive, so try leaving them out.
  ;; Following change is to let Johan play with clausal ATMS.
  (cond (*new*
	 ;; These four creations are unecessary but make debugging a lot easier.
	 (get-tms-node `(,eq . :FALSE))  
	 (get-tms-node `(,lt . :FALSE))  
	 (get-tms-node `(,gt . :FALSE))  
	 (get-tms-node `(,un . :FALSE))  
	 (tms::oneof (list (get-tms-node `(,un . :TRUE))
			   (get-tms-node `(,gt . :TRUE))
			   (get-tms-node `(,lt . :TRUE))
			   (get-tms-node `(,eq . :TRUE)))))
	(t (adb:justify `(,gt . :FALSE) `((,lt . :TRUE)) 'ineq-tax)
	   (adb:justify `(,eq . :FALSE) `((,lt . :TRUE)) 'ineq-tax)
	   (adb:justify `(,lt . :FALSE) `((,gt . :TRUE)) 'ineq-tax)
	   (adb:justify `(,eq . :FALSE) `((,gt . :TRUE)) 'ineq-tax)
	   (adb:justify `(,lt . :FALSE) `((,eq . :TRUE)) 'ineq-tax)
	   (adb:justify `(,gt . :FALSE) `((,eq . :TRUE)) 'ineq-tax)
	   (adb:justify `(,lt . :FALSE) `((,un . :TRUE)) 'ineq-tax)
	   (adb:justify `(,eq . :FALSE) `((,un . :TRUE)) 'ineq-tax)
	   (adb:justify `(,gt . :FALSE) `((,un . :TRUE)) 'ineq-tax)
	   (adb:justify `(,eq . :TRUE) `((,gt . :FALSE)
					 (,lt . :FALSE)
					 (,un . :FALSE)) 'ineq-tax))))

(defun destructure-individuals (ref)
  ;;; Enforces conventions of quantity reference
  (cond ((null ref)
	 (error  "~% Null individual impossible -- DESTRUCTURE-INDIVIDUAL"))
	((not (listp ref)) (if (eq ref 'zero) nil
			       (error  "~%~A bad quantity reference." ref)))
	((member (car ref) '(s m D A)) (destructure-individuals (cadr ref)))
	((Quantity-type? (car ref)) (quantity-individual-args ref))
	(t (error  "~%~A bad quantity reference." ref))))

;;;; Building INEQ records
;;

(defvar *ineqs* nil) ;; List of inequality records
(proclaim '(special *ineqs*))
(adb::Init-Form (setq *ineqs* nil))

(defvar *zero-in-lattice* nil) ;; Internal switch, for experimental purposes only.

(defun add-qrel-index-information (n1 n2 ltf gtf eqf unf)
  ;; Now build the INEQ struct for the pair, and incrementally install justifications
  ;;  regarding transitivity.  
  ;; IMPORTANT: This should NEVER be called with N1 = N2 = ZERO!
  (let ((gt (get-tms-node (cons gtf ':TRUE)))
	(eq (get-tms-node (cons eqf ':TRUE)))
	(lt (get-tms-node (cons ltf ':TRUE)))
	(un (get-tms-node (cons unf ':TRUE)))
	(ineq nil) (n1-str nil) (n2-str nil))
    (setq n1-str (fetch-number n1)) ;Get structs for numbers
    (setq n2-str (if (eq n2 'zero) *zero-struct* (fetch-number n2)))
    ;; Build struct to hold the various properties for easy reference.
    (setq ineq (make-ineq :n1 n1-str :n2 n2-str :string (compute-ineq-string n1 n2)
			  :> gt := eq :< lt
			  :bt un
			  :~= (get-tms-node (cons eqf ':FALSE))
			  :~> (get-tms-node (cons gtf ':FALSE))
			  :~< (get-tms-node (cons ltf ':FALSE))
			  :~bt (get-tms-node (cons unf ':FALSE))))
    (push ineq *ineqs*)
    ;Index the relationship under each local.
    (push (cons n2-str ineq) (qnum-ineqs1 n1-str))
    (push (cons n1-str ineq) (qnum-ineqs2 n2-str))
    (setf (ineq-type ineq)
	  (case (qnum-role n1-str)
	    (:A (case (qnum-role n2-str)
		  (:A ':A)
		  (:Mag ':Mag)
		  (:Special ':Special)
		  (:D ':Mixed)
		  (t (error "~% Bad number role ~A for ~A." (qnum-role n2-str)
			    n2-str))))
	    (:D (case (qnum-role n2-str)
		  (:A ':Mixed)
		  (:Mag ':Mag)
		  (:Special ':Special)
		  (:D ':D)
		  (t (error "~% Bad number role ~A for ~A." (qnum-role n2-str)
			    n2-str))))
	    (:Mag ':Mag)
	    (t (error "~% Bad number role ~A for ~A." (qnum-role n1-str) n1-str))))
    ;; Unless the inequality is a relationship with zero, incrementally add transitivity
    ;; justifications for it
    (unless (and (not *zero-in-lattice*)
		 (eq n2-str *zero-struct*))
      (search-for-new-cycles ineq))
    ineq))

;;;; Accessing specialized structs

;; Grab assumes it is there already.  Fetch creates stuff as needed.
(defun grab-number (as) (gethash as *qnum-table*))
(defun grab-quantity (as) (gethash as *quant-table*))
(defun grab-ineq (qn1 qn2) (cdr (assoc qn2 (qnum-ineqs1 qn1))))

(defun fetch-number (form)
  (let ((num (grab-number form)))
    (cond (num num)
	  ((eq (number-subtype form) ':Magnitude)
	   (fetch-number (cadr form))
	   (construct-magnitude form))
	  (t (construct-quantity (quantity-of form))
	     (grab-number form)))))

(defun number-subtype (form)
  (cond ((not (listp form)) ':SPECIAL)
	((eq (car form) 'm) ':MAGNITUDE)
	(t ':NUMBER)))

(defun fetch-quantity (form)
  (or (grab-quantity form)
      (construct-quantity (quantity-of form))))

(defun fetch-ineq (n1 n2) ;; For  easy toplevel access
  (or (grab-ineq (fetch-number n1) (fetch-number n2))
      (multiple-value-bind (ignore ineq)
	  (construct-qrel n1 n2)
	ineq)))

(defun unordered-fetch-ineq (n1 n2 &aux pair)
  (setq pair (find-qrel n1 n2))
  (fetch-ineq (car pair) (cadr pair)))

(defun ineq-neg-rel-node (ineq rel)
  (case rel
	(:> (ineq-~> ineq))
	(:= (ineq-~= ineq))
	(:< (ineq-~< ineq))
	((:? :bt :un) (ineq-~bt ineq))
	(t (error "~%Bad relation: ~A; ... INEQ-NEG-REL-NODE." rel))))

(defun ineq-rel-asn (ineq rel)
  (case rel
	((:> >) (ineq-gt-a ineq))
	((:= =) (ineq-eq-a ineq))
	((:< <) (ineq-lt-a ineq))
	((? bt un :? :bt :un) (ineq-bt-a ineq))
	(t (error "~%Bad relation: ~A; ... INEQ-REL-ASN." rel))))

(defun ineq-rel-node (ineq rel &optional (qnum1 (ineq-n1 ineq)))
  (unless (eq qnum1 (ineq-n1 ineq))
    (setq rel (opposite-relation rel)))
  (case rel
	((:> > :GREATER-THAN GREATER-THAN) (ineq-> ineq))
	((:= = :EQUAL-TO EQUAL-TO) (ineq-= ineq))
	((:< < :LESS-THAN LESS-THAN) (ineq-< ineq))
	((? bt un :? :bt :un :UNRELATED) (ineq-bt ineq))
	(t (error "~%Bad relation: ~A; ... INEQ-REL-NODE." rel))))

    
(defun assume-inequality (n1 n2) ;; Assume the numbers are in the right order
  (let ((ineq (fetch-ineq n1 n2)))
    (unless ineq
      (error  "~%No inequality struct exists between ~A and ~A."
	      n1 n2))
    (tms::assume-node (ineq-< ineq))
    (setf (ineq-lt-a ineq) (tms::find-assumption-for (ineq-< ineq)))
    (tms::assume-node (ineq-= ineq))
    (setf (ineq-eq-a ineq) (tms::find-assumption-for (ineq-= ineq)))
    (tms::assume-node (ineq-< ineq))
    (setf (ineq-gt-a ineq) (tms::find-assumption-for (ineq-> ineq)))
    (tms::assume-node (ineq-bt ineq))
    (setf (ineq-bt-a ineq) (tms::find-assumption-for (ineq-bt ineq)))))

;;;;Canonicalizing inequalities
;;;  Rather than deal with several equivalent forms, the arguments to
;;;  an order relationship must take a canonical form.  When ZERO is
;;;  an argument it is always the second argument.  Otherwise, only
;;;  assert statements about a pair of numbers in alphabetical order.

(defun find-qrel-ordering (n1 n2)
  ;Returns canonical form for a pair of numbers
  (unless (and n1 n2)
    (error "NIL impossible number: ~A, ~A" n1 n2))
  (if (eq n1 'zero) (values n2 n1)
      (if (eq n2 'zero) (values n1 n2)
	  (if (adb::alphalessp n1 n2) (values n1 n2)
	      (values n2 n1)))))

(defun find-qrel (n1 n2) ;when only lists will do.
  (multiple-value-list (find-qrel-ordering n1 n2)))

(defun in-order? (n1 n2) ;test for screw-ups
  (if (eq n1 'zero) (if (eq n2 'zero) t nil)
      (adb::alphalessp n1 n2)))

(defun find-qrel-assertion (n1 n2 rel &aux qrel)
  ;;; This function is for user-level code and QP syntaxers
  ;;; if you don't (or can't) know if your ordering is correct,
  ;;; run the expression through this procedure to get the right version. 
  (setq qrel (find-qrel n1 n2))
  (case rel
    ((equal-to =) (cons 'Equal-to qrel))
    ((less-than lt <) (if (eq n1 (cadr qrel)) (cons 'greater-than qrel)
		(cons 'less-than qrel)))
    ((greater-than gt >) (if (eq n1 (cadr qrel)) (cons 'less-than qrel)
		(cons 'greater-than qrel)))
    ((lt-eq <= <-=) (if (eq n1 (cadr qrel)) (cons 'gt-eq qrel)
			(cons 'lt-eq qrel)))
    ((gt-eq >= >-=) (if (eq n1 (cadr qrel)) (cons 'lt-eq qrel)
			(cons 'gt-eq qrel)))
    ((unrelated ?) (cons 'unrelated qrel))
    (t (error "~% ~A not possible value for ~A" rel qrel))))

(proclaim '(special adb::*flip-truth-value*))

(defun find-qrel-assertion1 (n1 n2 rel) 
;;****** Not clear find-qrel-assertion1 is needed anymore.
;; If it is, syntaxer probably needs cleaning up.
  (if (and (boundp 'adb::*flip-truth-value*)
	  adb::*flip-truth-value*)
      `(,(find-qrel-assertion n1 n2 rel) . :FALSE)
      `(,(find-qrel-assertion n1 n2 rel) . :TRUE)))

(defun make-qrel-assertion (n1 n2 rel)
  `(,(find-qrel-assertion n1 n2 rel) . :TRUE))

;;; Accessor and helper functions

(proclaim '(special *derived-qstate*))

;;; ****** Do pushnew of n's instead of nums, then do mapcar at end.

(defun gather-qstate-numbers (&optional (all? nil) &aux result)
  (dolist (ineq (gather-assumed-qstate))
    (pushnew (ineq-num1 ineq) result :test #'equal)
    (pushnew (ineq-num2 ineq) result :test #'equal))
  (when all?
    (dolist (ineq *derived-qstate*)
      (pushnew (ineq-num1 ineq) result :test #'equal)
      (pushnew (ineq-num2 ineq) result :test #'equal)))
  result)

(defun gather-qstate-qnums (&optional (all? nil) &aux result)
  (dolist (ineq (gather-assumed-qstate))
    (pushnew (ineq-n1 ineq) result)
    (pushnew (ineq-n2 ineq) result))
  (when all?
    (dolist (ineq *derived-qstate*)
      (pushnew (ineq-n1 ineq) result)
      (pushnew (ineq-n2 ineq) result)))
  result)

(defun gather-quantity-numbers ()
  (mapcar #'(lambda (quant) `(A ,quant)) (gather-quantities)))

(defun opposite-order-relation (rel)
  (case rel
    (> '<)
    (< '>)
    (= '=)
    (t (cerror "" "~%   ~A bad ordering relationship -- opposite-order-relation"
	       rel))))

; More debugging goodies

(proclaim '(special *misordered*))

(defun test-inequalities-for-order ()
  (setq *misordered*
	(remove-if #'(lambda (ineq)
		    (equal (cdar ineq) (find-qrel (cadar ineq) (caddar ineq))))
		(append (adb::fetch '((Less-than . ?x) . ?y))
			(adb::fetch '((Greater-than . ?x) . ?y))
			(adb::fetch '((Equal-to . ?x) . ?y))))))

;;;; Accessors for QPE sets of quantities

(proclaim '(special *QC-Qstate* *Resolve-Qstate* *Derived-Qstate*
		    *Limit-Qstate*))

(defun gather-assumed-qstate (&optional (re-order? nil)) 
  (if re-order? ;; heuristically re-order for speed
      ;; Theory is that assumptions at the beginning of each list might
      ;; depend on those later in each list, but not vice-versa.  Therefore
      ;; reversing them will make the more primitive distinctions first.
      (append (reverse *qc-qstate*) (reverse *resolve-qstate*) (reverse *limit-qstate*))
      (append *QC-Qstate* *Resolve-Qstate* *Limit-Qstate*)))

(defun constant-ineq? (ineq)
  (or (tms::true? (ineq-< ineq))
      (tms::true? (ineq-> ineq))
      (tms::true? (ineq-= ineq))
      (tms::true? (ineq-bt ineq))))

(defun filter-ineq-choice-sets (ineqs)
  ;; Assumes you can nuke the input list
  (delete-if #'(lambda (ineq)
		 (and (null (ineq-choice-set ineq))
		      (fixed-ineq? ineq))) ineqs))

(defun gather-derived-qstate () *Derived-Qstate*)

(defun gather-total-qstate () (append (gather-assumed-qstate)
				      (gather-derived-qstate)))

(defvar *quantities* nil "List of possible quantities")
;; This list gets constructed bottom-up, as quantities are created.
(adb::init-form (setq *quantities* nil))
(defun gather-quantities () *quantities*)

;;;; Debugging utilities

(defun show-ineq (n1 n2)
  (let ((ineq (fetch-ineq n1 n2)))
    (cond (ineq (show-ineq1 ineq))
	  (t (format t
	      "~%Relationship between ~A and ~A not known to be interesting."
		     n1 n2)))
    ineq))

(defun show-ineq1 (ineq) ;; make externally available
  (format t "~%For ~A:" ineq)
  (format t "~%    >: ~A" (tms::label (ineq-> ineq)))
  (format t "~%    =: ~A" (tms::label (ineq-= ineq)))
  (format t "~%    <: ~A" (tms::label (ineq-< ineq))))

(defun show-number (n)
  (let ((qnum (if (qp-number? n) n (fetch-number n))))
    (cond (qnum 
	   (format t "~% For ~A:" qnum)
	   (format t "~%   Above: ~A" (mapcar #'car (qnum-ineqs1 qnum)))
	   (format t "~%   Below: ~A" (mapcar #'car (qnum-ineqs2 qnum))))
	  (t (format t "~%~A not a known number...")))))

(defun show-quantity (q)
  (let ((quant (if (qp-quantity? q) q (fetch-quantity q))))
    (cond (quant
	   (format t "~% For ~A:" quant)
	   (show-number (quant-amount quant))
	   (show-number (quant-derivative quant)))
	  (t (format t "~% ~D not a known quantity...")))))

(defun show-qstate (&optional (stream *standard-output*))
  (format stream "~%From Quantity Conditions (~D):" (length *QC-qstate*))
  (dolist (ineq *QC-Qstate*)
    (format stream "~%    ~40<~A~>~40<~A~>" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (format stream "~%From Influence Resolution (~D):" (length *resolve-qstate*))
  (dolist (ineq *resolve-qstate*)
    (format stream "~%    ~40<~A~>~40<~A~>" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (format stream "~%From Limit Analysis (~D):" (length *limit-qstate*))
  (dolist (ineq *limit-qstate*)
    (format stream "~%    ~40<~A~>~40<~A~>" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (format stream "~%Derived (~D):" (length *derived-qstate*))
  (dolist (ineq *derived-qstate*)
    (format stream "~%    ~40<~A~>~40<~A~>" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (format stream "~%Total = ~D." (+ (length *QC-qstate*)
			       (length *resolve-qstate*)
			       (length *limit-qstate*)
			       (length *derived-qstate*))))

(defun show-quantities (&optional (stream *standard-output*)
				  &aux (counter 0))
  (format stream "~% Quantities:")
  (dolist (q (sort (copy-list (gather-quantities)) #'adb::alphalessp))
    (format stream "~%   ~A" (adb::lprinc-string q)) (incf counter))
  (format t "~%   ~D quantities total." counter))

;;;;; Interface for installing inequality assumptions

(defvar *ineq-asn-table* nil) ;; key = inequality assumption, item = INEQ struct

(adb::init-form 
  (cond (*ineq-asn-table* (clrhash *ineq-asn-table*))
	(t (setq *ineq-asn-table* (make-hash-table :test #'eq)))))

(defun Create-inequality-assumption (nn1 nn2 &optional (reason ':ASSUMED))
  (multiple-value-bind (n1 n2) ;; Just being careful.
      (find-qrel-ordering nn1 nn2)
    (create-inequality-assumption1 n1 n2 
				   (fetch-ineq n1 n2)
				   reason)))

(defun assume-ineq-if-needed (ineq &optional (informant ':ASSUMED))
  (unless (ineq-assumed? ineq)
    (create-inequality-assumption1 (ineq-num1 ineq)
				   (ineq-num2 ineq)
				   ineq informant)))

(defun create-inequality-assumption1 (n1 n2 ineq informant)
    (let ((lt `((less-than ,n1 ,n2) . :TRUE))
	  (gt `((greater-than ,n1 ,n2) . :TRUE))
	  (eq `((equal-to ,n1 ,n2) . :TRUE))
	  (un `((unrelated ,n1 ,n2) . :TRUE)))
      (adb::oneof (list lt gt eq un)) ;; Tell ATMS this is a closed class of assumptions
      ;;; Now cache assumptions for limit analysis.
      ;;; Johan's optimization: skip assumptions known to be false. 
      (setf (ineq-assumed? ineq) informant)
      (when (adb::fact-assumption lt)
	(setf (ineq-lt-a ineq) (adb::fact-assumption lt))
	(push (ineq-lt-a ineq) (ineq-choice-set ineq))
	(setf (gethash (ineq-lt-a ineq) *ineq-asn-table*) ineq))
      (when (adb::fact-assumption eq)
	(setf (ineq-eq-a ineq) (adb::fact-assumption eq))
	(push (ineq-eq-a ineq) (ineq-choice-set ineq))
	(setf (gethash (ineq-eq-a ineq) *ineq-asn-table*) ineq))
      (when (adb::fact-assumption gt)
	(setf (ineq-gt-a ineq) (adb::fact-assumption gt))
	(push (ineq-gt-a ineq) (ineq-choice-set ineq))
	(setf (gethash (ineq-gt-a ineq) *ineq-asn-table*) ineq))
      (when (adb::fact-assumption un)
	(setf (ineq-bt-a ineq) (adb::fact-assumption un))
	(push (ineq-bt-a ineq) (ineq-choice-set ineq))
	(setf (gethash (ineq-bt-a ineq) *ineq-asn-table*) ineq))
      (adb:run)
      ineq))

(defun lookup-ordering (N1 N2 env)
  (unless (eq N1 (car (find-qrel N1 N2)))
    (error "~% ~A & ~A are not in order!!" N1 N2))
  (let* ((qnum1 (fetch-number N1))
	 (qnum2 (fetch-number N2))
	 (ineq (grab-ineq qnum1 qnum2)))
    (when ineq
      (let ((rel (inequality-under ineq env)))
	(case rel
	  (:greater-than '>)
	  (:less-than '<)
	  (:equal-to '=)
	  (:unrelated '?)
	  (:unknown nil))))))


