;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; QPE print utilities

(proclaim '(special *DUMPING* *Report-Mode*))

(defvar *Dumping* nil)
(defvar *Report-Mode* :ASCII) ;; Other values are :LATEX and :TROFF
;; WARNING -- :TROFF will not be supported much longer.

(defvar *dump-lh-details* t) ;; Printing mode switch
(defvar *dump-conjunctive-lhs* t) 

;; Use list instead of big string for easier on-line editing.
(defvar *latex-report-header*
  '("\\batchmode"
"\\documentstyle[12pt,alltt]{article}"
"\\pagestyle{myheadings}"
"\\makeatletter"
"\\newcommand{\\singlespacing}{\\let\\CS=\\@currsize\\renewcommand{\\baselinestretch}{0.5}\\small\\CS}"
"\\newcommand{\\normalspacing}{\\singlespacing}"
"\\newcommand{\\border}{\\rule{\\textwidth}{0.04in}}"
"\\makeatother"
"\\textheight 720pt"
"\\textwidth 500pt"
"\\topmargin -0.5in"
"\\oddsidemargin -0.5in"
"\\evensidemargin -0.5in"
"\\begin{document}"
"\\singlespacing"
""))

;; Some generally useful helpers for printing

(defun make-lp-binding-list (ll)
  (cons 'binding-list (mapcar #'(lambda (pair) 
				  (cons 'Binding (cons (car pair) (cdr pair))))
			      ll)))

(defun get-inequality-relation-name (thing)
  (case thing
    (GREATER-THAN (if *dumping* (case *report-mode*
				  ((:ASCII :TROFF) '>)
				  (:LATEX " \\( > \\) ")) ">"))
    (LESS-THAN (if *dumping* (case *report-mode*
				  ((:ASCII :TROFF) "<")
				  (:LATEX " \\( < \\) ")) "<"))
    (EQUAL-TO (if *dumping* (case *report-mode*
				  ((:ASCII :TROFF) "=")
				  (:LATEX " \\( = \\) ")) "="))
    (i> (if *dumping* (case *report-mode*
				  ((:ASCII :TROFF) "i>")
				  (:LATEX " \\( i> \\) ")) "i>"))
    (i< (if *dumping* (case *report-mode*
				  ((:ASCII :TROFF) "i<")
				  (:LATEX " \\( i< \\) ")) "i<"))
    (UNRELATED (cond (*dumping*
		      (case *REPORT-MODE*
			(:ASCII '*)
			(:LATEX "\\( \\downarrow \\)")
			(:TROFF '|\\(da|)))
		     (t #+Symbolics ' ;dagger
			#-Symbolics '*
			)))
    (T '?)))

;;;; More LaTeX helpers

(defmacro if-LaTeX (&rest stuff)
  `(when (eq *report-mode* :LATEX) ,@ stuff))

(defun LaTeX-start (stream environment) (format stream "~%\\begin{~A}" environment))
(defun LaTeX-end (stream environment) (format stream "~%\\end{~A}" environment))
(defun LaTeX-line-divider (stream) (format stream " \\\\"))

(defmacro With-LaTeX-Tabbing (stream &rest line-specs)
  `(progn (format ,stream "~%\\begin{tabbing}")
	  ,@ (mapcar #'(lambda (line-spec) `(format ,stream ,(car line-spec)
						    ,@ (cdr line-spec)))
		     line-specs)
	  (format ,stream "~%\\end{tabbing}")))

(defmacro With-LaTeX-centering (stream &rest line-specs)
  `(progn (format ,stream "~%\\begin{tabbing}")
	  ,@ (mapcar #'(lambda (line-spec) `(format ,stream ,(car line-spec)
						    ,@ (cdr line-spec)))
		     line-specs)
	  (format ,stream "~%\\end{tabbing}")))


;;;;Printing tables of envs.

(defun decode-env-assumptions (env &aux result)
  (dolist (ass (tms::env-assumptions env))
    (push (cons (tms::class-datum (tms::assumption-variable ass))
		(tms::assumption-value ass)) result))
  (sort result #'(lambda (x y) (adb::alphalessp (car x) (car y)))))

(defun translate-cdi-active-env (env &aux bare)
  (setq bare (decode-env-assumptions env))
  (dolist (entry bare)
    (rplaca entry (cadar entry)) ;flush the "(ACTIVE )" bit
    (rplacd entry (if (eq (cdr entry) ':TRUE) 'A 'I)))
  bare)

(defun make-cdi-solution-table (envs &aux table)
  (dolist (env envs)
    (push (cons (tms::env-unique (car env))
		(translate-cdi-active-env (car env))) table))
  (setq table (sort table #'(lambda (x y) (< (car x) (car y)))))
  (cons (mapcar #'car (cdar table)) ;the labels for the columns
	table))

;;;; Printing situations

(defun print-sclass (sit &optional (stream *terminal-io*) (full-cdi-names? t))
  (format stream "~%Sclass ~A, ~A situations:" (sclass-name sit)
	  (length (sclass-assumptions sit)))
  (format stream "~%   Status = ~A, Duration = ~A"
	  (sclass-status sit) (sclass-duration sit))
  (format stream "~%   IS:")
  (adb::lprinc (cons 'set-members
		(append (sclass-individuals sit)
			(sclass-view-instances sit)
			(sclass-process-instances sit))) stream)
  (cond (full-cdi-names?
	 (Let ((active-vis (remove-if-not #'(lambda (pair) (eq (cdr pair) 'ACTIVE))
				       (sclass-view-structure sit)))
	       (first? t))
	   (cond (active-vis
		  (format stream "~%    VS: ")
		  (dolist (vi-pair active-vis)
		    (cond (first? (setq first? nil))
			  (t (format stream "~%        ")))
		    (format stream "~A: " (car vi-pair))
		    (adb::lprinc (lookup-vi (car vi-pair)) stream)))
		 (t (format stream "~%        No active individual views."))))
	 (let ((active-pis (remove-if-not #'(lambda (pair) (eq (cdr pair) 'ACTIVE))
				       (sclass-process-structure sit)))
	       (first? t))
	   (cond (active-pis
		  (format stream "~%    PS: ")
		  (dolist (pi-pair active-pis)
		    (cond (first? (setq first? nil))
			  (t (format stream "~%        ")))
		    (format stream "~A: " (car pi-pair))
		    (adb::lprinc (lookup-pi (car pi-pair)) stream)))
		 (t (format stream "~%        No active processes.")))))
	(t (format stream "~%   VS: ") 
	   (adb::lprinc (make-lp-binding-list (sclass-view-structure sit)) stream)
	   (format stream "~%   PS: ")
	   (adb::lprinc (make-lp-binding-list
			  (sclass-process-structure sit)) stream)))
  sit)

(defun print-situation (env &optional (stream *terminal-io*) (full-cdi-names? t)
			    &aux sit)
  (unless (env-sclass? env)
    (error "~% ~A not a situation." env))
  (setq sit (env-sclass env))
  (format stream "~%Situation ~A, (One of ~D)" (sclass-name sit)
	  (length (sclass-assumptions sit)))
  (format stream "~%   Status = ~A, Duration = ~A"
	  (sclass-status sit) (sclass-duration sit))
  (format stream "~%   IS:")
  (adb::lprinc (cons 'set-members
		(append (sclass-individuals sit)
			(sclass-view-instances sit)
			(sclass-process-instances sit))) stream)
  (cond (full-cdi-names?
	 (Let ((active-vis (remove-if-not #'(lambda (pair) (eq (cdr pair) 'ACTIVE))
				       (sclass-view-structure sit)))
	       (first? t))
	   (cond (active-vis
		  (format stream "~%    VS: ")
		  (dolist (vi-pair active-vis)
		    (cond (first? (setq first? nil))
			  (t (format stream "~%        ")))
		    (adb::lprinc (lookup-vi (car vi-pair)) stream)))
		 (t (format stream "~%        No active individual views."))))
	 (let ((active-pis (remove-if-not #'(lambda (pair) (eq (cdr pair) 'ACTIVE))
				       (sclass-process-structure sit)))
	       (first? t))
	   (cond (active-pis
		  (format stream "~%    PS: ")
		  (dolist (pi-pair active-pis)
		    (cond (first? (setq first? nil))
			  (t (format stream "~%        ")))
		    (adb::lprinc (lookup-pi (car pi-pair)) stream)))
		 (t (format stream "~%        No active processes.")))))
	(t (format stream "~%   VS: ") 
	   (adb::lprinc (make-lp-binding-list (sclass-view-structure sit)) stream)
	   (format stream "~%   PS: ")
	   (adb::lprinc (make-lp-binding-list
			  (sclass-process-structure sit)) stream)))
  (print-env env stream)
  sit)

(proclaim '(special *Silence-zero-Ds-values*))

(defun print-ds-values (sit &optional (stream *terminal-io*) &aux printed-any? count)
  (setq count 0)
  (cond ((sclass? sit)
	 (dolist (q-pair (sclass-ds-values sit))
	   (incf count)
	   (unless (and *Silence-zero-Ds-values*
			(= (cdr q-pair) 0))
	     (unless printed-any? (setq printed-any? t))
	     (format stream "~%   ")
	     (adb::lprinc (cons `(s (D ,(car q-pair))) (cdr q-pair)) stream)))
	 (unless printed-any?
	   ;; Either there weren't any, zeros were being suppressed and they were all
	   ;; zero, or nothing was known.
	   (cond ((sclass-unknown-ds-values sit)
		  (cond ((and *silence-zero-ds-values* (> count 0))
			 (format stream "~% All known Ds values are zero."))
			(t (format stream "~%  No Ds values known."))))
		 ((and *silence-zero-ds-values* (> count 0))
		  (format stream "~% All Ds values are zero."))
		 (t (format stream "~% No quantities exist here.")))))
	(t ;; assume it is an environment
	 (let ((slist (list sit))
	       (val 0))
	   (dolist (q (gather-quantities))
	     (setq val (get-ds-value Q slist))
	     (unless val (setq val '??))
	     (unless (and *Silence-zero-Ds-values*
			  (numberp val)
			  (= val 0))
	       (unless printed-any? (setq printed-any? t))
	       (format stream "~%   ")
	       (adb::lprinc `((s (D ,Q)) . ,val))))
	   (unless printed-any? (format stream "~%  "))))))

;;;; Printing limit hypotheses

(defun princ-lh (lh &optional (details? nil)
		    (stream *terminal-io*)(say-name? t))
  (cond ((limit-hypothesis? lh)
	 (cond ((lh-single-change lh)
		(princ-single-change-lh lh stream say-name?))
	       (t ;; Conjunctive change
		(format stream "~A: Conjunction of ~A"
			(lh-name lh)
			(adb::lprinc-string
			  (cons 'set-members
				(mapcar #'(lambda (lh1) (lh-name lh1))
					(lh-singles lh)))))
		(when (or details? (not *dumping*))
		  (dolist (lh1 (lh-singles lh))
		    (cond ((eq *report-mode* :LATEX)
			   (format stream "~%\\newline ~%\\hspace{0.2in}~%"))
			   (t (format stream "~% ")))
		    (princ-single-change-lh lh1 stream say-name?))
		  (if-LaTeX (format stream "~%\\newline "))))))
	(t (error "~% ~A not a limit hypothesis." lh))))

(defun princ-single-change-lh (lh stream &optional (say-name? t))
  (let ((entry (car (lh-relations lh))))
    (cond (say-name? 
	   (format stream "~A:[~A~A]~A~%~A~A~A~A~A"
	    (case *report-mode*
	      (:ASCII (lh-name lh))
	      (:LATEX (format nil "{\\tt ~A}" (lh-name lh))))
	    (get-inequality-relation-name (second (lh-D-rels lh)))
	    (get-inequality-relation-name (third (lh-D-rels lh)))
	      (case *report-mode*
		(:ASCII (qnum-string (ineq-n1 (car entry))))
		(:LATEX (concatenate 'string "{\\tt " (qnum-string (ineq-n1 (car entry)))
				     "}")))
	      (case *report-mode*
		(:ASCII "          ")
		(:LATEX "\\hspace{\\parindent}"))
	      (case *report-mode*
		(:ASCII (order-short-string (cadr entry)))
		(:LATEX (concatenate 'string "{\\tt " (order-short-string (cadr entry))
				     "}")))
	      (case *report-mode*
		(:ASCII (qnum-string (ineq-n2 (car entry))))
		(:LATEX (concatenate 'string "{\\tt " 
				     (qnum-string (ineq-n2 (car entry))) "}")))
	    (case *report-mode*
	      (:ASCII "-->")
	      (:LATEX "\\( \\longrightarrow \\) "))
	    (get-inequality-relation-name (caddr entry))))
	  (t (if-LaTeX (format stream "{\\tt "))
	     (format stream "~A"
	  (let ((str (adb::lprinc-string `((,(cadr entry)
					    ,(view-qnum (ineq-n1 (car entry)))
					    ,(view-qnum (ineq-n2 (car entry)))) . :TRUE))))
	    (case *report-mode*
	      (:ASCII str)
	      (:LATEX  (concatenate 'string "{\\tt " str "}")))))
	     (if-LaTeX (format stream "}"))
	     (format stream "~A~A"
		     (case *report-mode*
		       (:ASCII "-->")
		       (:LATEX "\\( \\longrightarrow \\) "))
		     (get-inequality-relation-name (caddr entry)))))))

(defun print-lh (lh &optional (details? nil) (stream *terminal-io*))
  (terpri stream)
  (if-LaTeX (LaTeX-start stream "footnotesize") (LaTeX-start stream "alltt"))
  (princ-lh lh details? stream)
  (if-LaTeX (LaTeX-end stream "alltt") (LaTeX-end stream "footnotesize"))
  (if-LaTeX (LaTeX-start stream "footnotesize") (LaTeX-start stream "alltt"))
  (unless (conjunctive-lh? lh) (format stream "~%   Start: ~A" (lh-start-envs lh)))
  (format stream "~%     End: ~A" (lh-end-envs lh))
  (if-LaTeX (LaTeX-end stream "alltt") (LaTeX-end stream "footnotesize"))
  (when details?
    (if-LaTeX (LaTeX-start stream "scriptsize") (LaTeX-start stream "alltt"))
      (dolist (e (lh-start-envs lh)) (adb::print-env e stream))
      (dolist (e (lh-end-envs lh)) (adb::print-env e stream))
    (if-LaTeX (LaTeX-end stream "alltt") (LaTeX-end stream "scriptsize")))
  lh)

;;;; Summarizing limit hypothesis information

(defun show-lhs (&optional (details? nil)
			   (*report-mode* :ASCII) (stream *terminal-io*))
  (unless details?
    (if-LaTeX (LaTeX-start stream "tabbing")))
  (dolist (lh *limit-hypotheses*)
    (when (or (not *dumping*)
	      *dump-conjunctive-lhs*
	      (lh-single-change lh))
    (terpri stream)
    (princ-lh lh details? stream)
    (unless details? (if-LaTeX (LaTeX-line-divider stream)))))
  (unless details?
    (if-LaTeX (LaTeX-end stream "tabbing")))
  (length *limit-hypotheses*))

(defun lh-stats (&optional (stream *standard-output*) (asns? nil))
  (case *report-mode*
    ((:ASCII :TROFF) (format stream "~% LIMIT ANALYSIS STATISTICS"))
    (:LATEX (format stream "~%{\\center {\\bf Limit Analysis Statistics}}~%~%")))
  (multiple-value-bind (count total ave max min)
      (lhtable-statistics *situation-lh-table*)
    (if-LaTeX (LaTeX-Start stream "tabbing"))
    (format stream 
"~% ~A situations, ~D potential transitions (Max = ~D, Min = ~D, Ave =~3F)"
	    count total max min ave))
  (if-LaTeX (LaTeX-Line-Divider stream))
  (multiple-value-bind (count total ave max min)
      (lhtable-statistics *first-next-states*)
    (format stream 
 "~% Before Continuity: ~D potential transitions (Max = ~D, Min = ~D, Ave =~3F)"
  total max min ave))
  (if-LaTeX (LaTeX-Line-Divider stream))
 (multiple-value-bind (count total ave max min)
      (lhtable-statistics *contin-next-states*)
    (format stream 
 "~% Before ECL: ~D potential transitions (Max = ~D, Min = ~D, Ave =~3F)"
	    total max min ave))
  (if-LaTeX (LaTeX-Line-Divider stream))
  (multiple-value-bind (count total ave max min)
      (lhtable-statistics *next-state-table*)
    (format stream 
	    "~% Final result: ~D transitions (Max = ~D, Min = ~D, Ave =~3F)"
	    total max min ave))
  (if asns? (classify-ineq-assumptions stream))
  (if-LaTeX (LaTeX-end stream "tabbing")))

;;;; Global dump of limit hypotheses

(proclaim '(special *limit-hypotheses* *single-limit-hypotheses*))

(defun dump-lhs (&optional (stream *terminal-io*))
  (case *Report-Mode*
    (:ASCII
      (format stream
	      "~%There are ~D limit hypotheses, ~D singletons and ~D conjunctive.~%"
	      (length *limit-hypotheses*)
	      (length *single-limit-hypotheses*)
	      (- (length *limit-hypotheses*)
		 (length *single-limit-hypotheses*))))
    (:LATEX (format stream
      "~%\\begin{center}~%{\\bf Details of Limit Hypotheses} \\\\" )
	    (format stream
      "~%{\\bf There are ~D limit hypotheses, ~D singletons and ~D conjunctive.}~%\\end{center}~%"
	      (length *limit-hypotheses*)
	      (length *single-limit-hypotheses*)
	      (- (length *limit-hypotheses*)
		 (length *single-limit-hypotheses*)))))
  (dolist (lh *limit-hypotheses*)
    (when (or *dump-conjunctive-lhs* (lh-single-change lh))
      (terpri stream)
      (print-lh lh t stream))))

(defun summarize-LHs (&optional (stream *terminal-io*))
  (print-divider stream)
  (case *Report-Mode*
    (:ASCII (format stream "~%    --- Summary of Limit Hypotheses --- ~%"))
    (:LATEX (format stream "~%\\begin{center}~%{\\bf Summary of Limit Hypotheses}~%\\end{center}"))
    (:TROFF (format stream "~%.ce~%\fBLimit Hypotheses\f(fx~%~%")))
  (show-lhs nil *report-mode* stream)
  (print-divider stream))

;;;; Printers for debugging

(defun Qstate-under-env (env)
  (format t "~%For ~A:" env)
  (dolist (ineq (gather-assumed-qstate))
    (let ((loser? t) ;; *****inefficient, but this is for debugging
	  (n-pair (list (ineq-num1 ineq) (ineq-num2 ineq)))) 
      (dolist (rel '(less-than greater-than equal-to unrelated))
	(when (adb::implied-by? `((,rel ,@ n-pair) . :TRUE) env)
	  (adb::lprint `(,rel ,@ n-pair))
	  (setq loser? nil)))
      (if loser? (format t "~%Nothing for ~A, ~A" (adb::lprinc-string (car n-pair))
			 (adb::lprinc-string (cadr n-pair)))))))

;;;; Dumping procedures

(defvar *dump-situations* t)

(defun dump-situation (sit &optional (stream *terminal-io*))
  (if-LaTeX (LaTeX-start stream "footnotesize")(LaTeX-start stream "alltt"))
  (print-sclass sit stream)
  (if-LaTeX (LaTeX-end stream "footnotesize")(LaTeX-end stream "alltt"))
  (case *Report-Mode*
    (:ASCII (format stream "~%~%       -*- Ds Values -*- "))
    (:LATEX (format stream "~%~%\\vspace{0.1in}~%~%\\begin{center}~%{\\bf {\\tt Ds} values}~%\\end{center}~%")
	    (LaTeX-start stream "footnotesize")(LaTeX-start stream "alltt"))
    (:TROFF (format stream "~%~%       -*- Ds Values -*- ")))
  (print-ds-values sit stream)
  (if-LaTeX (LaTeX-end stream "alltt") (LaTeX-end stream "footnotesize"))
  (when *dump-situations*
    (case *Report-Mode*
      (:ASCII (format stream "~%~%       -*- Environments -*-"))
      (:LATEX (format stream "~%~%\\vspace{0.1in}~%~%\\begin{center}~%{\\bf Environments}~%\\end{center}~%")
	      (LaTeX-start stream "scriptsize") (LaTeX-start stream "alltt"))
      (:TROFF (format stream "~%~%       -*- Environments -*-")))
    (display-sclass-assumptions sit stream)
    (if-LaTeX (LaTeX-end stream "alltt") (LaTeX-end stream "scriptsize"))))


(defun display-sclass-assumptions (sclass &optional (stream *standard-output*)
					  &aux common residues)
  (cond ((not (and *differential-situation-display*
		   (cdr (sclass-assumptions sclass))))
	 (dolist (e1 (sclass-assumptions sclass))
	   (terpri stream) (adb::print-env e1 stream)))
	(t
	  (unless (get-sclass-prop sclass ':common)
	    (setq common (adb::envs-common-env (sclass-assumptions sclass)))
	    (setq residues (mapcar #'(lambda (sit)
				       (cons sit (adb::env-residue-env sit common)))
				   (sclass-assumptions sclass)))
	      (put-sclass-prop sclass common ':common)
	      (put-sclass-prop sclass residues ':diffs))
	  (setq common (get-sclass-prop sclass ':common))
	  (format stream "~%In common: ")
	  (adb::print-env common stream)
	  (format stream "~%Differences: (~D)" (length (sclass-assumptions sclass)))
	  (dolist (entry (get-sclass-prop sclass ':diffs))
	    (format stream "~% ~A + ~A = ~A" (cdr entry) common (car entry))
	    (adb::print-env (cdr entry) stream)))))

(defun print-new-page (stream) 
  (case *Report-Mode*
    (:ASCII (format stream "~%") (write-char #\page stream))
    (:LATEX (format stream "~%\\newpage~%"))
    (:TROFF (format stream "~%.bp~%"))))

;;;; Dumping envisionments
(proclaim '(special *next-state-table*
		    *QPE-version-number*
		    *situation-lh-table*
		    adb::*report-mode*))

(defun summarize-location (&optional (stream *standard-output*))
  (case *report-mode*
    (:ASCII (format stream "~%Run on: ~A, a ~A (~A) running ~A."
		    (machine-instance) (machine-type) (machine-version)
		    (software-type))
	    (format stream "~%        ~A, ~A" (lisp-implementation-type)
		    (lisp-implementation-version))
	    (format stream "~%        ~A" (long-site-name)))
    (:LATEX (format stream "~%\\begin{tabbing}")
	    (format stream "~% Run on: \\= ~A, a ~A (~A) running ~A. \\\\"
		    (machine-instance) (machine-type) (machine-version)
		    (software-type))
	    (format stream
 "~% \\> ~A, \\begin{minipage}[t]{5in}~A. \end{minipage}\\\\" 
 (hide-latex-schars (lisp-implementation-type))
 (hide-latex-schars (lisp-implementation-version)))
	    (format stream "~%\\end{tabbing}"))))

(defun dump-envisioner-results (test-name fname)
  (proclaim '(special *ps-vs*))
  (let ((the-time (summarize-current-time))
	(*dumping* t)
	(adb::*report-mode* qpe::*report-mode*)) ;; Pass down information
  (with-open-file (fout fname :direction :output)
    (case *Report-Mode*
      (:ASCII (format fout "~% ~A~% ~A" test-name the-time)
	      (format fout "~% QPE (version ~A) Envisionment" *version-number*)
	      (format fout "~%   ATMoSphere version ~A, ATMS version ~A"
		      adb::*version-number* tms::*version*))
      (:LATEX (dolist (line *latex-report-header*)
		(format fout "~%~A" line))
	      (LaTeX-start fout "center")
	      (format fout "~%~A \\\\~% ~A \\\\"
		       test-name the-time)
	      (format fout "~% QPE (version ~A) Envisionment \\\\" *version-number*)
	      (format fout "~% ATMoSphere version ~A, ATMS version ~A."
		      adb::*version-number* tms::*version*)
	      (LaTeX-end fout "center")))
    (summarize-location fout)
    (summarize-qpe-statistics test-name fout)
    (append-user-supplied-reports fout)
    (print-new-page fout)
    (cond ((eq *report-mode* :LATEX) (LaTeX-cdis fout))
	  (t (format fout "~% ---- View and Process Instances ----")
	     (show-cdis fout)))
    (print-new-page fout)
    (print-divider fout)
    (cond ((eq *report-mode* :LATEX)
	   (format fout "~%\\begin{center}~%{\\bf Union of Quantity Space information}~%\\end{center}~%"))
	  (t (format fout "~% ---- Union of Quantity Space information ----")))
    (if (eq *report-mode* :LATEX) (latex-qstate fout) (show-qstate fout))
    (print-divider fout)
    ;;; Brief description of lh's & transitions.
    (when (and (boundp '*limit-hypotheses*)
	       *limit-hypotheses*)
      (print-new-page fout)
      (if-LaTeX (LaTeX-start fout "small"))
      (summarize-lhs fout)
      (if-LaTeX (LaTeX-end fout "small")))
    (when (and (boundp '*next-state-table*)
	       *next-state-table*)
	  (print-new-page fout)
	  (print-next-state-table fout))
    (map-sclasses #'(lambda (sit) 
		      (print-new-page fout)
		      (dump-situation sit fout))
		    *ps-vs*)
    ;; Dump information about limit hypotheses
    (when (and (boundp '*limit-hypotheses*)
	       *limit-hypotheses*
	       *dump-lh-details*)
      (print-new-page fout)
      (dump-lhs fout))
    ;; Dump information about influences
    (print-new-page fout)
    (if-LaTeX (LaTeX-start fout "alltt"))
    (show-possible-directs fout)
    (if-LaTeX (LaTeX-end fout "alltt"))
    (print-new-page fout)
    (if-LaTeX (LaTeX-start fout "alltt"))
    (show-possible-indirects fout)
    (if-LaTeX (LaTeX-end fout "alltt"))
    (if-LaTeX (LaTeX-end fout "document")))))

;;;; Some support routines

(defun print-divider (fout)
  (case *report-mode*
    (:ASCII (format fout "~%===================================="))
    (:LATEX (format fout "~%~%\\border~%~%"))
    (:TROFF (format fout "~%===================================="))))

(defun count-envisioned-situations ()
  (let ((class-counter 0)
	(sit-counter 0))
    (map-sclasses #'(lambda (s)
		      (incf class-counter)
		      (setq sit-counter (+ (length (sclass-assumptions s))
					   sit-counter))))
    (values sit-counter class-counter)))

(defvar *possible-user-report-hooks* nil) ;; List of (<unique name> <string> <procedure>)
(defvar *user-report-hooks* nil) ;; Actual procedures chosen this time

(defun append-user-supplied-reports (&optional (stream *standard-output*)) 
  ;; Each user report will appear on a seperate page
  (dolist (user-report-generator *user-report-hooks*)
    (print-new-page stream)
    (funcall user-report-generator stream)))

(defmacro defUser-Report-Generator (name string fun)
  ;; Use name to allow easy editing of the list.
  `(let ((entry (assoc ',name *possible-user-report-hooks*)))
     (unless entry (setq entry (list ',name nil nil))
	     (push entry *possible-user-report-hooks*))
     (setf (cadr entry) ,string)
     (setf (caddr entry) (function ,fun))))

;;;; Printing QPE statistics
(defun summarize-qpe-statistics (&optional (test-name "current")
					   (fout *standard-output*))
  (print-divider fout)
  (case *report-mode*
    (:ASCII
     (format fout "~%     Envisoner results, ~A" test-name)
     (multiple-value-bind (sit-count class-count)
	 (count-envisioned-situations)
       (format fout "~%~% ~D situations, in ~D classes" sit-count class-count))
     (format fout "~% ~D Domain files: " (length *domain-files*))
     (dolist (physics-file *domain-files*)
       (format fout "~%     ~A" physics-file))
     (format fout "~% ~D Example files: " (length *scenario-files*))
     (dolist (example-file *scenario-files*)
       (format fout "~%     ~A"  example-file))
     (cond (*Augment-files*
	    (format fout "~% ~D User augments: " (length *Augment-Files*))
	    (dolist (file *Augment-Files*)
	      (format fout "~%     ~A" file)))
	   (t (format fout "~% No user augments declared."))))
    (:LATEX
     (multiple-value-bind (sit-count class-count)
       (count-envisioned-situations)
       (With-LaTeX-Centering fout ("~%Envisioner results, ~A \\\\" test-name)
			     ("~%~D situations, in ~D sclasses. \\\\" sit-count class-count))
       (format fout "~%\\begin{tabbing}")
       (format fout "~% ~D Domain files: \\= \\\\" (length *domain-files*))
       (mapcar #'(lambda (dfile)
		   (format fout "~% \\> ~A \\\\" (hide-latex-schars dfile)))  *domain-files*)
       (format fout "~% ~D Scenario files: \\= \\\\" (length *scenario-files*))
       (mapcar #'(lambda (dfile)
		   (format fout "~% \\> ~A \\\\" (hide-latex-schars dfile))) *scenario-files*)
       (cond (*Augment-files*
	      (format fout "~% ~D Domain files: \\= \\\\" (length *domain-files*))
	      (mapcar #'(lambda (dfile)
			  (format fout "~% \\> ~A \\\\" (hide-latex-schars dfile)))
		      *Augment-files*))
	     (t (format fout "~% No user augments declared.")))
       (format fout "~%\\end{tabbing}"))))
  (print-divider fout)
  (print-QPE-modes fout)
  (print-divider fout)
  (lh-stats fout)
  ;; Put ADB statistics on next page.
  (unless (eq fout *terminal-io*) (print-new-page fout))
  (print-divider fout)
  (if-LaTeX (LaTeX-start fout "alltt"))
  (adb::print-statistics fout)
  (if-LaTeX (LaTeX-end fout "alltt"))
  (print-divider fout))

;;;; Specialized print routines for envisionment dumps

(defun print-next-state-table (&optional (stream *terminal-io*))
  (let ((entries (remove-if #'(lambda (pair) (null (cdr pair)))
			 *next-state-table*)))
    (case *report-mode*
      (:ASCII (print-divider stream)
	      (format stream "~% ----- STATE TRANSITION TABLE ------ ~%")
	      (format stream "~%~D states out of ~D have candidate transitions.~%"
		      (length entries) (count-situations)))
      (:LATEX (With-LaTeX-Centering stream ("{\\bf State Transition Table}"))))
      (if-LaTeX (LaTeX-start stream "footnotesize")(LaTeX-start stream "alltt"))
    (dolist (entry entries)
      (format stream "~% ~A :   " (car entry))
      (adb::lprinc
	(make-lp-binding-list
	  (mapcar #'(lambda (pair) (cons (lh-name (car pair)) (cdr pair)))
			   (cdr entry)))
		 stream))
      (if-LaTeX (LaTeX-end stream "alltt")(LaTeX-end stream "footnotesize"))      
      (print-divider stream)))

(defun print-situation-lh-table (&optional (stream *terminal-io*) &aux entries)
  (setq entries (remove-if #'(lambda (pair) (null (cdr pair))) 
			   *situation-lh-table*))
  (print-new-page stream)
  (case *report-mode*
      (:ASCII (format stream "~%     STATE -- LIMIT HYPOTHESIS TABLE  ~%~%~%"))
      (:LATEX (With-LaTeX-Centering stream ("{\\bf State --- Limit Hypothesis Table}"))))
  (format stream "~%~D states out of ~D have associated limit hypotheses.~%"
	  (length entries) (count-situations))
    (if-LaTeX (LaTeX-start stream "alltt"))
  (dolist (entry entries)
    (format stream "~% ~A :   " (car entry))
    (adb::lprinc (cons 'set-members 
		  (mapcar #'(lambda (lh) (lh-name lh)) (cdr entry)))
	    stream))
    (if-LaTeX (LaTeX-end stream "alltt")))

;;;; Printing modes

(proclaim '(special *QPE-MODES*))

(defun print-QPE-modes (&optional (stream *terminal-io*))
  ;;; Shows mode information
  (case *report-mode*
    (:ASCII
       (format stream "~%QPE MODES")
       (dolist (mentry *QPE-Modes*)
	 (format stream "~% ~A" (if (eval (car mentry))
				    (caddr mentry)
				    (cadddr mentry)))))
    (:LaTeX 
     (format stream "~%~%\\begin{tabbing}")
     (format stream "~%QPE \\= Modes: \\\\")
     (mapcar #'(lambda (mentry)
		 (format stream "~% \\> ~A \\\\" (if (eval (car mentry))
						     (caddr mentry)
						     (cadddr mentry))))
	     *QPE-modes*)
     (format stream "~%~%\\end{tabbing}"))))

;;;; Yet more debugging stuff

(defun summarize-Ds-values (&optional (stream *standard-output*))
  (let ((table (mapcar #'(lambda (q) ;; (<q> <-1> <0> <1> <bt>)
			   (list q 0 0 0 0)) (gather-quantities))))
    (setq table (sort table #'adb::alphalessp :key #'car))
    (map-sclasses
      #'(lambda (scl)
	  (let ((Ds-table (sclass-Ds-values scl))
		(nsits (length (sclass-assumptions scl)))
		(answer nil))
	  (dolist (entry table)
	    (setq answer (assoc (car entry) Ds-table
				:test #'equal))
	    (cond ((null answer) (incf (cadr (cdddr entry)) nsits))
		  (t (case (cdr answer)
		       (-1 (incf (cadr entry) nsits))
		       (0 (incf (caddr entry) nsits))
		       (1 (incf (cadddr entry) nsits)))))))))
    (format stream "~%~3:< ~>~45:<Quantity~>~5:<-1~>~5:<0~>~5:<1~>~5:<BT~>")
    (dolist (entry table) 
      (format stream "~%~3:< ~>~45:<~A~>~5:<~D~>~5:<~D~>~5:<~D~>~5:<~D~>"
	      (adb::lprinc-string (car entry)) (cadr entry) (caddr entry) (cadddr entry)
	      (cadr (cdddr entry))))))

;;;; Ancillary print functions

(defun latex-qstate (&optional (stream *standard-output*))
  (format stream "~%~%{\\bf From Quantity Conditions (~D):}~%\\vspace{0.1in}~%" (length *QC-qstate*))
  (LaTeX-start stream "center")
  (LaTeX-start stream "tabular")
  (format stream "{ll}~%")
  (dolist (ineq *QC-Qstate*)
    (format stream "~%{\\tt ~A} & {\\tt ~A} \\\\" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (LaTeX-end stream "tabular")
  (LaTeX-end stream "center")
  (format stream "~%{\\bf From Influence Resolution (~D):}~%\\vspace{0.1in}~%" (length *resolve-qstate*))
  (LaTeX-start stream "center")
  (LaTeX-start stream "tabular")
  (format stream "{ll}~%")
  (dolist (ineq *Resolve-Qstate*)
    (format stream "~%{\\tt ~A} & {\\tt ~A} \\\\" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (LaTeX-end stream "tabular")
  (LaTeX-end stream "center")
  (format stream "~%{\\bf From Limit Analysis (~D):}~%\\vspace{0.1in}~%" (length *limit-qstate*))
  (LaTeX-start stream "center")
  (LaTeX-start stream "tabular")
  (format stream "{ll}~%")
  (dolist (ineq *Limit-Qstate*)
    (format stream "~%{\\tt ~A} & {\\tt ~A} \\\\" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (LaTeX-end stream "tabular")
  (LaTeX-end stream "center")
  (format stream "~%{\\bf Derived (~D):}~%\\vspace{0.1in}~%" (length *derived-qstate*))
  (LaTeX-start stream "center")
  (LaTeX-start stream "tabular")
  (format stream "{ll}~%")
  (dolist (ineq *derived-Qstate*)
    (format stream "~%{\\tt ~A} & {\\tt ~A} \\\\" (adb::lprinc-string (ineq-num1 ineq))
	    (adb::lprinc-string (ineq-num2 ineq))))
  (LaTeX-end stream "tabular")
  (LaTeX-end stream "center")
  (format stream "~%{\\bf Total = ~D.}" (+ (length *QC-qstate*)
			       (length *resolve-qstate*)
			       (length *limit-qstate*)
			       (length *derived-qstate*))))


(defun LaTeX-cdis (&optional (stream *standard-output*))
  (format stream "~%\\begin{center}~%{\\bf View and Process Instances}~%\\end{center}~%")
  (format stream "~%~%\\vspace{0.1in}~%~%")
  (LaTeX-cdi-list stream *pi-list* "Process instances:" "No process instances." )
  (format stream "~%~%\\vspace{0.1in}~%~%")
  (LaTeX-cdi-list stream *vi-list* "View instances:" "No view instances."))

(defun LaTeX-cdi-list (&optional (stream *standard-output*) cdi-list pos-header nega)
  (cond (cdi-list
	 (format stream "~%{\\bf ~A}" pos-header)
	 (LaTeX-start stream "tabbing")
	 (dolist (entry (reverse cdi-list))
	   (format stream "~%{\\tt ~A} $=$ {\\tt ~A} \\\\"
		   (car entry) (adb::lprinc-string (caddr entry))))
	 (LaTeX-end stream "tabbing"))
	(t (format stream "~%{\\bf ~A}" nega))))

;;;; Some LaTeX-oriented printing utilities.

(defvar *tab-spaces* "        ")

(defvar *latex-substitutions* 
  `((#\Tab . ,*tab-spaces*)
    ;; The order of these is important!
    (#\\ . "\\(\\backslash\\)")
    (#\< . "\\(<\\)")
    (#\> . "\\(>\\)")
    (#\# . "\\#")
    (#\& . "\\&")
;;    (#\$ . "\\$") Assume not in source.  O.W., replace backslash
;; stuff by substring replacement.
    (#\{ . "\\{")
    (#\} . "\\}")
    (#\% . "\\%")))

(defun hide-latex-schars (line)
  ;; Puts backslashes in front of characters latex
  ;; treats as special.  Also removed tabs.
  (dolist (sub *latex-substitutions* line)
	  (setq line (replace-char-by-string (car sub) (cdr sub) line))))

(defun replace-char-by-string (char sub str)
  (do ((pos (position char str))
       (inc (length sub)))
      ((null pos) str)
      (setq str (concatenate 'string (subseq str 0 pos)
			     sub
			     (subseq str (1+ pos))))
      (setq pos (position char str :start (+ pos inc)))))

(defun order-short-string (rel)
  (case rel
    (greater-than (adb::get-gt-symbol))
    (less-than (adb::get-lt-symbol))
    (equal-to (adb::get-eq-symbol))
    (unrelated (adb::get-unrelated-symbol))))



;;;; Some old stuff
;; The next few functions are dead right now,
;; but something like them will be needed when
;; domain-sets are re-introduced.

(defun cwa-form? (form) (or (eq (car form) 'SET-CWA) (eq (cadr form) 'set-cwa)))

(defun get-cwa-parts (form)
  (cond ((eq (car form) 'set-cwa)
	 (values (cadr form) (cadddr form) (caddr form) 
		 (cadr (cdddr form))
		 (caddr (cdddr form))))
	((eq (cadr form) 'set-cwa)
	 (values (car form) (cadddr form) (caddr form) 
		 (cadr (cdddr form))
		 (caddr (cdddr form))))
	(t (cerror "" "~A not a closed-world assumption..." form))))

(defun members-statement? (form)
  (and (symbolp (cadr form)) 
       (get (cadr form) '**LOCAL-ELEMENT-RELATION**)))

(defun membership-statement? (form) 
  (and (symbolp (cadr form)) 
       (get (cadr form) '**LOCAL-TOTAL-RELATION**)))

(defun get-set-members-statement-parts (form)
  (values (car form) (cadr form) (caddr form)))

(defun get-set-membership-statement-parts (form)
  (values (car form) (get (cadr form) '**LOCAL-TOTAL-RELATION**) (caddr form)))

;;;; Tabulating PS-VS information

;;Only designed for small examples.  Need industrial-grade tables for big ones

(defun tabulate-psvs (&optional (stream *standard-output*)
				&key (active-char "A")
				(inactive-char "I")(bozo-char "?"))
  (labels ((lookup-cdi-sclass-status (cdi sclass)
	     (let ((entry (assoc cdi (sclass-process-structure sclass))))
	       (cond (entry (case (cdr entry)
			      (ACTIVE active-char)
			      (INACTIVE inactive-char)
			      (t bozo-char)))
		     ((setq entry (assoc cdi (sclass-view-structure sclass)))
		      (case (cdr entry)
			(ACTIVE active-char)
			(INACTIVE inactive-char)
			(t bozo-char)))
		     (t bozo-char)))))
    (let ((cdis (append (sort (mapcar #'car *vi-list*) #'adb::alphalessp)
			(sort (mapcar #'car *pi-list*) #'adb::alphalessp)))
	  (sclasses (list-all-sclasses)))
      (setq sclasses (sort sclasses #'adb::alphalessp))
      (format stream "~%~6<~>")
      (dolist (cdi cdis)
	(format stream "~4<~A~>" cdi))
      (dolist (sclass sclasses)
	(format stream "~%~6<~A~>" (sclass-name sclass))
	(dolist (cdi cdis)
	  (format stream "~4<~A~>" (lookup-cdi-sclass-status cdi sclass)))))))
