;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;;  Influence Resolution
;;
;; Operates by finding for each quantity all the ways that it
;; could be influenced, and for each such set of influences, determining
;; what the Ds value could be.  This is a more efficient scheme than resolving
;; each quantity once per situation, since typically there is a great deal of
;; overlap in the patterns of influence.

;; Doing this efficiently means caching.  The table *DP-INFLUENCERS*
;; is augmented whenever an I+ or I- appears.  The entries in this table take
;; the form
;;
;; (<quantity> . (<ACTIVE statement> . <direct influence statements>) ...)
;;
;;  {Older versions had the influence statement as the sub-key, which made it
;;  inconvenient to deal with processes that provided more than one influence on
;;   a quantity}
;;
;; The table *IP-INFLUENCERS* has a similar format; entries look like
;;
;;  (<quantity> . (<predicate statement> . <qualitative proportionalities>)...)
;;
;; Once again, it is possible for a particular definition to attach more than one
;; constraint to a particular quantity.  Notice that the key, which is interpreted as
;; the antecedent for the influences that make up the item for the subindex, must
;; be a predicate statement and not a function.

;; Unfortunately, while the direction of causality is strictly enforced for any particular
;; time, QP theory does not forbid it from being different at distinct times.  This means
;; that causality does not impose a time-independent global ordering, and thus there is no  
;; distinguished place to start.  For simplicity, we simply install justifications
;; for the results of influence resolution assuming that all "input" Ds values to Qprops
;; and all inequalities between rates are possible. 

;; For every direct influencer N, N comp ZERO will be searched.
;; By becoming part of the state description, we will guarentee that the sign
;; is known for every situation.  Sloppy models which let them float will be slower,
;; thus upping the ante for making one's model more precise.

;;; CHANGES WITH THIS VERSION (2.6)

;; Given quantity Q, the quantity Net-Influence(Q) will always exist whenever Q is
;; directly influenced.  


;;;;; Variables

(defvar *DP-INFLUENCERS* nil "direct P-influencers")
(defvar *IP-INFLUENCERS* nil "indirect P-influencers")
(adb::Init-Form (setq *DP-INFLUENCERS* nil))
(adb::Init-Form (setq *IP-INFLUENCERS* nil))

(defvar *resolve-qstate* nil
  "Assumptions about rates introduced during influence resolution.")
(adb::Init-Form (setq *resolve-qstate* nil))
;;; Resolving ambiguous direct influences requires adding assumptions about the relationship
;;; between flow rates.  This table caches this information, since proofs and transitivity violations  
;;; involving these inequalities must be found to finish augmenting state descriptions
;;;  before limit analysis. 

(defvar *R-Completion-Sclasses* nil "Situations created as R-completions")
(adb::Init-Form (setq *R-Completion-Sclasses* nil))
;;; Build up a list of situations created as R-completions for debugging purposes.

(defun directly-influenced? (q)
  (assoc q *DP-INFLUENCERS* :test #'EQUAL))

(defun indirectly-influenced? (q)
  (assoc q *IP-INFLUENCERS* :test #'EQUAL))

(defun uninfluenced? (q)
  (not (or (directly-influenced? q)
	   (indirectly-influenced? q))))

;;;; Entry into influence resolution

(defun resolve-influences ()
  (When-Careful
    (setq *QPE-Program-Counter* "Resolve"))
  (run-qpe-events :BEFORE :IR)
  ;;; First install the basic justifications, independent of situation.
  ;;; This requires going through the influencers tables and resolving each entry
  ;;; in all possible ways.
  (resolve-all-direct-influences)
  (resolve-all-indirect-influences)
  ;;; Then figure out the circumstances when each quantity is uninfluenced,
  ;;; and install the appropriate justifications.  (Could speed this up by keeping the
   ;;; influencer tables in sorted order, and doing a synchronized walk down them....)
  (dolist (q (gather-quantities)) (construct-uninfluenced-justification q))
  (unless *incremental-transitivity* (process-transitivity))
  (construct-Sclasses)
  (expand-derived-qstate-by-derivatives *resolve-qstate*)
  (run-qpe-events :AFTER :IR))

(defun resolve-all-direct-influences ()
  (dolist (entry *DP-INFLUENCERS*)
    (resolve-direct-influences-on (car entry) (cdr entry))))

(defun resolve-all-indirect-influences ()
  (dolist (entry *IP-INFLUENCERS*)
    (resolve-indirect-influences-on (car entry) (cdr entry))))

;;;; Updating Sclasses with results of influence resolution so far

(defun Add-unambiguous-Ds-information (&aux quantities)
;;This function records the known and unknown Ds values
;;(ignoring quantities that don't exist within the situation)
;;and updates the situation's status flag appropriately. 
;; FORMAT CHANGE: The key is now the quantity itself, rather than Ds[q].
;; ...saves a few cons cells, makes constructing keys and alists easier.
  (setq quantities (gather-quantities))
    ;;;There are certain domain models,  so far a subset of those which peg some
    ;;; Ds value to be constant over time, that cause situations to be discovered as nogood
    ;;; as a consequence of the resolution process.  Here we prune them out.
  (dolist (sit *PS-VS*)
    (do ((qs (cdr quantities) (cdr qs))
	 (q (car quantities) (car qs))
	 (unknowns nil)
	 (envs (sclass-assumptions sit))
	 (knowns nil)
	 (val nil))
	((null q) 
	 (setf (sclass-Ds-values sit)
	       (setq knowns (nreverse knowns))) ;assumes sorted
	 (setf (sclass-unknown-Ds-values sit) (setq unknowns (nreverse unknowns))) 
	 (adb::if-debug-flag resolve
	   (format t "~%In ~A, ~D knowns, ~D unknowns." (sclass-name sit)
		 (length knowns) (length unknowns)))
	 (if unknowns (setf (sclass-status sit) 'PS-VS-COMPLETE)
	     (setf (sclass-status sit) 'R-COMPLETE)))
      (setq val (get-Ds-value q envs))
      (if val
	  (unless (eq val 'BT) (push (cons q val) knowns))
	  (push q unknowns)))))

(defun get-Ds-value (q envs) ;; Low down and dirty version
  (let ((Ds-form '((s (D q)) . val))
	(Q-form '((Quantity q) . :TRUE))
	(node nil)) ;; constant
    (setf (cadar Q-form) q)
    (setq node (get-tms-node Q-form))
    (unless (every #'(lambda (e) (tms::true-in? node e)) envs)
      (When-Careful
      (if (some #'(lambda (e) (tms::true-in? node e)) envs)
	  (add-recorder-event :Mixed-Quantity-Existence q envs)))
      (return-from get-Ds-value 'BT))
    (setf (car (cdadar Ds-form)) q)
    (dolist (val '(BT -1 0 1))
      (setf (cdr Ds-form) val)
      (setq node (get-tms-node Ds-form))
      (if (every #'(lambda (e) (tms::true-in? node e)) envs)
	  (return val)))))

;;;; Resolving uninfluenced quantities

(defun construct-uninfluenced-justification (q &aux ante-sets justs)
  ;enforces law that uninfluenced quantity has Ds=0. 
  (setq ante-sets 
	(nconc (mapcar #'car
		       (cdr (assoc q *DP-INFLUENCERS* :test #'equal)))
	       (mapcar #'car
		       (cdr (assoc q *IP-INFLUENCERS* :test #'equal)))))
  (setq justs (adb::compute-negated-antecedents ante-sets))
  (cond ((null justs)
	    ;; In this case nothing ever influences the quantity.  Thus we can assert
	   ;; globally that it never changes.
	   (adb::if-debug-flag uninfluenced
			  (format t "~%    IR: ~A never changes."
				  (adb::lprinc-string q)))
	   ;; Of course, if it depends on some individuals, then we must
	   ;; limit the justification
	   ;; to be true only when those individuals exist.
	   (adb::justify `((Equal-To (D ,q) ZERO) . :TRUE)
			 `(((quantity ,q) . :TRUE))
			 ':Uninfluenced-when-existing))
	  (t ;; Previous version distinguished several cases, such as figuring out
	   ;; if the quantity could really exist when the justification held.  Since 
	   ;; the cost is one more justification per quantity, this version just
	   ;; build the justification and is done with it.
	   (adb::if-debug-flag uninfluenced 
	    (format t "~%     IR: ~A uninfluenced when:" (adb::lprinc-string q))
	    (dolist (just justs)
	      (format t "~%   ~A" (adb::lprinc-string (cons 'AND just))))
	    (format t "."))
	   (dolist (just justs)
	     (adb::justify `((Equal-To (D ,q) ZERO) . :TRUE)
			(cons `((quantity ,q) . :TRUE) just)
			':Uninfluenced-Case)))))

;;;; Resolving direct influences

;; 1. Create the Net-Influence quantity for Q.
;; 2. Close all I+ and I- statements.
;; 3. For each consistent combination of I+ and I- statements,
;;    3.1 If there are no influencers, do nothing.
;;    3.2 If there is one positive influencer, Net-Influence = influencer.
;;    3.3 If there is one negative influencer, Net-Influence = - influencer.
;;    3.4 If there is one negative and one positive influencer,
;;           Net-Influence = Pinfluencer - Ninfluencer.
;;    3.5 If there are multiple influences, install sign constraints and mark
;;           ambiguous cases.

(adb::after-init-form 
  (progn (defQuantity-Type Net-Influence constant individual)
	 ;; All right, this is stretching things, ontologically speaking.
	 ;; Every quantity except special quantities have to be a property of something.
	 ;; Making QPE an individual is just a foil to keep from introducing
	 ;; yet more special quantities, which are painful to handle.
	 (adb::assert! '(Individual QPE))
	 (adb::assert! '(Exists QPE))))

(defvar *NI-quantities* nil) ;; Cache of NI quantities
(adb::init-form (setq *NI-quantities* nil))

(defun net-influence-ineq? (ineq)
  (when (eq (ineq-n2 ineq) *zero-struct*)
    (let ((form (lisp-form (qnum-assert (ineq-n1 ineq)))))
      (and (listp form)
	   (listp (cadr form))
	   (eq (caadr form) 'Net-influence)))))

;;;; Computing direct influences via net-influence quantities

(defun resolve-direct-influences-on (q directs &aux q-form ni ni-form)
  ;; Create the net influence quantity
  (setq ni (list 'NET-INFLUENCE q 'QPE)
	q-form `((Quantity ,q) . :TRUE) 
	ni-form `((Quantity, ni) .  :TRUE))
  (push ni *NI-quantities*)
  (assume-inequality-if-needed `(A ,ni) 'ZERO)
  (dolist (entry directs)
    (adb::justify ni-form (cons q-form (car entry)) :DI-consequence)
    (let ((this-ante (car entry)))
;;	    (cons q-form (cons ni-form (car entry))))) ;; EXPERIMENT---JWC
      (dolist (inf-statement (cdr entry))
	(adb::close-fact inf-statement :TRUE)
	(let ((sign (case (car inf-statement) (I+ ':+) (I- ':-))))
	  (add-indirect-influence-on
	    ni (quantity-of (third inf-statement)) sign this-ante)))))
  (adb::justify (cons (find-qrel-assertion `(A ,ni) `(D ,q) 'Equal-To) :TRUE)
		(list q-form ni-form) :DI-effect)
  ;(adb::assert! (cons (find-qrel-assertion `(A ,ni) `(D ,q) 'Greater-Than) :FALSE))
  ;(adb::assert! (cons (find-qrel-assertion `(A ,ni) `(D ,q) 'Less-Than) :FALSE))
  (adb::close-fact (car ni-form) :TRUE)
  (dolist (solution  ;; Go for conciseness this version
	    (adb::find-all-consistent-maximal-combinations
	      (mapcan #'(lambda (entry)
			  (mapcar #'(lambda (inf)
				      (list (cons inf :TRUE)
					    (cons inf :FALSE)))
				  (cdr entry))) directs)))
    (install-DI-consequences q ni ni-form solution)
    (adb::run)))

(defun add-indirect-influence-on (q influencer sign antecedents
				    &optional (informant :Internal-Qprop-Addition))
  (let ((form (list (case sign (:+ 'Qprop) (:- 'Qprop-)) q influencer)))
    (adb::justify (cons form ':TRUE) (cons `((Quantity ,influencer)
					    . :TRUE) antecedents)
		 informant)
    (adb::add-to-nested-alist (list q antecedents) form *IP-Influencers*)))

;;;; Installing direct influences consequences

(defun install-DI-consequences (q ni ni-form inf-set &aux pluses minuses common antes)
  ;; Must sort out number of influences of each type, do case analysis.
  (setq antes (cons ni-form (cons `((quantity ,q) . :TRUE) inf-set)))
  (dolist (inf inf-set)
    (unless (eq (cdr inf) :FALSE) ;; ignore this one
      (case (caar inf)
	(I+ (push (third (car inf)) pluses))
	(I- (push (third (car inf)) minuses)))))
  ;; subtract identical terms, if any
  (setq common (intersection pluses minuses :test #'equal)
	pluses (set-difference pluses common :test #'equal)
	minuses (set-difference minuses common :test #'equal))
  (adb::if-debug-flag Resolve
    (format t "~%     IR: ~A combination of ~A (I+) and ~A (I-)."
	    (adb::lprinc-string q) 
	    (adb::lprinc-string `(set ,@ pluses)) 
	    (adb::lprinc-string `(set ,@ minuses))))
  (install-sum-consequences `(A ,ni) pluses minuses antes))

;;;; Installing consequences of sums

(defun install-sum-consequences (result pluses minuses antes)
  ;; Called by resolve and by sum code
  (cond ((and (null pluses) (null minuses))
	 ;; Ds = 0, due to cancellation 
	 (adb::if-debug-flag adder
			     (format t "~%     Sum: ~A=0, via cancellation."
				     (adb::lprinc-string result)))
	 (adb::justify `((Equal-To ,result ZERO) . :TRUE) antes
		       :Sum-of-Zero))
	((or (cdr minuses) (cddr pluses) (and minuses (cdr pluses)))
	 ;; Have to perform case split on signs of influencers
	 (install-sum-sign-constraints result pluses minuses antes nil)
	 (install-sum-sign-constraints result minuses pluses antes t))
	((null pluses) ;; Single negative input
	 (adb::if-debug-flag adder
			     (format t "~%     Sum: ~A single negative: ~A"
				     (adb::lprinc-string result) 
				     (adb::lprinc-string (car minuses))))
	 (adb::justify `((Rcorrespondence :ordered ;for efficiency
					  (zero ,result)(,(car minuses) zero))
			 . :TRUE)
		       antes :Sum-single--input))
	(minuses ;; Must be one plus and one minus.
	 (adb::if-debug-flag adder
	     (format t "~%     Sum: ~A opposing paired inputs: ~A and ~A."
		     (adb::lprinc-string result) (adb::lprinc-string (car pluses))
		     (adb::lprinc-string (car minuses))))
	 (adb::justify `((rcorrespondence :ORDERED
					  (,(car pluses) ,result)
					  (,(car minuses) ZERO)) . :TRUE)
		       antes :Sum--+-Case)
	 (adb::justify `((rcorrespondence :ORDERED
					  (,result ZERO)
					  (,(car pluses) ,(car minuses))) . :TRUE)
		       antes :Sum-+--Case)
	 (when *Resolve-simple-multiple-directs*
	   (assume-inequality-if-needed (car pluses) (car minuses))))
	((null (cdr pluses)) ;; Single positive input
	 (adb::if-debug-flag adder
           (format t "~%     Sum: ~A single positive: ~A"
		   (adb::lprinc-string result) (adb::lprinc-string (car pluses))))
	 (adb::justify `((rcorrespondence :ORDERED
					  (,result ZERO)
					  (,(car pluses) ZERO)) . :TRUE)
		       antes :Sum-single-+input))
	(t ;; Must be two pluses
	 (adb::if-debug-flag adder
	     (format t "~%     Sum: ~A two + inputs: ~A, ~A."
		     (adb::lprinc-string result) (adb::lprinc-string (car pluses))
		     (adb::lprinc-string (cadr pluses))))
	 (adb::justify `((rcorrespondence :ORDERED
					  (,result ,(car pluses))
					  (,(cadr pluses) ZERO)) . :TRUE)
		       antes :Sum-++-Case)
	 (adb::justify `((rcorrespondence :ORDERED
					  (,result ,(cadr pluses))
					  (,(car pluses) ZERO)) . :TRUE)
		       antes :Sum-++-Case))))

;;;; Installing sign constraints

(defun install-sum-sign-constraints (result pluses minuses antes-in switch?)
  ;; Since we have always ensured that the signs of the pluses and minuses
  ;; will be assumptions, we can compute consistent combinations of signs
  ;; Devious negation logic by jcollins, filthy structure sharing by kdf.
  (let* ((pls (mapcar #'(lambda (p) `((NIL ,p ZERO) . :FALSE)) pluses))
	 (mns (mapcar #'(lambda (p) `((NIL ,p ZERO) . :FALSE)) minuses))
	 (output `((NIL ,result ZERO) . :TRUE))
	 (antes (append antes-in pls mns))(count 0))
    (dolist (rel '(Greater-Than Less-Than))
      (catch 'Just-cannot-win
      (let* ((o-rel (case rel (Greater-Than 'Less-than)(t 'Greater-Than)))
	     (net-rel (if switch? o-rel rel)))
	(setf (caar output) net-rel)
	(dolist (m mns) (setf (caar m) rel)
		(if (adb::false? m) (throw 'just-cannot-win count)))
	(setq count 0)
	(dolist (p pls) (setf (caar p) o-rel)
		(if (adb::false? p) (incf count)))
	(if (> count 1) (throw 'Just-cannot-win count))
	(dolist (p pls)
	  (when (or (= count 0) ;; All negated ones are consistent
		    (adb::false? p)) ;; This was the only loser
	    (setf (caar p) rel) (setf (cdr p) :TRUE)
	    (unless (adb::false? p) ;; avoid fruitless justifications
	      (adb::justify output antes :Sum-sign-constraint))
	    (setf (caar p) o-rel) (setf (cdr p) :FALSE))))))
    (unless switch?
      (setf (cdr output) :FALSE)
      (dolist (rel '(greater-than less-than))
	(setf (caar output) rel)
	(catch 'just-cannot-win
	(let ((o-rel (case rel (Greater-Than 'Less-Than) (t 'Greater-Than))))
	  (dolist (m mns) (setf (caar m) o-rel)
		  (if (adb::false? m) (throw 'just-cannot-win m)))
	  (dolist (p pls) (setf (caar p) rel)
		  (if (adb::false? p) (throw 'just-cannot-win p)))
	  (adb::justify output antes :Sum-negated-sign-constraint)))))))

;; Some helpers for debug flags
(defun string-for-Ds (q val) (adb::lprinc-string (cons (list 's (list 'D q)) val)))
(defun string-for-and (conjs) (adb::lprinc-string (cons 'AND conjs)))

;;;; Resolving ambiguous direct influences

(proclaim '(special *Resolve-Simple-Multiple-Directs*))

;(defun resolve-ambiguous-direct-influences (q pluses minuses ante-table envs
;					      &optional (Flag-ambig? t)
;					      &aux antecedents)
;  (cond ((and (null (cdr pluses)) (null (cdr minuses)))
;	 ;; special case of one plus and one minus. 
;	 ;; The flag is there because sometimes even this much work is too much.
;	 ;; Assumptions are forever.
;	 (when *Resolve-Simple-Multiple-Directs*
;	   (let ((plus nil) (minus nil) (swap? nil))
;	     (cond ((and (eq (caar pluses) 'I+)
;			 (eq (caar minuses) 'I-))
;		    ;; Both are greater than zero
;		    (setq plus (caddr (car pluses))
;			  minus (caddr (car minuses))
;			  swap? nil))
;		   ((and (eq (caar pluses) 'I-)
;			 (eq (caar minuses) 'I+))
;		    ;; Both are less than zero
;		    (setq plus (caddr (car pluses))
;			  minus (caddr (car minuses))
;			  swap? t))
;		   (t ;; mixed signs, so must take magnitudes
;		    (setq plus `(m , (caddr (car pluses)))
;			  minus `(m ,(caddr (car minuses)))
;			  swap? nil)))
;	   (adb::if-debug-flag Resolve
;	    (format t "~%    IR: ~A, via ~A, ~A"
;		    (string-for-Ds q '(set-members -1 0 1))
;		    (adb::lprinc-string (car pluses))
;		    (adb::lprinc-string (car minuses))))
;	   (assume-inequality-if-needed plus minus)
;	   (setq antecedents (make-influence-antecedents-from-table ante-table))
;	   (adb::justify `((Greater-Than (D ,q) ZERO) . :TRUE)
;			  (cons (cons (find-qrel-assertion plus minus
;					  (if swap? '< '>))
;				      ':TRUE)
;				antecedents)
;			  'Single-Ambiguous-Direct-Influences)
;	   (adb::justify `((Equal-To (D ,q) ZERO) . :TRUE)
;			  (cons (cons (find-qrel-assertion plus minus '=)
;				      ':TRUE)
;				antecedents)
;			  'Single-Ambiguous-Direct-Influences)
;	   (adb::justify `((Less-Than (D ,q) ZERO) . :TRUE)
;			  (cons (cons (find-qrel-assertion plus minus
;					     (if swap? '> '<))
;				      ':TRUE)
;				antecedents)
;			  'Single-Ambiguous-Direct-Influences)
;	   (when flag-ambig?
;	     (install-ambig-res-flag q :FALSE antecedents)))))
;	(t (resolve-multiple-ambiguous-direct-influences
;	     q pluses minuses ante-table envs))))

;;;; Resolving multiple ambiguous direct influences

;;; There are several ways to do this.  

;; One technique is to label the bipartite graph formed by the plus and minus sets,
;; finding all combinations for which the result is fixed.  Assumptions are made
;; as needed, in order to minimize the total number of them.  This was used with
;; great success in GIZMO, but tends to introduce ALOT of assumptions in QPE.
;; Worse yet, even when transitions involving magnitudes are supported, some legal
;; transitions will be pruned because the change will be to a dominating set resulting
;; in the opposite sign value (falken pointed this out, and w/jcollins suggested the scheme
;;; below).
;
;(proclaim '(special *Resolve-Multiple-Directs-By-Labeling*
;		    *Resolve-Multiple-Directs-By-NI-quantity*))
;
;; Another strategy is to interpose a new quantity, NET-INFLUENCE,
;; between the direct influences and the quantity being influenced.

;(defun resolve-multiple-ambiguous-direct-influences
;       (q pluses minuses ante-table envs)
;  (cond (*Resolve-Multiple-Directs-By-NI-quantity*
;	 (Introduce-NI-quantity-for q pluses minuses ante-table envs))
;       (*Resolve-Multiple-Directs-by-labeling*
;    (let ((N+ (length pluses))
;	  (N- (length minuses))
;	  (P-mags (mapcar #'(lambda (Is) `(m ,(caddr Is))) pluses))
;	  (N-mags (mapcar #'(lambda (Is) `(m ,(caddr Is))) minuses))
;	  (antecedents (make-influence-antecedents-from-table ante-table)))
;    ;; **** Add careful mode box that keeps distribution information about
;    ;; **** multiple directs.
;    (cond ((= N+ N-) ;;; Must check for dominance both ways, and for cancellation.
;	   (find-dominating-set 'Greater-Than q antecedents envs nil
;				N-mags P-mags)
;	   (find-dominating-set 'Less-Than q antecedents envs nil
;				P-mags N-mags)
;	   (find-cancellation-set q antecedents envs
;				  N-mags P-mags))
;	  ((< N+ N-) ;; Find dominating set for the positive influences
;	   (find-dominating-set 'Less-Than q antecedents envs T
;				P-mags N-mags))
;	  (t (find-dominating-set 'Greater-Than q antecedents envs T
;				  N-mags P-mags)))))))

(defun assume-inequality-if-needed (n1 n2)
  (let ((ineq (apply #'fetch-ineq (find-qrel n1 n2))))
    (unless (member ineq *QC-Qstate*)
      (unless (member ineq *Resolve-Qstate*)
	(assume-ineq-if-needed ineq :DI-Assumption)
	(adb::run) ;; run the rules, just in case.
	(push ineq *Resolve-Qstate*)))))

;;;; Finding labellings for resolving multiple direct influences

;(defun find-cancellation-set (q antes envs Negs Poss)
;  ;;; Finds all combinations of equality assumptions that will allow
;  ;;; cancellation of direct influences.
;  (cond ((null Negs) ;; Got one!
;	 (adb::if-debug-flag Resolve
;	   (format t "~%   IR: ~A, via multiple cancellation:" (string-for-Ds q 0))
;	   (dolist (ante antes)
;	     (when (eq (caar ante) 'EQUAL-TO)
;	     (format t "~%        ~A" (adb::lprinc-string ante)))))
;	 (adb::justify `((Equal-to (D ,q) ZERO) . :TRUE)
;		      antes ':Multiple-Direct-Influence-Cancellation))
;	(t (dolist (P Poss)
;	     (assume-inequality-if-needed (car Negs) P)
;	     (let ((form (cons (find-qrel-assertion (car Negs) P 'Equal-To)
;			       :TRUE))
;		   (nenvs nil))
;	       (setq nenvs (tms::generic-union-envss envs (adb::environments form)))
;	       (when nenvs ;; Fine so far
;		 (find-cancellation-set q (cons form antes) nenvs
;					(cdr negs) (remove p Poss))))))))
;
;(defun find-dominating-set (final-rel q antes envs less-than? minors majors)
;  (cond ((null minors)
;	 (when less-than? ;; At least one less-than
;	   (adb::if-debug-flag Resolve
;	     (format t "~%     IR: ~A, via" (string-for-Ds q (case final-rel
;							       (GREATER-THAN 1)
;							       (LESS-THAN -1)
;							       (EQUAL-TO 0)
;							       (t 'BOGOSITY))))
;	     (dolist (ante antes)
;	       (when (member (caar ante) '(GREATER-THAN LESS-THAN EQUAL-TO))
;		 (format t "~%        ~A"
;			 (adb::lprinc-string ante)))))
;	   (adb::justify `((,final-rel (D ,q) ZERO) . :TRUE)
;			antes ':Multiple-Direct-Influence-Domination)))
;	(t (dolist (MJ majors)
;	     (assume-inequality-if-needed (car minors) MJ)
;	     (let ((form (cons (find-qrel-assertion (car minors) MJ 'Equal-To)
;			       :TRUE))
;		   (nenvs nil))
;	       (setq nenvs (tms::generic-union-envss envs (adb::environments form)))
;	       (when nenvs
;		 (find-dominating-set final-rel q (cons form antes) nenvs
;				      less-than? (cdr minors) (remove MJ majors))))
;	     (let ((form (cons (find-qrel-assertion (car minors) MJ 'Less-Than)
;			       :TRUE))
;		   (nenvs nil))
;	       (setq nenvs (tms::generic-union-envss envs (adb::environments form)))
;	       (when nenvs
;		 (find-dominating-set final-rel q (cons form antes) nenvs
;				      T (cdr minors) (remove MJ majors))))))))

;;;; Resolving indirect influences
;;; As with direct influences, we search through the alternatives for both the sets of
;;; influences and for the Ds values to the inputs of these qualitative proportionalities.
;;; However, if we get an ambiguity, we simply note this fact, hoping
;;; that some other kind of knowledge will be used to ascertain this value
;;; (If not, QPE will simply generate all the possibilities).

(defun resolve-indirect-influences-on (q ip-influencers)
  (let ((qform `((quantity ,q) . :TRUE)))
    (find-indirect-influence-sets q ip-influencers (adb::environments qform)
				  nil (list qform) nil)))

(defun find-indirect-influence-sets (q potential-indirects envs result antecedents
				       any?)
  (cond ((null envs)) ;; Inconsistent branch, punt.
	((null potential-indirects)
	 (when any? 
	   (establish-indirect-influence-result q result antecedents)))
	(t ;; Now we do the search.  First, the case where the particular contributor we are
	 ;; looking at isn't active.
	 (let* ((antes (flip-truths (caar potential-indirects)))
		(new-envs nil))
	   ;; Any one being false is sufficient to wipe out the influence, 
	   ;; hence must branch to get the effect of an OR.
	   (dolist (neg-ante antes)
	     (when (setq new-envs
			 (tms::generic-union-envss envs (adb::environments neg-ante)))
	       (find-indirect-influence-sets q (cdr potential-indirects)
					     new-envs result
					     (cons neg-ante antecedents) any?)))
	   (setq antes (caar potential-indirects)
		 new-envs (adb::combine-facts-with-envs antes envs))
	   (when new-envs
	     (search-indirect-influence-inputs q (cdar potential-indirects)
					       (cdr potential-indirects) new-envs
					       result (append antes antecedents)
					       t))))))

;;;; Searching out possible values for the constrainers
;; Cannot do env filtering, since we do not know every possible order to establish
;; these facts in.  So generate all possibilities, subject to ambiguity filtering (see below).

(defun search-indirect-influence-inputs (q qprops potential-indirects
					   envs result antecedents any?)
  (cond ((eq result 'AMBIG)
	 ;; Optimization: If the result is ambiguous, then install the ambiguous
	 ;; justification right away.  Notice that the antecedents need not be minimal,
	 ;; since they will depend on the order of the qprops in the particular IP-influencer
	 ;; sets.  However, this technique should still generate many 
	 ;; fewer justifications than previously.
		(establish-indirect-influence-result q result antecedents))
	((null qprops) ;; Return to the main search
	 (find-indirect-influence-sets q potential-indirects
						envs result antecedents any?))
	(t (let* ((qprop (car qprops))
		  (rest (cdr qprops))
		  (num (caddr qprop)))
     ;; Must explore each possible input value.  Simplest case is zero, which contributes
     ;; nothing.
     ;; ***** Potential optimization: if you can prove that a certain value is required,
     ;; ***** then we can avoid searching any of the other possibilities.
	     (search-indirect-influence-inputs q rest potential-indirects
			       envs (update-result qprop 'zero result)
					       (append `(((s (D ,num)) . 0)
							 (,qprop . :TRUE))
						          antecedents) any?)
	     (search-indirect-influence-inputs q rest potential-indirects
			     envs (update-result qprop 'minus result)
			     (append `(((s (D ,num)) . -1) (,qprop . :TRUE))
				     antecedents) any?)
	     (search-indirect-influence-inputs q rest potential-indirects
			     envs (update-result qprop 'plus result)
			     (append `(((s (D ,num)) . 1) (,qprop . :TRUE))
				   antecedents) any?)))))

;;;;; Helpers for resolving indirect influences
(defun update-result (qprop input-sign old-result)
  (let ((contribution (case (car qprop)
			(QPROP input-sign)
			(QPROP- (case input-sign
				  (PLUS 'MINUS)
				  (MINUS 'PLUS)
				  (ZERO 'ZERO)))
			(t (error
   "~% ~A invalid qualitative proportionality indicator.~
~%     Input sign = ~A, Old result = ~A" qprop input-sign old-result)))))
    (cond ((null old-result) contribution) ;; First one to establish sign
	  ((eq old-result 'AMBIG) 'AMBIG) ;; fixed point due to information loss.
	  ((eq old-result contribution) contribution) ;; Everything is cool.
	  ((eq contribution 'ZERO) old-result) ;; pass it on
	  ((eq old-result 'ZERO) contribution) ;; pass it on.
	  (t 'AMBIG)))) ;; anything else loses information.

(defun establish-indirect-influence-result (q result antecedents)
  (when result ;; null result means no qprops, handled elsewhere.
    (adb::if-debug-flag indirectly-influenced
		       (format t "~%     IR: ~A indirectly influenced to ~A via:"
			       (adb::lprinc-string q) RESULT)
		       (dolist (ante antecedents)
			 (format t "~%          ~A" (adb::lprinc-string ante))))
      (cond ((eq result 'AMBIG) ;; special case
	   (adb::justify `((AMBIGUOUS-INDIRECT-INFLUENCES ,q) . :TRUE)
			antecedents 'Indirect-Influence))
	  (t (adb::justify `((,(case result
				(PLUS 'GREATER-THAN)
				(MINUS 'LESS-THAN)
				(ZERO 'EQUAL-TO)
			      (T (error 
  "~%~A bad indirect influence result for ~A, with ~A"
  result q antecedents)))
			     (D ,q) ZERO) . :TRUE)
			  antecedents 'Indirect-Influence)))))

;;;; Resolving ambigous influences
;;
;; Now go down each sclass and find out which of them
;; have not been completely resolved.  

(defun resolve-ambiguous-Ds-values ()
  (run-qpe-events :BEFORE :IR-AMBIG)
  (dolist (sit *PS-VS*)
    (unless (eq (sclass-status sit) 'R-COMPLETE)
      (resolve-ambiguous-influences sit)
      (unless (sclass-r-completions sit)
	;; Should be some consistent alternative.
	;; I don't THINK this can happen, but if it does the matter bears further thought.
	;; Either it's a symptom of a particular modeling error, or it is yet one more way
	;; that constraints can be imposed (in which case a nogood is needed to clobber the envs
	;; involved).
	(error "~%Gee, ~A has no consistent R-Completions.~%Please call Forbus."
	       sit))
      ;;;; DEBUGGING
;      (when (some #'(lambda (e) (tms::env-contradictory e))
;		  (sclass-assumptions sit))
;	(break " ~A has just lost at least one situation." sit))
      ))
  (run-qpe-events :AFTER :IR-AMBIG))

;;; If we have been unable to determine the Ds value for
;; a quantity, then we must branch by finding the consistent
;; sets of Ds values for the unknown quantities.  This is accomplished as
;; follows:  

    ;;; ASSUMPTION: Quantity info and influences are constant over sclasses.
    ;;; By definition, every situation in the Sclass has identical PS/VS.  Thus any information about
    ;;; quantities which follows from the PS/VS must be constant within the Sclass.  But what if
    ;;; it depended on some inequality as well?  Then that inequality, to be in here, must make some
    ;;; difference, and should be expressed as part of an individual view. 

;;;; Finding the influences that occur in a particular situation

(proclaim '(special order-table))

(defun compute-situation-influence-table (env)
  (let ((directs nil)
	(indirects nil)
	(quantities nil)
	(constrainers nil)
	(order-table nil)
	(qform '((Quantity q) . :TRUE)))
     ;; Filter out those quantities which don't exist in the Sclass
    (dolist (q (gather-quantities))
      (setf (car (cdar qform)) q)
      (if (adb::implied-by? qform env) (push q quantities)))
    (setq constrainers (mapcar #'(lambda (q) (cons q nil)) quantities))
    ;;; Walk down DP-influencers and IP-influencers, checking each quantity to see if
    ;;; it exists in the sclass.  If it does, then check which, if any, of the influences
    ;;; associated with it are implied by the sclass' envs.
    (dolist (entry *DP-influencers*)
      (when (member (car entry) quantities :test #'equal)
	;; Quantity exists, now must see if any of the direct influences do.
	(let* ((q (car entry))
	       (direct-entry (cons q nil)))
	  (dolist (subentry (cdr entry))
	    (when (every #'(lambda (ante) (adb::implied-by? ante env))
			 (car subentry))
	        ;; add these direct influences to the entry being built
	      (setf (cdr direct-entry)
		    (append (cdr subentry) (cdr direct-entry)))))
	  (when (cdr direct-entry) (push direct-entry directs)))))
    (dolist (entry *IP-influencers*)
      (when (member (car entry) quantities :test #'equal)
	;; Might have indirect influences, must check
	(let* ((q (car entry))
	       (indirect-entry (cons q nil)))
	  (dolist (subentry (cdr entry))
	    (when (every #'(lambda (ante) (adb::implied-by? ante env))
			 (car subentry))
	      (setf (cdr indirect-entry)
		    (append (cdr subentry) (cdr indirect-entry)))
	      (dolist (qprop (cdr subentry))
		(let ((pair (assoc (caddr qprop) constrainers :test #'equal)))
		  (unless pair
		    (error
  "~%QPE error: ~A does not exist when ~A is believed via ~A."
  (adb::lprinc-string (caddr qprop)) (adb::lprinc-string qprop)
  (adb::lprinc-string (car subentry))))
		  (setf (cdr pair) (cons q (cdr pair)))))))
	  (when (cdr indirect-entry)
	    (push indirect-entry indirects)))))
    (When-careful 
      (let ((overlap (intersection directs indirects :key #'car
				 :test #'equal)))
      (when overlap
	(Add-recorder-Event :Simultaneous-Direct-Indirect-Influences
			  env overlap directs indirects))))
    ;;Now it's time to construct the order table
    (setq order-table (mapcar #'(lambda (q) (cons q -1)) quantities))
    (dolist (q-pair directs)
      ;; Walk outward from direct influences
      (walk-and-mark-constrainer-depth (car q-pair) constrainers 0 nil))
    (if-debug-flag resolve-ambig
       (format t "~%For situation ~A:~%   Directs:" env)
       (print-list-with-commas (mapcar #'lprinc-string directs))
       (format t "~%   Indirects:")
       (print-list-with-commas (mapcar #'lprinc-string indirects)))
    (values directs indirects (remove-if #'(lambda (pair) (= (cdr pair) -1))
					 order-table))))

(defun print-list-with-commas (ll &optional (stream *standard-output*))
  (format stream "~%  ~{~<~%  ~1:; ~A~>~^,~}.~%" ll))

;;;; Resolving ambiguous Ds values for a particular sclass 

(defun resolve-ambiguous-influences (sit &aux unknowns env)
  (adb::if-debug-flag Resolve-Ambig
		 (format t "~%   Attempting to resolve ambiguous influences on ~A"
			 (sclass-name sit))
		 (format t "~% Unknown Ds values:")
		 (print-list-with-commas (mapcar #'lprinc-string
						 (sclass-unknown-ds-values sit))))
  (multiple-value-bind (directs indirects order-table)
      (compute-situation-influence-table 
	(setq env (car (sclass-assumptions sit))))
    (unless directs ;; Reality check
      (error 
       "~%Ambiguous Ds values detected w/o direct influences:~% Sit =~A, Ambiguous = ~A;~%indirects = ~A."
       env (adb::lprinc-string (cons 'set-members
				     (sclass-unknown-ds-values (env-sclass env))))
       (adb::lprinc-string (cons 'set-members (mapcan #'cdr indirects)))))
    ;the first problem is to find the order (which is situation-dependent)
    ;in which the ambiguous cases must be resolved.  The order is important
    ;because the choices can depend on each other.
    (when (null order-table)
      (error "This domain model is exceedingly brain-damaged in some way.~
~%Probably it's a loop in the qualitative proportionalities.~
~%But maybe it's a typo.  Who knows?~
~%If it is something other than these, please let Forbus know when you find out.~
~%The problem arose in sclass ~A, where the influences are:~%directs: ~A~%indirects: ~A
QPE ERROR" sit (if directs
		    (adb::lprinc-string (cons 'set-members (mapcar #'car directs)))
		    "{}")
                (if indirects
		    (adb::lprinc-string
		      (cons 'set-members (mapcar #'car indirects)))
		    "{}")))
    (setq unknowns (take-ordering-from-pair-table (sclass-unknown-Ds-values sit)
						  order-table))
      ;;; DEBUGGING -- See debug file.
;;;  (test-ambiguous-influence-code sit unknowns)
  (try-all-Ds-possibilities unknowns sit (sclass-assumptions sit) nil)))

;;;; Computing optimal order of resolution for ambiguous quantities.

(defvar *influence-order-table* nil) ;; for debugging only
(defvar *unknown-quantity* nil)

(defun cdr-< (x y) (< (cdr x) (cdr y)))

(defun take-ordering-from-pair-table (items order-table &aux entry result)
  (dolist (item items)
    (setq entry (assoc item order-table :test #'equal))
    (cond (entry (setq result (ordered-insert entry #'cdr-< result)))
	  (t (setq *influence-order-table* order-table
		   *unknown-quantity* item)
	     (error "Uncertain of influences on ~A in ~A."
		    item order-table))))
  (mapcar #'car result))

(proclaim '(special order-table))

(defun walk-and-mark-constrainer-depth (q constrainers index path)
  (declare (special order-table))
  (let ((old-index (assoc q order-table :test #'equal)))
    ;;;We have to carry around the path to catch utter brain damage.
    ;;;In particular, QP theory forbids loops in the Qprops.
    (cond ((member q path :test #'equal)
	   (error   
"~%This domain model violates QP theory.
The Qprops which hold at any particular time cannot form loops.
This violates the interpretation of Qprop as impuning causality
The path in question is: ~A
~%QPE ERROR" path))
	  ((> (cdr old-index) index)) ;do nothing; longer path has been found.
	 (t (rplacd old-index index)
	    (dolist (nq (cdr (assoc q constrainers :test #'equal)))
	      (walk-and-mark-constrainer-depth nq constrainers
					       (1+ index) (cons q path)))))))

;;;; Filling out unknown values
;;Start down the list of unknown Ds values in order.
;;; See if it has been figured out (since earlier choices may constraint
;;; later choices).  If so, go on.
;;; If it is still unknown, we add in an assumption about it and branch.
;;;      (Must filter for consistency, of course.)
;;; When finished, add the r-extension sclass.

(defun try-all-Ds-possibilities (unknowns parent envs new-values &aux new-value q)
  (cond ((null (setq envs (delete-if #'(lambda (x) (tms::env-contradictory x))
				     envs)))) ;; We lost on this branch.
	((null unknowns);; The envs now represent a complete, consistent solution.
	 ;; Thus they are an R-Completion, and must be indexed appropriately.
	 (build-r-extension-sclass parent envs new-values))
	((setq new-value (Get-Ds-value (setq q (car unknowns)) envs))
	 ;some previous assumption fixed it, so go on.
	 (try-all-Ds-possibilities (cdr unknowns) parent envs
				   (cons (cons q new-value)
					 new-values)))
	(t ;;; See if there is any special knowledge that will help.
	   ;; It is assumed that these hooks will only add justifications, not
	   ;; new assumptions.  A resolver may be able to disambiguate some envs
	   ;; but not all.  That's okay.  Unless there is a single consistent answer
	   ;; for them all we want to branch, and they will be sorted out by
	   ;; TRY-IF-CONSISTENT-DS-EXTENSION.
	   (dolist (resolver *resolve-hooks*)
	     (when (resolver-applicable? resolver (car unknowns) envs)
	       (adb::if-debug-flag resolve-hooks 
				   "~%     IR: Trying ~A on ~A for ~A."
				   (car resolver) envs (adb::lprinc-string q))
	       (funcall (resolver-code resolver) (car unknowns) envs)
	       ;; No assumptions are made about the value(s) returned by
	       ;; the resolver code.  Instead, check to see if we have come
	       ;; to a single answer
	       (when (setq new-value (Get-Ds-value q envs))
		 ;; We won
		 (adb::if-debug-flag resolve-hooks
		      (format t "~%     IR: ~A resolved to ~A."
			      (adb::lprinc-string q) new-value))
		 (return-from try-all-Ds-possibilities
		   (try-all-Ds-possibilities (cdr unknowns) parent envs
					     (cons (cons q new-value) new-values))))))
	   ;; Otherwise, must branch on the different alternatives.
	   (let ((deriv `(D ,q))
		 (ineq nil))
	     (assume-inequality-if-needed deriv 'ZERO)
	     (setq ineq (fetch-ineq deriv 'ZERO))
	     ;; While the outcome can't be fixed  -- o.w., the third clause above
	     ;; would have triggered -- any of the assumptions might be false.
	     (when (ineq-lt-a ineq)
	       (try-if-consistent-Ds-extension unknowns parent envs new-values
					       (ineq-lt-a ineq) -1))
	     (when (ineq-gt-a ineq)
	       (try-if-consistent-Ds-extension unknowns parent envs new-values
					       (ineq-gt-a ineq) 1))
	     (when (ineq-eq-a ineq)
	       (try-if-consistent-Ds-extension unknowns parent envs new-values
					       (ineq-eq-a ineq) 0))))))

(defun try-if-consistent-Ds-extension (unks parent envs new-values asn val)
  ;; ***** Install as macrology when performance tuning??
  (let ((new-envs nil) (new-env nil))
    (dolist (old-env envs)
      ;; These have to be real envs, since they are solutions.
      (setq new-env (tms::cons-env old-env asn))
      (unless (tms::env-contradictory new-env) (push new-env new-envs)))
    (when new-envs 
      (try-all-Ds-possibilities (cdr unks) parent
				new-envs
				(cons (cons (car unks) val)
				      new-values)))))

;;;; Domain-specific influence resolution hooks

;; Often there is some knowledge about a particular domain or situation
;; that can be used in resolving ambiguous influences.  This code provides
;; an easy way to add such knowledge, as well as to experiment with augmentations.

(defvar *resolve-hooks* nil)
(adb::init-form (setq *resolve-hooks* nil))

(defmacro defResolver (name documentation test code)
  `(push (list ',name ,documentation (function ,test) (function ,code))
	 *resolve-hooks*))

;; ****** ensure the existing resolve hooks are printed out with QPE reports.
;; ****** add a similar capability to limit analysis.

(defun resolver-applicable? (resolver q envs) (funcall (third resolver) q envs))

(defun resolver-code (resolver) (fourth resolver))

;;;; Creating R-completions.
;; Situations and S-classes are S-complete and/or R-complete.
;; R-complete -> S-complete, but not vice versa.

(defun build-r-extension-sclass (parent envs new-values &aux sit)
;;must build the new situation, inheriting everything but the Ds
;;values from the old.  The Ds values which were previously unknown
;;are now tested to make sure that they are known, and inserted into
;;the Ds-value record.  Appropriate pointers are also installed between
;;the parent and the r-completion.
  (setq sit (clone-sclass parent))
  (setf (sclass-assumptions sit) envs)
  (setf (sclass-unknown-Ds-values sit) nil)
  (setf (sclass-status sit) 'R-complete)
  (setf (sclass-r-completion-of sit) parent)
  (setf (sclass-r-completions parent) (cons sit (sclass-r-completions parent)))
  (dolist (e envs) (tms::put-env-prop e sit 'sclass))
  ;;Here we upgrade the list of Ds values associated with the S-class.
  ;;Assumes the known and unknown Ds values are in lexigraphic order for speed.
  (splice-in-Ds-values sit new-values))

(defun splice-in-Ds-values (sit new-values)
  ;; ****** Need to get rid of this alphalessp
  (setf (sclass-Ds-values sit)
	(sort (nconc (copy-list new-values)
		     (sclass-ds-values sit))
	      #'(lambda (xp yp)
		  (adb::alphalessp (car xp) (car yp))))))

;;;; Displaying closed-world tables for influence resolution
;;
;;; Very handy for debugging domain models.

(defun show-possible-directs (&optional (stream *standard-output*))
  (format stream "~%Table of possible direct influencers.")
  (dolist (entry *dp-influencers*)
    (format stream "~%~% ~A : " (adb::lprinc-string (car entry)))
    (dolist (pair (cdr entry))
      (format stream "~%     ~A contributes:" 
	      (adb::lprinc-string (cond ((cdar pair) (cons 'AND (car pair)))
				   (t (caar pair)))))
      (dolist (contribution (cdr pair))
	      (format stream "~%     ~A"
		      (adb::lprinc-string contribution))))))

(defun show-possible-indirects (&optional (stream *standard-output*))
  (format stream "~%Table of possible indirect influencers.")
  (dolist (entry *ip-influencers*)
    (format stream "~%~% ~A : " (adb::lprinc-string (car entry)))
    (dolist (pair (cdr entry))
      (format stream "~%     ~A contributes:" 
	      (adb::lprinc-string (cond ((cdar pair) (cons 'AND (car pair)))
				   (t (caar pair)))))
      (dolist (contribution (cdr pair))
	      (format stream "~%     ~A"
		      (adb::lprinc-string contribution))))))

;;;; Products code

;; From Jcollins.  Installed as a resolve hook for simplicity.

(adb::after-init-form (defResolver RATIO-RESOLVER "Uses ratio information."
			(lambda (q envs) (and *resolve-ratios* (ratio? q)))
			(lambda (q envs) (dolist (env envs)
					   (resolve-ratio-influences q env)))))

(defvar *resolve-ratios* t)
(defvar *ratio-justs* nil)

(defun resolve-ratio-influences (q env &aux f-ectv-numer f-ectv-denom
				   numer-directs denom-directs result pairs
				   a-n-directs a-d-directs)
  (if-debug-flag resolve-ratios
     (format t "~%     RR: Trying to resolve ratio: ~A in ~A" q env))
  (setq *ratio-justs* nil)
  (let* ((entry (cdr (assoc q *ratios* :test #'EQUAL)))
	 (form (caar entry)) ;; Entry: ((/rel ,ratio ,numer ,denom) . ,just)
	 (numer (third form))
	 (denom (fourth form)))
    (unless (implied-by? (setq form (cons form :TRUE)) env)
	(return-from resolve-ratio-influences (values nil :NOT-A-RATIO)))
    (push form *ratio-justs*)
    (setq f-ectv-numer (find-directly-influenced-equivalent numer env)
	  f-ectv-denom (find-directly-influenced-equivalent denom env))
    (if-debug-flag resolve-ratios
		   (format t "~%  RR: Effective-numer for ~A is: ~A" q f-ectv-numer)
		   (format t "~%  RR: Effective-denom for ~A is: ~A" q f-ectv-denom))
    (unless (and F-ectv-numer F-ectv-denom)
      (return-from resolve-ratio-influences nil))
    (setq numer-directs (active-directs (car f-ectv-numer) env)
	  denom-directs (active-directs (car f-ectv-denom) env))
      (if-debug-flag resolve-ratios
		     (format t "~%    RR: Numer-directs for ~A are: ~A" q numer-directs)
		     (format t "~%    RR: Denom-directs for ~A are: ~A" q denom-directs))
    (setq a-n-directs (prepare-ratio-directs-for-pairing numer-directs (cdr f-ectv-numer))
	  a-d-directs (prepare-ratio-directs-for-pairing denom-directs (cdr f-ectv-denom)))
    (if-debug-flag resolve-ratios
		   (format t "~%    RR: Prepared-n-directs are ~A" a-n-directs)
		   (format t "~%    RR: Prepared-d-directs are ~A" a-d-directs))
    (setq pairs (pair-up-ratio-influencers a-n-directs a-d-directs env))
    (if-debug-flag resolve-ratios
		   (format t "~%~10T Pairs are ~A" pairs))
    (when (setq result (disambiguate-ratio q pairs env))
      (let ((relation (case result (+1 'GREATER-THAN) (-1 'LESS-THAN) (0  'EQUAL-TO))))
	(if-debug-flag resolve-ratios
	  (format t "~%    RR: ~A"
		  (adb::lprinc-string `(-> (AND ((Quantity ,Q) . :TRUE) ,@ *ratio-justs*)
					   ((,relation (D ,q) ZERO) . :TRUE)))))
	(justify `((,relation (D ,q) ZERO) . :TRUE)
		 `(((quantity ,q) . :TRUE) ,@ *ratio-justs*)
		 :RESOLVED-RATIO)
	relation))))

;(defun find-directly-influenced-equivalent (qty env &aux entry form equiv-qtys)
;  (cond ((directly-influenced? qty) qty)
;	((setq entry (cdr (assoc qty *EQUIVALENTS* :test #'EQUAL)))
;	 (dolist (pair entry)
;	   (setq form `(,(cdr pair) . :TRUE))
;	   (when (and (directly-influenced? (car pair))
;		      (implied-by? form env))
;	     (push form *ratio-justs*)
;	     (push (car pair) equiv-qtys)))
;	 (cond ((cdr equiv-qtys)
;		(format t "~% Multiple equivalents of ~A are directly-influenced in ~A: ~A"
;		       qty env equiv-qtys)
;		(car equiv-qtys)) ;For now just return the first one.
;	       ((car equiv-qtys))
;	       (t (format t "~% No equivalent of ~A is directly-influenced." qty) nil)))
;	((indirectly-influenced? qty)
;	 (format t "~% ~A is indirectly-influenced, but has no equivalent." qty) nil)
;	(t nil)))


;;;EXPERIMENT: Try following Qprops to find directly-influenced equivalent.

(defun find-directly-influenced-equivalent (qty env &optional (sign 1)
						&aux entry ineq opp inf-qty equiv-qtys)
  (if-debug-flag resolve-ratios
		   (format t "~%  Looking for directly-influenced equivalent of ~A;" qty))
  (cond ((directly-influenced? qty) (cons qty sign))
	((setq entry (cdr (assoc qty *IP-INFLUENCERS* :test #'EQUAL)))
	 (dolist (set entry)
	   (dolist (inf (cdr set))
	     (setq inf-qty (third inf))
	     (case (car inf)
	       (qprop (setq ineq (unordered-fetch-ineq `(D ,qty) `(D ,inf-qty)))
		      (when (eq (inequality-under ineq env) :EQUAL-TO)
			(push (adb:view-node (ineq-= ineq)) *ratio-justs*)
			(push inf-qty equiv-qtys)))
	       (qprop- (setq opp `((opposite ,@ (find-qrel `(D ,qty) `(D ,inf-qty))) . :TRUE))
		       (when (and (referent opp) (implied-by? opp env))
			 (if-debug-flag resolve-ratios
			   (format t "~% Found opposite: ~A" opp))
			 (setq sign (- sign))
			 (push opp *ratio-justs*)
			 (push inf-qty equiv-qtys))))))
	 (cond ((cdr equiv-qtys)
		(error "~% Multiple equivalents of ~A were found in ~A: ~A"
		       qty env equiv-qtys))
	       ((car equiv-qtys)
		(find-directly-influenced-equivalent (car equiv-qtys) env sign))
	       (t (if-debug-flag resolve-ratios
		   (format t "~% No equivalent of ~A is directly-influenced in ~A." qty env))
		  nil)))))


(defun prepare-ratio-directs-for-pairing (influences init-sign &aux directs entry)
  ;; Filters out identical positive and negative influences, then returns an Alist of
  ;; ( ... (Qty sign) (Qty sign) ... ) , where sign is 1 for I+ and -1 for I-. and 0 for both.
  (dolist (influence influences directs)
    (let* ((qty (quantity-of (third influence)))
	   (sign (case (car influence) (I+ init-sign) (I- (- init-sign))
		       (t (error "~% Bad influence type: ~A" influence)))))
      (if (setq entry (assoc qty directs :test #'EQUAL))
	  (if (eq sign (cadr entry))
	      (error "~% Two identical directs: ~A" influence)
	      (setf directs (delete entry directs))) ; Found both I+ & I-
	  (push (list qty sign) directs)))))

(defun pair-up-ratio-influencers (numer-directs denom-directs env &aux entry)
;; Numer-directs & Denom-directs are of the form: ( ... (Qty sign) (Qty sign) ... ),
;; where sign is 1 for I+ & -1 for I-.
;; Returns an Alist of
;; ( ... (dd sign . ((nd1 sign1 . other-mult1) (nd2 sign2 . other-mult2) ... )) ... )
  (dolist (nd numer-directs denom-directs)
    (cond ((product? (car nd))
	   (multiple-value-bind (mult other-mult)
	       (get-product-multipliers (car nd) env) ;;**** Should include env here!!
	     (cond ((setq entry (or (assoc mult denom-directs :test #'EQUAL)
				    (assoc other-mult denom-directs :test #'EQUAL)))
		    (push `((QPRODUCT ,(car nd) ,mult ,other-mult) . :TRUE) *ratio-justs*)
		    (unless (equal mult (car entry))
		      (setq other-mult mult))
		    (setf (cddr nd) other-mult
			  (cddr entry) (cons nd (cddr entry))))
		   (t (push (list nil nil nd) denom-directs)))))
	  (t (push (list nil nil nd) denom-directs)))))

(defun disambiguate-ratio (q pairs env &aux (dir 0) new-dir pair)
  (dolist (entry pairs dir)
    (dolist (numer (cddr entry))
      (setq pair (list (car entry) (cadr entry) numer))
      (unless (and (setq new-dir (net-pair-influence q pair env))
		   (setq dir (update-net-influence dir new-dir)))
	(return-from disambiguate-ratio nil)))))

(defun active-directs (q env &aux active form)
  (dolist (set (cdr (assoc q *DP-INFLUENCERS* :test #'EQUAL)) active)
    (dolist (influence (cdr set))
      (push (setq form (cons influence :TRUE)) *ratio-justs*)
      (if (implied-by? form env)
	  (push influence active)
	  (rplacd form :FALSE)))))

(defun update-net-influence (old new)
  (cond ((zerop new) old) ;no change
	((zerop old) new)
	((= old new) old)))

(defun net-pair-influence (q pair env &aux q1 q2 influence-sign influencer)
  (let* ((denom-direct (car pair))
	 (denom-inf-sign (second pair))
	 (numer-direct (car (third pair)))
	 (numer-inf-sign (second (third pair)))
	 (other-mult (cddr (third pair))))
    (if-debug-flag resolve-ratios
       (format t "~% For pair: ~A, sign is ~A" pair numer-inf-sign))
    (cond ((and numer-inf-sign denom-inf-sign)
	   (check-ratio-pair q numer-inf-sign denom-inf-sign)
	   (setq q1 other-mult q2 q
		 influence-sign denom-inf-sign
		 influencer denom-direct))
	  (numer-inf-sign
	    (setq q1 numer-direct q2 'zero
		  influence-sign numer-inf-sign
		  influencer numer-direct))
	  (denom-inf-sign
	    (setq q1 'zero q2 denom-direct
		  influence-sign denom-inf-sign
		  influencer denom-direct))
	  (t (error "~% No influence for pair: ~A" pair)))
    (if-debug-flag resolve-ratios
       (format t "~% Q1 is ~A; Q2 is ~A" q1 q2)
       (format t "~% Influencer: ~A; Influence-sign: ~A" influencer influence-sign))
    (cond ((or (equal q1 q2)
	       (zerop influence-sign)) 0)
	  (t (let* ((pair-ineq (unordered-fetch-ineq (make-amount q1) (make-amount q2)))
		    (pair-order-sign (if (eq (ineq-n1 pair-ineq) (quant-amount (grab-quantity q2))) -1 +1))
		    (pair-rel (inequality-under pair-ineq env))
		    (inf-ineq (grab-ineq (quant-amount (grab-quantity influencer)) *zero-struct*))
		    (inf-rel (inequality-under inf-ineq env)))
	       (when (eq pair-rel :UNRELATED)
		 (error "Unrelated ratio pair: ~A"
			(view-node (ineq-rel-node pair-ineq pair-rel))))
	       (when (eq inf-rel :UNRELATED)
		 (error "Unrelated influence: ~A" influencer))
	       (unless (or (eq pair-rel :UNKNOWN)
			   (eq inf-rel :UNKNOWN))
		 (push (view-node (ineq-rel-node pair-ineq pair-rel)) *ratio-justs*)
		 (push (view-node (ineq-rel-node inf-ineq inf-rel)) *ratio-justs*)
		 (* pair-order-sign
		    influence-sign
		    (relation-sign pair-rel)
		    (relation-sign inf-rel))))))))

(defun relation-sign (rel)
  (case rel
    (:GREATER-THAN +1)
    (:EQUAL-TO 0)
    (:LESS-THAN -1)
    (T (error "Bad relation: ~A" rel))))

(defun check-ratio-pair (q numer-inf-sign denom-inf-sign)
  (unless (eq numer-inf-sign denom-inf-sign)
    (error "~% Opposite influences on numerator and denominator for ~A" q)))

(defun get-product-multipliers (prod env &aux x y found?)
  (let ((entry (cdr (assoc prod *products* :test #'EQUAL))))
    (unless entry 
      (error "~% ~A not a product --- get-product-multipliers" prod))
    (do ((items (cdr entry) (cdr items))
	 (item (car entry) (car items)))
	((null item)
	 (unless found?
	   (error "~% ~A not a product in ~A --- get-product-multipliers" prod env))
	 (values x y))
      (when (tms::true-in? (get-tms-node (cons (car item) :TRUE)) env)
	(when found?
	  (error "~% ~A has multiple entries: ~A --- get-product-multipliers" prod entry))
	(setq x (third (car item))
	      y (fourth (car item))
	      found? t)))))

(defun product? (q)
  (assoc q *products* :test #'equal))

(defun ratio? (q)
  (assoc q *ratios* :test #'equal))
