;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; Syntactic translations for QP theory
;;;
;;
;;; The definitions in here are the second pass at a truely implementation-independent
;;; formal QP-based domain language.
;;; The goal is to insulate the modeler as much as possible from the underlying
;;; system. 

;;; The contents of this file are broken down into translations of:
;;;
;;; Logical connectives (and, or ->, iff, not, taxonomy, at-most-one)
;;; Existence statements (Introduces Uniquely-introduces There-is There-is-Unique)
;;; Local specifications (Local, Local-name)
;;; Rules (QPE:RULE)
;;; Quantities and inequalities
;;; Influences and qualitative proportionalities (I+, I-, Qprop, Qprop-)
;;; Algebra (Q=, =, and friends)
;;;

;;;; Translations of logical operators
;;; These are mostly diked out for now.  The reason is that the automatic translation
;;; facilities in ADB are turned off for the indefinite future.  Sigh.

(defTranslation :== (cons (cadr form) (caddr form)))

(defTranslation And
		(cons 'And (remove-redundant-subexpressions pieces 'AND))
  :locals  ((pieces (delete nil (mapcar 'trans-rel1 (cdr form))))))

(defTranslation When
		(let ((pieces (delete nil (mapcar #'trans-rel1 (cddr form)))))
		  `(implies (AND ,@ **ANTECEDENTS**)
			    (AND ,@ (remove-redundant-subexpressions pieces 'AND))))
  :locals ((**ANTECEDENTS** (union (list (trans-rel1 (cadr form)))
				    **ANTECEDENTS**))))

(deftranslation eval
		NIL ;; A jcollins hack
  :results `((,form *antecedents* :SYNTAXER)))

(deftranslation quote (cadr form)) ;; Another jcollins hack

(deftranslation set NIL
  :eval (let  ((args (cddr form))
	       (mod :AFTER)
	       (op :SCENARIO-EXPANSION)
	       (pri 100))
	  (when (car args) ;; Modifier
	    (setq mod (car args))
	    (when (cadr args)
	      (setq op (cadr args))
	      (when (third args)
		(setq pri (third args)))))
	  `((schedule-qpe-event (function (lambda ()
					    (adb::close-set ,(quotize (cadr form)))
					    (adb::run)))
				,mod ,op ,pri))))

(deftranslation Every `((Every ,(second form) ,(third form)
			      ,(trans-rel1 (fourth form))) . :TRUE))

(deftranslation Some `((Some ,(second form) ,(third form)
			      ,(trans-rel1 (fourth form))) . :TRUE))

;;;; Some things that aren't here, plus helpers

(defTranslation Or
;		`(Or ,@ (delete nil (mapcar 'trans-rel1 (cdr form))))
		(error "OR not allowed in QPE syntax yet.  Sorry.~%~A"
		       form))

(defTranslation Taxonomy
;		`(Taxonomy ,@ (delete nil (mapcar 'trans-rel1 (cdr form))))
		(error "TAXONOMY not allowed in QPE syntax yet.  Sorry.~%~A"
		       form))

(defTranslation At-Most-One
;		`(At-Most-One ,@ (delete nil (mapcar 'trans-rel1 (cdr form))))
		(error "AT-MOST-ONE not allowed in QPE syntax yet.  Sorry.~%~A"
		       form))

(defTranslation Not (flip-truths (trans-rel1 (cadr form))))

(defTranslation -> ;`(-> ,(trans-rel1 (cadr form))
		   ;	,(trans-rel1 (caddr form)))
  (error "-> not allowed in QPE syntax yet.  Sorry.~%~A"
		       form))

(defTranslation iff `(iff ,(trans-rel1 (cadr form))
		    	  ,(trans-rel1 (caddr form))))

(defun remove-redundant-subexpressions (subexpressions type)
  (cond ((null subexpressions) nil)
	((not (listp subexpressions)) subexpressions)
	((not (listp (car subexpressions)))
	 (cons (car subexpressions)
	       (remove-redundant-subexpressions (cdr subexpressions) type)))
	((equal (caar subexpressions) type)
	 (append (cdar subexpressions)
		 (remove-redundant-subexpressions (cdr subexpressions) type)))
	(t (cons (car subexpressions)
		 (remove-redundant-subexpressions (cdr subexpressions) type)))))

(defun distribute-if-needed (pattern form)
  ;Pattern has a key denoted by **KEY**
  ;if <form> is an AND or OR expression, it
  ;distributes the expression amoung the terms
  (cond ((and (listp form)
	      (member (car form) '(AND OR)))
	 (cons (car form)
	       (mapcar #'(lambda (subform)
			  (distribute-if-needed pattern subform))
		       (cdr form))))
	(t (subst form '**KEY** pattern :test #'equal))))

;;;; Translating existence statements
;;
;; Currently these are not allowed outside of CDI's.  Outside,
;; all one has to do is use the EXISTS predicate.

;THERE-IS - states logical existence of an individual
;           for example, (There-is (C-L ?s ?c)
;                         ((C-L ?s ?c) substance ?s)
;                         ((C-L ?s ?c) container ?c)) will yield
;                        (AND (individual (C-L ?s ?c))
;                             ....), justified by the existence of the
;                                    current description


(defTranslation Introduces (trans-rel1 (cons 'There-is (cdr form))))

(defTranslation Uniquely-Introduces (trans-rel1 (cons 'There-is-Unique (cdr form))))

(defTranslation There-Is
		(progn (unless **CDI**
			 (error "THERE-IS cannot be used outside of CDI's.~%~A."
				form))
		       (distribute-if-needed '(And ((Individual **KEY**) . :TRUE)
						   ((Exists **KEY**) . :TRUE))
					     (cadr form))))

(defTranslation There-Is-Unique
		(progn (unless **CDI**
			 (error "THERE-IS-UNIQUE cannot be used outside of CDI's.~%~A"
				form))
		         (setq *creates-individual* t) ;Inform world about this
		       (trans-rel1 `(Individual ,(cadr form))))
  :results `((((exists ,(cadr form)) . :TRUE) ,**ANTECEDENTS**) ;justifications
	     ,@ (mapcar #'(lambda (ante)
			    `(((exists ,(cadr form)) . :FALSE) (,ante)))
			(flip-translation-antecedents))))

(defTranslation Only-During (trans-rel1 (cadr form))
  :results (mapcar #'(lambda (ante)
		       `((,(cadr form) . :FALSE) (,ante)))
		   (flip-translation-antecedents)))

;; Translating rules and expressions to evaluate
;These don't appear as assertions but instead are evaluated.
;; Notice that SRULE is used to provide syntactic sugar

(defTranslation Rule nil :eval (list (cons 'ADB::SRULE (cdr form))))

;;; ***** Add translations for rjustify, refer-to as well?

;;;; Translating quantity declarations and inequalities

(deftranslation Quantity
		(cond ((or (not (listp (cdr form)))
			   (null (cadr form))
			   (cddr form))
		       (error "~%Badly formed quantity statement: ~A." form))
		      (t (unless (variable? (cadr form))
			   (check-quantity-declaration (cadr form)))
			 `((quantity ,(cadr form)) . :TRUE)))
  :results (if **CDI** ;; A CDI only has a quantity when it is active 
	      (mapcar #'(lambda (ante)
			  `(((Quantity ,(cadr form)) . :FALSE) (,ante)))
		      (flip-translation-antecedents))))

;the "eval" is so the assert translator knows to evaluate it at run-time

(deftranslation less-than `(:eval (find-qrel-assertion1 ,(adb::quotize
							 (cadr form))
						      ,(adb::quotize
							 (caddr form)) '<)))

(deftranslation equal-to `(:eval (find-qrel-assertion1 ,(adb::quotize
							(cadr form))
						     ,(adb::quotize
							(caddr form)) '=)))

(deftranslation greater-than `(:eval (find-qrel-assertion1 ,(adb::quotize
							    (cadr form))
							 ,(adb::quotize
							    (caddr form)) '>)))

;;;; Translating influences and qualitative proportionalities

;;The form of these remain the same.  However, appropriate entries must be made to
;;    QPE's tables.

(defTranslation I+ `(,form . :TRUE)
  :results `(( (,form . :FALSE) ,(flip-truths **ANTECEDENTS**)))
  :eval `(;; Refer to the influenced quantity, just in case.
	  ;; (Otherwise, not all quantities may be around at closing time.
	  (fetch-quantity ,(adb::quotize (cadr form)))
	  ;; Caches direct influence in the closed world table
	  (adb::add-to-nested-alist
	    ,(adb::quotize (list (cadr form) **ANTECEDENTS**))
	    ,(if (eq **CDI** 'PROCESS) 
		 (adb::quotize form)
		 (error "~A cannot be used outside a process."
			form))
	    *DP-INFLUENCERS*)))

(defTranslation I- `(,form . :TRUE)
  :results `(( (,form . :FALSE) ,(flip-truths **ANTECEDENTS**)))
  :eval `(;; Refer to the influenced quantity, just in case.
	  ;; (Otherwise, not all quantities may be around at closing time.
	  (fetch-quantity ,(adb::quotize (cadr form)))
	  ;; Caches direct influence in the closed world table
	  (adb::add-to-nested-alist ,(adb::quotize (list (cadr form) **ANTECEDENTS**))
			   ,(if (eq **CDI** 'PROCESS) 
				(adb::quotize form)
				      (error "~A cannot be used outside a process."
					      form))
			   *DP-INFLUENCERS*)))

(Deftranslation CTrans  ;(CTrans source-quant dest-quant rate)
		;; "Continuous Transfer"
		;; Due to BCF.  Sugar for a common pattern of abstraction.
   (trans-rel1 `(And (I- ,(second form) ,(fourth form))
		     (I+ ,(third form)  ,(fourth form))))
  :results `(((,form . :TRUE) ,**ANTECEDENTS**)))

;;;; Qualitative proportionalities

(defvar *careful-qprops?* nil) ;; Internal mode switch
  ;; When non-nil, introduce nogoods which ensure that
  ;; qprop is never believed when you don't believe each
  ;; quantity exists.  I think it's probably a bad idea,
  ;; jcollins claims "no experienced modeller would ever do that".

(deftranslation Qprop+ (trans-rel1 `(Qprop ,@ (cdr form))))

(defTranslation Qprop NIL 
  ;must index this under the potential indirect influencers.
  :results `(((,form . :TRUE) ,**antecedents** ,*default-informant*))
  :eval `(;; Mention quantities so they are definitely around at closing time.
	  (fetch-quantity ,(adb::quotize (cadr form)))
	  (fetch-quantity ,(adb::quotize (caddr form)))
	  (adb::add-to-nested-alist-unique
	    ,(adb::quotize (list (cadr form) **antecedents**))
	    ,(adb::quotize form)
	    *IP-INFLUENCERS*)
	  ;Qprop only holds when the quantities exist.
	  ;This says something stronger, that it is inconsistent for the quantity to not
	  ;exist when a qprop involving it is true.  Just being careful...
	  ;; 12/31/87 -- Is this code needed?
	  (when *careful-qprops?*
	    (adb::nogood ,(adb::quotize
			    `((,form . :TRUE) ((quantity ,(cadr form)) . :FALSE))))
	    (adb::nogood ,(adb::quotize
			    `((,form . :TRUE) ((quantity ,(caddr form)) . :FALSE)))))))

(defTranslation Qprop- NIL
  ;must index this under the potential indirect influencers.
  :results `(((,form . :TRUE) ,**antecedents** ,*default-informant*))
  :eval `(;; Mention quantities so they are definitely around at closing time.
	  (fetch-quantity ,(adb::quotize (cadr form)))
	  (fetch-quantity ,(adb::quotize (caddr form)))
	  (adb::add-to-nested-alist-unique
	    ,(adb::quotize (list (cadr form) **ANTECEDENTS**))
	    ,(adb::quotize form)
	    *IP-INFLUENCERS*)
	  ;Qprop only holds when the quantities exist.
	  (when *careful-qprops?*
	    (adb::nogood ,(adb::quotize
			    `((,form . :TRUE) ((quantity ,(cadr form)) . :FALSE))))
	    (adb::nogood ,(adb::quotize
			    `((,form . :TRUE) ((quantity ,(caddr form)) . :FALSE)))))))

;;; Explicit functions are now expanded at translation time.
;;; This should speed up instantiation considerably.
(defTranslation Function-Spec (multiple-value-bind
				(j-form name value arguments)
				  (parse-explicit-function (cdr form))
				(dolist (qprop (cddr form)) (trans-rel1 qprop))
				`(:EVAL ,(adb::quotize j-form)))
  :eval (multiple-value-bind (ignore name value arguments)
	    (parse-explicit-function (cdr form))
	  `((install-explicit-function-specs ,(adb::quotize name)
					     ,(adb::quotize value)
					     ,(adb::quotize arguments)))))

;;;; Correspondences

;;; Correspondences are now expanded at translation time as well.
;; Leave in the correspondence as an explicit fact, for debugging purposes.
;; Could probably affect a slight speedup by dropping them from explicit
;; representation in the database. 

(defTranslation Correspondence
		(cons form ':TRUE)
  :eval (let ((pairs (check-correspondence-pairs (cddr form))))
		  (unless pairs
		    (error "Bad correspondence form:~%~A." form))
		  (append 
		      (mapcar #'(lambda (pair)
				  `(install-derived-qstate-pair
				     ,(adb::quotize (car pair))
				     ,(adb::quotize (cadr pair))))
			      (cdr form))
		      (mapcar #'(lambda (j)
				  `(adb::justify ,(adb::quotize (trans-rel1 (car j)))
						,(adb::quotize 
						   (cons (cons form :TRUE)
							 (mapcar #'trans-rel1 (cdr j))))
						'Correspondence))
			      (make-correspondence-justifications (cadr form)
								  (cddr form))))))

(Deftranslation ordered-correspondence
  `(,form . :TRUE)
  ;; Due to Jcollins
  :eval `((install-derived-qstate-pair ,(quotize (car (cadr form)))
				       ,(quotize (cadr (cadr form))))
	  (install-derived-qstate-pair ,(quotize (car (caddr form)))
				       ,(quotize (cadr (caddr form)))))
  :results (let ((**antecedents** `((,form . :TRUE))))
	     (make-ordered-correspondence-results (second form) (third form))))

(defun replace-quantity-by-number (form selector)
  (cond ((null form) nil)
	((not (listp form)) ;hit bottom - must be quantity
	 (if (eq form 'zero) form
	     (list selector form)))
	((algebraic-operator? (car form))
	 (cond ((eq (car form) 'SUM-OF) ;special operator
		(list (car form)
		      (cons (cadr form) selector)
		      (caddr form)))
	       (t (cons (car form)
			(mapcar #'(lambda (f)
				   (replace-quantity-by-number f selector))
				(cdr form))))))
	((member (car form) '(:A :D)) ;already a number
	 (error "Improperly mixed algebraic expression - ~A." form))
	(t (list selector form))))

;;;; Algebra
;
;;These transformations can be used anywhere.
;;Assumptions are
;;1. + and - are binary operators
;;2. = or Q= indicates the type of the ground terms (i.e., either numbers or quantities)
;;3. Equality statements which translate to relational operators must look like
;;   assignment statements, i.e., the thing on the left is "the result of" the thing
;;   on the right.  (causally)


;;; First, some declarations.  To allow user extensions, must be able to declare
;;; new operators and relational forms to be expanded.

(defmacro DefRelational-Operator (symb inf)
  `(progn (setf (get ',symb 'Relational-Form)  ',inf)))

(defun form-to-be-relationalized? (form)
  (and form (listp form) (symbolp (car form))
       (get (car form) 'Relational-form)))

(defun relational-form-of (operator &aux relational-form)
  (unless (symbolp operator)
    (error "~% ~A can't be relational operator -- must be a symbol." operator))
  (setq relational-form  (get operator 'Relational-form))
  (cond ((and relational-form 
	      (symbolp relational-form)) relational-form)
	(t (error "~%~A not a relational operator." relational-form))))


(Defmacro DefAlgebraic-Operator (symb) 
  `(progn (setf (get ',symb 'Algebraic-Operator)  T)
	  ',symb))

(defun algebraic-operator? (symb) (and (symbolp symb)
				       (get symb 'Algebraic-Operator)))

;; Operators supplied with default syntax
(DefAlgebraic-Operator +)
(DefAlgebraic-Operator -)
(DefAlgebraic-Operator =)
(DefAlgebraic-Operator Sum-Of)
(DefAlgebraic-Operator Qprop)
(DefAlgebraic-Operator Qprop-)

(defRelational-Operator + ++rel) ;prevent duplicate expansion
(defRelational-Operator - --rel)

;; Some jcollins augments (experimental)
(DefAlgebraic-Operator Q-)
(DefRelational-Operator Q- Q--rel)

;;;; Equality 
(defTranslation Q=
		(trans-rel1 `(And ,@ (unless (form-to-be-relationalized? (third form))
				       `((Qprop ,(cadr form) ,(third form))
					 (quote (,form . :TRUE))))
				  ,(replace-quantity-by-number
				     (cons '= (cdr form)) 'A)
				  ,(replace-quantity-by-number
				     (cons '= (cdr form)) 'D)))
  :eval
  (unless (form-to-be-relationalized? (third form))
    `((install-equivalent-quantity-pair ,(quotize form)))))

(defTranslation =
		(cond ((form-to-be-relationalized? (caddr form))
		       (trans-rel1 `(,(relational-form-of (car (caddr form)))
				     ,(cadr form)
				     ,@ (cdr (caddr form)))))
		      (t (let ((t1 (trans-rel1 (cadr form)))
			       (t2 (trans-rel1 (caddr form))))
			   `(:eval (find-qrel-assertion1 ,(adb::quotize t1)
							 ,(adb::quotize t2)
							 '=))))))

;;These are support for jcollins' products resolution code.
(defvar *equivalents* nil)
(adb:init-form (setq *equivalents* nil))

(defun install-equivalent-quantity-pair (form)
  (let ((Q1 (second form))
	(Q2 (third form)))
    (add-alist-value Q1 (cons Q2 form) *equivalents*)
    (add-alist-value Q2 (cons Q1 form) *equivalents*)))

;;;; Simple addition and subtraction 

(DefTranslation ++rel ;addition relation
		(cond ((and (eq 'A (car (cadr form)))
			    (eq 'A (car (caddr form)))
			    (eq 'A (car (cadddr form))))
		       ;if +rel between amounts of something, assert Qprop
		       ;between the quantities involved
		       (trans-rel1 `(And (+rel ,@ (cdr form))
					 (Qprop ,(cadr (cadr form)) ,(cadr (caddr form)))
					 (Qprop ,(cadr (cadr form))
						 ,(cadr (cadddr form))))))
		      (t (trans-rel1 `(+rel ,@ (cdr form))))))

;(DefTranslation --rel ;subtraction relation
;		(cond ((and (eq 'A (car (cadr form)))
;			    (eq 'A (car (caddr form)))
;			    (eq 'A (car (cadddr form))))
;		       ;if -rel between amounts of something, assert Qprop
;		       ;between the quantities involved
;		       (trans-rel1 `(And (-rel ,@ (cdr form))
;					 (Qprop ,(cadr (cadr form)) ,(cadr (caddr form)))
;					 (Qprop- ,(cadr (cadr form))
;						  ,(cadr (cadddr form))))))
;		      (t (trans-rel1 `(-rel ,@ (cdr form))))))


(DefTranslation --rel ;subtraction relation
   (cond ((eq (third form) 'ZERO) ; Opposite-sign.
	  (trans-rel1 `(--rel ,(second form) ,(fourth form))))
	 ((null (cdddr form)) ; Unary minus (opposite-sign).
	  (cond ((some #'(lambda (num) (and (listp num) (eq 'A (car num)))) (cdr form))
		 (trans-rel1 `(And (opposite-sign ,@ (cdr form))
				   (Qprop- ,(second (second form))
					   ,(second (third form))))))
		(t (trans-rel1 `(opposite-sign ,@ (cdr form))))))
	 ((some #'(lambda (num) (and (listp num) (eq 'A (car num)))) (cdr form))
	  ;if -rel between amounts of something, assert Qprop & Qprop-
	  ;between the quantities involved
	  (trans-rel1 `(And (+rel ,(third form) ,(second form) ,(fourth form))
			    (Qprop ,(second (second form))
				   ,(second (third form)))
			    (Qprop- ,(second (second form))
				    ,(second (fourth form)))
			    (Qdiff ,(second (second form))
				   ,(second (third form))
				   ,(second (fourth form))))))
	 (t (trans-rel1 `(+rel ,(third form) ,(second form) ,(fourth form))))))


(Deftranslation +rel ; Addition operation on numbers ;; X = Y + Z
; OLD VERSION
;		(let ((X (cadr form))
;		      (Y (caddr form))
;		      (Z (cadddr form)))
;		(trans-rel1 `(And (Correspondence (,X ,Y) (,Z ZERO))
;				  (Correspondence (,X ,Z) (,Y ZERO)))))
		nil
  :eval `((install-derived-qstate-pair ,(quotize (cadr form))
				       ,(quotize (caddr form)))
	  (install-derived-qstate-pair ,(quotize (cadr form))
				       ,(quotize (cadddr form))))
  :results `(,@ (make-ordered-correspondence-results `(,(cadr form) ,(caddr form))
						     `(,(cadddr form) ZERO))
	     ,@ (make-ordered-correspondence-results `(,(cadr form) ,(cadddr form))
						     `(,(caddr form) ZERO))
	     ((:eval `((opposite ,@ (find-qrel ,(quotize (caddr form))
					       ,(quotize (cadddr form)))) . :TRUE))
	      (((equal-to ,(cadr form) ZERO) . :TRUE) ,@ **antecedents**)
	      :OPPOSITES)))

(Deftranslation -rel ; Subtraction operation on numbers ;; X = Y - Z
		(let ((X (cadr form))
		      (Y (caddr form))
		      (Z (cadddr form)))
		(trans-rel1 `(And (Correspondence (,X ZERO) (,Y ,Z))
				  (Correspondence (,X ,Y) (,Z ZERO))
				  ;; This is the old definition.  Not as powerful
				  ;; (Correspondence (,X ZERO) (,Y ZERO) (,Z ZERO))
				  ))))



;;;; Experimental relational operators (from Jcollins)

(DefTranslation Q--rel ;Qualitative Subtraction Relation using only one corresp.
		(cond ((and (eq 'A (car (cadr form)))
			    (eq 'A (car (caddr form)))
			    (eq 'A (car (cadddr form))))
		       ;if -rel between amounts of something, assert Qprop
		       ;between the quantities involved
		       (trans-rel1 `(And (Q-rel ,@ (cdr form))
					 (Qprop ,(cadr (cadr form)) ,(cadr (caddr form)))
					 (Qprop- ,(cadr (cadr form))
						  ,(cadr (cadddr form))))))
		      (t (trans-rel1 `(Q-rel ,@ (cdr form))))))

(Deftranslation Q-rel ; Qualitative subtraction operation on numbers
		(let ((X (cadr form))
		      (Y (caddr form))
		      (Z (cadddr form)))
		(trans-rel1 `(Ordered-Correspondence (,X ZERO) (,Y ,Z)))))

(Deftranslation opposite-sign
   (trans-rel1 `(And (ordered-correspondence (,(second form) ZERO)
					     (ZERO ,(third form)))
		     (,form . :TRUE)))
  :results
    `(((:eval (make-qrel-assertion
		(list 'm ,(adb:quotize (second form)))
		(list 'm ,(adb:quotize (third form)))
		'=))
       (,@ **antecedents**))))

(DefTranslation +Qrel ;Qualitative (constraint) Addition Relation (w/o causal links)
   (trans-rel1 `(And (+rel ,(make-amount (second form))
			   ,(make-amount (third form))
			   ,(make-amount (fourth form)))
		     (+rel ,(make-derivative (second form))
			   ,(make-derivative (third form))
			   ,(make-derivative (fourth form))))))

;;;; Products

(Defalgebraic-operator *)
(defalgebraic-operator /)
(defalgebraic-operator *+)
(defalgebraic-operator /+)
(defalgebraic-operator *0+)
(defalgebraic-operator /0+)

(defrelational-operator * **rel)
(defrelational-operator / /rel)
(defrelational-operator *+ *+rel)
(defrelational-operator /+ /+rel)
(defrelational-operator *0+ *0+rel)
(defrelational-operator /0+ /0+rel)

(defvar *products* nil)
(defvar *ratios* nil)

;;; PRODUCTS:

(defTranslation **rel
	(cond ((or (eq 'A (car (second form)))
		   (eq 'A (car (third form)))
		   (eq 'A (car (fourth form))))
	       (cond ((equal (third form) (fourth form))
		      (trans-rel1 `(*rel-square ,(second form) ,(third form))))
		     (t (trans-rel1 `(And (*rel-common ,(second form)
						       ,(third form)
						       ,(fourth form))
					  (*rel-symmetric ,(second form)
							  ,(third form)
							  ,(fourth form))
					  (*rel-symmetric ,(second form)
							  ,(fourth form)
							  ,(third form)))))))
	      (t (trans-rel1 `(*rel-D0 ,(second form)
				       ,(third form)
				       ,(fourth form))))))

(deftranslation *rel-D0
     nil
  :results `((((equal-to ,(second form) zero) . :TRUE)
	      (((equal-to ,(third form) zero) . :TRUE)
	       ((equal-to ,(fourth form) zero) . :TRUE)
	       ,@ **ANTECEDENTS**) :CONSTANT-PRODUCT)))

(defTranslation *rel-square
   (progn
     (let ((**antecedents**
	     (cons `((greater-than ,(third form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop ,(second (second form))
			   ,(second (third form)))))
     (let ((**antecedents**
	     (cons `((less-than ,(third form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop- ,(second (second form))
			    ,(second (third form)))))
     nil)
    :results `((((less-than ,(second form) ZERO) . :FALSE)
		(,@ **ANTECEDENTS**) :SQUARE)
	       (((greater-than ,(second form) ZERO) . :TRUE)
		(((greater-than ,(third form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :SQUARE)
	       (((greater-than ,(second form) ZERO) . :TRUE)
		(((less-than ,(third form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :SQUARE)
	       (((equal-to ,(second form) ZERO) . :TRUE)
		(((equal-to ,(third form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :SQUARE)
	       (((equal-to ,(third form) ZERO) . :TRUE)
		(((equal-to ,(second form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :SQUARE)
	       (((equal-to ,(third form) ZERO) . :FALSE)
		(((equal-to ,(second form) ZERO) . :FALSE)
		 ,@ **ANTECEDENTS**) :SQUARE)))

(defTranslation *rel-common
    (trans-rel1 `(Qproduct ,(second (second form))
			   ,(second (third form))
			   ,(second (fourth form))))
    :results `((((greater-than ,(second form) ZERO) . :TRUE)
		(((greater-than ,(third form) ZERO) . :TRUE)
		 ((greater-than ,(fourth form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :PRODUCT)
	       (((greater-than ,(second form) ZERO) . :TRUE)
		(((less-than ,(third form) ZERO) . :TRUE)
		 ((less-than ,(fourth form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :PRODUCT))
    :eval (when (or (eq 'A (car (second form)))
		    (eq 'A (car (third form)))
		    (eq 'A (car (fourth form))))
	  `((insert-product ,(quotize (second (second form)))
			    ,(quotize (second (third form)))
			    ,(quotize (second (fourth form)))
			    ,(quotize (car **ANTECEDENTS**))))))
	       
(defTranslation *rel-symmetric
   (progn		
     (let ((**antecedents**
	     (cons `((greater-than ,(fourth form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop ,(second (second form))
			   ,(second (third form)))))
     (let ((**antecedents**
	     (cons `((less-than ,(fourth form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop- ,(second (second form))
			    ,(second (third form)))))
     (trans-rel1 `(*rel-0 ,(second form) ,(third form) ,(fourth form)))))

(defTranslation *rel-0
		nil
   :results `((((equal-to ,(second form) ZERO) . :TRUE)
	       (((equal-to ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((less-than ,(second form) ZERO) . :TRUE)
	       (((greater-than ,(third form) ZERO) . :TRUE)
		((less-than ,(fourth form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((greater-than ,(fourth form) ZERO) . :TRUE)
	       (((greater-than ,(second form) ZERO) . :TRUE)
		((greater-than ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((less-than ,(fourth form) ZERO) . :TRUE)
	       (((greater-than ,(second form) ZERO) . :TRUE)
		((less-than ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((less-than ,(fourth form) ZERO) . :TRUE)
	       (((less-than ,(second form) ZERO) . :TRUE)
		((greater-than ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((greater-than ,(fourth form) ZERO) . :TRUE)
	       (((less-than ,(second form) ZERO) . :TRUE)
		((less-than ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((equal-to ,(fourth form) ZERO) . :TRUE)
	       (((equal-to ,(second form) ZERO) . :TRUE)
		((equal-to ,(third form) ZERO) . :FALSE)
		,@ **ANTECEDENTS**) :PRODUCT)))

;;; This is a simplified version of PRODUCT
;;; which assumes there are NO NEGATIVE QUANTITIES.

(defTranslation *0+rel
   (when (or (eq 'A (car (second form)))
	     (eq 'A (car (third form)))
	     (eq 'A (car (fourth form))))
     (trans-rel1 `(And (*0+rel-common ,(second form)
				      ,(third form)
				      ,(fourth form))
		       (*0+rel-symmetric ,(second form)
					 ,(third form)
					 ,(fourth form))
		       (*0+rel-symmetric ,(second form)
					 ,(fourth form)
					 ,(third form))
		       (Qproduct ,(second (second form))
				 ,(second (third form))
				 ,(second (fourth form))))))
  :eval (when (or (eq 'A (car (second form)))
		  (eq 'A (car (third form)))
		  (eq 'A (car (fourth form))))
	  `((insert-product ,(quotize (second (second form)))
			    ,(quotize (second (third form)))
			    ,(quotize (second (fourth form)))
			    ,(quotize (car **ANTECEDENTS**))))))

(defTranslation *0+rel-common
		nil
    :results `((((greater-than ,(second form) ZERO) . :TRUE)
		(((greater-than ,(third form) ZERO) . :TRUE)
		 ((greater-than ,(fourth form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :PRODUCT)))
	       
(defTranslation *0+rel-symmetric
   (progn
     (let ((**antecedents**
	     (cons `((greater-than ,(fourth form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop ,(second (second form))
			   ,(second (third form)))))
     (trans-rel1 `(*0+rel-0 ,(second form) ,(third form) ,(fourth form)))))

(defTranslation *0+rel-0
		nil
   :results `((((equal-to ,(second form) ZERO) . :TRUE)
	       (((equal-to ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((greater-than ,(fourth form) ZERO) . :TRUE)
	       (((greater-than ,(second form) ZERO) . :TRUE)
		((greater-than ,(third form) ZERO) . :TRUE)
		,@ **ANTECEDENTS**) :PRODUCT)
	      (((equal-to ,(fourth form) ZERO) . :TRUE)
	       (((equal-to ,(second form) ZERO) . :TRUE)
		((equal-to ,(third form) ZERO) . :FALSE)
		,@ **ANTECEDENTS**) :PRODUCT)))

;;; RATIOS:

(defTranslation /rel
   (when (or (eq 'A (car (second form)))
	     (eq 'A (car (third form)))
	     (eq 'A (car (fourth form))))
     (trans-rel1 `(And (/rel-common ,(second form)
				    ,(third form)
				    ,(fourth form))
		       (*rel-0 ,(third form)
			       ,(second form)
			       ,(fourth form))
		       (*rel-0 ,(third form)
			       ,(fourth form)
			       ,(second form))
		       (Qratio ,(second (second form))
			       ,(second (third form))
			       ,(second (fourth form))))))
  :eval (when (or (eq 'A (car (second form)))
		  (eq 'A (car (third form)))
		  (eq 'A (car (fourth form))))
	  `((insert-ratio ,(quotize (second (second form)))
			  ,(quotize (second (third form)))
			  ,(quotize (second (fourth form)))
			  ,(quotize (car **ANTECEDENTS**))))))

(defTranslation /rel-common
   (progn
     (let ((**antecedents**
	     (cons `((greater-than ,(fourth form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop ,(second (second form))
			   ,(second (third form)))))
     (let ((**antecedents**
	     (cons `((less-than ,(fourth form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop- ,(second (second form))
			    ,(second (third form)))))
     (let ((**antecedents**
	     (cons `((greater-than ,(third form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop- ,(second (second form))
			    ,(second (fourth form)))))
     (let ((**antecedents**
	     (cons `((less-than ,(third form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop ,(second (second form))
			   ,(second (fourth form)))))
     nil)
    :results `((((greater-than ,(third form) ZERO) . :TRUE)
		(((greater-than ,(second form) ZERO) . :TRUE)
		 ((greater-than ,(fourth form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :RATIO)
	       (((greater-than ,(third form) ZERO) . :TRUE)
		(((less-than ,(second form) ZERO) . :TRUE)
		 ((less-than ,(fourth form) ZERO) . :TRUE)
		 ,@ **ANTECEDENTS**) :RATIO)))

;;; This is a simplified version of RATIO
;;; which assumes there are NO NEGATIVE QUANTITIES.

(defTranslation /0+rel
   (when (or (eq 'A (car (second form)))
	     (eq 'A (car (third form)))
	     (eq 'A (car (fourth form))))
     (trans-rel1 `(And (/0+rel-common ,(second form)
				       ,(third form)
				       ,(fourth form))
		       (*0+rel-0 ,(third form)
				 ,(second form)
				 ,(fourth form))
		       (*0+rel-0 ,(third form)
				 ,(fourth form)
				 ,(second form))
		       (Qratio ,(second (second form))
			       ,(second (third form))
			       ,(second (fourth form))))))
  :eval (when (or (eq 'A (car (second form)))
		  (eq 'A (car (third form)))
		  (eq 'A (car (fourth form))))
	  `((insert-ratio ,(quotize (second (second form)))
			  ,(quotize (second (third form)))
			  ,(quotize (second (fourth form)))
			  ,(quotize (car **ANTECEDENTS**))))))

(defTranslation /0+rel-common
   (progn
     (let ((**antecedents**
	     (cons `((greater-than ,(fourth form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop ,(second (second form))
			   ,(second (third form)))))
     (let ((**antecedents**
	     (cons `((greater-than ,(third form) ZERO) . :TRUE)
		   **antecedents**)))
       (trans-rel1 `(Qprop- ,(second (second form))
			    ,(second (fourth form))))))
  :results `((((greater-than ,(third form) ZERO) . :TRUE)
	      (((greater-than ,(second form) ZERO) . :TRUE)
	       ((greater-than ,(fourth form) ZERO) . :TRUE)
	       ,@ **ANTECEDENTS**) :RATIO)))

; This is for comparing coefficients to ONE:

'''(defTranslation *rel-1
   (trans-rel1 `(And (IFF (equal-to ,(third form) ONE)
			  (equal-to ,(second form) ,(fourth form)))
		     (-> (And (greater-than ,(third form) ONE)
			      (greater-than ,(fourth form) ZERO))
			 (greater-than ,(second form) ,(fourth form)))
		     (-> (And (less-than ,(third form) ONE)
			      (less-than ,(fourth form) ZERO))
			 (greater-than ,(second form) ,(fourth form)))
		     (-> (And (greater-than ,(third form) ONE)
			      (less-than ,(fourth form) ZERO))
			 (less-than ,(second form) ,(fourth form)))
		     (-> (And (less-than ,(third form) ONE)
			      (greater-than ,(fourth form) ZERO))
			 (less-than ,(second form) ,(fourth form)))
		     
		     )))


;; This is an extremely simplified version of PRODUCT
;; which assumes that ALL Qty's ARE POSITIVE:

(defTranslation *+rel
   (when (or (eq 'A (car (second form)))
	     (eq 'A (car (third form)))
	     (eq 'A (car (fourth form))))
     (trans-rel1 `(And (greater-than ,(second form) ZERO)
		       (Qprop ,(second (second form))
			      ,(second (third form)))
		       (Qprop ,(second (second form))
			      ,(second (fourth form))) 
		       (Qproduct ,(second (second form))
				 ,(second (third form))
				 ,(second (fourth form))))))
  :eval (when (or (eq 'A (car (second form)))
		  (eq 'A (car (third form)))
		  (eq 'A (car (fourth form))))
	  `((insert-product ,(quotize (second (second form)))
			    ,(quotize (second (third form)))
			    ,(quotize (second (fourth form)))
			    ,(quotize (car **ANTECEDENTS**))))))

;; This is an extremely simplified version of RATIO
;; which assumes that ALL Qty's ARE POSITIVE:

(defTranslation /+rel
   (when (or (eq 'A (car (second form)))
	     (eq 'A (car (third form)))
	     (eq 'A (car (fourth form))))
     (trans-rel1 `(And (greater-than ,(second form) ZERO)
		       (Qprop ,(second (second form))
			      ,(second (third form)))
		       (Qprop- ,(second (second form))
			       ,(second (fourth form))) 
		       (Qratio ,(second (second form))
			       ,(second (third form))
			       ,(second (fourth form))))))
  :eval (when (or (eq 'A (car (second form)))
		  (eq 'A (car (third form)))
		  (eq 'A (car (fourth form))))
	  `((insert-ratio ,(quotize (second (second form)))
			  ,(quotize (second (third form)))
			  ,(quotize (second (fourth form)))
			  ,(quotize (car **ANTECEDENTS**))))))

(defun insert-product (prod mult1 mult2 just)
  (add-alist-value prod `((QPRODUCT ,prod ,mult1 ,mult2) . ,just) *products*)
  (setf (get (car prod) ':PRODUCT?) T))

(defun insert-ratio (ratio numer denom just)
  (add-alist-value ratio `((QRATIO ,ratio ,numer ,denom) . ,just) *ratios*)
  (if-debug-flag resolve-ratios
     (format t "~%   RR: Inserting ratio: ~A" ratio))
;  (setf (get (car ratio) ':RATIO?) T)
  )

(adb::init-form (progn (dolist (entry *products*)
			(remprop (caar entry) ':PRODUCT?))
		      (setq *products* nil)))

(adb::init-form (progn (dolist (entry *ratios*)
			(remprop (caar entry) ':RATIO?))
		      (setq *ratios* nil)))