;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: QPE; -*-

(in-package 'QPE)

;;;; QPE syntaxer
;;;
;;;  This code takes "user-friendly" (or at least, "not horrific") statements
;;; out of domain models and QPE rules and turns them into something that
;;; ADB can deal with directly.  Often side-effects are required to inform QPE
;;; about various things, and this code provides a mechanism to install them.
;;; The intent is to insulate the user from the internal details of QPE as much as    
;;; possible.

;;;  For simplicity and ease of extension, most of the syntax is defined via
;;; data-driven translation functions.  These functions can expand code in various  
;;; ways and provide a means of doing side-effects.
;;; It also provides a mechanism for user-defined extensions.  Such files should be
;;; declared using the DefAugment-File mechanism below.

(proclaim '(special **ANTECEDENTS** **CDI**))
  ;; These variables are used by TRANSLATE-RELATIONS to store the antecedent forms.

(proclaim '(special *result-forms* *evaluation-forms*))
 ;; These variables are used as accumulators for translator results.
 ;; Result-forms are things to be asserted.
 ;; evaluation-forms are things to be eval'ed at run-time.

(proclaim '(special *local-names* *creates-individual* *default-informant*))
 ;; Local names are names introduced by translator constructs. 
(defvar *default-informant* ":QPE") ;; used by syntaxer in translations

(defvar *Augment-Files* nil "Files containing user-supplied code.")

(defmacro DefAugment-File (file &key (init nil))
  ;; :Init signifies that the file should be loaded whenever ADB is initialized
  ;; Default is to load the file once.
  `(progn (load ,file)
	  (pushnew ,file *Augment-Files* :test #'string-equal)
	  ,@ (if init `((adb::Init-Form (load ,file))))))

(defmacro DeclareAugment-File (file)
  `(pushnew ,file *Augment-Files* :test #'string=))

(defun reload-augment-files () ;; Just in case.
  (dolist (file *Augment-Files*) (load file)))

;;;; Internal interface to QPE syntaxer

(defun translate-relations (form &optional (antecedents nil) (cdi? nil)
				 (default-informant ":syntaxer")
				 &aux new-form)
;; Translates a list of expressions into more primitive terms during cdi definition
  (setq **ANTECEDENTS** antecedents) ;list of antecedents
  (setq **CDI** cdi?) ;; type of CDI, if in that scope.
  ;; This distinction is a holdover from GIZMO.  The right thing may be to always pass in
  ;; a list of antecedent forms.  Using a list means that the trigggers for a rule can be used
  ;; by the translator, as well as just single statements.
  (setq *default-informant* default-informant)
  (setq *result-forms* nil
	*evaluation-forms* nil
	*local-names* nil
	*creates-individual* nil)
  (setq *local-names* (find-all-local-names form))
  (setq new-form (substitute-local-names form *local-names*))
  (values (delete nil
		  (remove-conjunctions
		   (cond ((not (listp new-form)) (trans-rel1 new-form)) ;do the translation
			 (t (mapcar #'trans-rel1 new-form)))))
	  *result-forms*
	  *evaluation-forms*
	  *local-names*
	  *creates-individual*))

(defun flip-translation-antecedents () (flip-truths **ANTECEDENTS**))


;;;; Installing local names
;;
;; For elegance, certain forms are translated at run-time to be relative
;; to the given named description.  A common example is FLOW-RATE
;; in a process definition.  One could force the modeller to write
;; (FLOW-RATE ?SELF), but that is ugly. 
;;
;; Finding and replacing local names must be done before the rest of the
;; syntaxing process.  Postponing it will make accurate substitution difficult
;; (or even impossible) because of the form changes required as a consequence
;; of partial evaluation.        

(defun substitute-local-names (form local-names) 
  (adb::first-order-sublis local-names form))

(defun find-all-local-names (form)
  (mapcar #'(lambda (name) (cons name (list name '?SELF)))
	  (get-all-local-names form nil)))

(defun get-all-local-names (form locals)
  (cond ((null form) locals)
	((not (listp form)) locals)
	((eq (car form) 'Quantity) ;only local creator at the moment
	 (unless (or (listp (cadr form))
		     (variable? (cadr form)))
	   (pushnew (cadr form) locals))
	 locals)
	(t ;; must recurse carefully, since we only want to substitute
	 ;; for Quantity (or other local creators, if they are added.)
	 (when (listp (cdr form))
	   (dolist (subform form)
	     (setq locals (get-all-local-names subform locals))))
	   locals)))

;;;; Case analysis

(defun trans-rel1 (form &aux translation-function)
  (cond ((null form) nil)
	((adb::variable? form)
	 (cond ((adb::bound-variable? form) form)
	       (t (error "~%~A illegal free variable in expression" form))))
	((not (listp form)) form) ;do nothing
	((symbolp (car form))
	 ;attempt a translation.  The priorities are: 
	 ; If there is a translation function, use it.
	 ; If the symbol is marked as a function, do nothing.
	 ; Otherwise, assume it is a predicate, and wrap appropriate
	 ; "truth value" around it to make it into truth-functional form.
	 (setq translation-function (get-translation-form (car form)))
	 (cond (translation-function (funcall translation-function form))
	       ((name-of-function? (car form)) form)
	       ((eq (car form) 'not) (cons (cadr form) ':FALSE))
	       (t (cons form ':TRUE))))
	;; Do NOT tree walk!  Only do this on top-level expressions, thank you!
	;; if this is needed on sub-expressions, the translator will call it explicitly.
	(t form)))
 
(defun name-of-function? (thing)
  (cond ((symbolp thing) (get thing 'QP-function-name))
	((and (listp thing)
	      (symbolp (car thing))) (get (car thing) 'QP-function-name))
	(t (adb::get-assert-prop (adb::referent thing t) 'qp-function-name))))

(defmacro defFunction-name (thing)
  `(progn 'compile
	  ,(cond ((symbolp thing) `(setf (get ',thing 'qp-function-name)  t))
		 (t `(adb::put-assert-prop (adb::referent ',thing t) t 'qp-function-name)))))

;; Wire in the standard functions involving quantities
(defFunction-name A)
(defFunction-name D)
(defFunction-name s)
(defFunction-name m)

(defun get-translation-form (sym) (get sym 'qp-translation-form))

;;;; Defining translators

(defmacro DefTranslation (sym translation
			  &key (results nil) (locals nil) (eval nil))
  `(progn (setf (get ',sym 'QP-translation-form)
		,(cond (locals
			;; If there are local names, make sure they are bound
			;; before doing anything.  In any case, if there are results or eval-forms
			;; then make sure they are added to the global accumulators.  The
			;; value to be returned is the translation itself.
			`(function (lambda (form)
				  (let ,locals
				    ,@ (cond (results
					      `((setq *result-forms*
						      (append ,results
							      *Result-forms*))))
					     (t nil))
				       ,@ (cond (eval
						 `((setq *evaluation-forms*
							 (append ,eval
								 *evaluation-forms*))))
						(t nil))
				       ,translation))))
		       (t `(function (lambda (form)
				    ,@ (cond (results
					      `((setq *result-forms*
						      (append ,results
							      *result-forms*))))
					     (t nil))
					  ,@ (cond (eval
						    `((setq *evaluation-forms*
							    (append ,eval
								    *evaluation-forms*))))
						   (t nil))
					  ,translation)))))
	  ',sym)) ;return the name of the symbol translated as the value.

(defun remove-conjunctions (forms)
  ;;; For each form, walk down tree to gather
  ;;; non-AND leaves.  Append results.
  (mapcan #'flatten-AND-Tree forms))

(defun flatten-and-tree (tree)
  (cond ((null tree) nil)
	((not (listp tree)) (list tree))
	((eq (car tree) 'AND)
	 (mapcan #'flatten-AND-tree (cdr tree)))
	(t (list tree))))

;;;; Debugging the syntaxer
;;
;; Past experience has shown that debugging translators can be
;; increadibly painful.  By providing the ability to translate small
;; groups of statements, this debugger should speed the process noticably.
;; It should also help modelers who want to know what their statements
;; REALLY expand into, in case they want (or need) deeper knowledge of
;; how QPE works.

(defun debug-QPE-syntaxer ()
  ;; This debugger is keyboard driven for portability 
  (do ((done? nil)
       (forms nil)
       (antes '((**FOO** . :TRUE)))
       (cdi nil)
       (print-all? nil)
       (each-break? nil))
      (done? t)
    (format t "~%Syntaxer (F, O, Q, ?):")
    (case (read-char)
      (#\?
       (format t "~%This is the QPE Syntaxer debugger.  Commands are:")
       (format t "~%  F, f -- Expand a form.")
       (format t "~%  O, o -- Select various options")
       (format t "~%  Q, q -- Leave the QPE Syntaxer debugger.")
       (format t "~%  ?    -- This message."))
      ((#\F #\f)
       (format t "~% Form to expand:")
       (setq forms (read))
       (multiple-value-bind (translation result-forms
			     eval-forms local-names
			     creator?)
	   (translate-relations (list forms) antes cdi)
	 (if print-all? 
	     (if creator? (format t "~% Creates an individual.")
		 (format t "~% Does not create an individual.")))
	 (if print-all?
	     (if local-names (format t "~% The local names are ~A." local-names)
		 (format t "~% No local names found.")))
	 (format t "~% Translation is:")
	 (dolist (tr translation)
	   (pprint tr) ;Oh, for the days of LPRINT!
	   (if each-break? (read-char)))
	 (format t "~% Result forms are:")
	 (dolist (re result-forms)
	   (pprint re)
	   (if each-break? (read-char)))
	 (format t "~% Evaluation forms are:")
	 (dolist (ef eval-forms)
	   (pprint ef)
	   (if each-break? (read-char)))))
      ((#\O #\o) ;; Query about options
       ;; Provide predicate form
       ;; Provide CDI form
       ;; Turn breaking after each form on or off
       ;; Turn total printing on or off.
       (do ((options-done? nil))
	   (options-done? t)
	 (format t "~% Options are:")
	 (format t "~%  A, a -- Change antecedents (Currently ~A)."
		 antes)
	 (format t "~%  C, c -- Change CDI flag (Currently ~A)."
		 cdi)
	 (format t "~%  B, b -- Wait after typing each expanded form (Currently ~A)."
		 (if each-break? "ON" "OFF"))
	 (format t "~%  P, p -- Print all information (Currently ~A)."
		 (if print-all? "ON" "OFF"))
	 (format t "~%  Q, q -- Quit, and return to debugger.") 
	 (case (read-char)
	   ((#\Q #\q) (setq options-done? t))
	   ((#\A #\a)
	    (format t "~% Please type in list of new antecedents:")
	    (do ((new-form (read)))
		((listp new-form) (setq antes new-form))
	      (unless (listp new-form)
		(format t "~% The antecedents must be a list.  Please try again.")
		(setq new-form (read)))))
	   ((#\C #\c)
	    (format t "~% Please type in symbol or NIL for CDI flag:")
	    (setq cdi (read)))
	   ((#\B #\b) (setq each-break? (not each-break?)))
	   ((#\P #\p) (setq print-all? (not print-all?))))))
      ((#\Q #\q) (setq done? t)))))
