;;; -*- Mode: Lisp; Syntax: Common-lisp; Package: QPE; -*-

(in-package 'QPE)

;;; QPE toplevel drivers

;; These hooks do not assume any particular mode of interaction.
;; That is the province of various interface systems.
;; The default is now character-oriented. 

;;; This version uses the following ADB rule packets:
;;   :QPE-rules - these are the default rules to be loaded very time.
;;   :QPE-Domain-Model -- consisting of the forms in the domain files.
;;   :QPE-Scenario-Model -- consisting of the forms in the scenario files.
;; 
;; These force a few changes in the system interface.
;; 1. If you change the list of files which comprise the domain model, the interface
;;    must clear the appropriate packet.  Otherwise, QPE will continue to use the old
;;    model.
;; 2. If the interface wants to manipulate the construction of the domain model or
;;    scenario model itself, it can arrange these packets any way it likes, and
;;    the routines in this file will simply use them.

(proclaim '(special *Assume-PCs*))

(defun clear-qpe-packets ()
  ;; force stuff to be reloaded
  (adb::clear-packet :QPE-rules)
  (adb::clear-packet :QPE-Domain-Model)
  (adb::clear-packet :QPE-Scenario-Model))

(adb::init-form (setq tms::*h45* nil))

(defun e (num) (tms::e num))

(proclaim '(special *QPE-rules-files*))

(defun reload-QPE-rules ()
  (unless (adb::packet-defined? :QPE-rules)
    (dolist (rules-file *QPE-rules-files*)
      (adb::load-in-packet rules-file :QPE-rules)))
  (adb::use-packet :QPE-Rules))

;;;; Flags to ensure preconditions are met.
;; By checking to see if previous steps of an analysis
;; have been done at this level, any extra interfaces don't  
;; need to be bothered with it.

(proclaim '(special *QPE-initialized*
		    *domain-loaded*
		    *Scenario-loaded*
		    *analysis-complete*))

(defvar *QPE-initialized* nil)
(defvar *Domain-loaded* nil)
(defvar *Scenario-loaded* nil)
(defvar *Analysis-complete* nil)

;;;; Start up

(defun start-qpe ()
  (clear-switches *domain-loaded*
		  *scenario-loaded*
		  *analysis-complete*)
  (adb::init t)
  (reload-qpe-rules)
  (set-switch *QPE-initialized* t))

(defun load-domain ()
  (if-switches-set ((*QPE-initialized* . :QPE-NOT-INITIALIZED))
   (unless (adb::packet-defined? :QPE-Domain-Model)
     (dolist (physics-file *domain-files*)
	    (adb::load-in-packet physics-file :QPE-Domain-Model
				 *user-package*)))
   (adb::use-packet :QPE-Domain-Model)
   (set-switch *Domain-Loaded* t)))

(defun load-scenario ()
  (if-switches-set ((*QPE-Initialized* . :QPE-NOT-INITIALIZED)
		    (*Domain-Loaded* . :DOMAIN-NOT-LOADED)) 
   (run-qpe-events :BEFORE :LOAD-DOMAIN)		   
   (unless (adb::packet-defined? :QPE-Scenario-Model)
     (dolist (scenario-file *scenario-files*)
       (adb::load-forms-into-packet  scenario-file :QPE-Scenario-Model
				     *user-package*)))
   (adb::use-packet :QPE-Scenario-Model)
   (set-switch *scenario-loaded* t)
   (run-qpe-events :AFTER :LOAD-DOMAIN)))

(defun expand-scenario () 
  (if-switches-set ((*QPE-Initialized* . :QPE-NOT-INITIALIZED)
		    (*Domain-Loaded* . :DOMAIN-NOT-LOADED)
		    (*Scenario-Loaded* . :SCENARIO-NOT-LOADED))
    (expand-scenario-using-domain-model)))

(proclaim '(special *assume-pcs*))

(defun envision (&optional (pre-expanded? nil) &aux real-start start end)
  (setq real-start (mark-time))
  (unless pre-expanded?
    (adb::if-debug-flag text-highlights
			(format t "~%Expanding Scenario..."))
    (adb::if-debug-flag timers  ;; assumes timers utility
			(setq start (setq real-start (mark-time))))
    (expand-scenario)
    (adb::if-debug-flag timers (setq end (mark-time))
			(say-time (make-time-difference end start)
				  :header "Scenario expansion")))
  (adb::if-debug-flag text-highlights
	     (format t "~%Installing QC/PC assumptions..."))
  (adb::if-debug-flag timers (setq start (mark-time)))
  (Install-QC-PC-assumptions *Assume-Pcs*)
  (adb::if-debug-flag timers
   (setq end (mark-time))
   (say-time (make-time-difference end start) :header "Installing QC/PC"))
  (adb::if-debug-flag text-highlights 
    (format t "~%Resolving influences..."))
  (adb::if-debug-flag timers (setq start (mark-time)))
  (resolve-influences)
  (adb::if-debug-flag timers
   (setq end (mark-time))
   (say-time (make-time-difference end start) :header "Influence resolution"))
  (adb::if-debug-flag text-highlights
   (format t "~%Performing Limit analysis..."))
  (adb::if-debug-flag timers (setq start (mark-time)))
  (limit-analysis)
  (adb::if-debug-flag timers
   (setq end (mark-time))
   (say-time (make-time-difference end start) :header "Limit analysis")
   (say-time (make-time-difference end real-start) :header "Total"))
  (set-switch *analysis-complete* t))

;;;; QPE debugging flags

;; DO NOT CHANGE ORDER W/O CHANGING MANUAL.

(adb::defDebugFlag adb::count-extensions
		   "Incrementally monitor number of interpretations.")
(adb::defDebugFlag adb::justifications "Show all ADB justifications (v. long!)")
(adb::defDebugFlag adb::nogoods "Show all ADB nogoods (v. long!)")
(adb::defDebugFlag adb::close "Show closing of facts")
(adb::defDebugFlag ECL "Show results of Equality Change Law")
(adb::defDebugFlag zero-existence "Checking continuity across changes in existence")
(adb::defDebugFlag continuity "Show results of continuity checking")
(adb::defDebugFlag limit-details "Details of limit analysis")
(adb::defDebugFlag limit "Limit analysis")
(adb::defDebugFlag resolve-ambig
		   "Show justifications for ambiguous influence resolutions")
(adb::defDebugFlag indirectly-influenced
		   "Show justifications for indirect influences")
(adb::defDebugFlag directly-influenced
		   "Show justifications for direct influences")
(adb::defDebugFlag uninfluenced "Show justifications for no influences")
(adb::defDebugFlag Resolve "Influence Resolution")
(adb::defDebugFlag basics "Sclass construction")
(adb::defDebugFlag pcs "Operations involving choice sets")
(adb::defDebugFlag Expand-break "Break after scenario expansion.")
(adb::defDebugFlag cdi-break "Break on finding process or view instance")
(adb::defDebugFlag cdi-asserts "Details of PI and VI consequences")
(adb::defDebugFlag cdi-installer "Installing consequences of PI's and VI's")
(adb::defDebugFlag cdi-finder "Finding process and view instances")
(adb::defDebugFlag quantity "Quantity system")
(adb::defDebugFlag timers "Time QPE operations")
(adb::defDebugFlag Text-Highlights "Print notice of envisioning phases")
