;;; -*- Mode: LISP; Syntax: Common-lisp; Package: QPE -*-

(in-package 'QPE)

;;;; Transitivity inferences

;; Consider the graph of INEQ's as an undirected graph.
;;  Every possible justification and nogood
;; is determined by the cycles in this graph.
;;  In this version, adding an inequality to the graph
;; includes finding all the cycles it makes with the exisiting graph,
;; and creating the appropriate
;; justifications.  No consequent calls are required, since everything which
;; can be deduced given already
;; existing information is deduced.

(defvar *cycles* nil) ;; Cache
(adb:init-form (setq *cycles* nil))

(defvar *incremental-transitivity* nil) ;; Internal mode switch
(defvar *transitivity-processed* nil) ;; Go incremental from here on out
(adb::Init-form (setq *transitivity-processed* nil))

(defun search-for-new-cycles (ineq &aux first-steps n1 n2)
  (setq n1 (ineq-n1 ineq)
	n2 (ineq-n2 ineq))
  (if-debug-flag inequality-search
	(format t "~%  INEQ: Searching from ~A for ~A"
		(qnum-string n1) (qnum-string n2)))
  ;; N1 and N2 are qnums.
  ;; Start by initalizing the paths                  Dotted pairs of ((qnum . ineq) ... )
  (setq first-steps (get-ineqs-search-new-steps (list (cons n1 ineq)) n2))
  ;search from n1 to n2
  (do ((paths first-steps (nconc new-paths (cdr paths))) ;depth-first search
       (step nil) ;current search element
       (new-paths nil nil) ;next prospects
       (new-cycles nil))
      ((null paths)
       (when new-cycles
	 (setf (ineq-new-cycles ineq) new-cycles) ;; Cache
	 (setf *cycles* (append new-cycles *cycles*)))
       (if (or *incremental-transitivity* *transitivity-processed*)
	   (install-ineq-transitivity ineq) ;; Creates appropriate justifications
	   (unless *zero-in-lattice*
	     (if (eq (ineq-type ineq) :MAG)
		 (justify-transitivity-for-mag ineq)
		 (justify-transitivity-across-zero ineq)))))
    (setq step (car paths))
    (cond ((eq (caar step) n2)
	   (push step new-cycles))
	  (t ;; Otherwise, carry on with the search.
	   (setq new-paths (get-ineqs-search-new-steps step n2))))))

;;;; Cycle-finding helpers

(defun get-ineqs-search-new-steps (step n2 &aux qnum ineq marker)
  ;; Step is a list of Dotted pairs of ((qnum . ineq) ... )
  (setq qnum (caar step)
	ineq (cdar step)
	marker (cons t t))
  (dolist (entry (cddr step))
    (put-qnum-prop (car entry) marker :MARKER))
  (do ((entries (qnum-ineqs1 qnum) (qnum-ineqs2 qnum))
       (count 0 (1+ count))
       new-steps other-qnum other-ineq closing-edge)
      ((eq count 2) (if closing-edge
			(list (cons closing-edge step))
			new-steps))
    (dolist (entry entries)
      (setq other-qnum (car entry)
	    other-ineq (cdr entry))
      (cond ((eq (get-qnum-prop other-qnum :marker) marker)
	     (return-from get-ineqs-search-new-steps nil))
	    (closing-edge)
	    ((eq other-ineq ineq))
	    ((eq other-qnum N2)
	     (setq closing-edge entry))
	    ((and (not *zero-in-lattice*)
		  (eq other-qnum *zero-struct*)))
	    (t (push (cons entry step) new-steps))))))

;;; Valid-cycle? is not currently called, since the paths are checked as they're grown. 
;;; This code is worth saving, just in case we decide to forego using cycles until all the facts are in.
;;; A cycle may be invalidated by a later relation, so it would be worth checking them before installing
;;; all those justifications. Also cycles-of-assumptions may not be recognized until all assumptions are in;
;;; and nogoods are MUCH cheaper than justifications (there are fewer of them.)

(defun valid-cycle? (cycle &aux marker qnum ineq)
  ;Checks to see if cycle has a chord.  Since search is breadth-first, shortest elements will be found first.
  ; Thus the chord marks a sub-cycle which should already have been found, (on the side including the
  ; new ineq), and the subcycle on the other side will have been found by some previous step.  Hence a
  ; cycle with a chord can add no new information.
  (setq marker (cons t t))
  (dolist (entry cycle)
    (put-qnum-prop (car entry) marker :MARKER)
    (put-ineq-prop (cdr entry) marker :MARKER))
  (dolist (pair cycle)
    (setq qnum (car pair))
    (dolist (entry (qnum-ineqs1 qnum))
      (setq ineq (cdr entry))
      (when (and (eq (get-qnum-prop (ineq-n2 ineq) :marker) marker)
		 (not (eq (get-ineq-prop ineq :MARKER) marker)))
	(if-debug-flag cycles
	  (format t "~% INEQ:Found a losing cycle:") (print-cycle cycle))
	(return-from valid-cycle? (values nil ineq))))
    (dolist (entry (qnum-ineqs2 qnum))
      (setq ineq (cdr entry))
      (when (and (eq (get-qnum-prop (ineq-n1 ineq) :marker) marker)
		 (not (eq (get-ineq-prop ineq :MARKER) marker)))
	(if-debug-flag cycles
		       (format t "~% INEQ: Found a losing cycle:")
		       (print-cycle cycle))
	(return-from valid-cycle? (values nil ineq)))))
  (values t nil))

;;;; Installing transitivity justifications

(defun process-transitivity (&aux (count 0) (total 0) start end)
  (adb::if-debug-flag trans
     (format T "~%     TRANS: Beginning to process transitivity cycles..."))
  (adb::if-debug-flag timers (setq start (mark-time)))
  (dolist (cycle *cycles*)
    (incf total)
    (when (valid-cycle? cycle)
      (incf count) (install-cycle-transitivity cycle)))
  (setq *transitivity-processed* t)
  (adb::if-debug-flag trans
		      (format t "     TRANS: ~A out of ~A were processed." count total))
  (adb::if-debug-flag timers
		      (setq end (mark-time))
		      (say-time (make-time-difference end start)
				:header "Transitivity (included in Inf Res)")))

(defun install-ineq-transitivity (ineq)
  (if-debug-flag cycles
    (when (ineq-new-cycles ineq)
      (format t "~% INEQ:Found ~D cycle~:P for: ~A"
	      (length (ineq-new-cycles ineq)) ineq)))
  (dolist (cycle (ineq-new-cycles ineq))
    (install-cycle-transitivity cycle))
  (unless (or *zero-in-lattice*
	      (eq (ineq-type ineq) :MAG))
    (justify-transitivity-across-zero ineq)))

;;;; Nogoods due to transitivity

(defun install-cycle-transitivity (cycle)
  (cond ((cycle-of-assumptions? cycle)
	 (install-transitive-nogoods cycle))
	(t (install-transitive-justifications cycle))))

;; When the ineq is fixed (either by being assumed or known as true) then there is
;; little point in installing justifications: nogoods will do as well.

(defun install-transitive-nogoods (cycle &aux quants qs-node)
  (if-debug-flag transitive-nogoods
		 (format t "~% INEQ: Installing transitive nogoods for:")
		 (print-cycle cycle))
  (setq quants (mapcar #'(lambda (entry) (qnum-quantity (car entry))) cycle)
	qs-node (get-tms-node `((quantities ,@(mapcar #'view-quant quants))
				. :TRUE)))
  (adb::justify-w-nodes qs-node (mapcar #'quant-just-node quants) :CYCLE-EXISTS)
  (do ((entries-after (cdr cycle) (cdr entries-after))
       (entry (car cycle) (car entries-after))
       (entries-before nil (cons entry entries-before))
       just)
      ((null entry))
    (dolist (rel '(:> :<))
      (setq just `(,(make-hard-ineq-rel entry rel)
		   ,@ (mapcar #'(lambda (e)
				  (make-soft-ineq-rel e rel))
			      entries-after)
		   ,@ (nreverse
			(mapcar #'(lambda (e)
				    (make-soft-ineq-rel e rel))
				entries-before))
		   ,qs-node))
      (unless (some #'tms::false? just)
	(adb::nogood-w-nodes just :Transitivity-nogood)))))

;;;; Installing transitivity justifications
(defun install-transitive-justifications (cycle &aux quants qs-node)
  (if-debug-flag transitive-justs
      (format t "~% INEQ: Installing transitive nogoods for:")
      (print-cycle cycle))
  (setq quants (mapcar #'(lambda (entry) (qnum-quantity (car entry))) cycle)
	qs-node (get-tms-node `((quantities ,@(mapcar #'view-quant quants))
				. :TRUE)))
  (adb::justify-w-nodes qs-node (mapcar #'quant-just-node quants) :CYCLE-EXISTS)
  (do ((entries-after (cdr cycle) (cdr entries-after))
       (entry (car cycle) (car entries-after))
       (entries-before nil (cons entry entries-before)))
      ((null entry))
    (unless (fixed-ineq? (cdr entry))
      (install-negative-inequality-justifications
	entries-before entry entries-after qs-node)
      (install-transitive-inequality-justifications
	entries-before entry entries-after qs-node)
      (install-transitive-inequality-justifications
	entries-after entry entries-before qs-node))))

(defun install-negative-inequality-justifications (entries1 entry entries2 qs-node)
  (dolist (rel '(:> :<))
    (let ((node (make-soft-ineq-rel entry (opposite-relation rel)))
	  (nodes `(,@ (mapcar #'(lambda (e)
				  (make-soft-ineq-rel e rel))
			      entries2)
		   ,@ (nreverse
			(mapcar #'(lambda (e)
				    (make-soft-ineq-rel e rel))
				entries1))
		   ,qs-node)))
    (unless (some #'tms::false? nodes)
      (adb::justify-w-nodes node nodes :Transitivity-Negation)))))

(defun install-transitive-inequality-justifications
       (entries1 entry entries2 qs-node)
  (do ((entries1-after (cdr entries1) (cdr entries1-after))
       (entry1 (car entries1) (car entries1-after))
       (entries1-before nil (cons entry1 entries1-before))
       node nodes)
      ((null entry1))
    (dolist (rel '(:> :<))
      (setq node (make-hard-ineq-rel entry (opposite-relation rel))
	    nodes `(,@ (mapcar #'(lambda (e)
				   (make-soft-ineq-rel e rel))
			       entries2)
		      ,@ (nreverse  ; Now splice in the reverse of entries1:
			 (mapcar #'(lambda (e)
				     (make-soft-ineq-rel e rel))
				 entries1-after))
		    ,(make-hard-ineq-rel entry1 rel)
		    ,@ (mapcar #'(lambda (e)
				   (make-soft-ineq-rel e rel))
			       entries1-before)
		    ,qs-node))
      (unless (some #'tms::false? nodes)
	(adb::justify-w-nodes node nodes :Transitivity)))))


;;;;  Using sign information

;; When ZERO not in the graph, must add extra justifications to propagate the
;; equivalent of sign information.  (By convention, signs themselves are not used in
;; other justifications, just inequality relationships with zero.)

(defun justify-transitivity-across-zero (ineq &aux N1-ineq-zero N2-ineq-zero
				        N1-quantity-just N2-quantity-just opp-rel
					node1 node2)
  (setq N1-ineq-zero (grab-ineq (ineq-N1 ineq) *zero-struct*)
	N2-ineq-zero (grab-ineq (ineq-N2 ineq) *zero-struct*)
	N1-quantity-just (quant-just-node (qnum-quantity (ineq-N1 ineq)))
	N2-quantity-just (quant-just-node (qnum-quantity (ineq-N2 ineq))))
  (dolist (rel '(:> :< :=))
    (setq opp-rel (opposite-relation rel))
    (unless (tms::false? (setq node2 (ineq-rel-node ineq ':=)))
      (unless (tms::false?
		(setq node1 (ineq-rel-node N1-ineq-zero rel)))
	(adb::justify-w-nodes (ineq-rel-node N2-ineq-zero rel) 
			     `(,node1 ,node2)
			     :ZERO-TRANSITIVITY))
      (unless (tms::false?
		(setq node1 (ineq-rel-node N2-ineq-zero rel)))
	(adb::justify-w-nodes (ineq-rel-node N1-ineq-zero rel) 
			     `(,node1 ,node2)
			     :ZERO-TRANSITIVITY)))
    (unless (or (tms::false?
		  (setq node2 (ineq-rel-node N2-ineq-zero ':=)))
		(tms::false? 
		  (setq node1 (ineq-rel-node N1-ineq-zero rel))))
	(adb::justify-w-nodes (ineq-rel-node ineq rel)
			     `(,node1 ,node2)
			     :TRANSITIVITY-ACROSS-ZERO))
    (when (eq rel ':=) (return (values)))
    (unless (tms::false? n1-quantity-just)
      (unless (or (tms::false?
		  (setq node1 (ineq-neg-rel-node N1-ineq-zero opp-rel)))
		(tms::false?
		  (setq node2 (ineq-rel-node ineq opp-rel))))
	(adb::justify-w-nodes (ineq-rel-node N2-ineq-zero rel) 
			     `(,N1-quantity-just
			       ,node1 ,node2)
			     :ZERO-TRANSITIVITY))
      (unless (or (tms::false?
		    (setq node1 (ineq-neg-rel-node N1-ineq-zero opp-rel) ))
		(tms::false?
		  (setq node2 (ineq-rel-node N2-ineq-zero opp-rel))))
	(adb::justify-w-nodes (ineq-rel-node ineq rel)
			     `(,N1-quantity-just
			       ,node1 ,node2)
			     :TRANSITIVITY-ACROSS-ZERO)))
    (unless (or (tms::false? N2-quantity-just)
		(tms::false?
		  (setq node1 (ineq-neg-rel-node N2-ineq-zero opp-rel)))
		(tms::false?
		  (setq node2 (ineq-rel-node ineq rel))))
      (adb::justify-w-nodes (ineq-rel-node N1-ineq-zero rel) 
		      `(,N2-quantity-just
		       ,node1 ,node2)
		      :ZERO-TRANSITIVITY))))

;;;; Installing information about cycles involving magnitudes
;; (Suggested by BCF)

(defun justify-transitivity-for-mag (ineq &aux N1-ineq-zero N2-ineq-zero
					  node1 node2)
  (setq N1-ineq-zero (grab-ineq (ineq-N1 ineq) *zero-struct*)
	N2-ineq-zero (grab-ineq (ineq-N2 ineq) *zero-struct*))
  (dolist (rel '(:> :=))
    (unless (tms::false? (setq node2 (ineq-rel-node ineq ':=)))
      (unless (tms::false?
		(setq node1 (ineq-rel-node N1-ineq-zero rel)))
	(adb::justify-w-nodes (ineq-rel-node N2-ineq-zero rel) 
			     `(,node1 ,node2)
			     :ZERO-TRANSITIVITY))
      (unless (tms::false?
		(setq node1 (ineq-rel-node N2-ineq-zero rel)))
	(adb::justify-w-nodes (ineq-rel-node N1-ineq-zero rel) 
			     `(,node1 ,node2)
			     :ZERO-TRANSITIVITY)))
    (unless (or (tms::false?
		  (setq node2 (ineq-rel-node N2-ineq-zero ':=)))
		(tms::false? 
		  (setq node1 (ineq-rel-node N1-ineq-zero rel))))
	(adb::justify-w-nodes (ineq-rel-node ineq rel)
			     `(,node1 ,node2)
			     :TRANSITIVITY-ACROSS-ZERO)))
    (unless (or (tms::false?
		  (setq node2 (ineq-rel-node N1-ineq-zero ':=)))
		(tms::false?
		  (setq node1 (ineq-rel-node N2-ineq-zero ':<))))
      (adb::justify-w-nodes (ineq-rel-node ineq ':<) (list node1 node2)
			    :TRANSITIVITY-VIA-ZERO)))

;;;; Some useful utilities.

(defun print-cycles (&optional (print-each? t) (cycles *cycles*)
		     &aux (invalid-count 0) (nogood-count 0))
  (dolist (cycle cycles
		 (format t "~% ~D cycle~:P, ~D invalid, ~D known or assumed."
			 (length cycles) invalid-count nogood-count))
    (unless (valid-cycle? cycle)
      (incf invalid-count))
    (when (cycle-of-assumptions? cycle)
      (incf nogood-count))
    (when print-each? (print-cycle cycle) (terpri))))

(defun print-cycle (cycle)
  (dolist (entry cycle)
    (format t "~%   ~A" (car entry)))
  (multiple-value-bind (valid? chord)
      (valid-cycle? cycle)
  (unless valid?
    (format t "~%*!@&$^%---INVALID---^@%#*&*!!")
    (format t "~%     Chord: ~S" (ineq-string chord)))))

(defun cycle-of-assumptions? (cycle)
  ;; Cycle is a list of Dotted pairs: ((qnum . ineq) ... )
  (every #'(lambda (entry) (fixed-ineq? (cdr entry))) cycle))

(defun fixed-ineq? (ineq) (or (ineq-assumed? ineq) (constant-ineq? ineq)))

;;; Some helpers used below
(defun make-soft-ineq-rels (entry rel)
  (list (make-soft-ineq-rel entry rel)
	(ineq-neg-rel-node (cdr entry) :BT)))

(defun make-soft-ineq-rel (entry rel)
  (let ((qnum (car entry))
	(ineq (cdr entry)))
    (when (eq qnum (ineq-n1 ineq))
      (setq rel (opposite-relation rel)))
     (ineq-neg-rel-node ineq rel)))

(defun make-hard-ineq-rel (entry rel)
  (let ((qnum (car entry))
	(ineq (cdr entry)))
    (when (eq qnum (ineq-n2 ineq))
      (setq rel (opposite-relation rel)))
    (ineq-rel-node ineq rel)))

(defun view-quant (quant) (adb::lisp-form (quant-assert quant)))

